#!/usr/bin/python
#
# converter between permission being defined per cifs share in the smb.conf
# file to volume-based permissions
# 
# TODO
#	what if cdb is not up?
#	execption handling for the cdb call

import commands, os, re, getopt, sys, syslog

SCRIPT_NAME = os.path.basename(sys.argv[0])

VOLUME_NAME			= "volume"
VOLUME_ID			= "id"
FILE_CREATE_MODE	= "file create mode"
DIR_CREATE_MODE		= "dir create mode"

# params in smb.conf
SMB	= {}
SMB_PATH_PARAM			= "path"
SMB[FILE_CREATE_MODE]	= "force create mode"
SMB[DIR_CREATE_MODE]	= "force directory mode"

# params in the cdb
FSD = {}
FSD_PARAM_PREFIX		= "params.fs.exavols.exavolumes"
FSD[FILE_CREATE_MODE]	= "samba_file_creation_mode_for_unix_volume"
FSD[DIR_CREATE_MODE]	= "samba_dir_creation_mode_for_unix_volume"

# errors
ERROR_MESSAGE_SMB_CONF_FILE_NOT_FOUND = "Error: Unable to open smb.conf file. Exiting."
ERROR_MESSAGE_CDB_NOT_UPDATED = "Error: failed to update CDB with following command:\n%s"
 
ERROR_CODE_ILLEGAL_PARAM = 3
ERROR_CODE_SMB_CONF_FILE_NOT_FOUND = 4
ERROR_CODE_CDB_NOT_UPDATED = 5

QUERY_CDB_FOR_VOLUMES_COMMAND = "paramstree params.fs.exavols.exavolumes |grep .name"
# mock for checking locally:
#QUERY_CDB_FOR_VOLUMES_COMMAND = "echo 'params.fs.exavols.exavolumes.642400.name ccsnap\nparams.fs.exavols.exavolumes.642352.name vol0\nparams.fs.exavols.exavolumes.642368.name Common\nparams.fs.exavols.exavolumes.642384.name prod'"
COMMAND_FOR_CDB_CHANGE = "setparam"

# globals
def set_debug_level(level):
	globals()["DEBUG_LEVEL"] = level

def get_debug_level():
	return int(globals()["DEBUG_LEVEL"])

def set_pretend(val):
	globals()["PRETEND"] = val

def get_pretend():
	return globals()["PRETEND"]

def set_smb_conf_filename(name):
	globals()["SMB_CONF_FILENAME"] = name

def get_smb_conf_filename():
	return globals()["SMB_CONF_FILENAME"]

# utilities
def debug(level, message):
	"print the message to the console and write to syslog"
	if (get_debug_level() >= level):
		print message
		syslog.syslog(syslog.LOG_ERR, message)

def is_header(line):
	return line[0]=='['

def get_new_header(line):
	return line.strip('[]\n')

def parse_param_line(line):
	split_line = line.split("=")
	return split_line[0].strip(), split_line[1].strip()

def get_volume_name_from_path(path):
	return re.sub("^.*/exavol/", "", path).split("/")[0]


class create_mode_converter:
	"""
	converts the file and directory force create mode from being written in the
	smb.conf file and logically cifs-share based to being volume based, and
	written in the cdb
	"""
	
	def read_smb_file(self):
		"opens the smb.conf file"
		try:
			debug(1, ">>> about to open %s file for read" % get_smb_conf_filename())
			smb_conf_file=open(get_smb_conf_filename(), "rt")
		except IOError:
			print SMB_CONF_FILE_NOT_FOUND_ERROR
			sys.exit(ERROR_CODE_SMB_CONF_FILE_NOT_FOUND)

		self.smb_conf_lines = smb_conf_file.readlines()
		smb_conf_file.close()

	def __init__(self):
		# members
		self.params = {}			# all parameters from smb.conf
		self.shares = {}			# only share create mode params
		self.volumes = {}			# all volumes in the system
		self.cdb_commands = []		# commands to be run for updating the cdb
	
		self.read_smb_file()

	def parse_smb_conf(self):
		"parse share create modes from smb.conf"
		current_header = ""
		for line in self.smb_conf_lines:
			if is_header(line):
				current_header = get_new_header(line)
				self.params[current_header] = {}
				debug(2, "[%s]" % current_header)
			elif current_header != "":
				if line.strip() != "": # not an empty line
					current_param, current_value = parse_param_line(line)
					self.params[current_header][current_param] = current_value
					debug(2, "%s[%s]=%s" % (current_header, current_param, current_value))

	def fill_volume_ids(self):
		"query the CDB for the volume name and ids, and populate the volumes dict"
		cdb_query_output = commands.getoutput(QUERY_CDB_FOR_VOLUMES_COMMAND)
		for line in cdb_query_output.splitlines():
			id, name = line.replace("params.fs.exavols.exavolumes.", "").replace(".name", "").split()
			debug(2, "VOLUME %s, ID %s" % (name, id))
			self.volumes[name] = {VOLUME_ID: id}
	
	def set_volume_create_mode(self, volume_name, mode, value):
		"sets the volume's given create mode to the value given"
		if self.volumes[volume_name].has_key(mode):
			debug(0, "%s: Warning: more than one share affects \"%s\" on volume %s\n	Current setting (%s) will not be replace by new setting (%s)" % (SCRIPT_NAME, mode, volume_name, self.volumes[volume_name][mode], value))
		else:
			self.volumes[volume_name][mode] = value

	def fill_share_create_modes(self):
		"iterate over the params and fill the shares with the create mode params"
		for share_name in self.params:
			if share_name != "global":
				self.shares[share_name] = {}
				for param_name, param_value in self.params[share_name].iteritems():
					debug(2, ">>> iteration over share params: %s = %s" % (param_name, param_value))
					if param_name == SMB_PATH_PARAM:
						self.shares[share_name][VOLUME_NAME] = get_volume_name_from_path(param_value)
					elif param_name == SMB[FILE_CREATE_MODE]:
						self.shares[share_name][FILE_CREATE_MODE] = param_value
					elif param_name == SMB[DIR_CREATE_MODE]:
						self.shares[share_name][DIR_CREATE_MODE] = param_value
			else: # global
				for (param_name, param_value) in self.params[share_name].iteritems():
					if param_name in (SMB[FILE_CREATE_MODE], SMB[DIR_CREATE_MODE]):
						debug(0, "%(script_name)s: Warning: ignoring \"%(parameter)s\" in global section." % {"script_name": SCRIPT_NAME, "parameter": param_name})
					
	def fill_volume_create_modes(self):
		"iterate over the shares and decide which create modes best suit each volume"
		for share_name in self.shares:
			debug(1, ">>> converting share %s to volume create mode" % share_name)
			share_defined_params = self.shares[share_name].keys()
			volume_name			 = self.shares[share_name][VOLUME_NAME] 
			for mode in (FILE_CREATE_MODE, DIR_CREATE_MODE):
				if mode in share_defined_params:
					self.set_volume_create_mode(volume_name, mode, self.shares[share_name][mode])

	def build_volume_create_modes(self):
		"build volume create modes data struct"
		self.fill_volume_ids()
		self.fill_share_create_modes()
		debug(2, ">>> self.shares after population:\n %s" % self.shares)
		debug(2, ">>> self.volumes before population:\n %s" % self.volumes)
		self.fill_volume_create_modes()
		debug(2, ">>> self.volumes after population:\n %s" % self.volumes)

	def prepare_cdb_commands(self):
		for volume_name in self.volumes:
			changed_modes = self.volumes[volume_name].keys()
			for mode in (FILE_CREATE_MODE, DIR_CREATE_MODE):
				if mode in changed_modes:
					volume_id = self.volumes[volume_name][VOLUME_ID]
					value	  = eval("0" + self.volumes[volume_name][mode])
					command = "%(command)s %(prefix)s.%(vol_id)s.%(changed_mode)s %(new_mode)s" % {"command": COMMAND_FOR_CDB_CHANGE, "prefix": FSD_PARAM_PREFIX, "vol_id": volume_id, "changed_mode": FSD[mode], "new_mode": value}
					debug(1, ">>> setparam command: %s" % command)
					self.cdb_commands.append(command)
		debug(1, ">>> commands:\n %s" % self.cdb_commands)

	def save_volume_create_modes(self):
		"update the cdb with the keys and values matching the new force create modes"
		for command in self.cdb_commands:
			ret = os.system(command)
			if ret != 0:
				debug(0, ERROR_MESSAGE_CDB_NOT_UPDATED % command)
				sys.exit(ERROR_CODE_CDB_NOT_UPDATED)

	def make_smb_conf_backup(self):
		"back up the smb.conf file"
		debug(1, ">>> Backing up smb.conf")
		os.system("cp %s{,.backup}" % get_smb_conf_filename())

	def remove_force_create_mode_from_smb_conf(self):
		"edit the smb.conf file to remove the force create mode lines"
		os.system("sed -i '/%s/d' %s" % (SMB[FILE_CREATE_MODE], get_smb_conf_filename()))
		os.system("sed -i '/%s/d' %s" % (SMB[DIR_CREATE_MODE], get_smb_conf_filename()))

	def clean_smb_conf(self):
		"clean smb.conf from all the old params"
		self.make_smb_conf_backup()
		self.remove_force_create_mode_from_smb_conf()

	def convert(self):
		self.parse_smb_conf()
		self.build_volume_create_modes()
		self.prepare_cdb_commands()
		if not get_pretend():
			self.save_volume_create_modes()
			self.clean_smb_conf()

def usage():
	print """usage: %s [OPTION]...
Converts the "force create mode" parameters from being share-based to being
volume-based by setting the appropriate cdb keys.

Optional arguments:
	-d, --debug=LEVEL			print debug information
	-h, --help					show this screen
	-p, --pretend				do not actually change configuration
	-s, --smbconfigfile=PATH	use specified smb.conf file instead of the default one
""" % SCRIPT_NAME


def main():
	try:
		opts, args = getopt.getopt(sys.argv[1:], "hpd:s:", ["help", "pretend", "debug=", "smbconfigfile="])
	except getopt.GetoptError:
		# print usage and exit
		usage()
		sys.exit(ERROR_CODE_ILLEGAL_PARAM)

	# defaults
	set_debug_level(0)
	set_pretend(False)
	set_smb_conf_filename("/cluster/config/samba/smb.conf")

	for o, a in opts:
		debug(2, "handling command line args %s (=%s)" % (o, a))
		if o in ("-h", "--help"):
			usage()
			sys.exit()
		if o in ("-p", "--pretend"):
			set_pretend(True)
		if o in ("-d", "--debug"):
			set_debug_level(a)
		if o in ("-s", "--smbconfigfile"):
			set_smb_conf_filename(a)

	conv = create_mode_converter()
	conv.convert()

if __name__ == '__main__':
	main()
