#!/bin/bash
#
# CHECK_PROTO_PATCH trigger - check protocols patch pre-checkin trigger
#
# Check whether the comment contains a valid patch name for samba.
# If no, suggest to specify a patch name, for example:
#	smb_patch_3.0.32_2008-11-13-10:29:39
# If such patch exists, apply its name as attribute to the checked-in version.
#
# Install the trigger as follows (on the /vobs/vendor only):
#    cleartool mktrtype -element -all -preop checkin \
#	-c 'Check protocols patch pre-checkin' \
#	-execunix /ccase/triggers/check_proto_patch.trg \
#	-execwin 'cmd.exe /C ' \
#	CHECK_PROTO_PATCH@/vobs/vendor
#
# Besides, the PatchName attribute type should be created in /vobs/vendor:
#    ct mkattype -c 'Attribute type for patch name' PatchName@/vobs/vendor
#

# early exit point

[ "${IGNORE_PROTO_PATCH_TRIGGER}" -o "${IGNORE_ALL_TRIGGERS}" ] && exit 0
[ "${DEBUGTRIGGER}" ] && { echo === $0 ===; set -xv; }	## debug output

# enforce patch name only on main branches
case "${CLEARCASE_BRTYPE}" in
    austin|buick|chevy|ducati|del4|elise|local|main) :;;
    *) exit 0;;
esac

# select smb files to process
file="${CLEARCASE_PN}"
proto=
case "${file}" in
    */vobs/vendor/smb/exanet*)  exit 0;;
    */vobs/vendor/smb/*)        proto="smb";;
    *)                          exit 0;;
esac

cleartool=${ATRIAHOME:-/usr/atria}/bin/cleartool
PATH=/bin:/usr/bin:/usr/sbin
pattern="^Patch: \(${proto}_patch_[^[:blank:],;]*\).*"

# patch name should appear in the 1st comment line in any place
patch=$(echo "${CLEARCASE_COMMENT}" | sed -n "s/${pattern}/\1/p" | tail -1)

# if patch name was not specified in the comment, request it now
if [ -z "${patch}" ]; then
    clearprompt=${ATRIAHOME:-/usr/atria}/bin/clearprompt
    tmpfile=${TMPDIR:-/tmp}/$$
    prompt="Please enter ${proto} patch name:"
    if [ "${CLEARCASE_XC}" ]; then	## under xclearcase use clearprompt
	${clearprompt} text -out "${tmpfile}" -prompt "${prompt}"
	if [ -r "${tmpfile}" ]; then
	    patch=$(head -1 "${tmpfile}")
	    rm -f "${tmpfile}"
	else
	    exit 1
	fi
    else				## in text mode do not use clearprompt
	read -p "${prompt} " patch 
    fi
fi

# validate patch file existence
patch_file="${EXAROOT}/vendor/${proto}/exanet/patch/${patch}"
if [ ! -f "${patch_file}" ]; then
    echo "***Patch file ${patch_file} does not exist"
    exit 1
fi

# apply attribute
${cleartool} mkattr -rep -nc PatchName "\"${patch}\"" "${file}"

### the last line ###
