#include "includes.h"
#include "exasmb_id_mapping.h"
#include "smbutil/sid_struct.h"
#include "smbutil/exasmb_vfsfsclient.h"
#include "exasmb_if.h"

BOOL exasmb_is_connected_to_id_mapper()
{
	static BOOL is_connection_established = False;

	if ( ! is_connection_established ) {
		exasmb_try_getsmbid( &is_connection_established );
	}
	return is_connection_established;
}

static 
const char* strerror_extended_by_vfs_messages( int errno_ex )
{
	static char unknown_err_buf[100];
	
	if(errno_ex >= 0)
		return strerror(errno_ex);
	switch(errno_ex)
	{
		case VFS_FATAL_ERROR:
			return "VFS_FATAL_ERROR";
		case VFS_OPERATION_FATAL_ERROR:
			return "VFS_OPERATION_FATAL_ERROR";
		case VFS_CLIENT_DISCONNECTED_ERROR:
			return "VFS_CLIENT_DISCONNECTED_ERROR";
		case VFS_UNEXPECTED_ERROR:
			return "VFS_UNEXPECTED_ERROR";
		case VFS_UNKNOWN_REQUEST_TYPE:
			return "VFS_UNKNOWN_REQUEST_TYPE";
		case VFS_INTERNAL_FS_PROBLEM:
			return "VFS_INTERNAL_FS_PROBLEM";
		case VFS_READ_ERROR:
			return "VFS_READ_ERROR";
		case VFS_WRITE_ERROR:
			return "VFS_WRITE_ERROR";
		case VFS_PERM_DENIED:
			return "VFS_PERM_DENIED";
		case VFS_NOT_SUPPORTED:
			return "VFS_NOT_SUPPORTED";
	}

	/* Error not translated... */
	snprintf(unknown_err_buf, sizeof(unknown_err_buf), "Unknown error #%d", errno_ex);
	return unknown_err_buf;
}

static
const char* sid_struct_to_string_static( const sid_struct* p_sid_struct )
{
	static fstring buffer;
	return exasmb_vfs_sid_to_string( p_sid_struct, buffer, sizeof( buffer ) );
}

static
int verify_DOM_SID_and_sid_struct_compatibility()
{
	static bool verified_once = false;
	static int verification_ret = 0;
	static int verification_errno = 0;

	DOM_SID a_dom_sid;
	sid_struct a_sid_struct;

	if ( ! verified_once ) {
		if ( sizeof(a_dom_sid.sid_rev_num) != sizeof(a_sid_struct._revision) ||
			 sizeof(a_dom_sid.num_auths) != sizeof(a_sid_struct._sub_authority_count) ||
			 sizeof(a_dom_sid.id_auth) != sizeof(a_sid_struct._identifier_authority._value) ||
			 sizeof(a_dom_sid.id_auth[0]) != sizeof(a_sid_struct._identifier_authority._value[0]) ||
			 sizeof(a_dom_sid.sub_auths) != sizeof(a_sid_struct._sub_authority) ||
			 sizeof(a_dom_sid.sub_auths[0]) != sizeof(a_sid_struct._sub_authority[0]) ) {
			DEBUG(0,( "ERROR: DOM_SID and sid_struct are not compatible.\n" ));
			verification_errno = ENOSYS;
			verification_ret = -1;
		} else {
			verification_ret = 0;
		}
		verified_once = true;
	}

	errno = verification_errno;
	return verification_ret;
}

static
int convert_DOM_SID_to_sid_struct( const DOM_SID* sid_in, sid_struct* sid_out )
{
	int i = 0;

	if ( sid_in == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: sid_in is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}
	if ( sid_out == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: sid_out is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}

	int ret_verify_compatib = verify_DOM_SID_and_sid_struct_compatibility();
	if ( ret_verify_compatib != 0 ) {
		return ret_verify_compatib;
	}

	sid_out->_revision = sid_in->sid_rev_num;
	sid_out->_sub_authority_count = sid_in->num_auths;
	for ( i = 0 ; i < SID_AUTHORITY_SIZE ; ++i ) {
		sid_out->_identifier_authority._value[i] = sid_in->id_auth[i];
	}
	for ( i = 0 ; i < sid_in->num_auths ; ++i ) {
		sid_out->_sub_authority[i] = sid_in->sub_auths[i];
	}

	return 0;
}

static
int convert_sid_struct_to_DOM_SID( const sid_struct* sid_in, DOM_SID* sid_out )
{
	int i = 0;

	if ( sid_in == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: sid_in is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}
	if ( sid_out == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: sid_out is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}

	int ret_verify_compatib = verify_DOM_SID_and_sid_struct_compatibility();
	if ( ret_verify_compatib != 0 ) {
		return ret_verify_compatib;
	}

	sid_out->sid_rev_num = sid_in->_revision;
	sid_out->num_auths = sid_in->_sub_authority_count;
	for ( i = 0 ; i < SID_AUTHORITY_SIZE ; ++i ) {
		sid_out->id_auth[i] = sid_in->_identifier_authority._value[i];
	}
	for ( i = 0 ; i < sid_in->_sub_authority_count ; ++i ) {
		sid_out->sub_auths[i] = sid_in->_sub_authority[i];
	}

	return 0;
}

int exasmb_map_sid_to_temp_id( const DOM_SID* sid, uint32_t* temp_id_out )
{
	int ret_int = 0;
	sid_struct sid_struct_aux;
	static int failure_dbg_level = 2;
	static int success_dbg_level = 10;
	int dbg_level = 0;

	if ( sid == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: sid is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}
	if ( temp_id_out == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: temp_id_out is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}

	ret_int = convert_DOM_SID_to_sid_struct( sid, &sid_struct_aux );
	if ( ret_int != 0 ) {
		DEBUG(0,( "Failed to convert DOM_SID (%s) to sid_struct. errno: %d (%s).\n", 
				  sid_string_static(sid), errno, strerror_extended_by_vfs_messages(errno) ));
		return ret_int;
	}

	DEBUG(success_dbg_level,( "Calling extern exasmb_vfs_map_sid_to_temp_id with SID: %s...\n",
							  sid_string_static(sid) ));

	ret_int = exasmb_vfs_map_sid_to_temp_id( &sid_struct_aux, temp_id_out );

	dbg_level = ((ret_int == 0) ? success_dbg_level : failure_dbg_level);
	DEBUG(dbg_level,( "Call to exasmb_map_sid_to_temp_id completed. Return value: %d", 
					  ret_int ));
	if ( ret_int == 0 ) {
		DEBUGADD(success_dbg_level,( " (SUCCESS)" ));
		DEBUGADD(success_dbg_level,( ", SID %s mapped to TEMP ID %u",
									 sid_string_static(sid), *temp_id_out ));
	} else {
		DEBUGADD(failure_dbg_level,( " (FAILURE), SID: %s, errno: %d (%s)", 
									 sid_string_static(sid), errno, strerror_extended_by_vfs_messages(errno) ));
	}
	DEBUGADD(dbg_level,( ".\n" ));

	return ret_int;
}

int exasmb_map_temp_id_to_sid( uint32_t temp_id, DOM_SID* sid_out )
{
	int ret_int = 0;
	sid_struct sid_struct_aux;
	static int failure_dbg_level = 2;
	static int success_dbg_level = 10;
	int dbg_level = 0;

	if ( sid_out == NULL ) {
		DEBUG(0,( "ERROR: Invalid argument: sid_out is NULL.\n" ));
		errno = EINVAL;
		return -1;
	}

	DEBUG(success_dbg_level,( "Calling exasmb_vfs_map_temp_id_to_sid with TEMP ID: %u...\n",
							  temp_id ));

	ret_int = exasmb_vfs_map_temp_id_to_sid( temp_id, &sid_struct_aux );

	dbg_level = ((ret_int == 0) ? success_dbg_level : failure_dbg_level);
	DEBUG(dbg_level,( "Call to exasmb_map_temp_id_to_sid completed. Return value: %d", ret_int ));
	if ( ret_int == 0 ) {
		DEBUGADD(success_dbg_level,( " (SUCCESS)" ));
		DEBUGADD(success_dbg_level,( ", TEMP ID %u mapped to SID %s.\n",
									 temp_id, sid_struct_to_string_static(&sid_struct_aux) ));
	} else {
		DEBUGADD(failure_dbg_level,( " (FAILURE), TEMP ID %u, errno: %d (%s).\n", 
									 temp_id, errno, strerror_extended_by_vfs_messages(errno) ));
		return ret_int;
	}

	ret_int = convert_sid_struct_to_DOM_SID( &sid_struct_aux, sid_out );
	if ( ret_int != 0 ) {
		DEBUG(0,( "TEMP ID %u was successfully mapped to SID %s but failed to convert sid_struct to DOM_SID. errno: %d (%s).\n", 
				  temp_id, sid_struct_to_string_static(&sid_struct_aux), errno, strerror_extended_by_vfs_messages(errno) ));
		return ret_int;
	}

	return 0;
}

