#include "includes.h"
#include "exasmb_types.h"
#include "exasmb_if.h"
extern int              last_lock_counter;

static void exasmb_if_br_check_error(exasmb_message_error err, const char * what, uint64_t dev, uint64_t inode)
{
	switch (err) {
	case EXASMB_OK:
		return;
	case EXASMB_FATAL_ERROR:
	case EXASMB_LRLOCK_ERROR:
	case EXASMB_ERROR:
		DEBUG(0, ("%s failed: err %d dev %u inode %llx - Terminating\n",
			  what, err, (unsigned) dev, (long long) inode));
		break;
	default:
		DEBUG(0, ("UNKNOWN ERROR - %s failed: err %d dev %u inode %llx - Terminating\n",
			  what, err, (unsigned) dev, (long long) inode));
		break;
	}

	smb_panic(what);
}

static void
pack_br_entry(exasmb_br_lock *exasmb_br_entry, const struct lock_struct *smb_br_entry)
{
	SMB_ASSERT(smb_br_entry->lock_flav == WINDOWS_LOCK);
	exasmb_br_entry->smbid = exasmb_getsmbid();
	exasmb_br_entry->smbpid = smb_br_entry->context.smbpid;
	exasmb_br_entry->tid = smb_br_entry->context.tid;
	exasmb_br_entry->pid = smb_br_entry->context.pid.pid;
	exasmb_br_entry->node = smb_br_entry->context.pid.node;
	exasmb_br_entry->fnum = smb_br_entry->fnum;
	exasmb_br_entry->start = smb_br_entry->start;
	exasmb_br_entry->size = smb_br_entry->size;
	exasmb_br_entry->lock_type = smb_br_entry->lock_type;
}

static void
unpack_br_entry(const exasmb_br_lock *exasmb_br_entry, struct lock_struct *smb_br_entry)
{
	smb_br_entry->context.smbid = exasmb_br_entry->smbid;
	smb_br_entry->context.smbpid = exasmb_br_entry->smbpid;
	smb_br_entry->context.tid = exasmb_br_entry->tid;
	smb_br_entry->context.pid = node_pid_to_procid(exasmb_br_entry->node, exasmb_br_entry->pid);
	smb_br_entry->fnum = exasmb_br_entry->fnum;
	smb_br_entry->start = exasmb_br_entry->start;
	smb_br_entry->size = exasmb_br_entry->size;
	smb_br_entry->lock_type = exasmb_br_entry->lock_type;
	smb_br_entry->lock_flav = WINDOWS_LOCK;
}

#if 0
/* unused */
static void
pack_br_array(exasmb_br_lock *exasmb_br_array, struct lock_struct *smb_br_array, int size)
{
	int i;

	for (i = 0; i < size; i++) {
		pack_br_entry(&exasmb_br_array[i], &smb_br_array[i]);
	}
}
#endif

static void
unpack_br_array(exasmb_br_lock *exasmb_br_array, struct lock_struct *smb_br_array, int size)
{
	int i;

	for (i = 0; i < size; i++) {
		unpack_br_entry(&exasmb_br_array[i], &smb_br_array[i]);
	}
}

/************************************************************************************************/
/*                                  Byte range locks                                            */
/************************************************************************************************/

static BOOL
exasmb_if_br_lock_lr_int(files_struct *fsp, BOOL do_lock)
{
	exasmb_br_lock_lr_request request;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.lr, False);
	 
	request.lock_count = last_lock_counter;

	if (do_lock) {
		verify_no_double_lock (fsp->dev,fsp->inode);
	}	
	do {
		err = do_lock
			? exasmb_br_lock_lr(client, &request)
			: exasmb_br_unlock_lr(client, &request);
	} while (err == EXASMB_COMM_ERROR);
	exasmb_if_br_check_error(err, do_lock ? "exasmb_if_br_lock_lr" : "exasmb_if_br_unlock_lr", fsp->dev, fsp->inode);

	if (do_lock) {
		exasmb_lock_increase(fsp->dev,fsp->inode);
	} else {
		exasmb_lock_decrease(fsp->dev,fsp->inode);
	}

    
	return True;
}

BOOL
exasmb_if_br_lock_lr(files_struct *fsp)
{
	DEBUG(10, ("exasmb_if_br_lock_lr: %s\n", fsp->fsp_name));
	return exasmb_if_br_lock_lr_int(fsp, True);
}

BOOL
exasmb_if_br_unlock_lr(files_struct *fsp)
{
	DEBUG(10, ("exasmb_if_br_unlock_lr: %s\n", fsp->fsp_name));
	return exasmb_if_br_lock_lr_int(fsp, False);
}

int32_t
exasmb_if_br_get_locks(files_struct *fsp, struct lock_struct **pp_locks, uint32_t num_of_entries, uint64_t cookie_in, uint64_t *cookie_out)
{
	exasmb_br_get_locks_request request;
	exasmb_br_get_locks_reply reply;
	dlm_client client;
	int32_t num_br_locks;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.lr, -1);
	request.cookie = cookie_in;
	request.num_locks = MIN(num_of_entries, MAX_FETCH_NUM_ENTRIES);

	do {
		reply.body.entries = NULL;

		err = exasmb_br_get_locks(client, &request, &reply);
	} while( err == EXASMB_COMM_ERROR);

	exasmb_if_br_check_error(err, "exasmb_if_br_get_locks", fsp->dev, fsp->inode);

	*cookie_out = EXASMB_COOKIE_ENDDATA;
	num_br_locks = reply.header.num_entries;

	if (num_br_locks > 0) {
		/* Paranoia */
		if (!reply.body.entries) {
			   DEBUG(1,("No entries returned in reply;"
				    " locks %d cookie %06llx\n",
				    (int) num_br_locks,
				    (long long) reply.header.cookie));
			   return -1;
		}

		*pp_locks = malloc(num_br_locks * sizeof(struct lock_struct));
		if (!*pp_locks) {
			DEBUG(0, ("Unable to allocate lock entries"));
			return -1;
		}

		unpack_br_array(reply.body.entries, *pp_locks, num_br_locks);

		*cookie_out = reply.header.cookie;
	}

	return num_br_locks;
}

BOOL
exasmb_if_br_add_lock(files_struct *fsp, const struct lock_struct *entry)
{
	exasmb_br_add_lock_request request;
	exasmb_br_lock *plock = NULL;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.header.lr, False);
	request.header.num_entries = 1;

	plock = (exasmb_br_lock *) exasmb_dlm_client_get_req_buf(client, sizeof(exasmb_br_lock));
	if (!plock) {
		DEBUG(0, ("Unable to allocate request buffer"));
		return False;
	}

	pack_br_entry(plock, entry);
	request.body.entries = plock;

	do {
		err = exasmb_br_add_lock(client, &request);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_br_check_error(err, "exasmb_if_br_add_lock", fsp->dev, fsp->inode);

	exasmb_dlm_client_release_req_buf(client, plock);

	return True;
}


BOOL
exasmb_if_br_delete_lock(files_struct *fsp, const struct lock_struct *entry)
{
	exasmb_br_delete_lock_request request;
	exasmb_br_chk_lock_request chkreq;
	exasmb_message_error err;
	exasmb_br_lock *plock;
	dlm_client client;

	client = exasmb_if_client_open();

	plock = (exasmb_br_lock *) exasmb_dlm_client_get_req_buf(client, sizeof(exasmb_br_lock));
	if (!plock) {
		DEBUG(0, ("Unable to allocate request buffer"));
		return False;
	}

	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, chkreq.header.lr, False);
	chkreq.header.num_entries = 1;
	pack_br_entry(plock, entry);
	chkreq.body.entries = plock;

	do {
		err = exasmb_br_chk_lock(client, &chkreq);
	} while (err == EXASMB_COMM_ERROR);

	if(err == EXASMB_NOLOCK_ERROR) {
		exasmb_dlm_client_release_req_buf(client, plock);
		return False;
	}

	exasmb_if_br_check_error(err, "exasmb_if_br_delete_lock", fsp->dev, fsp->inode);

	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.header.lr, False);
	request.header.num_entries = 1;
	pack_br_entry(plock, entry);
	request.body.entries = plock;

	do {
		err = exasmb_br_delete_lock(client, &request);
	} while ( err == EXASMB_COMM_ERROR );

	exasmb_if_br_check_error(err, "exasmb_if_br_delete_lock", fsp->dev, fsp->inode);

	exasmb_dlm_client_release_req_buf(client, plock);

	return True;
}

BOOL
exasmb_if_file_close(files_struct *fsp)
{
	exasmb_file_close_request request;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();
	request.fnum = fsp->fnum;

	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.lr, False);

	do {
		err = exasmb_file_close(client, &request);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_br_check_error(err, "exasmb_if_file_close", fsp->dev, fsp->inode);

	return True;
}
