#include "includes.h"
#include "exasmb_types.h"
#include "exasmb_if.h"
#include "smb_macros.h"
#include <sys/types.h>
#include <limits.h>

/* this is to be able to clear the TDB_CLEAR_IF_FIRST flag */
#include "tdb/common/tdb_private.h"
#include "events/EventsClientAccess_c.h"


/* Exastore kernel defines 18 as OOM protection. */
#define PR_SET_OOM_PROTECT 18

#define SLEEP_TIME_IN_SEC 5

static int		smb_instances;
static int		*smb_instance_id;
static int64_t		*smb_instance_version;
static dlm_client	h_client = NULL;
static int32_t		g_smbid = -1;

int              last_lock_counter = 0;
static uint64_t last_dev = 0;
static uint64_t last_inode =0;

const char *smbd_cleaner_pid_file = "smbd_dlmcleaner";
const char *env_instances_str = "EXANET_SMB_INSTANCES";
const char *env_vers_base_path_str = "EXANET_SMB_VERS_BASE_PATH";
const char *exa_smb_id_str = "EXANET_SMB_ID";
static char **env_id_str;
const char *env_root_path = "EXANET_SMB_ROOT_PATH";

static int filedes[2];
static pid_t child_monitor = 0;
typedef struct {
	pid_t pid;
	int status;
} proc_exit_status;

typedef struct fnfs_fh_t {
	unsigned short size;
	unsigned char data[64];
} fnfs_fh;

typedef struct nfs_getfh_t {
	 int fd;
	 fnfs_fh *fh;
} nfs_getfh;

/*
 * get nfs file handle
 */
int exasmb_get_nfs_fh(int fd, unsigned short* len, char nfsfh[64] )
{
	static int dev_exastore_fd = -1;
	nfs_getfh nfs_fh;

	if (dev_exastore_fd == -1) {
		 dev_exastore_fd = open ("/dev/exastore", 0);
	}

	DEBUG(5,("exasmb_get_nfs_fh: fd %d \n", fd));

	nfs_fh.fd = fd;
	nfs_fh.fh = (fnfs_fh *) malloc (sizeof (fnfs_fh));
	nfs_fh.fh->size = 0;

	if (ioctl(dev_exastore_fd, _IOR('e', 1000, nfs_getfh), &nfs_fh) == -1) {
		DEBUG(1,("exasmb_get_nfs_fh: ioctl failed - %m\n"));
		free (nfs_fh.fh);
		return -1;
	}

	DEBUG(5,("exasmb_get_nfs_fh:enter2, copying %d bytes to buffer\n", nfs_fh.fh->size));

	memcpy(nfsfh, nfs_fh.fh->data, nfs_fh.fh->size);
	if (len) {
		*len = nfs_fh.fh->size;
		DEBUG(5,("Set len - size was %d\n", *len));
	}
	else
		DEBUG(5,("unable to set len - the pointer is NULL\n"));
 
	if( DEBUGLVL( 5 )) {
		int i_fh;
		pstring nfs_fh_string,temp_str;

		nfs_fh_string[0] = '\0';
		for (i_fh=0; i_fh<32; i_fh++) {                
			snprintf (temp_str,PSTRING_LEN,"%s%.2X %s",
				i_fh>0 ? nfs_fh_string : "", (unsigned char) nfsfh[i_fh], (1+i_fh)%8==0 ? "\n" : "");
			strncpy (nfs_fh_string,temp_str, PSTRING_LEN);
		}
		DEBUG(5,("exasmb_get_nfs_fh: This is the nfs file handle we got:\n%s\n", nfs_fh_string)); 
	}
	free (nfs_fh.fh);
	return 0;
}

static void
exasmb_cleaner_child_exit(pid_t pid, int status, void * private_data) 
{
	/* Important note: DEBUG messages have been removed to a deadlock with syslog.
	   Please try to keep it that way. NS. */

	if (WIFEXITED(status)) {
		return;
	}
	if (WIFSTOPPED(status)) {
		/* Are we really sure we want to exit if we stop? That might leave
		   a DLM cleaner hanging without a daddy. */
	}
	_exit(1);
	return; /* avoid warnings */
}

/* Write the process information of the dead process into the pipe. */
static void
exasmb_child_exit(pid_t pid, int status, void * private_data)
{
	/* Again, important note, DEBUG messages where not essential have been removed to avoid deadlocks. */

	if (child_monitor) {
		if (pid == child_monitor) {
			DEBUG(0,("PANIC: DLM cleaner master process died. Exiting.\n"));
			char panic_reason[] = {"DLM cleaner master process is dead.\n"};
			/* Dump supportability event on such disaster */
			clientAccess_cifsAllClientsDisconnection_event(panic_reason);
	   		exit_server(panic_reason);
		}
		ssize_t n,left = sizeof(proc_exit_status);
		proc_exit_status stat;
		char *buf = (char*) &stat;
		
		/* DEBUG(5, ("exasmb_child_exit: pid %u status %d\n", (unsigned) pid, status));  */
		
		stat.pid = pid;
		stat.status = status;
		do {
			n = write(filedes[1], buf, left);
			if (n == left)
				return;
			if (n == -1) {
				if (errno != EINTR) {
					DEBUG(0, ("exasmb_child_exit: write failed"
							  " %d pid %u status %d\n",
							  errno, (unsigned) pid, status));
					usleep(777);
				}
			} else {
				buf += n;
				left -= n;
			}
		} while ( 1 );
	} else {
		DEBUG(0, ("exasmb_child_exit: unexpected in %u (pid %u status %d)\n",
				  (unsigned) sys_getpid(), (unsigned) pid, status));
	}
}


static TDB_CONTEXT *tdb; // key is pid

static void
init_dlm_cleanup_tdb(int flags)
{
	if (!tdb)
		tdb = tdb_open_log(lock_path("dlm_cleaners.tdb"), 0, flags,
						   O_RDWR | O_CREAT, 0644);
	if (!tdb)
		smb_panic("failed to init dlm_cleaners.tdb");
}

static void
deinit_dlm_cleanup_tdb(void)
{
	tdb_close(tdb);
	tdb = NULL;
}

void
register_dlm_cleanup_pid(void)
{
	pid_t needs_cleaning=EXASMB_CLEANER_NEEDS_CLEANING;
	pid_t pid;

	TDB_DATA dbuf;
	TDB_DATA kbuf;
	
	init_dlm_cleanup_tdb(TDB_DEFAULT);
	pid = sys_getpid();
	kbuf.dptr = (char*) &pid;
	kbuf.dsize = sizeof(pid);
	dbuf.dptr=(char*)&needs_cleaning;
	dbuf.dsize=sizeof(needs_cleaning);

	if (tdb_store(tdb, kbuf, dbuf, TDB_INSERT) != 0) {
		char *reason = NULL;
		asprintf(&reason, "failed to insert %u to dlm_cleaners.tdb with error %s\n",
				 (unsigned) pid, tdb_errorstr(tdb));
		smb_panic(reason);
	}

	deinit_dlm_cleanup_tdb();
	DEBUG(1, ("registered pid %u\n", pid));
}

static void
unregister_dlm_cleanup_pid(pid_t pid)
{
	TDB_DATA key;
	key.dptr = (char*)&pid;
	key.dsize = sizeof(pid);
	if (tdb_delete(tdb, key) != 0) {
		if (tdb_error(tdb) == TDB_ERR_NOEXIST) {
			DEBUG(1,("did not find pid %u in tdb\n", pid));
			return;
		}

		DEBUG(0,("failed to delete pid %u from tdb %s\n",
			(unsigned) pid, tdb_errorstr(tdb)));
	}
}

BOOL
need_dlm_cleanup(pid_t pid)
{
	TDB_DATA key;
	key.dptr = (char*)&pid;
	key.dsize = sizeof(pid);

	return tdb_exists(tdb,key);
}

static BOOL
is_being_cleaned(pid_t smbdpid) {
	int ret = False;
	TDB_DATA key;
	TDB_DATA data;

	DEBUG(10, ("is_being_cleaned() entry.\n"));
	key.dptr=(char*)&smbdpid;
	key.dsize=sizeof(smbdpid);
	data = tdb_fetch(tdb,key);
	if (data.dptr == NULL) {
		DEBUG(5,("data.dptr is NULL\n"));
		return False;
	}
	DEBUG(5,("is_being_cleaned() - data fetched is - pid : %u\n", (pid_t)(*data.dptr)));
	if ((pid_t)(*data.dptr) == EXASMB_CLEANER_IS_CLEANING) {
		ret = True;
	}
	SAFE_FREE(data.dptr);
	return ret;
}

static void 
mark_cleaning_pid(pid_t smbdpid, pid_t cleanerpid)
{
	TDB_DATA dbuf;
	TDB_DATA kbuf;

	kbuf.dptr=(char*)&smbdpid;
	kbuf.dsize=sizeof(smbdpid);
	dbuf.dptr=(char*)&cleanerpid;
	dbuf.dsize=sizeof(cleanerpid);
	if ( tdb_store(tdb,kbuf,dbuf,TDB_REPLACE) != 0 ) {
		DEBUG(0,("Couldn't mark cleaning pid as running in tdb_cleaners.\n"));
	}
}

void
exasmb_init_cleanup(void)
{
	ssize_t n, left;
	pid_t pid, child;
	int status;
	dlm_client cln;
	exasmb_cleanup_request request;
	exasmb_message_error err;
	proc_exit_status stat;
	char *buf;
	int i;
	
	if (child_monitor)
		return;

	if (pipe(filedes) < 0) {
		DEBUG(1, ("pipe failed: %d\n", errno));
		return;
	}

	if (lp_use_external_lock_db()) {
		init_dlm_cleanup_tdb(TDB_CLEAR_IF_FIRST|TDB_DEFAULT); // fail before fork
	}

	if ((pid = sys_fork()) == (pid_t)-1) {
		char *reason = NULL;
		asprintf(&reason, "failed to fork dlm cleanup process - %m\n");
		smb_panic(reason);
	}

	if (pid) {
		/* This is the parent. pid is the pid of monitor process */
		child_monitor = pid;

		sigchld_register(exasmb_child_exit, NULL);

		DEBUG(3, ("master cleaner pid is %d\n", child_monitor));
		return;
	}

	/* The code below is the "master cleaner". It is executed ONLY if we are using DLM */

	tdb->flags &= ~TDB_CLEAR_IF_FIRST;
	tdb_reopen(tdb);

	CatchChild();
	sigchld_register(exasmb_cleaner_child_exit, NULL);
	
	/* write the master cleaner pid to /tmp/smbd_dlmcleaner.pid. This is required
	   in order to protect this pid from being killed during stop of smbd service */
	pidfile_create(smbd_cleaner_pid_file);
	
	while (1) {
		/* read the struct proc_exit_status from pipe { */

		left = sizeof(proc_exit_status);
		buf = (char*) &stat;

		do {
			n = read(filedes[0], buf, left);
			if (n == left)
				break;
			if (n == -1) {
				if (errno != EINTR) {
					DEBUG(0, ("exasmb_init_cleanup: read failed %d\n", errno));
					usleep(777);
				}
			} else {
				buf += n;
				left -= n;
			}
		} while ( 1 );
		/* } */

		pid = stat.pid;
		status = stat.status;

		if (WIFSIGNALED(status)) {
			DEBUG(1, ("child: %u exited with signal %d\n",
					pid, WTERMSIG(status)));
		} else if (WIFEXITED(status)) {
			if (WEXITSTATUS(status)) {
				DEBUG(1, ("child: %u exited with error %d\n",
						pid, WEXITSTATUS(status)));
			} else {
				DEBUG(1, ("child: %u exited clean\n", pid));
			}
		} else {
			DEBUG(1, ("pid: %u exited with status %d\n", pid, status));
		}

		if (!need_dlm_cleanup(pid)) {
			DEBUG(1,("pid %d does not need cleanup.\n",pid));
			continue;
		}

		if (is_being_cleaned(pid)) {
			DEBUG(1,("pid %u is already being cleaned.\n",pid));
			continue;
		}

		mark_cleaning_pid(pid,EXASMB_CLEANER_IS_CLEANING);

		while ((child = sys_fork()) == (pid_t)-1) {
			DEBUG(0, ("fork() failed - %m\n"));
			sleep(1);
		}

		if (child != 0)
			continue; /* parent */

		/* child */
		
		DEBUG(1, ("child: cleaning %u\n", pid));
		tdb_reopen(tdb);

		for (i = 0; i < smb_instances; ++i) {
			BOOL fatal=True;
			request.pid = pid;
			g_smbid = smb_instance_id[i];
			while (fatal) {
				cln = exasmb_if_client_open();
				do {
					err = exasmb_cleanup(cln, &request);
					/* sleep ? */
				} while (err == EXASMB_COMM_ERROR);

				/* We need to destroy the client in cases we have OK or FATAL, so do it now. */

				exasmb_if_client_close();             
				switch (err) {
				case EXASMB_FATAL_ERROR:
					DEBUG(0,("received an EXASMB_FATAL_ERROR from dlm. err=%d, pid=%u, instance id: %d, instance index: %d",err,pid,g_smbid,i));
					break;
				case EXASMB_OK:
					DEBUG(2,("Cleanup of pid %u, instance id: %d successful\n",pid,g_smbid));
					fatal=False;
					break;
				default:
					DEBUG(0,("received an unknown error from dlm on exasmb_cleanup. treating it as a fatal error. err=%d, pid=%u, instance id: %d, instance index: %d",err,pid,g_smbid,i));
					break;
				}
			}
		}

		DEBUG(1, ("child: %u cleanup completed\n", pid));
		unregister_dlm_cleanup_pid(pid);
		exit(0);
	}
}

/* Dina: Prevent the lock of two (or more) different fsp's at the same time */
void
exasmb_lock_increase(uint64_t dev, uint64_t inode) {
	
	DEBUG(3, ("Locking name: id: dev %lld inode %lld, current counter: %d\n", dev, inode , last_lock_counter));
	/* trying to lock a new fsp lock resource */

	verify_no_double_lock(dev,inode);

	last_dev = dev;
	last_inode = inode;
	last_lock_counter++;	
	DEBUG(3, ("Locking name: id: dev %lld inode %lld, current counter: %d\n", dev, inode , last_lock_counter));
}

void
exasmb_lock_decrease(uint64_t dev, uint64_t inode)
{

    DEBUG(3, ("UnLocking name: id: dev %lld inode %lld, current counter: %d\n", dev, inode , last_lock_counter));
	if (last_lock_counter > 0) {
		--last_lock_counter;
	} else
		DEBUG(0, ("Trying to unlock not locked resource,  id:  dev %lld inode %lld \n",dev, inode));
}

/* Dina: Increase lock sequence number */
uint64_t
exasmb_get_lock_sequence(){
	static uint64_t lock_sequence_id = 1;

	DEBUG(3, ("return %lld\n", lock_sequence_id + 1));
	return ++lock_sequence_id;
}

void
verify_no_double_lock(uint64_t dev, uint64_t inode)
{    
	DEBUG(3, ("verify_no_double_lock: dev %lld inode %lld , current counter: %d\n", dev, inode , last_lock_counter));
	if ( (last_lock_counter > 0) &&  ((dev != last_dev) || (inode != last_inode) )) {
		/* Already has a locked fsp resource*/
		DEBUG(0,("Trying to lock which is a new resource(dev %lld,inode %lld) when other resource(dev %lld , inode %lld ) is already locked\n", dev,inode,last_dev,last_inode));

				 
		smb_panic("Trying to lock additional lock resource");
	}
}

/* Shlomi: This whole function is based on assumptions about external variables, exports 
   and knowledge of the cluster configuration. It probably needs a full rewrite... */
int32_t
exasmb_try_getsmbid( BOOL* is_successful_out )
{
	DEBUG(3,("exasmb_getsmbid: enter\n"));
	if (g_smbid == -1) {
		char path[PATH_MAX];
		char name[PATH_MAX];
		int len;
		char *s =getenv(env_root_path);

		/* Verify the env variable exists and it's not an empty string... */
		if (s)
			safe_strcpy(name, s, sizeof(name)-1);
		else 
			goto panic_smbid;
		if ((len = strlen(name)) == 0)
			goto panic_smbid;

		if (name[len - 1] != '/')
		{
			safe_strcat(name, "/", sizeof(name)-1);
			len = strlen(name);
		}
		safe_strcat(name, client_addr(), sizeof(name)-1);
		if (strcmp(name+len, "0.0.0.0") == 0) /* Skip env_root_path during comparation! */
		{
			if ( is_successful_out == NULL ) {
				DEBUG(1,("Client disconnected & fd is invalid. Going to EXIT now, this is expected\n"));
				/* goto panic_smbid; */
				exit(1);
			} else {
				*is_successful_out = False;
			}
			return -1;
		}
		int r = readlink(name, path, sizeof(path) -1); /* -1 to leave space for NULL terminator*/
		if (r == -1 || r == 0 || r >= sizeof(path))
		{
			/*	bug22056 - Daddy smbd process cannot find this link for the very first access (it is not exist yet) -
				called from exasmb_is_connected_to_id_mapper() with non-NULL is_successful_out.
				That's ok for Daddy process, don't make a MAJOR to debug log! */
				
			DEBUG( (is_successful_out == NULL) ? 0 : 4, ("readlink(%s) failed, r %d, errno %d path '%s'\n",
				name, r, errno, ((r > 0 && r < sizeof(path)) ? path : "<UNDEFINED>") ));
			goto panic_smbid;
		}
		path[r] = '\0';
		DEBUG(3,("readlink(%s) r %d path '%s'\n",name,r,path));
		s = strrchr(path, '/');
		s = s ? s+1 : path;
		if (!isdigit(*s))
		{
			DEBUG(0,("samba domain ID is not a number, it is %s\n",s));
			goto panic_smbid;
		}
		g_smbid = atoi(s);
		DEBUG(3,("instance %d smbid %d\n",atoi(s),g_smbid));
	}

	if ( is_successful_out != NULL ) {
		*is_successful_out = True;
	}
	return g_smbid;
panic_smbid:
	if ( is_successful_out == NULL ) {
		DEBUG(0,("PANIC: could not find CIFS root path...\n"));
		smb_panic("assert failed");
	} else {
		*is_successful_out = False;
	}
	return -1; /* never happens, just remove a warning */
}

int32_t
exasmb_getsmbid(void)
{
	return exasmb_try_getsmbid( NULL );
;
}

static uint64_t 
get_smb_instance_version(int inst)
{
	/* This should be changed to use the CDB */
	pstring version_file_path;
	BOOL exists, success = False;
	int64_t ver = 0;
	int fd;
	   
	if (smb_instance_version[inst] > 0) 
		return smb_instance_version[inst];

	pstrcpy(version_file_path, getenv(env_vers_base_path_str));
	pstrcat(version_file_path, getenv(env_id_str[inst]));  
   
	exists = (!access(version_file_path, F_OK));

	if ((fd = open(version_file_path, O_RDWR | O_CREAT | O_SYNC, 0666)) < 0) {
		DEBUG(1,("Unable to open smb version file (%s)\n", version_file_path));
		return ~((uint64_t)0);
	}

	if (exists) {
		if (read(fd, &ver, sizeof(int64_t)) != sizeof(int64_t))  {
			struct stat stbuf;
			DEBUG(1,("Unable to read smb version file\n"));
			if (fstat(fd, &stbuf) != 0 || stbuf.st_size != 0) {
				close(fd);
				return ~((uint64_t)0);
			}
			DEBUG(0,("%s is zero length!!!\n", version_file_path));
		}
	}

	ver++;
	if (lseek(fd, 0, SEEK_SET) == 0) {
		if (write(fd, &ver, sizeof(int64_t)) == sizeof(int64_t))
			success = True;
	}

	close(fd);

	if (!success) {
		DEBUG(1,("Error while modifying smb version file\n"));
		return ~((uint64_t)0);
	}
	DEBUG(3,("read vers. %lld successfuly\n", ver));
	return ver;
}

static BOOL
allocate_smb_instances_globals(void)
{
	static BOOL is_alocated = False;
	int i;

	DEBUG(5,("entry\n"));
	if (is_alocated) {
		DEBUG(3,("Already allocated\n"));
		return True;
	}
	smb_instance_id = SMB_MALLOC_ARRAY(int, smb_instances);
	if (smb_instance_id == NULL) {
		DEBUG(0,("Failed to allocate smb_instance_id array, error - %d (%s)\n", errno,strerror(errno)));
		return False;
	}
	smb_instance_version = SMB_CALLOC_ARRAY(int64_t, smb_instances);
	if (smb_instance_version == NULL) {
		DEBUG(0,("Failed to allocate smb_instance_version array, error - %d (%s)\n", errno,strerror(errno)));
		SAFE_FREE(smb_instance_id);
		return False;
	}

	env_id_str = SMB_MALLOC_ARRAY(char *, smb_instances);
	if (env_id_str == NULL) {
		DEBUG(0,("Failed to allocate smb_instance_version array, error - %d (%s)\n", errno,strerror(errno)));
		SAFE_FREE(smb_instance_id);
		SAFE_FREE(smb_instance_version);
		return False;
	}
	for (i=0; i < smb_instances; i++) {
		/* 20 should be enough for int32 */
		int id_str_len = sizeof(exa_smb_id_str) + 20;
		env_id_str[i] = SMB_MALLOC_ARRAY(char, id_str_len);
		if (env_id_str[i] == NULL) {
			DEBUG(0,("Failed to allocate env_id_str[%d], error - %d (%s)\n",i, errno,strerror(errno)));
			int j;
			for (j=0; j<i; j++) {
				SAFE_FREE(env_id_str[j]);
			}
			SAFE_FREE(smb_instance_id);
			SAFE_FREE(smb_instance_version);
			SAFE_FREE(env_id_str);
			return False;
		}
	   memset(env_id_str[i], '\0', id_str_len);
	   snprintf(env_id_str[i], id_str_len, "%s%d", exa_smb_id_str, i);
	}
	is_alocated = True;
	return True;
}

BOOL 
exasmb_if_initiate_connection(void)
{
	pid_t *cleaner_childs;
	uint64_t *versions;

	int i, status, children = 0;
	pid_t child = -1;
	exasmb_message_error err;

	smb_instances = atoi(getenv(env_instances_str));
	if (!smb_instances) {
		DEBUG(0,("Failed to get smb_instances\n"));
		return False;
	}
	if (smb_instances < 0) {
		DEBUG(0,("Invalid smb_instances value - %d\n", smb_instances));
		return False;
	}
	/* allocate smb_instance stuff*/
	if (!allocate_smb_instances_globals()) {
		DEBUG(0,("Failed to allocate globals\n"));
		return False;
	}

	/*allocate cleaner_childs and versions*/
	cleaner_childs = SMB_CALLOC_ARRAY(pid_t, smb_instances);
	if (cleaner_childs == NULL) {
		DEBUG(0,("Failed to allocate cleaner_childs array, error - %d (%s)\n", errno,strerror(errno)));
		return False;
	}
	versions = SMB_MALLOC_ARRAY(uint64_t, smb_instances);
	if (versions == NULL) {
		DEBUG(0,("Failed to allocate versions array, error - %d (%s)\n", errno,strerror(errno)));
		SAFE_FREE(cleaner_childs);
		return False;
	}
	
again_:
	for (i = 0; i < smb_instances; ++i) {
		if (smb_instance_version[i] || cleaner_childs[i])
			continue;
		versions[i] = get_smb_instance_version(i);
		DEBUG(5,("versions[%d] is %lld\n", i, versions[i]));
		if (versions[i] == -1){
			DEBUG(0,("Failed to get smb instance %d version\n", i));
			SAFE_FREE(cleaner_childs);
			SAFE_FREE(versions);
			return False;
		}
		smb_instance_id[i] = atoi(getenv(env_id_str[i]));
		DEBUG(5, ("Got smb_instance_id[%d] for variable %s\n", smb_instance_id[i], env_id_str[i]));
		while ((child = sys_fork()) == (pid_t)-1) {
			DEBUG(0, ("fork() failed - %m\n"));
			sleep(1);
		}
		if (child == 0) {
			exasmb_connect_request request;
			dlm_client client;  
			BOOL rc = False;

			request.conn.vers = versions[i];
			request.conn.inst_id = smb_instance_id[i];
			request.conn.inst_name = "";
			client = exasmb_dlm_client_create(smb_instance_id[i]);
			DEBUG(1,("connecting... %d\n", smb_instance_id[i]));
			if (client) {
				while (!rc) {

					err=exasmb_connect(client,&request);

					switch (err) {
					case EXASMB_COMM_ERROR:
						sleep(SLEEP_TIME_IN_SEC);
						continue;
						break;
					case EXASMB_OK:
						DEBUG(1,("exasmb_connect success for %d\n", smb_instance_id[i]));
						rc=True;
						break;

					case EXASMB_FATAL_ERROR:
					case EXASMB_ERROR:
						DEBUG(0,("exasmb_connect returned err=%d. Instance: %d - Aborting connect attempt.\n",err,smb_instance_id[i]));
						_exit(1);
						break;
					default:
						DEBUG(0,("UNKNOWN ERROR - exasmb_connect returned err=%d. Instance: %d - Aborting connect attempt.\n",err,smb_instance_id[i]));
						_exit(1);
						break;
					}            
				}
				exasmb_dlm_client_destroy(client);
				exit(0);
			} else {
				DEBUG(0,("could not create client for DLM. Aborting this connection attempt.\n"));
				_exit(1);
			}
		} else {
			cleaner_childs[i] = child;
			++children;
		}
	}

	DEBUG(5,("Setting cleaner_childs[%d] with %u\n", i, child ));
	while (children && (child = sys_waitpid((pid_t)-1 ,&status, 0)) > 0) {
		sigchld_dispatch(child, status);
		for(i = 0; i < smb_instances && (cleaner_childs[i] != child); i++)
			;
		if (i == smb_instances) {
			DEBUG(1, ("Child terminated %u with status %d\n", child, status));
			continue;
		}
		--children;
		if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) {
			smb_instance_version[i] = versions[i];
			DEBUG(1, ("domain %d version: %llu\n", smb_instance_id[i], versions[i]));
		} else {
			DEBUG(1,("cleaner chiled exit status %u\n", status));
			cleaner_childs[i] = 0;
			goto again_;
		}
	}
	if (children) {
		char *msg;
		asprintf(&msg, "sys_waitpid failed: children %d - %m\n", children);
		smb_panic(msg);
	}
	DEBUG(1, ("g_smbid: %u\n", g_smbid));

	SAFE_FREE(cleaner_childs);
	SAFE_FREE(versions);
	return True;
}

dlm_client 
exasmb_if_client_open(void)
{
	if (g_smbid == -1)
		exasmb_getsmbid();
	if (h_client == NULL) {
		h_client = exasmb_dlm_client_create(g_smbid);
	}
	
	return h_client;
}

void 
exasmb_if_client_close(void)
{
	if (h_client)
		exasmb_dlm_client_destroy(h_client);  

	h_client = NULL;
}
