#include "includes.h"

/*
note: there can be several callback registered with the same name. The last one added will be invoked,
allowing to treat this kind of like a stack by temporary overriding callbacks.
*/

static struct gp_hooks_cb_fns
{
	struct gp_hooks_cb_fns *next, *prev;

	char* name;
	int (*fn)(void* private_data, const char* name, unsigned int argc, const char** argv, char* out_buf, unsigned int out_buf_size);
	void *private_data;
} *gp_hooks_cb_fns;

static struct gp_hooks
{
	struct gp_hooks *next, *prev;

	char* param;
	void (*fn)(
		const char* param,
		const char* name,
		unsigned int argc, const char** argv,
		int (*cb)(const char* name, unsigned int argc, const char** argv, char* out_buf, unsigned int out_buf_size)
	);
	void* so_handle;
} *gp_hooks;

void* gp_hooks_register_cb(
	const char* name,
	int (*fn)(void* private_data, const char* name, unsigned int argc, const char** argv, char* out_buf, unsigned int out_buf_size),
	void *private_data
	)
{
	struct gp_hooks_cb_fns *pfn;

	pfn = SMB_MALLOC_P(struct gp_hooks_cb_fns);

	if(!pfn) {
		DEBUG(0,("gp_hooks_register_cb: Not enough memory. malloc failed!\n"));
		return NULL;
	}

	ZERO_STRUCTPN(pfn);
	pfn->name = SMB_STRDUP(name);
	/* note: SMB_STRDUP panics in case of allocation failure */
	pfn->fn = fn;
	pfn->private_data = private_data;

	DLIST_ADD(gp_hooks_cb_fns, pfn);

	return pfn;
}

void gp_hooks_deregister_cb(void* hook)
{
	struct gp_hooks_cb_fns *fn = (struct gp_hooks_cb_fns *)hook;

	DLIST_REMOVE(gp_hooks_cb_fns, fn);
	SAFE_FREE(fn->name);
	SAFE_FREE(fn);
}

void gp_hooks_deregister_cb_by_name(const char* name)
{
	struct gp_hooks_cb_fns *fn, *next;

	for(fn = gp_hooks_cb_fns; fn; fn = next) {
		next = fn->next;
		if (strcmp(fn->name, name) == 0) {
			gp_hooks_deregister_cb(fn);
			return;
		}
	}
}

void gfree_gp_hooks(void)
{
	{
		struct gp_hooks_cb_fns *fn, *next;

		/* delete the gp_hooks_cb_fns list */
		fn = gp_hooks_cb_fns;
		while(fn) {
			next = fn->next;
			gp_hooks_deregister_cb(fn);
			fn = next;
		}
	}
	{
		struct gp_hooks *hook, *next;

		/* delete the gp_hooks list */
		hook = gp_hooks;
		while(hook) {
			next = hook->next;
			DLIST_REMOVE(gp_hooks, hook);
			sys_dlclose(hook->so_handle);
			SAFE_FREE(hook->param);
			SAFE_FREE(hook);
			hook = next;
		}
	}
}

static int gp_hooks_invoke_cb(const char* name, unsigned int argc, const char** argv, char* out_buf, unsigned int out_buf_size)
{
	struct gp_hooks_cb_fns *fn, *next;

	for(fn = gp_hooks_cb_fns; fn; fn = next) {
		next = fn->next;
		if (strcmp(fn->name, name) == 0) {
			return fn->fn(fn->private_data, name, argc, argv, out_buf, out_buf_size);
		}
	}

	return -1;
}

static char* extract_str(const char** p, char delimiter)
{
	const char* pos = strchr(*p, delimiter);
	const char* p0 = *p;

	if(pos) {
		*p = pos + 1;
	} else {
		pos = *p + strlen(*p);
		*p = pos;
	}

	return SMB_STRNDUP(p0, pos - p0);
}

static BOOL gp_hooks_init_hook(const char* descriptor)
{
	const char* p = descriptor;
	char* so_name;
	char* func_name;
	struct gp_hooks* hook;
	const char* error;

	hook = SMB_MALLOC_P(struct gp_hooks);

	if(!hook) {
		DEBUG(0,("gp_hooks_init_hook: Not enough memory. malloc failed!\n"));
		return False;
	}

	ZERO_STRUCTPN(hook);

	so_name = extract_str(&p, ':');
	func_name = extract_str(&p, ':');
	hook->param = SMB_STRDUP(p);

	hook->so_handle = sys_dlopen(so_name, RTLD_LAZY);
	error = sys_dlerror();
	if(!hook->so_handle) {
		DEBUG(0, ("gp_hooks_init_hook: sys_dlopen failed for %s: %s\n", so_name, error ? error : "?"));

		SAFE_FREE(so_name);
		SAFE_FREE(func_name);
		SAFE_FREE(hook->param);

		return False;
	}

	hook->fn = sys_dlsym(hook->so_handle, func_name);
	error = sys_dlerror();
	if(error) {
		DEBUG(0, ("gp_hooks_init_hook: Error trying to resolve symbol '%s' in %s: %s\n", func_name, so_name, error));

		sys_dlclose(hook->so_handle);

		SAFE_FREE(so_name);
		SAFE_FREE(func_name);
		SAFE_FREE(hook->param);

		return False;
	}

	SAFE_FREE(so_name);
	SAFE_FREE(func_name);

	DLIST_ADD(gp_hooks, hook);

	return True;
}

BOOL gp_hooks_init(void)
{
	const char **gp_hooks, **hook;

	gp_hooks = lp_gp_hooks();

	if(!gp_hooks || !gp_hooks[0])
		return True;

	for(hook = gp_hooks; *hook; hook++) {
		if(!gp_hooks_init_hook(*hook)) {
			DEBUG(0, ("gp_hooks_init: gp_hooks_init_hook failed for %s\n", *hook));
			return False;
		}
	}

	return True;
}

void gp_hooks_invoke(const char* name, unsigned int argc, const char** argv)
{
	struct gp_hooks *hook, *next;

	for(hook = gp_hooks; hook; hook = next) {
		next = hook->next;

		hook->fn(hook->param, name, argc, argv, gp_hooks_invoke_cb);
	}
}

struct gp_hooks_cb_get_str
{
	void* hook;
	char* str;
	int len;
};

static int gp_hooks_cb_get_str_destructor(struct gp_hooks_cb_get_str* p)
{
	if(p->hook)
		gp_hooks_deregister_cb(p->hook);
	SAFE_FREE(p->str);
	return 0;
}

static int gp_hooks_cb_get_str(void* private_data, const char* name, unsigned int argc, const char** argv, char* out_buf, unsigned int out_buf_size)
{
	struct gp_hooks_cb_get_str* ctx = (struct gp_hooks_cb_get_str*)private_data;

	if(out_buf_size <= ctx->len) {
		memcpy(out_buf, ctx->str, out_buf_size);
	} else {
		memcpy(out_buf, ctx->str, ctx->len);
		out_buf[ctx->len] = '\0';
	}

	return ctx->len;
}

void* gp_hooks_register_get_str_cb(TALLOC_CTX *mem_ctx, const char *name, const char *fmt, ...) PRINTF_ATTRIBUTE(3,4);
void* gp_hooks_register_get_str_cb(TALLOC_CTX *mem_ctx, const char *name, const char *fmt, ...)
{
	va_list ap;
	struct gp_hooks_cb_get_str* ctx;

	ctx = TALLOC_ZERO_P(mem_ctx, struct gp_hooks_cb_get_str);
	if(ctx == NULL) {
		DEBUG(0, ("gp_hooks_register_get_str_cb: talloc failed\n"));
		return NULL;
	}

	va_start(ap, fmt);
	ctx->len = vasprintf(&ctx->str, fmt, ap);
	va_end(ap);

	if(ctx->len == -1) {
		TALLOC_FREE(ctx);
		DEBUG(0, ("gp_hooks_register_get_str_cb: vasprintf failed\n"));
		return NULL;
	}

	talloc_set_destructor(ctx, gp_hooks_cb_get_str_destructor);

	ctx->hook = gp_hooks_register_cb(name, gp_hooks_cb_get_str, ctx);
	if(!ctx->hook) {
		TALLOC_FREE(ctx);
		return NULL;
	}

	return ctx;
}
