#include "includes.h"

#include <poll.h>

#include "messages_inter.h"

static int proxy_to_use = 0;
static unsigned int udp_seq;
static int client_socket;
in_addr_t* node_addrs;
int this_node;
int node_count;
int messaging_port;

int message_this_node(void)
{
	return this_node;
}

void message_pick_next_proxy(void)
{
	proxy_to_use++;
}

static BOOL find_iface(in_addr_t addr, struct iface_struct* ifaces, int count)
{
	int i;

	for(i = 0; i < count; i++)
		if(ifaces[i].ip.s_addr == addr)
			return True;

	return False;
}

BOOL message_init_inter(void)
{
	struct iface_struct ifaces[MAX_INTERFACES];
	int iface_count;

	{
		struct servent *se = getservbyname(UDP_SERV_NAME, "udp");

		if(!se) {
			DEBUG(0, ("ERROR: could not find entry for '" UDP_SERV_NAME "' service\n"));
			return False;
		}

		messaging_port = ntohs(se->s_port);
	}

	iface_count = get_interfaces(ifaces, MAX_INTERFACES);

	{
		in_addr_t na[MAX_NODES];
		char *p0, *p1;

		p0 = lp_messaging_nodes();
		p1 = strchr(p0, '*');

		if(!p1) {
			DEBUG(0, ("ERROR: 'messaging nodes' should contain '*' to specify substitution position\n"));
			return False;
		}

		node_count = 0;

		while(node_count < MAX_NODES) {
			fstring name;
			in_addr_t addr;

			snprintf(name, sizeof(name), "%.*s%d%s", p1 - p0, p0, node_count, p1 + 1);
			addr = interpret_addr(name);

			if(addr == 0)
				break;

			if(find_iface(addr, ifaces, iface_count))
				this_node = node_count;

			na[node_count++] = addr;
		}

		if(node_count == 0) {
			DEBUG(0, ("ERROR: no hosts matching the '%s' pattern found\n", lp_messaging_nodes()));
			return False;
		}

		node_addrs = SMB_CALLOC_ARRAY(in_addr_t, node_count);
		if(!node_addrs) {
			DEBUG(0, ("ERROR: SMB_CALLOC_ARRAY failed\n"));
			return False;
		}

		memcpy(node_addrs, na, sizeof(in_addr_t) * node_count);
	}

	if(this_node == -1) {
		DEBUG(0, ("ERROR: could not find this node among nodes defined in 'messaging nodes'\n"));
		return False;
	}

	DEBUG(3, ("this is node number %d, number of nodes %d\n", this_node, node_count));

	return True;

	/* !!! do we need any clean-up for partial initialization? */
}

void message_done_inter(void)
{
	SAFE_FREE(node_addrs);
}

static BOOL message_send_udp_to_node(int node, int port_offset, const void *buf, size_t size)
{
	ssize_t sent;
	struct sockaddr_in addr;

#ifdef UDP_DROP_PROB
	if(rand() % 1000 < 10 * UDP_DROP_PROB)
		return True;
#endif

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = node_addrs[node];
	addr.sin_port = htons(messaging_port + port_offset % lp_messaging_servers());

	sent = sys_sendto(client_socket, buf, size, 0, (struct sockaddr *)&addr, sizeof(addr));
	if(sent != size) {
		DEBUG(1, ("ERROR: message_send_udp_to_node: sendto to %s:%d failed: %s\n", ADDR_IPPORT(&addr), strerror(errno)));
		return False;
	}

	return True;
}

static bool message_maybe_init_child(void)
{
	static pid_t prev_self_pid = 0;
	struct timeval current_time;

	/* note: sys_getpid is caching, thus cheap to call multiple times */
	if(sys_getpid() == prev_self_pid)
		return True;

	prev_self_pid = sys_getpid();

	GetTimeOfDay(&current_time);

	/* note: possible overflow is ok as we just need some more or less
	   unique starting value which would be better than 0 */
	udp_seq = 1000000 * current_time.tv_sec + current_time.tv_usec;

	client_socket = socket(PF_INET, SOCK_DGRAM, 0);
	if(client_socket == -1)	{
		DEBUG(0, ("ERROR: failed to create UDP socket: %s\n", strerror(errno)));
		return False;
	}

	set_close_on_fork(client_socket, 1);

	return True;
}

static void adjust_resend_time(int attempt, struct timeval* resend_time)
{
	unsigned int t = UDP_RETRY_INTERVAL;
	int i = attempt - 1;

	while(i > 0 && t < UDP_RETRY_INTERVAL_LIMIT)
	{
		t *= 2;
		i--;
	}

	if(t > UDP_RETRY_INTERVAL_LIMIT)
		t = UDP_RETRY_INTERVAL_LIMIT;

	resend_time->tv_usec += t * 1000;

	resend_time->tv_sec += resend_time->tv_usec / 1000000L;
	resend_time->tv_usec %= 1000000L;
}

NTSTATUS message_send_pid_internal_inter(struct process_id pid, int msg_type,
						const void *buf, size_t len,
						BOOL duplicates_allowed,
						unsigned int timeout)
{
	char packet_buf[UDP_BUF_SIZE];
	struct udp_request_header * udp_header;
	struct send_message_request * request_header;
	ssize_t to_send;
	struct timeval current_time, end_time, resend_time;
	int port_offset;
	int do_send;
	int last_send;
	int attempt = 0;

	if(!message_maybe_init_child())
		return NT_STATUS_UNSUCCESSFUL;

	if(!lp_messaging_servers()) {
		DEBUG(1, ("ERROR: message_send_pid to remote node %d, while inter-node messaging is not enabled\n", pid.node));
		return NT_STATUS_INVALID_HANDLE;
	}

	if(pid.node < 0 || pid.node >= node_count) {
		DEBUG(1, ("ERROR: message_send_pid to non-existant node %d (there are only %d nodes)\n", pid.node, node_count));
		return NT_STATUS_INVALID_HANDLE;
	}

	to_send = sizeof(struct udp_request_header) + sizeof(struct send_message_request) + len;

	if(to_send > UDP_BUF_SIZE) {
		DEBUG(0, ("ERROR: message_send_pid: message size (%d) exceeds limit set for inter-node messaging\n", len));
		return NT_STATUS_NO_MEMORY;
	}

	udp_header = (struct udp_request_header *)packet_buf;
	request_header = (struct send_message_request *)(packet_buf + sizeof(struct udp_request_header));

	udp_header->type = UDP_MSG_SEND_MESSAGE;
	udp_header->seq = udp_seq++;
	udp_header->source_pid = sys_getpid();

	DEBUG(10, ("INFO: message_send_pid_internal_inter: from pid %d to node %d pid %d seq %u\n", udp_header->source_pid, pid.node, pid.pid, udp_header->seq));

	request_header->src_pid = procid_self();
	request_header->dst_pid = pid;
	request_header->msg_type = msg_type;
	request_header->duplicates_allowed = duplicates_allowed;
	request_header->timeout = timeout;
	request_header->len = len;

	if(len > 0)
		memcpy(packet_buf + sizeof(struct udp_request_header) + sizeof(struct send_message_request), buf, len);

	port_offset = proxy_to_use;

	GetTimeOfDay(&current_time);
	do_send = 1;
	last_send = 0;

	if(timeout) {
		end_time.tv_sec = current_time.tv_sec + timeout;
		end_time.tv_usec = current_time.tv_usec;
	}

	resend_time = current_time;

	while(1) {
		SMB_BIG_INT poll_timeout;
		int ret;
		short int revents;

		if(do_send) {
			/* note: any send failure is ignored (the function itself will do the logging).
			   we pretend that we did send and will do retry when we fail to get a reply.
			*/
			message_send_udp_to_node(pid.node, port_offset, packet_buf, to_send);
			attempt++;

			GetTimeOfDay(&current_time);

			adjust_resend_time(attempt, &resend_time);

			if(timeout && usec_time_diff(&resend_time, &end_time) >= 0)
				last_send = 1;

			do_send = 0;
		}

		poll_timeout = usec_time_diff(last_send ? &end_time : &resend_time, &current_time);

		if(poll_timeout > 0) {
			struct pollfd fds[1];

			fds[0].fd = client_socket;
			fds[0].events = POLLIN;
			ret = poll(fds, 1, poll_timeout / 1000);
			revents = fds[0].revents;
		} else {
			ret = 0;
		}

		if(ret == -1) {
			if(errno != EINTR) {
				DEBUG(1, ("ERROR: message_send_pid: poll failed: %s\n", strerror(errno)));
				return NT_STATUS_UNSUCCESSFUL;
			}
		} else if(ret == 0) {
			if(last_send) {
				DEBUG(2, ("message_send_pid: timed out waiting for peer reply (attempt = %d)\n", attempt));
				return NT_STATUS_IO_TIMEOUT;
			} else {
				do_send = 1;
			}
		} else if(ret == 1 && revents & POLLIN) {
			struct {
				struct udp_reply_header msg;
				/* the purpose of dummy is to detect UDP packets longer than struct udp_reply_header */
				int dummy;
			} in_packet;
			ssize_t got;
			struct sockaddr_in from;
			ssize_t from_size = sizeof(from);

			got = sys_recvfrom(client_socket, &in_packet, sizeof(in_packet), 0, (struct sockaddr *)&from, &from_size);
			if(got == -1) {
				/* we treat error in recv the same way as bad packet, as this might be just transient condition */
				DEBUG(1, ("ERROR: message_send_pid_internal_inter: recvfrom failed: %s\n", strerror(errno)));
				goto packet_done;
			}

			if(got != sizeof(in_packet.msg)) {
				DEBUG(0, ("message_send_pid_internal_inter: dropping packet from %s:%d having wrong size\n", ADDR_IPPORT(&from)));
				goto packet_done;
			}

			if(from.sin_addr.s_addr != node_addrs[pid.node] || in_packet.msg.seq != udp_header->seq) {
				DEBUG(3, ("message_send_pid_internal_inter: dropping stray packet from %s:%d seq %u\n",
					ADDR_IPPORT(&from), in_packet.msg.seq));
				goto packet_done;
			}

			DEBUG(10, ("INFO: message_send_pid_internal_inter: got reply to seq %u => 0x%08x\n",
				udp_header->seq, NT_STATUS_V(in_packet.msg.result)));

			return in_packet.msg.result;
		packet_done: ;
		} else {
			DEBUG(0, ("ERROR: message_send_pid_internal_inter: unexpected poll result: ret = %d, revents = %d\n", ret, revents));
		}

		if(!do_send)
			GetTimeOfDay(&current_time);

		if(do_send)
			DEBUG(10, ("INFO: message_send_pid_internal_inter: re-sending packet to node %d seq %u (attempt = %d)\n",
				pid.node, udp_header->seq, attempt));

	}
}
