#include "includes.h"

#include <poll.h>

#include "messages_inter.h"

BOOL in_message_server = False;
extern in_addr_t* node_addrs;
extern int this_node;
extern int messaging_port;

struct message_server_udp_ctx {
	int sock;
	struct sockaddr_in addr;
	socklen_t addr_size;
};

static BOOL message_server_udp_reply_raw(struct message_server_udp_ctx * ctx, void * buf, ssize_t size)
{
	ssize_t sent;

#ifdef UDP_DROP_PROB
	if(rand() % 1000 < 10 * UDP_DROP_PROB)
		return True;
#endif

	/* note: EAGAIN is not treated in any special way. If it happens, which indicates that there is
	   something like buffer exhaustion, we just rely on client resending the request after timeout.
	*/

	sent = sys_sendto(ctx->sock, buf, size, 0, (struct sockaddr *)&ctx->addr, ctx->addr_size);
	if(sent != size) {
		DEBUG(1, ("ERROR: sendto to %s:%d failed: %s\n", MESSAGE_SERVER_UDP_CTX_IPPORT(ctx),
			strerror(errno)));
		return False;
	}

	return True;
}

static TDB_DATA message_server_make_udp_reply(unsigned int seq, NTSTATUS result, char* extra, unsigned int extra_len)
{
	struct udp_reply_header rep;
	TDB_DATA r;

	rep.seq = seq;
	rep.result = result;
	rep.extra = extra_len;

	r.dsize = sizeof(rep) + extra_len;
	r.dptr = malloc(r.dsize);

	memcpy(r.dptr, &rep, sizeof(rep));
	if(extra_len)
		memcpy(r.dptr + sizeof(rep), extra, extra_len);

	return r;
}

static TDB_DATA message_server_process_send_msg(struct message_server_udp_ctx * ctx, unsigned int seq, ssize_t size,
					    struct send_message_request * msg)
{
	NTSTATUS result;

	if(size < sizeof(struct send_message_request)) {
		DEBUG(1, ("INFO: dropping malformed 'send msg' from %s:%d: too short, only %d bytes\n",
			MESSAGE_SERVER_UDP_CTX_IPPORT(ctx), size));
		return make_tdb_data((char *)0, 0);
	}

	if(size != sizeof(struct send_message_request) + msg->len) {
		DEBUG(1, ("INFO: dropping malformed 'send msg' from %s:%d: wanted %d bytes, got %d bytes\n",
			MESSAGE_SERVER_UDP_CTX_IPPORT(ctx), sizeof(struct send_message_request) + msg->len, size));
		return make_tdb_data((char *)0, 0);
	}

	if(msg->dst_pid.node != this_node) {
		DEBUG(1, ("INFO: dropping stray 'send msg' from %s:%d: to node %d, but this is node %d\n",
			MESSAGE_SERVER_UDP_CTX_IPPORT(ctx), msg->dst_pid.node, this_node));
		return make_tdb_data((char *)0, 0);
	}

	result = message_send_pid_internal(
		msg->src_pid,
		msg->dst_pid,
		msg->msg_type,
		msg + 1, msg->len,
		msg->duplicates_allowed,
		msg->timeout);

	DEBUG(10, ("INFO: remote message from pid %d node %d to pid %d seq %u timeout %d => 0x%08x\n",
		msg->src_pid.pid, msg->src_pid.node, msg->dst_pid.pid, seq, msg->timeout, NT_STATUS_V(result)));

	return message_server_make_udp_reply(seq, result, NULL, 0);
}

static void message_server_child(int sock, int pipe_fd)
{
	struct seen {
		in_addr_t source_addr;
		pid_t source_pid;
		unsigned int seq;
	} seen[MESSAGE_HIST_SIZE] = {{0}};
	int seen_i = 0;
	TDB_CONTEXT * seen_tdb;

	seen_tdb = tdb_open_log(NULL, 0, TDB_INTERNAL, O_RDWR|O_CREAT, 0644);
	if(!seen_tdb) {
		DEBUG(0, ("FATAL: message_server_child: failed to create internal tdb\n"));
		_exit(1);
	}

	while(1) {
		struct message_server_udp_ctx ctx;
		char buf[UDP_BUF_SIZE];
		ssize_t got;
		struct udp_request_header * msg = (struct udp_request_header *)buf;

		ctx.sock = sock;
		ctx.addr_size = sizeof(ctx.addr);

		while(1) {
			struct pollfd fds[2];

			fds[0].fd = sock;
			fds[0].events = POLLIN;

			fds[1].fd = pipe_fd;
			fds[1].events = POLLIN;

			if(poll(fds, 2, -1) == -1) {
				if(errno != EINTR) {
					DEBUG(1, ("ERROR: message_server_child: poll failed: %s\n", strerror(errno)));

					_exit(1);
				} else {
					continue;
				}
			}

			if(fds[1].revents & POLLHUP) {
				DEBUG(2, ("message_server_child: parent has terminated, so will we\n"));

				_exit(1);
			}

			if(fds[0].revents & POLLIN) {
				break;
			}

			DEBUG(0, ("message_server_child: unexpected poll result.\n"));
		}

		got = sys_recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&ctx.addr, &ctx.addr_size);

		if(got == -1) {
			DEBUG(1, ("ERROR: message_server_child: recvfrom failed: %s\n", strerror(errno)));

			continue;
		}

		if(got < sizeof(struct udp_request_header)) {
			DEBUG(1, ("INFO: message_server_child: dropping malformed message from %s:%d: too short, only %d bytes\n",
				MESSAGE_SERVER_UDP_CTX_IPPORT(&ctx), got));
			continue;
		}

		{
			struct seen rec = {ctx.addr.sin_addr.s_addr, msg->source_pid, msg->seq};
			TDB_DATA rep;

			if(tdb_exists(seen_tdb, make_tdb_data((char *)&rec, sizeof(rec)))) {
				DEBUG(2, ("INFO: message_server_child: duplicate message from %s:%d pid %d seq %u\n",
					MESSAGE_SERVER_UDP_CTX_IPPORT(&ctx), msg->source_pid, msg->seq));

				rep = tdb_fetch(seen_tdb, make_tdb_data((char *)&rec, sizeof(rec)));
			} else {
				switch(((struct udp_request_header *)buf)->type) {
					case UDP_MSG_SEND_MESSAGE:
						rep = message_server_process_send_msg(
							&ctx,
							msg->seq,
							got - sizeof(struct udp_request_header),
							(struct send_message_request *)(buf + sizeof(struct udp_request_header)));
						break;
					default:
						DEBUG(0, ("INFO: message_server_child: dropping message from %s:%d: unknown type %d\n",
							MESSAGE_SERVER_UDP_CTX_IPPORT(&ctx), msg->type));

						rep = make_tdb_data((char *)0, 0);
				}

				/* if slot is used, source_pid != 0*/
				if(seen[seen_i].source_pid) {
					if(tdb_delete(seen_tdb, make_tdb_data((char *)(seen + seen_i), sizeof(struct seen)))) {
						DEBUG(0, ("FATAL: message_server_child: failed to delete data from internal tdb\n"));
						_exit(1);
					}
				}

				if(tdb_store(seen_tdb, make_tdb_data((char *)&rec, sizeof(rec)), rep, TDB_INSERT)) {
					DEBUG(0, ("FATAL: message_server_child: failed to add data to internal tdb\n"));
					_exit(1);
				}

				seen[seen_i] = rec;
				seen_i = (seen_i + 1) % MESSAGE_HIST_SIZE;
			}

			if(rep.dsize) {
				message_server_udp_reply_raw(&ctx, rep.dptr, rep.dsize);
				SAFE_FREE(rep.dptr);
			}
		}
	}

	/* unreachable */
	/* tdb_close(seen_tdb); */
}

static int server_pipe_fd;
typedef struct {
	pid_t pid;
	int start_count;
	struct timeval start_time;
} server_rec;

static server_rec* message_servers;


static BOOL message_server_init(int n)
{
	pid_t pid;
	struct timeval now;

	GetTimeOfDay(&now);
	if(message_servers[n].start_count && usec_time_diff(&now, &message_servers[n].start_time) > SERVER_STARTED_THRESHOLD * 1000000ll) {
		DEBUG(10, ("Previous instance of inter-node messaging server no %d has worked sufficiently long."
			   " Resetting restart counter.\n", n));
		message_servers[n].start_count = 0;
	}

	if(message_servers[n].start_count > SERVER_START_LIMIT) {
		DEBUG(0, ("ERROR: Number of restarts for inter-node messaging server no %d has been exceeded\n", n));

		return False;
	}

	if((pid = sys_fork()) == (pid_t)-1) {
		DEBUG(0, ("ERROR: Failed to fork inter-node messaging server: %s\n", strerror(errno)));

		return False;
	}

	if(pid) {
		/* parent */

		DEBUG(3, ("forked a message proxy server no %d: pid=%d\n", n, pid));

		message_servers[n].pid = pid;
		message_servers[n].start_count++;
		message_servers[n].start_time = now;
	} else {
		/* child */
		int sock = -1;

		if(message_servers[n].start_count) {
			/* the child has an un-incremented value, as ++ takes place after fork */

			/* we sleep for min(2 ** (restarts - 1), SERVER_START_DELAY_LIMIT) seconds */
			int delay = 1 << (message_servers[n].start_count - 1);
			if(delay > SERVER_START_DELAY_LIMIT)
				delay = SERVER_START_DELAY_LIMIT;

			/* the sleep is intentionally before the DEBUG(), as we want it to be the first thing after fork
			   to minimize chances of possibly crashing again too soon */
			sleep(delay);

			DEBUG(3, ("inter-node messaging server no %d restart has been intentionally delayed for %d s\n",
				n, delay));
		}

		sock = open_socket_in(SOCK_DGRAM, messaging_port + n, /* dbglevel*/ 0, node_addrs[this_node], /* reuse addr */ True);

		if(sock == -1) {
			DEBUG(0, ("ERROR: Failed to get UDP socket (port %d) for inter-node messaging server: %s\n", 
				messaging_port + n, strerror(errno)));

			_exit(1);
		}

		if(message_tdb_reopen() != 0)
		{
			DEBUG(0, ("ERROR: tdb_reopen failed in inter-node messaging server child: %s\n", strerror(errno)));

			_exit(1);
		}

		in_message_server = True;

		message_server_child(sock, server_pipe_fd);

		_exit(0);
	}

	return True;
}

static void message_server_sigchld(pid_t pid, int status, void* private_data)
{
	int i;

	for(i = 0; i < lp_messaging_servers(); i++)
		if(message_servers[i].pid == pid)
			break;

	if(i == lp_messaging_servers())
		return;

	DEBUG(0, ("ERROR: inter-node messaging server no %d child pid %d died unexpectedly with status code %d\n", i, pid, status));

	if(!message_server_init(i))
		exit(1);
}

BOOL message_init_inter_server(void)
{
	int i;
	int pipe_fd[2];

	message_servers = SMB_CALLOC_ARRAY(server_rec, lp_messaging_servers());
	if(!message_servers) {
		DEBUG(0, ("ERROR: SMB_CALLOC_ARRAY failed\n"));
		return False;
	}

	if(pipe(pipe_fd)) {
		DEBUG(0, ("ERROR: pipe failed for inter-node messaging server: %s\n",  strerror(errno)));

		return False;
	}

	server_pipe_fd = pipe_fd[0];

	sigchld_register(message_server_sigchld, NULL);

	set_close_on_fork(pipe_fd[1], 1);

	for(i = 0; i < lp_messaging_servers(); i++)
		if(!message_server_init(i))
			return False;

	return True;
}
