/* 
 *  Unix SMB/Netbios implementation.
 *  SEC_DESC handling functions
 *  Copyright (C) Jeremy R. Allison            1995-2003.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "includes.h"

/*******************************************************************
 Create the share security tdb.
 ********************************************************************/

#define SHARE_DATABASE_VERSION_V1 1
#define SHARE_DATABASE_VERSION_V2 2 /* version id in little endian. */

/* Map generic permissions to file object specific permissions */

static struct generic_mapping file_generic_mapping = {
        FILE_GENERIC_READ,
        FILE_GENERIC_WRITE,
        FILE_GENERIC_EXECUTE,
        FILE_GENERIC_ALL
};

BOOL open_share_info_tdb(const char* file_name, TDB_CONTEXT** ptdb)
{
	const char *vstring = "INFO/version";
	int32 vers_id;
	TDB_CONTEXT* ret_tdb = NULL;

	if (file_name == NULL) {
		DEBUG(0,("Invalid intput parameter: file_name = NULL.\n"));
		return False;
	}

	if (ptdb == NULL) {
		DEBUG(0,("Invalid output parameter: ptdb = NULL.\n"));
		return False;
	}

	become_root();
	ret_tdb = tdb_open_log(file_name,0,TDB_DEFAULT,O_RDWR|O_CREAT,0600);
	unbecome_root();
	if (!ret_tdb) {
		DEBUG(0,("Failed to open share info database %s. errno = %d(%s)\n",
			file_name, errno, strerror(errno) ));
		return False;
	}
 
	/* handle a Samba upgrade */
	tdb_lock_bystring(ret_tdb, vstring);

	/* Cope with byte-reversed older versions of the db. */
	vers_id = tdb_fetch_int32(ret_tdb, vstring);
	if ((vers_id == SHARE_DATABASE_VERSION_V1) || (IREV(vers_id) == SHARE_DATABASE_VERSION_V1)) {
		/* Written on a bigendian machine with old fetch_int code. Save as le. */
		tdb_store_int32(ret_tdb, vstring, SHARE_DATABASE_VERSION_V2);
		vers_id = SHARE_DATABASE_VERSION_V2;
	}

	if (vers_id != SHARE_DATABASE_VERSION_V2) {
		tdb_traverse(ret_tdb, tdb_traverse_delete_fn, NULL);
		tdb_store_int32(ret_tdb, vstring, SHARE_DATABASE_VERSION_V2);
	}
	tdb_unlock_bystring(ret_tdb, vstring);

	DEBUG(10,("Successfully opened share info database \"%s\".\n", file_name));
	*ptdb = ret_tdb;
	return True;
}

BOOL close_share_info_tdb(TDB_CONTEXT* tdb_to_close)
{
	int ret = 0;
	pstring name;

	if (DEBUGLEVEL >= 10) {
		pstrcpy(name, tdb_name(tdb_to_close));
	} else {
		pstrcpy(name, "<N/A>");
	}

	if (tdb_to_close == NULL) {
		DEBUG(0,("Invalid intput parameter: tdb_to_close = NULL.\n"));
		return False;
	}

	ret = tdb_close(tdb_to_close);
	if (ret != 0) {
		DEBUG(0,("Failed to close share info database \"%s\", errno = %d (%s)\n",
			name, errno, strerror(errno) ));
		return False;
	}

	DEBUG(10,("Successfully closed share info database \"%s\".\n", name));
	return True;
}

/*******************************************************************
 Fake up a Everyone, default access as a default.
 def_access is a GENERIC_XXX access mode.
 ********************************************************************/

SEC_DESC *get_share_security_default( TALLOC_CTX *ctx, size_t *psize, uint32 def_access)
{
	SEC_ACCESS sa;
	SEC_ACE ace;
	SEC_ACL *psa = NULL;
	SEC_DESC *psd = NULL;
	uint32 spec_access = def_access;

	se_map_generic(&spec_access, &file_generic_mapping);

	init_sec_access(&sa, def_access | spec_access );
	init_sec_ace(&ace, &global_sid_World, SEC_ACE_TYPE_ACCESS_ALLOWED, sa, 0);

	if ((psa = make_sec_acl(ctx, NT4_ACL_REVISION, 1, &ace)) != NULL) {
		psd = make_sec_desc(ctx, SEC_DESC_REVISION, SEC_DESC_SELF_RELATIVE, NULL, NULL, NULL, psa, psize);
	}

	if (!psd) {
		DEBUG(0,("Failed to make default SEC_DESC.\n"));
		return NULL;
	}

	return psd;
}

/*******************************************************************
 Pull a security descriptor from the share tdb.
 ********************************************************************/

static SEC_DESC* fetch_share_security_from_tdb( TALLOC_CTX *ctx,
																TDB_CONTEXT* tdb_ctx,
																const char *share_name,
																const char *tdb_file_path,
																size_t* psize )
{
	SEC_DESC* psd = NULL;
	prs_struct ps;
	fstring key;

	SMB_ASSERT( ctx != NULL );
	SMB_ASSERT( tdb_ctx != NULL );
	SMB_ASSERT( share_name != NULL );
	SMB_ASSERT( tdb_file_path != NULL );
	SMB_ASSERT( psize != NULL );

	*psize = 0;

	/* Fetch security descriptor from tdb */
	slprintf(key, sizeof(key)-1, "SECDESC/%s", share_name);

	if (tdb_prs_fetch(tdb_ctx, key, &ps, ctx)==0) {
		if (sec_io_desc("get_share_security_for_tdb", &psd, &ps, 1)) {
			*psize = sec_desc_size(psd);	
		} else {
			DEBUG(0,("Failed to parse secdesc from \"%s\" for share \"%s\".\n", tdb_file_path, share_name));
		}
		prs_mem_free(&ps);
	} else {
		DEBUG(10,("Failed to fetch secdesc from \"%s\" for share \"%s\".\n", tdb_file_path, share_name));
	}

	return psd;
}

SEC_DESC *get_share_security_for_tdb( TALLOC_CTX *ctx, const char *tdb_file_path,
				  const char *share_name, size_t *psize)
{
	TDB_CONTEXT* tdb_ctx = NULL;
	SEC_DESC *psd = NULL;

	*psize = 0;

	if (!open_share_info_tdb(tdb_file_path, &tdb_ctx)) {
		DEBUG(0,("Failed to open share info tdb \"%s\".\n", tdb_file_path));
		goto out;
	}

	psd = fetch_share_security_from_tdb(ctx, tdb_ctx, share_name, tdb_file_path, psize);
	if (psd == NULL) {
		DEBUG(2, ("Using default secdesc for share \"%s\".\n", share_name));
		psd = get_share_security_default(ctx, psize, GENERIC_ALL_ACCESS);
	} else {
		DEBUG(3,("Successfully got secdesc from \"%s\" for share \"%s\".\n", tdb_file_path, share_name));
	}

out:

	if ( tdb_ctx != NULL ) {
		close_share_info_tdb(tdb_ctx);
	}
	return psd;
}

SEC_DESC *get_share_security( TALLOC_CTX *ctx, const char *servicename,
			      size_t *psize)
{
	return get_share_security_for_tdb(ctx, lock_path("share_info.tdb"), servicename, psize);
}

/*******************************************************************
 Store a security descriptor in the share db.
 ********************************************************************/

BOOL set_share_security_for_tdb(const char *tdb_file_path, const char *share_name, SEC_DESC *psd)
{
	TDB_CONTEXT* tdb_ctx = NULL;
	prs_struct ps;
	TALLOC_CTX *mem_ctx = NULL;
	fstring key;
	BOOL ret = False;

	if (!open_share_info_tdb(tdb_file_path, &tdb_ctx)) {
		DEBUG(0,("Failed to open share info tdb \"%s\".\n", tdb_file_path));
		goto out;
	}

	mem_ctx = talloc_init("set_share_security_for_tdb");
	if (mem_ctx == NULL) {
		DEBUG(0,("Failed to allocate mem_ctx.\n"));
		goto out;
	}

	if (!prs_init(&ps, (uint32)sec_desc_size(psd), mem_ctx, MARSHALL)) {
		DEBUG(0,("Failed to init prs_struct.\n"));
		goto out;
	}
 
	if (!sec_io_desc("set_share_security_for_db", &psd, &ps, 1)) {
		DEBUG(0,("Failed to parse SEC_DESC.\n"));
		goto out;
	}
 
	slprintf(key, sizeof(key)-1, "SECDESC/%s", share_name);
 
	if (tdb_prs_store(tdb_ctx, key, &ps)==0) {
		DEBUG(3,("Successfully stored secdesc for \"%s\" in \"%s\".\n", 
				 share_name, tdb_file_path));
		ret = True;
	} else {
		DEBUG(1,("Failed to store secdesc for \"%s\" in \"%s\".\n", 
				 share_name, tdb_file_path));
	} 

out:
	/* Free malloc'ed memory */

	if (mem_ctx) {
		prs_mem_free(&ps);
		talloc_destroy(mem_ctx);
	}

	if (tdb_ctx != NULL) {
		close_share_info_tdb(tdb_ctx);
	}

	return ret;
}

BOOL set_share_security(const char *share_name, SEC_DESC *psd)
{
	return set_share_security_for_tdb(lock_path("share_info.tdb"), share_name, psd);
}

static BOOL has_sec_desc_for_share_in_tdb( TDB_CONTEXT* tdb_ctx,
														 const char *share_name,
														 const char *tdb_file_path )
{
	BOOL ret = False;
	TALLOC_CTX* mem_ctx = NULL;
	SEC_DESC* psd = NULL;
	size_t sd_size = 0;

	SMB_ASSERT( tdb_ctx != NULL );
	SMB_ASSERT( share_name != NULL );
	SMB_ASSERT( tdb_file_path != NULL );

	if (!(mem_ctx = talloc_init(FUNCTION_MACRO))) {
		DEBUG(0,("Failed to allocate mem_ctx.\n"));
		ret = False;
		goto out;
	}

	psd = fetch_share_security_from_tdb(mem_ctx, tdb_ctx, share_name, tdb_file_path, &sd_size);
	if (psd == NULL) {
		ret = False;
		goto out;
	}

	ret = True;

out:

	TALLOC_FREE(mem_ctx);
	return ret;
}

/*******************************************************************
 Delete a security descriptor.
********************************************************************/

BOOL delete_share_security_for_tdb(const char *tdb_file_path, const char *share_name)
{
	BOOL ret = False;
	TDB_CONTEXT* tdb_ctx = NULL;
	TDB_DATA kbuf;
	fstring key;

	if (!open_share_info_tdb(tdb_file_path, &tdb_ctx)) {
		DEBUG(0,("Failed to open share info tdb \"%s\".\n", tdb_file_path));
		goto out;
	}

	if (!has_sec_desc_for_share_in_tdb(tdb_ctx,share_name, tdb_file_path)) {
		DEBUG(10,("There's no entry for share \"%s\" in \"%s\", so nothing to delete.\n", 
				 share_name, tdb_file_path));
		ret = True;
		goto out;
	}

	slprintf(key, sizeof(key)-1, "SECDESC/%s", share_name);
	kbuf.dptr = key;
	kbuf.dsize = strlen(key)+1;

	if (tdb_trans_delete(tdb_ctx, kbuf) != 0) {
		DEBUG(0,("Failed to delete entry for share \"%s\" in \"%s\".\n", 
				 share_name, tdb_file_path));
		goto out;
	}

	DEBUG(3,("Successfully deleted entry for \"%s\" in \"%s\".\n", 
			 share_name, tdb_file_path));
	ret = True;

out:

	if (tdb_ctx != NULL) {
		close_share_info_tdb(tdb_ctx);
	}

	return ret;
}

BOOL delete_share_security(const struct share_params *params)
{
	return delete_share_security_for_tdb(lock_path("share_info.tdb"), 
										 lp_servicename(params->service));
}

static BOOL run_external_command_as_root(const char* command)
{
	BOOL ret = False;
	int rc_int = 0;

	become_root();
	rc_int = smbrun(command, NULL);
	unbecome_root();

	if (rc_int == 0) {
		DEBUG(10,("Command \"%s\" completed successfully.\n", command ));
		ret = True;
	} else {
		DEBUG(0,("Failed to run command \"%s\". Returned error = %d.\n", command, rc_int));
	}

	return ret;
}

BOOL delete_temp_share_info_tdb(const char* temp_share_info_file)
{
	pstring command;
	int rc_int = 0;
	BOOL rc_BOOL = False;

	rc_int = slprintf(command, sizeof(command)-1, "/bin/rm -f \"%s\"", temp_share_info_file);
	if (rc_int < 0) {
		DEBUG(0,("Failed to build command, slprintf errno = %d(%s)\n", errno, strerror(errno)));
		return False;
	}
	if (rc_int >= (sizeof(command)-1)) {
		DEBUG(0,("Failed to build command because it is too long (%d): \"/bin/rm -f \"%s\"\".\n", 
				 rc_int, temp_share_info_file));
		return False;
	}

	rc_BOOL = run_external_command_as_root(command);
	if (!rc_BOOL) {
		DEBUG(0,("Failed to run delete command: \"%s\".\n", command));
		return False;
	}

	DEBUG(3,("Successfully deleted temporary share info file \"%s\".\n", 
			 temp_share_info_file));
	return True;
}

BOOL create_temp_share_info_tdb(const char* temp_share_info_file)
{
	const char* org_share_info_file = NULL;
	int rc_int = 0;
	pstring command;

	org_share_info_file = lock_path("share_info.tdb");
	if (!file_exist(org_share_info_file, NULL)) {
		return delete_temp_share_info_tdb(temp_share_info_file);
	}

	/* duplicate file used for read  */
	rc_int = slprintf(command, sizeof(command)-1, "/bin/cp -f \"%s\" \"%s\"", 
					  org_share_info_file, temp_share_info_file);
	if (rc_int < 0) {
		DEBUG(0,("Failed to build copy command, slprintf errno = %d(%s)\n", errno, strerror(errno)));
		return False;
	}
	if ( rc_int >= (sizeof(command)-1) ) {
		DEBUG(0,("Failed to build copy command because it is too long (%d): \"/bin/cp -f \"%s\" \"%s\"\".\n", 
				 rc_int, org_share_info_file, temp_share_info_file));
		return False;
	}

	if (!run_external_command_as_root(command)) {
		DEBUG(0,("Failed to run copy command: \"%s\".\n", command));
		return False;
	}

	DEBUG(3,("Successfully copied share info from \"%s\" to temporary file \"%s\".\n", 
			 org_share_info_file, temp_share_info_file));

	return True;
}


/*
 * share_permission_tuple - struct to provide mapping:
 *		share permission <-> mask <-> human readable name
 */
struct share_permission_tuple {
	share_permission share_perm;
	uint32 mask;
	const char*	name;
};

/*
 * share_permission_tuples - table containing the mappings
 *		share permission <-> mask <-> human readable name
 */
static const struct share_permission_tuple share_permission_tuples[] = {
	{ SHARE_PERMISSION_FULL_CONTROL, SHARE_ALL_ACCESS,    "FULL CONTROL" },
	{ SHARE_PERMISSION_CHANGE,       SHARE_CHANGE_ACCESS, "CHANGE"       },
	{ SHARE_PERMISSION_READ,         SHARE_READ_ONLY,     "READ"         },
	{ SHARE_PERMISSION_NONE,         0x0,                 "NONE"         },
};

/*
 * Maps given share permission value to its corresponding mask.
 * Parameters:
 * 		share_perm - share permission value to map
 * Returns:
 *		if share_perm is valid share permission value then the mask corresponding to this value,
 *		otherwise - mask 0x0
 */
uint32 map_share_permission_to_mask(share_permission share_perm)
{
	unsigned int i;
	unsigned int tuples_count = STATIC_ARRAY_LENGTH(share_permission_tuples);
	for (i=0; i<tuples_count; ++i) {
		if (share_permission_tuples[i].share_perm == share_perm) {
			return share_permission_tuples[i].mask;
		}
	}

	DEBUG(0,("map_share_permission_to_name: cannot map share permission %u.\n", share_perm));
	return 0x0;
}

/*
 * Maps given mask value to its corresponding share permission value.
 * Parameters:
 * 		mask - mask value to map
 * Returns:
 *		if mask is valid mask for share permission value then the value corresponding to this mask,
 *		otherwise - SHARE_PERMISSION_INVALID
 */
share_permission map_mask_to_share_permission(uint32 mask)
{
	unsigned int i;
	unsigned int tuples_count = STATIC_ARRAY_LENGTH(share_permission_tuples);
	for (i=0; i<tuples_count; ++i) {
		if (CONTAINS_MASK(mask,share_permission_tuples[i].mask)) {
			return share_permission_tuples[i].share_perm;
		}
	}

	DEBUG(0,("map_mask_to_share_permission: cannot map mask 0x%x.\n", mask));
	return SHARE_PERMISSION_INVALID;
}

/*
 * Maps given share permission value to its corresponding human readable name.
 * Parameters:
 * 		share_perm - share permission value to map
 * Returns:
 *		if share_perm is valid share permission value then the human readable name corresponding to this value,
 *		otherwise - string "UNKNOWN"
 */
const char* map_share_permission_to_name(share_permission share_perm)
{
	unsigned int i;
	unsigned int tuples_count = STATIC_ARRAY_LENGTH(share_permission_tuples);
	for (i=0; i<tuples_count; ++i) {
		if (share_permission_tuples[i].share_perm == share_perm) {
			return share_permission_tuples[i].name;
		}
	}

	DEBUG(0,("map_share_permission_to_name: cannot map share permission %u.\n", share_perm));
	return "UNKNOWN";
}

/*
 * Gets maximum allowed share permission the user identified by the given token can have on the share sharename.
 * Parameters:
 * 		token - user token for authorization
 * 		sharename - name of the share to check permissions for
 * Returns:
 *		if share does not have security descriptor then SHARE_PERMISSION_FULL_CONTROL,
 *		otherwise maximum share permission allowed by share's security descriptor
 */
share_permission get_share_permission(const NT_USER_TOKEN *token, const char *sharename)
{
	uint32 granted;
	NTSTATUS status;
	TALLOC_CTX *mem_ctx = NULL;
	SEC_DESC *psd = NULL;
	size_t sd_size;
	BOOL ret_access_check = True;
	share_permission ret = SHARE_PERMISSION_NONE;

	if (!(mem_ctx = talloc_init("get_share_permission"))) {
		DEBUG(0,("Failed to allocate mem_ctx.\n"));
		return SHARE_PERMISSION_NONE;
	}

	psd = get_share_security(mem_ctx, sharename, &sd_size);

	/* if share does not have security descriptor return SHARE_PERMISSION_FULL_CONTROL */
	if (psd == NULL) {
		TALLOC_FREE(mem_ctx);
		DEBUG(5,("get_share_permission: security descriptor not found for share '%s'. Returning FULL_CONTROL.\n", sharename));
		return SHARE_PERMISSION_FULL_CONTROL;
	}

	/* find out all the permissions we may have on this share */
	ret_access_check = se_access_check(psd, 
									   token, 
									   MAXIMUM_ALLOWED_ACCESS, 
									   &granted, 
									   &status);

	if (ret_access_check) {
		/* if granted is MAXIMUM_ALLOWED_ACCESS it means no DACL so we have full control */
		if ( granted == MAXIMUM_ALLOWED_ACCESS ) {
			ret = SHARE_PERMISSION_FULL_CONTROL;
		} else {
			ret = map_mask_to_share_permission(granted);
		}
		DEBUG(10,("get_share_permission: share = '%s', granted = 0x%x. Returning %s.\n", sharename, granted, map_share_permission_to_name(ret)));
	} else {
		DEBUG(10,("get_share_permission: se_access_check returned False for share '%s'. Returning NONE.\n", sharename));
		ret = SHARE_PERMISSION_NONE;
	}

	talloc_destroy(mem_ctx);
	return ret;
}

/*
 * Checks if access with the given access allowed for share with the given share permission.
 * Parameters:
 *		desired_access_mask - mask for desired access, mask should have GENERIC bits converted to specific bits
 * 		share_perm - share permission to check against
 * Returns:
 *		True if the given share permission allows access with the desired mask otherwise False
 */
BOOL check_access_against_share_permission(uint32 desired_access_mask, share_permission share_perm)
{
	return CONTAINS_MASK(map_share_permission_to_mask(share_perm), desired_access_mask & SHARE_ALL_ACCESS);
}

/***************************************************************************
 Parse the contents of an acl string from a usershare file.
***************************************************************************/

BOOL parse_usershare_acl(TALLOC_CTX *ctx, const char *acl_str, SEC_DESC **ppsd)
{
	size_t s_size = 0;
	const char *pacl = acl_str;
	int num_aces = 0;
	SEC_ACE *ace_list = NULL;
	SEC_ACL *psa = NULL;
	SEC_DESC *psd = NULL;
	size_t sd_size = 0;
	int i;

	*ppsd = NULL;

	/* If the acl string is blank return "Everyone:R" */
	if (!*acl_str) {
		SEC_DESC *default_psd = get_share_security_default(ctx, &s_size, GENERIC_READ_ACCESS);
		if (!default_psd) {
			return False;
		}
		*ppsd = default_psd;
		return True;
	}

	num_aces = 1;

	/* Add the number of ',' characters to get the number of aces. */
	num_aces += count_chars(pacl,',');

	ace_list = TALLOC_ARRAY(ctx, SEC_ACE, num_aces);
	if (!ace_list) {
		return False;
	}

	for (i = 0; i < num_aces; i++) {
		SEC_ACCESS sa;
		uint32 g_access;
		uint32 s_access;
		DOM_SID sid;
		fstring sidstr;
		uint8 type = SEC_ACE_TYPE_ACCESS_ALLOWED;

		if (!next_token(&pacl, sidstr, ":", sizeof(sidstr))) {
			DEBUG(0,("parse_usershare_acl: malformed usershare acl looking "
				"for ':' in string '%s'\n", pacl));
			return False;
		}

		if (!string_to_sid(&sid, sidstr)) {
			DEBUG(0,("parse_usershare_acl: failed to convert %s to sid.\n",
				sidstr ));
			return False;
		}

		switch (*pacl) {
			case 'F': /* Full Control, ie. R+W */
			case 'f': /* Full Control, ie. R+W */
				s_access = g_access = GENERIC_ALL_ACCESS;
				break;
			case 'R': /* Read only. */
			case 'r': /* Read only. */
				s_access = g_access = GENERIC_READ_ACCESS;
				break;
			case 'D': /* Deny all to this SID. */
			case 'd': /* Deny all to this SID. */
				type = SEC_ACE_TYPE_ACCESS_DENIED;
				s_access = g_access = GENERIC_ALL_ACCESS;
				break;
			default:
				DEBUG(0,("parse_usershare_acl: unknown acl type at %s.\n",
					pacl ));
				return False;
		}

		pacl++;
		if (*pacl && *pacl != ',') {
			DEBUG(0,("parse_usershare_acl: bad acl string at %s.\n",
				pacl ));
			return False;
		}
		pacl++; /* Go past any ',' */

		se_map_generic(&s_access, &file_generic_mapping);
		init_sec_access(&sa, g_access | s_access );
		init_sec_ace(&ace_list[i], &sid, type, sa, 0);
	}

	if ((psa = make_sec_acl(ctx, NT4_ACL_REVISION, num_aces, ace_list)) != NULL) {
		psd = make_sec_desc(ctx, SEC_DESC_REVISION, SEC_DESC_SELF_RELATIVE, NULL, NULL, NULL, psa, &sd_size);
	}

	if (!psd) {
		DEBUG(0,("parse_usershare_acl: Failed to make SEC_DESC.\n"));
		return False;
	}

	*ppsd = psd;
	return True;
}
