/*
 * FS (file system) VFS module for Samba.
 *
 * Copyright (C) 2004, Dina Fine
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 This vfs layer requires the dev/exastore kernel module to be installed 
 which returns a proprietry exastore nfs handle (available through GPL)
 and the server to be iplemented (see external_acl_manager under the 
 source directory, also handles FS requests). 
*/

#include "includes.h"
#include "smbutil/exasmb_vfsfsclient.h"
#include "exasmb_if.h"
#include "smbutil/sid_struct.h"
#include "events/EventsClientAccess_c.h"

#define ALLOC_CHECK(ptr, label) do { if ((ptr) == NULL) { DEBUG(0, ("fs vfs: out of memory!\n")); errno = ENOMEM; goto label; } } while(0)
#define MAX_READ_SIZE	65536
#define MAX_READ_BUFFER_SIZE (MAX_READ_SIZE + 4096)
#define SHARED_READ_BLOCK_SIZE	8192
#define EXASTAT	exastat

static int vfs_fs_debug_level = DBGC_VFS;
extern uint32 (*get_snapshot_id)(char nfsfh[32]);
extern const char *sharedInBuffer;
extern int sharedInBufferSize;
extern int sharedInBufferID;
extern void (*my_seekdir)(DIR* dirp);

extern int (*get_snapshots_timestamps)( const char* volume_name, 
										uint32 max_number_of_snapshots,
										uint32 snapshots_timestamps[],
										uint32* number_of_snapshots );
extern int (*get_snapshot_name_by_timestamp)( const char* volume_name,
											  uint32 snapshot_timestamp,
											  int* snapshot_found,
											  char snapshot_name[] );

extern int (*get_antivirus_info)( const char* nfs_fh_data, 
								  uint32_t nfs_fh_len, 
								  uint32_t* p_val_out );
extern int (*set_antivirus_info)( const char* nfs_fh_data, 
								  uint32_t nfs_fh_len, 
								  uint32_t val );
extern int (*set_antivirus_info_guarded)( const char* nfs_fh_data, 
										  uint32_t nfs_fh_len, 
										  uint32_t val, 
										  uint32_t guard, 
										  uint32_t* p_val_out );


static char *sharedReadBuffer = NULL;
static int sharedReadBufferID = -1;
static int sharedReadBufferSize = 0;

#define exanet_dir_magic (((unsigned) 'e' << 24) | \
        ((unsigned) 'x' << 16) | ((unsigned) 'a' << 8) | 0xac)

typedef struct exa_dent_name {
        size_t len;
        char *name;
} exa_dent_name;


struct xattr_stat {
	struct stat stat;
	uint32_t createtime_sec;
	uint32_t createtime_nsec;
};
typedef struct exanet_dir {
        int magic;
        int path_len;
        char *path;
        const connection_struct *conn;
        int dirfd;
        void *buf;
        void *bufp;
        int buf_id;
        int buf_size;
        uint64_t cookie;
        int eod;
        exa_dent_name prev_name;
        struct xattr_stat *prev_stat;
        dev_t st_dev;
        blksize_t st_blksize;
        char nfs_fh[32];
        void* prev_dir_pos;
        short int override_seekdir;

} exanet_dir;

/* bug19657: join major message featue:
	to stop flooding by major messages we will join them togather */
#define NO_MSG_JOIN_THRESHOLD 10		/* Don't join the first N major messages */
#define MSG_JOIN_NUMBER 200				/* How many major messages should be joined in one */
#define MSG_JOIN_GROUP_TIMEOUT	(5*60)	/* Maximal  time (in secods) we will join messages to one */

enum MAJOR_MSG_JOIN_VERB {
	VERB_UNDEFINED = 0, 
	VERB_WRITE,
	VERB_PWRITE
};

static enum MAJOR_MSG_JOIN_VERB major_msg_join_verb = VERB_UNDEFINED;
static int join_msg_last_errno = 0;
static long int join_msg_err_count = 0;
static time_t join_msg_last_time = 0;	/* Start time for current group of joined messages */


#undef DBGC_CLASS
#define DBGC_CLASS vfs_fs_debug_level


static DIR *fs_opendir(struct vfs_handle_struct *handle, const char *fname, const char* mask, uint32 attr);
static SMB_STRUCT_DIRENT *fs_readdir(struct vfs_handle_struct *handle, DIR *dir);
static void fs_rewinddir(vfs_handle_struct *handle, DIR *dir);
static int fs_closedir(struct vfs_handle_struct *handle, DIR *dir);
 
ssize_t     fs_write(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data, size_t n);
ssize_t     fs_pwrite(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data,size_t n, SMB_OFF_T offset);
ssize_t     internal_write(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data,size_t n, const BOOL has_offset, SMB_OFF_T offset);
ssize_t     fs_sendfile(vfs_handle_struct *handle, int tofd, files_struct *fsp, int fromfd, const DATA_BLOB *hdr,SMB_OFF_T offset, size_t n);
ssize_t 	fs_read(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n);
ssize_t		fs_pread(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n, SMB_OFF_T offset);
int         fs_rename(vfs_handle_struct *handle, const char *from_name, const char *to_name);
int			fs_fsync(struct vfs_handle_struct *handle, files_struct *fsp, int fd);
int			fs_close(struct vfs_handle_struct *handle, files_struct *fsp, int fd);
static int fs_stat(struct vfs_handle_struct *handle, const char *fname, SMB_STRUCT_STAT *sbuf);
static int fs_fstat(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_STRUCT_STAT *sbuf);
static int fs_lstat(struct vfs_handle_struct *handle, const char *path, SMB_STRUCT_STAT *sbuf);
static int fs_ftruncate(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_OFF_T len);
SMB_BIG_UINT fs_disk_free(vfs_handle_struct *handle, const char *path, BOOL small_query, SMB_BIG_UINT *bsize, 
			       SMB_BIG_UINT *dfree, SMB_BIG_UINT *dsize);
static int fs_fsetxattr(struct vfs_handle_struct *handle, struct files_struct *fsp,int fd, const char *name,
						const void *value, size_t size, int flags);


static vfs_op_tuple fs_ops[] = {

    {SMB_VFS_OP(fs_opendir),	SMB_VFS_OP_OPENDIR,	    SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_readdir),	SMB_VFS_OP_READDIR,     SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(fs_rewinddir),	SMB_VFS_OP_REWINDDIR,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_closedir),	SMB_VFS_OP_CLOSEDIR,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_write),	SMB_VFS_OP_WRITE,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_pwrite),	SMB_VFS_OP_PWRITE,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_sendfile),	SMB_VFS_OP_SENDFILE,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(fs_read),		SMB_VFS_OP_READ,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(fs_pread),		SMB_VFS_OP_PREAD,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_fsync),	SMB_VFS_OP_FSYNC,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_close),	SMB_VFS_OP_CLOSE,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_rename),	SMB_VFS_OP_RENAME,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_stat),	SMB_VFS_OP_STAT,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(fs_fstat),	SMB_VFS_OP_FSTAT,	SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(fs_lstat),	SMB_VFS_OP_LSTAT,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(fs_ftruncate),	SMB_VFS_OP_FTRUNCATE,	SMB_VFS_LAYER_TRANSPARENT},	
	{SMB_VFS_OP(fs_disk_free),	SMB_VFS_OP_DISK_FREE,	SMB_VFS_LAYER_TRANSPARENT},	
    {SMB_VFS_OP(fs_fsetxattr),	SMB_VFS_OP_FSETXATTR,   SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(NULL),		SMB_VFS_OP_NOOP,	SMB_VFS_LAYER_NOOP}
};

static const char unified_share_name[] = "C$";

static BOOL is_unified_share( const struct vfs_handle_struct* handle )
{
   return 
      ( handle != NULL ) &&
      ( handle->conn != NULL ) &&
      StrCaseCmp( SERVICE( SNUM( handle->conn ) ), unified_share_name ) == 0;
}

static BOOL should_skip_vfs_fs( const struct vfs_handle_struct* handle, const char* name )
{
   static const char* names_to_skip[] = { ".", "/.", "./", "./.", "..", "/..", "../", "./..", "../." };
   static const size_t names_to_skip_size = STATIC_ARRAY_LENGTH( names_to_skip ) - 1;
   size_t i = 0;

   if ( ! is_unified_share( handle ) ) {
      return false;
   }

   for ( i = 0 ; i < names_to_skip_size ; ++i ) {
      if ( strcmp( name, names_to_skip[i] ) == 0 ) {
         return true;
      }
   }

   return false;
}

int mk_exanet_dir_magic(void *p)
{
	return ((int) p) ^ exanet_dir_magic;
}

static
void *
next_ent(void *p, exa_dent_name *name, struct xattr_stat **stbufp)
{
	if (!p)
		return NULL;
	char *pos = (char *) p;
	if (*(size_t*) pos == 0)
		return NULL;
	name->len = *(size_t *) pos;
	pos += sizeof(size_t);
	name->name = pos;
	if (name->len % 4)
		pos += (name->len & ~3) + 4;
	else
		pos += name->len;
	*stbufp = (struct xattr_stat *) pos;
	return pos + sizeof(struct xattr_stat);
}

static const char *strerror_extended_by_vfs_messages(int errno_ex)
{
	static char unknown_err_buf[100];
	
	if(errno_ex >= 0)
		return strerror(errno_ex);
	switch(errno_ex)
	{
		case VFS_FATAL_ERROR:
			return "VFS_FATAL_ERROR";
		case VFS_OPERATION_FATAL_ERROR:
			return "VFS_OPERATION_FATAL_ERROR";
		case VFS_CLIENT_DISCONNECTED_ERROR:
			return "VFS_CLIENT_DISCONNECTED_ERROR";
		case VFS_UNEXPECTED_ERROR:
			return "VFS_UNEXPECTED_ERROR";
		case VFS_UNKNOWN_REQUEST_TYPE:
			return "VFS_UNKNOWN_REQUEST_TYPE";
		case VFS_INTERNAL_FS_PROBLEM:
			return "VFS_INTERNAL_FS_PROBLEM";
		case VFS_READ_ERROR:
			return "VFS_READ_ERROR";
		case VFS_WRITE_ERROR:
			return "VFS_WRITE_ERROR";
		case VFS_PERM_DENIED:
			return "VFS_PERM_DENIED";
		case VFS_NOT_SUPPORTED:
			return "VFS_NOT_SUPPORTED";
	}

	/* Error not translated... */
	snprintf(unknown_err_buf, sizeof(unknown_err_buf), "Unknown error #%d", errno_ex);
	return unknown_err_buf;
}

/* bug19657: join major message featue:
	to stop flooding by major messages we will join them togather.
	Note:
	-First NO_MSG_JOIN_THRESHOLD messgaes are not joined
	-If messages are coming too slowly (see MSG_JOIN_GROUP_TIMEOUT), feature will be blocked
*/

static int is_print_major_msg_this_time(enum MAJOR_MSG_JOIN_VERB cur_verb, int cur_errno)
{
	if(major_msg_join_verb == cur_verb &&
		cur_errno == join_msg_last_errno &&
		join_msg_last_time + MSG_JOIN_GROUP_TIMEOUT > time(NULL)
	) { /* Same message as was last time (and in accptable timeout) ? */
		join_msg_err_count++;
	}
	else { /* Another kind of message... */
		if(join_msg_err_count>0) { /* Previous message was joined but still not prompted... */
			DEBUG(0,("Message with errno=%d, '%s' was repeated totally %ld times\n",
			join_msg_last_errno, strerror_extended_by_vfs_messages(join_msg_last_errno), join_msg_err_count));
		}
		/* Reset message history */
		major_msg_join_verb = cur_verb;
		join_msg_last_errno = cur_errno;
		join_msg_err_count = 0;
		join_msg_last_time = time(NULL);
	}
	return (join_msg_err_count < NO_MSG_JOIN_THRESHOLD ||
		(join_msg_err_count>NO_MSG_JOIN_THRESHOLD && (join_msg_err_count-NO_MSG_JOIN_THRESHOLD)%MSG_JOIN_NUMBER==0)
	);
}

static void print_major_msg_repeat_count_on_demand()
{
	if(join_msg_err_count>NO_MSG_JOIN_THRESHOLD &&
		(join_msg_err_count-NO_MSG_JOIN_THRESHOLD)%MSG_JOIN_NUMBER==0) {
		DEBUG(0,("Previous major message (errno=%d, '%s') was repeated %d times again (totally %ld times)\n",
			join_msg_last_errno, strerror_extended_by_vfs_messages(join_msg_last_errno), 
			MSG_JOIN_NUMBER, join_msg_err_count));

		/* Reset timeout for each printed group */
		join_msg_last_time = time(NULL);
	}
}

static exanet_dir *last_dir_read;

static
DIR *
fs_opendir(struct vfs_handle_struct *handle, const char *fname, const char* mask, uint32 attr)
{
	struct stat st;
    int ret_err;
	exanet_dir *dirp = NULL;

   if ( should_skip_vfs_fs( handle, fname ) ) {
      return SMB_VFS_NEXT_OPENDIR(handle, fname, mask, attr); 
   }

   dirp = calloc(1, sizeof(exanet_dir));
	if (dirp) {
		dirp->path = (*fname == '.') ? strdup("") : strdup(fname);
		dirp->path_len = strlen(dirp->path);
		dirp->conn = handle->conn;
		dirp->dirfd = sys_open(fname, O_RDONLY, 0444); // O_DIRECTORY is linux specific
		if (dirp->dirfd == -1){
            DEBUG(0, ("Failed to open %s, errno %d %s\n",fname, errno, strerror(errno)));
            goto cleanup;
        }
		if (fstat(dirp->dirfd, &st) == -1){
            DEBUG(0,("Fstat failed, errno %d %s\n", errno, strerror(errno)));
            goto cleanup;
        }
		if (!S_ISDIR(st.st_mode)) {
            /* Shlomi: This is not a real error. Sometimes Windows just try to query info
             * of files under other files, e.g. File1/view.dat although File1 is not a directory */
			errno = ENOTDIR;
            DEBUG(2,("%s is not a dir, errno %d %s\n",fname, errno, strerror(errno)));
			goto cleanup;
		}
		dirp->st_dev = st.st_dev;
		dirp->st_blksize = st.st_blksize;
		if (exasmb_get_nfs_fh(dirp->dirfd, NULL, dirp->nfs_fh) == -1){
            DEBUG(1,("get_nfs_fh failed, errno %d %s\n", errno, strerror(errno)));
            goto cleanup;
        }
		dirp->buf_size = 64*1024;
		dirp->buf_id = shmget(IPC_PRIVATE, dirp->buf_size,
							  IPC_CREAT|SHM_R|SHM_W|S_IROTH|S_IWOTH);
		if (dirp->buf_id == -1){
            DEBUG(0, ("failed to allocate shared memory, errno %d %s\n", errno, strerror(errno)));
            goto cleanup;
        }
		dirp->buf = shmat(dirp->buf_id, NULL, 0);
		if (shmctl(dirp->buf_id, IPC_RMID, NULL) < 0) {
            /* most we can do is print debug here*/
            DEBUG(0,("shmctl failed on id %d - a memory leak is possible, errno %d %s\n", dirp->buf_id, errno, strerror(errno)));
		}
		if (dirp->buf == (void *) -1){
            DEBUG(0,("Failed to attach to shared memory, id %d, errno %d %s\n", dirp->buf_id, errno, strerror(errno)));
            goto cleanup;
        }
		dirp->magic = mk_exanet_dir_magic(dirp);
		return (DIR*) dirp;
	}
cleanup:
    ret_err = errno;
	if (dirp) {
		if (dirp->buf) {
			if (dirp->buf != (void*) -1)
				shmdt(dirp->buf);
			dirp->buf = NULL;
		}
		if (dirp->dirfd && (dirp->dirfd != -1)) {
			close(dirp->dirfd);
			dirp->dirfd = -1;
		}
		SAFE_FREE(dirp->path);
		SAFE_FREE(dirp);
	}
	errno = ret_err;
	return NULL;
}

static
SMB_STRUCT_DIRENT *
fs_readdir(struct vfs_handle_struct *handle, DIR *dir)
{
	static SMB_STRUCT_DIRENT d_ent;
    int ret = 0;
	exanet_dir *dirp = (exanet_dir *) dir;

	if (dirp->magic != mk_exanet_dir_magic(dirp)) // not ours, pass it on
		return SMB_VFS_NEXT_READDIR(handle, dir); 

	// various asserts here
    if (dirp->override_seekdir == True) {
        DEBUG(5, ("fs_readdir: performing override_seekdir, prev_dir_pos=%p\n", dirp->prev_dir_pos));
        dirp->bufp = dirp->prev_dir_pos;
    }
    
    dirp->override_seekdir = False;
	last_dir_read = NULL;
	while (!dirp->eod) {
        dirp->prev_dir_pos = dirp->bufp;
		if ((dirp->bufp = next_ent(dirp->bufp, &dirp->prev_name,
								   &dirp->prev_stat)) == NULL) {
			if (dirp->cookie == ~0ull) {
				dirp->eod = True;
				break;
			}
            DEBUG(2, ("Calling to exasmb_vfs_dirplus\n"));
            ret = exasmb_vfs_dirplus(dirp->buf_id, dirp->buf_size, dirp->nfs_fh,
								   &dirp->cookie);
            if ( ret == -1) {
                DEBUG(0, ("exa_readdir failed errno %d %s\n", errno, strerror(errno)));
				dirp->eod = True; // hard error, what do we do?
				break;
			}
            if ( ret == -2) {
                DEBUG(0, ("exa_readdir fatal error, aborting\n"));
				exit_server("exa_readdir fatal error");
			}

            DEBUG(2, ("Finished exasmb_vfs_dirplus\n"));
			dirp->bufp = dirp->buf;
			continue;
		}
		memcpy(d_ent.d_name, dirp->prev_name.name, dirp->prev_name.len);
		d_ent.d_name[dirp->prev_name.len] = 0;
		last_dir_read = dirp;
		DEBUG(5, ("fs_readdir: dirp->path=%s, dirp->prev_name.name=%s dirp->st_dev=%x\n",
			dirp->path, dirp->prev_name.name, (unsigned int)dirp->st_dev ));
		return &d_ent;
	}
	return NULL;
}

/* In case there is a rewinddir call, we need to make sure the current buffer is not used (in case
   it's a big direcotry that spans across more than one buffer, and then read it from the start.
   this is done using resetting the cookie.
   Also need to check that EOD is resetted.
*/

static
void 
fs_rewinddir(vfs_handle_struct *handle, DIR *dir) {

	/* Convert to exanet_dir and make sure it's valid, otherwise skip */
	exanet_dir *dirp = (exanet_dir *) dir;
	if (dirp->magic != mk_exanet_dir_magic(dirp)) { // not ours
		SMB_VFS_NEXT_REWINDDIR(handle, dir);
		return;
	}
	dirp->bufp = 0;		/* cause an actual readdir call (don't use buffer)	*/
	dirp->cookie = 0;	/* cause the ctual call to start from the top		*/
	dirp->eod = False;	/* starting fresh...								*/
	return;
}

/* We need to implement seekdir in cases where the last entry from the readdirplus was not inserted 
   to the client buffer. This happens when the entry is too big and the client buffer is already full.
   What happens in higher layers is that in case of such failures, they perform seekdir to the last
   successful position. We need to keep the cookie/directory position for the next findnext request 
*/

void wrap_seekdir(DIR* dirp)
{
    DEBUG(5,("wrap_seekdir: enter, dirp %p\n", dirp));
    
    exanet_dir* mydirp = (exanet_dir*) dirp;
    if (mydirp->prev_dir_pos) {
        mydirp->override_seekdir = True;
    } 
    else
        DEBUG(5,("wrap_seekdir: prev_dir_pos is NULL\n"));
}

static
int
fs_closedir(struct vfs_handle_struct *handle, DIR *dir)
{
	exanet_dir *dirp = (exanet_dir *) dir;
	if (!dirp || (dirp->magic != mk_exanet_dir_magic(dirp))) // not ours, pass it on
		return SMB_VFS_NEXT_CLOSEDIR(handle, dir);
	if (dirp == last_dir_read)
		last_dir_read = NULL;
	shmdt(dirp->buf);
	close(dirp->dirfd);
	SAFE_FREE(dirp->path);
	ZERO_STRUCT(*dirp);
	SAFE_FREE(dirp);
	return 0;
}

ssize_t  fs_write(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data, size_t n){
   return internal_write(handle, fsp, fd, data, n, false, -1);
}


ssize_t  fs_pwrite(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data,size_t n, SMB_OFF_T offset){
    return internal_write(handle, fsp, fd, data, n, true, offset);
}
    

/*
    internal_write(...) - internal implementation for write. It contains all functionality required to fs_write() and fs_pwrite().
*/
ssize_t internal_write(vfs_handle_struct *handle, files_struct *fsp, int fd, const void *data,size_t n, const BOOL has_offset, SMB_OFF_T offset){
    ssize_t ret = -1;
    BOOL use_next_write = True;
    // check if the data comes from shared memory
    const char *cdata = (const char *) data;
    DEBUG(3,("fs_pwrite:enter file name %s offset %llu count %d has offset %d\n",fsp->fsp_name, offset ,n, has_offset));
    if (!lp_write_cache_size(SNUM(fsp->conn)) && cdata > sharedInBuffer &&
        !lp_disable_vfsfs_write() &&
            cdata < (sharedInBuffer + sharedInBufferSize - 1) && fsp->fh->fd == fd) 
    {
        use_next_write = False;
        size_t buff_offset = cdata - sharedInBuffer;
        off_t file_offset = offset;
        // use the unstable variable for unstable write flag
        int unstable = ((lp_allow_unstable_write_without_oplock() || EXCLUSIVE_OPLOCK_TYPE(fsp->oplock_type)) &&
                  !fsp->write_through && !fsp->o_sync && !lp_stable_write()) ? 1 : 0;
        int must_exist = fsp->need_sync;

        if (unstable)
            fsp->need_sync = True;

        if (lp_fake_read_write()){
            DEBUG(1, ("FAKE WRITE !!!"));
            ret = n;
        }else{
            DEBUG(3, ("Calling to exasmb_vfs_shwrite with sharedInBufferId: %d, count: %u,"
                      " buff_offset: %u, offset: %lld id: %d must exist: %d unstable: %d\n",
                       sharedInBufferID, n, buff_offset, file_offset, (int)fsp, must_exist, unstable));
            if(!has_offset) {
                // !!! todo: check how to use file_offset ???
                file_offset = lseek(fd, 0, SEEK_CUR);
            }
            if (fsp->nfs_fh[0] == 0) {
                DEBUG(1,("Invalid nfs file handle before shwrite, attempting to acquire...\n"));
                if (exasmb_get_nfs_fh(fd, NULL, fsp->nfs_fh) == -1) {
                    DEBUG(1,("FAILED acquiring nfs file handle before shwrite\n"));
                    return -1;
                }
            }
            bool update_mtime = true;
            if ( !null_timespec(fsp->pending_modtime)) {
                update_mtime = false; 
            }
                ret = exasmb_vfs_shwrite(sharedInBufferID, fsp->nfs_fh, file_offset, buff_offset, n, (int)fsp,
                                         &must_exist, unstable, update_mtime);
        }
        DEBUG(3, ("ret %d\n", ret));


        if (ret == -1) {

            /* If stable write returned with an error       
             * then the id was flushed and will not be exist in the next time*/
            if(!unstable) {
                fsp->need_sync = False;
            } else {
                /* unstable write failed, return fsp=>need_sync to previous state.
                 * note, exasmb_vfs_shwrite() changes must_ecist only if return !(-1)*/
                fsp->need_sync = must_exist;
            }

            if (errno > 0 || errno == VFS_PERM_DENIED || errno == VFS_UNEXPECTED_ERROR) {
				/* Protocols supportability: alert if errno==EROFS */
				alert_ro_fs( (fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name);	
                /* Non-fatal messages: join major messages */
                if(is_print_major_msg_this_time(VERB_PWRITE, errno)) {
                    switch(errno) {
                        case VFS_PERM_DENIED:
                            DEBUG(0,("exa_pwrite permission denied error\n"));
                        break;
                        case VFS_UNEXPECTED_ERROR:
                            DEBUG(0,("exa_pwrite unexpected error\n"));
                        break;
                        default:
                            DEBUG(0, ("exa_pwrite failed errno %d %s\n", errno, strerror(errno)));
                    }
                }
                print_major_msg_repeat_count_on_demand();

                if(errno == VFS_PERM_DENIED) errno = EACCES;

                return -1;
            }else if (errno == VFS_FATAL_ERROR) {
                DEBUG(0, ("exa_pwrite fatal error, aborting\n"));    
                exit_server("exa_pwrite fatal error");    
            }else if (errno == VFS_OPERATION_FATAL_ERROR) {
                DEBUG(0, ("exa_pwrite operation fatal error, aborting\n"));    
                exit_server("exa_pwrite fatal error");
            }else if (errno == VFS_CLIENT_DISCONNECTED_ERROR) {
                /* should never happen */
                DEBUG(0, ("exa_pwrite vfsclient disconnection error, aborting\n"));
                smb_panic("exa_pwrite vfsclient disconnection error");    
            } else if (errno == VFS_NOT_SUPPORTED) {
                DEBUG(3,("shwrite not implemented, using next\n"));
                use_next_write = True;
            }
        }

        if (!use_next_write) {
            if (!must_exist)
                fsp->need_sync = False;
            /* resetting the size to force a stat before the next sendfile */
            fsp->size_for_sendfile = 0;
            if(!has_offset) {            
                // update the file offset
    			file_offset = file_offset + ret;
    			lseek(fd, file_offset, SEEK_SET);
            }
            return ret;
        }
    }
    if (!use_next_write) {
        return ret;
    }
    /* use next_write*/
    int retries = 1;
    /* This should not happen normally, we'll add a debug (warning) to let us know why it happened */
    DEBUG(3,("\tlp_write_cache_size()=%d (should be 0)\n"
             "\tcdata (%p), sharedInBuffer(%p), cdata > sharedInBuffer ? %d (should be 1)\n"
             "\tlp_disable_vfsfs_write()=%d (should be 0)\n"
             "\tsharedInBufferSize=%d, cdata < (sharedInBuffer + sharedInBufferSize - 1 ? %d (should be 1)\n"
             "\tfsp->fh->fd=%d, fd=%d (should be the same)\n",
             lp_write_cache_size(SNUM(fsp->conn)),
             cdata, sharedInBuffer, cdata > sharedInBuffer ? 1 : 0,
             lp_disable_vfsfs_write(),
             sharedInBufferSize, cdata < (sharedInBuffer + sharedInBufferSize - 1) ? 1 : 0,
             fsp->fh->fd, fd) );
    do {
        if (fsp->need_sync) {
            fs_fsync(handle, fsp, fd);
        }
        if(has_offset) {
        
            DEBUG(3,("calling to the next pwrite, attempt number %d, offset %llu count %d\n",
                     retries, offset ,n));
            ret = SMB_VFS_NEXT_PWRITE(handle, fsp, fd, data, n, offset);
        } else {
            DEBUG(3,("calling to the next write, (offset 0) count %d\n", n));
            ret = SMB_VFS_NEXT_WRITE(handle, fsp, fd, data, n);
        }
        if (ret == -1) {
            DEBUG(0, ("(NEXT) pwrite failed errno %d %s, attempt %d\n", errno, strerror(errno), retries));
            retries++;
        } else
            SMB_VFS_NEXT_FSYNC(handle, fsp, fd);
    } while ( ret == -1 && retries < 4 );

    DEBUG(4,("written %lld\n", (long long)ret));
    /* resetting the size to force a stat before the next sendfile */
    fsp->size_for_sendfile = 0;         
    return ret;

}


void
detach_shared_read_memory()
{
	if (sharedReadBuffer) {
		DEBUG(2, ("Detaching read shared memory: %d\n", sharedReadBufferID));
		int err = shmdt(sharedReadBuffer);
		if (err != 0) {
			DEBUG(0,("Failed to detach shared memory: %d\n ", sharedReadBufferID));
		}
		sharedReadBufferID = -1;
		sharedReadBuffer = NULL;
		sharedReadBufferSize = 0;
	}

}

int
prepare_shared_read_mem(size_t size)
{
	
	DEBUG(10, ("size: %d, sharedReadBufferSize %d\n", size, sharedReadBufferSize));
	if (size > MAX_READ_SIZE) {
		DEBUG(0, ("Read count %d > MAX_READ_SIZE %d\n", size, MAX_READ_SIZE));
		return -1;
	}
	if (sharedReadBufferID > 0 && size <= sharedReadBufferSize) {
		return 0;
	}
	// make it 8K mult.
	int new_size_mod = size%SHARED_READ_BLOCK_SIZE;
	if (new_size_mod != 0) {
		size += SHARED_READ_BLOCK_SIZE - new_size_mod;
	}
	
    if (sharedReadBufferID > 0) {
		detach_shared_read_memory();
	}
    sharedReadBufferSize = size;
	DEBUG(2, ("Allocating shared read memory, size %d\n", sharedReadBufferSize));
	sharedReadBufferID = shmget(IPC_PRIVATE, sharedReadBufferSize, IPC_CREAT|SHM_R|SHM_W);
	DEBUG(5,("%d shareReadBufferID\n", sharedReadBufferID));
	if (sharedReadBufferID == -1) {
		DEBUG(0,("failed to allocate shared memory (%s)\n", strerror(errno)));
		return -1;
	}
	sharedReadBuffer = shmat(sharedReadBufferID, NULL, 0);
	if (shmctl(sharedReadBufferID, IPC_RMID, NULL) < 0) {
		DEBUG(0,("shmctl(sharedReadBufferID, IPC_RMID, NULL) failed, possible shared memory leak - (%s)\n",
				 strerror(errno)));
		if (sharedReadBuffer && sharedReadBuffer != (void *) -1) {
			shmdt(sharedReadBuffer);
		}
		return -1;
	}
	if (sharedReadBuffer == (void *) -1) {
		DEBUG(0,("Failed to attach to shared memory, id %d, errno %d %s\n", sharedReadBufferID, errno, strerror(errno)));
		return -1;
	}
	DEBUG(5,("%d sharedReadBuffer %p\n", sharedReadBufferID, sharedReadBuffer));
	return 0;
}
ssize_t
fs_read(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t count)
{
	ssize_t result;
	
	DEBUG(10, ("fs_read: file name: %s, count: %d\n", fsp->fsp_name, count));
	SMB_OFF_T file_offset = sys_lseek(fd, 0, SEEK_CUR);
	if (file_offset == (SMB_OFF_T) -1) {
		DEBUG(0,("fs_read(%s): failed to lseek current file position, errno %d (%s)\n", 
				 fsp->fsp_name, errno, strerror(errno)));
		return -1;
	}
	
	result = fs_pread(handle, fsp, fd, data, count, file_offset);
	if (result < 0) {
		return result;
	}
	file_offset += result;
	if ( (sys_lseek(fd, file_offset, SEEK_SET)) == (SMB_OFF_T) -1){
		DEBUG(0,("fs_read(%s): failed to lseek to the new file position, file_offset %lld errno %d (%s)\n", 
				  fsp->fsp_name, file_offset, errno, strerror(errno)));
		return -1;
	}
	return result;
}

ssize_t
fs_pread(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t count, SMB_OFF_T offset)
{
	ssize_t result = -1;
    DEBUG(2, ("fs_pread: file name: %s, count: %d, offset %llu\n", fsp->fsp_name, count, offset));
	/* retrieve nfs file handle*/
	if (fsp->nfs_fh[0] == 0) {
		DEBUG(1,("Invalid nfs file handle before read file, attempting to acquire...\n"));
		if (exasmb_get_nfs_fh(fd, NULL, fsp->nfs_fh) == -1) {
			DEBUG(0, ("fs_pread(%s), failed to get nfs_fh, fallback on NFS mount\n", fsp->fsp_name));
            result = SMB_VFS_NEXT_PREAD(handle, fsp, fd, data, count, offset);
			return result;
		}
	}
	if (prepare_shared_read_mem(count)) {
		DEBUG(0, ("fs_pread: file name: %s, count: %d, offset: %llu - failed to prepare share read memory.\n",
				  fsp->fsp_name, count, offset));
		result = SMB_VFS_NEXT_PREAD(handle, fsp, fd, data, count, offset);
		return result;
	}
    
	result = exasmb_vfs_read(fsp->nfs_fh, sharedReadBufferID, offset, count, (int)fsp);
	DEBUG(2, ("exanet_readfile: result %d\n", result));
	if (result >= 0) {
		// copy the result from the shared memory to the buffer
		memcpy(data, sharedReadBuffer, result);
		return result;
	}
	
    switch (errno) {
	case VFS_FATAL_ERROR:
		 DEBUG(0, ("fs_pread fatal error, aborting\n"));    
		 exit_server("fs_pread fatal error");
		 break;
	case VFS_UNEXPECTED_ERROR:
		DEBUG(0,("fs_pread unexpected error\n"));
		break;
	default:{
		int preserve_errno = errno;
		DEBUG(0, ("fs_pread failed errno %d %s\n", errno, strerror(errno)));
		errno = preserve_errno;
		break;
	}
	
	}
	return -1;
}


ssize_t fs_sendfile(struct vfs_handle_struct *handle, int tofd, files_struct *fsp, int fromfd, const DATA_BLOB *header, SMB_OFF_T offset, size_t count)
{
    ssize_t result;
    
	DEBUG(2, ("fs_sendfile: in tofd %u, file %s offset %llu count %d\n", tofd, fsp->fsp_name,offset,count));
    if (fsp->fh->fd == fromfd && !lp_disable_vfsfs_write()) {
		int flags = 0;
		ssize_t ret;
		size_t hdr_len = 0;
		if (header) {
			size_t total = 0;
			hdr_len = header->length;
			while (total < hdr_len) {
				ret = sys_send(tofd, header->data + total,hdr_len - total,
							   MSG_MORE);
				if (ret == -1)
					return -1;
				total += ret;
			}
		}
        if (lp_fake_read_write()){
            static void *buff = NULL;
            DEBUG(1, ("FAKE READ !!!"));
    		if (!buff) {
    			buff = mmap(0, 512*1024, PROT_READ, MAP_ANONYMOUS | MAP_PRIVATE,
    						0, 0);
    			DEBUG(0, ("buff %p error %s\n", buff, strerror(errno)));
    		}
    		ret = send(tofd, buff, count, 0);
        }else{                                                      
    		static int lastfd;
    		static int lastfdflags;
            if (fsp->nfs_fh[0] == 0) {
                DEBUG(1,("Invalid nfs file handle before sendfile, attempting to acquire...\n"));
                if (exasmb_get_nfs_fh(fromfd, NULL, fsp->nfs_fh) == -1) {
                    result = SMB_VFS_NEXT_SENDFILE(handle, tofd, fsp, fromfd, header, offset, count);
                    return result;
                }
            }
            if( DEBUGLVL( 3 )) {
                int i_fh;
                pstring nfs_fh_string,temp_str;

                nfs_fh_string[0] = '\0';
                DEBUG(3,("fs_sendfile: fsp %p, fromfd=%d\n", fsp, fromfd));
                for (i_fh=0; i_fh<32; i_fh++) {                
                    snprintf (temp_str,PSTRING_LEN,"%s%.2X %s",
                        i_fh>0 ? nfs_fh_string : "", (unsigned char) fsp->nfs_fh[i_fh], (1+i_fh)%8==0 ? "\n" : "");
                    strncpy (nfs_fh_string,temp_str, PSTRING_LEN);
                }
                DEBUG(3,("fs_sendfile: This is the nfs file handle we got:\n%s\n", nfs_fh_string));                
            }

    		if (tofd == lastfd)
    			flags = lastfdflags;
    		else {
    			if ((flags = fcntl(tofd, F_GETFL)) < 0)
    				smb_panic("fs_sendfile: could not get fd flags\n");
    			lastfdflags = flags;
    			lastfd = tofd;
    		}
    		ret = exasmb_vfs_sendfile(tofd, fsp->nfs_fh, offset, count, (int)fsp);
        }
		if (flags >= 0) {
			if (flags & O_NONBLOCK)
				DEBUG(4, ("fs_sendfile: fd %d entered non-blocking\n", tofd));
			if (fcntl(tofd, F_SETFL, flags & (~O_NONBLOCK)) < 0)
				smb_panic("fs_sendfile: failed to restore fd blocking state\n");
		}
        DEBUG(4, ("exanet_sendfile: ret %d\n", ret));
        if (ret == -1) {
            if (errno > 0) {
               DEBUG(0, ("fs_sendfile failed errno %d %s\n", errno, strerror(errno)));
               return -1;
           }else if (errno == VFS_FATAL_ERROR) {
               /* bug20836: when client closes its connection (brutally), we can have
                * 'write_vector failed' fatal error
                * in exasmb_acl_server::sendfile and VFS_FATAL_ERROR result code.
                *
                * Evgeny: I think we should not panic in this case,
                * but just to fail the operation and let samba to handle it further.
                */
               DEBUG(1, ("fs_sendfile fatal error, trying to keep samba working\n"));    
               /* exit_server("fs_sendfile fatal error"); */   
               return -1;
		   }else if (errno == VFS_OPERATION_FATAL_ERROR) {
				const char* nas_volume_name;
				const char* nas_path = nas_path_and_volname(
					(fsp->conn ? fsp->conn->connectpath : NULL),
					fsp->fsp_name, &nas_volume_name);

		   		extern struct current_user current_user;
		   		char failure_reason[] = {"fs_sendfile RPC request timed out (backoff algorithm abort); "
					"cannot send the body of SENDFILE reply, terminating..."};
					
		   	    /* RPC backoff algorithm has aborted request on timeout;
					we've already sent header of reply, so we cannot go on now - we can only terminate connection */
				clientAccess_cifsClientDisconnectionOnFile_event(current_user.ut.uid, client_addr(),
					nas_path, nas_volume_name, failure_reason);
				exit_server_cleanly(failure_reason);
           }else if (errno == VFS_UNEXPECTED_ERROR) {
                DEBUG(0,("fs_sendfile unexpected error\n"));
                return -1;
		   } else if (errno == VFS_NOT_SUPPORTED) {
			   DEBUG(3, ("fs_sendfile not implemented, calling next\n"));
			   result = SMB_VFS_NEXT_SENDFILE(handle, tofd, fsp, fromfd, header,
											  offset, count);
			   DEBUG(2, ("exanet_sendfile: result %d\n", result));
			   return result;
		   }
        }
		result = count + hdr_len;
	} else {
		result = SMB_VFS_NEXT_SENDFILE(handle, tofd, fsp, fromfd, header,
									   offset, count);
	}
	DEBUG(2, ("exanet_sendfile: result %d\n", result));
    return result;
}

int fs_fsync(struct vfs_handle_struct *handle, files_struct *fsp, int fd)
{
    int not_supported = 0;
    if (fsp->need_sync) {
        DEBUG(3, ("vfs_fs fs_fsync: file %s, fd %d needs sync\n",fsp->fsp_name,fd));
        int ret = exasmb_vfs_fsync(fsp->nfs_fh, (int)fsp);
		
        if (ret == -1) {
            if (errno > 0) {
                DEBUG(0, ("exa_fsync failed errno %d %s\n", errno, strerror(errno)));
				/* Protocols supportability: alert if errno==EROFS */
				alert_ro_fs( (fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name);	
                return -1;
            }else if (errno == VFS_FATAL_ERROR) {
                DEBUG(0, ("exa_fsync fatal error, aborting\n"));    
                exit_server("exa_fsync fatal error");    
            }else if (errno == VFS_CLIENT_DISCONNECTED_ERROR) {
                /* should never happen */
                DEBUG(0, ("exa_fsync vfsclient disconnection error, aborting\n"));
                smb_panic("exa_fsync vfsclient disconnection error");    
            }else if (errno == VFS_UNEXPECTED_ERROR) {
                DEBUG(0,("exa_fsync unexpected error\n"));
                return -1;
            } else if (errno == VFS_NOT_SUPPORTED) {
                DEBUG(3,("fsync not supported, using next\n"));
                not_supported = 1;
            }
        } else {
            fsp->need_sync = False;
        }
        if (!not_supported) {
            return ret;
        }
    } else {
        DEBUG(3, ("vfs_fs fs_fsync: file %s, fd %d doesn't need sync\n",fsp->fsp_name,fd));
    }
    return SMB_VFS_NEXT_FSYNC(handle, fsp, fd);
}

int fs_close(struct vfs_handle_struct *handle, files_struct *fsp, int fd)
{
	int sync_ret = 0;
	int ret;
	DEBUG(2,("Closing %s\n",fsp->fsp_name));
	if (fsp->need_sync) {
		// bad - should call the appropriate function !!!
		// sync_ret = exasmb_vfs_fsync(fsp->nfs_fh, (int)fsp);
		sync_ret = fs_fsync(handle, fsp, fd);
	}
	ret = SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
	DEBUG(2, ("fs_close: ret %d sync_ret %d\n", ret, sync_ret));
    if (ret == 0 && sync_ret == 0) {
        return 0;
    }    
    if (sync_ret == -1) {
        if (errno > 0) {
            DEBUG(0, ("exa_fsync failed errno %d %s\n", errno, strerror(errno)));
            return -1;
        }else if (errno == VFS_FATAL_ERROR) {
            DEBUG(0, ("exa_fsync fatal error, aborting\n"));    
            exit_server("exa_fsync fatal error");    
        }else if (errno == VFS_CLIENT_DISCONNECTED_ERROR) {
            /* should never happen */
            DEBUG(0, ("exa_fsync vfsclient disconnection error, aborting\n"));
            smb_panic("exa_fsync vfsclient disconnection error");    
        }else if (errno == VFS_UNEXPECTED_ERROR) {
            DEBUG(0,("exa_fsync unexpected error\n"));
            return -1;
        }
    }
	
	return ret;
}

#define SLASHES_NUM_TO_SEARCH 5 /* Number of slashes in cut-off pattern '/mnt/cifs/%IP/vol_name/' */
/*************************************************************************
Name: fsSetFullSharePath
Arguments: new_path - a new full path, old_path - client delivered relative
           path, int snum - connection snum
Description: add share directory to the path name
Return Value: 0 on success, -1 on failure
***************************************************************************/
int fsSetFullSharePath(connection_struct* conn, pstring new_path, const char *old_path, int snum)
{
    pstring share_path;
    const char *p;
    const char *old_name=NULL;
    int counter = 0;
	int slashes_num = SLASHES_NUM_TO_SEARCH;
    /*int len = 0; */
    /* cut off the header './' of name */
    if ( (old_name = strstr(old_path, "./")) == old_path) 
        old_name += 2;
    else
        old_name = old_path;

	/* In case of C$ - the path is /mnt/cifs/%IP/ */
	if (strequal(lp_servicename(snum), "C$")) {
		slashes_num = SLASHES_NUM_TO_SEARCH - 1;
	}
	
    pstrcpy(share_path, conn->connectpath);
    /* cut off /mnt/cifs/%IP/vol_name/ */

	/* ExaNet 2008 / Samba upgrade project / stabilization stage
	   Re-make of cut-off algorithm. */
	p = share_path;
	while(*p!=0 && counter < slashes_num) /* Go thru the path until all slashes found or EOL */
    {
        if(*p++ == '/') counter++;
    }
    if (counter < (slashes_num-1) || /* Not enough slashes in path?! */
        (counter == (slashes_num-1) && *(p-1)=='/')) /* '/mnt/cifs/%IP/vol_name' case should be also acceptable */
    {                                                          /* but '/mnt/cifs/%IP/' shouldn't */
        DEBUG(0,("Invalid share path format: '%s', but should be "
			"like /mnt/cifs/IP/vol_name/\n", share_path));
        return -1;
    }
    pstrcpy(new_path, p);
    if(*p != 0)
    {
        /* If last char is not '/', force it to be appended! */
        if (p[strlen(p)- 1] != '/') {
            pstrcat(new_path, "/");
        }
    }
    pstrcat(new_path, old_name);
    return 0;
} 

int  fs_rename(vfs_handle_struct *handle, const char *from_name, const char *to_name)
{
    static EXA_NFS_FH last_share_nfs_fh;
    static pstring last_share_path;
    static int last_share_fd = 0;
    pstring current_share_path;
    int snum =  SNUM(handle->conn);
    ssize_t ret;
	const char *orig_from_name;
	const char *orig_to_name;
    
    pstring from_path;
    pstring to_path;

    DEBUG(4, ("exanet_rename: from_name: %s to name: %s \n", from_name, to_name));
    if (lp_fs_case_sensitive())
        return SMB_VFS_NEXT_RENAME(handle, from_name, to_name);

    /* Now check if the source and destination files are the same name but different case.
       If it is true, no need to rename FS_API rename ! */
    if ( (fsSetFullSharePath(handle->conn, from_path, from_name, snum)) != 0 ||
         (fsSetFullSharePath(handle->conn, to_path, to_name, snum)) !=0){
        return -1;
    }
    DEBUG(4, ("exanet_rename:  from_path %s\n", from_path));
    DEBUG(4, ("exanet_rename:  to_path %s\n", to_path));
    if (StrCaseCmp(from_path, to_path) != 0) {
        DEBUG(4,("exanet_rename not same path, go to next\n"));
        return SMB_VFS_NEXT_RENAME(handle, from_name, to_name);
    }


    /* Now we know we must go to the FS to perform rename... */
    pstrcpy(current_share_path, handle->conn->connectpath);
    
    if (!last_share_fd || strcmp (current_share_path, last_share_path)) {
        /* First we close any possibly opened fd, to another share */
        if (last_share_fd > 0)
            close(last_share_fd);

        /* Now open an fs to the current share and get the nfs file handle for it */
        pstrcpy(last_share_path, handle->conn->connectpath);
        last_share_fd = open(last_share_path, O_RDONLY);
        if (last_share_fd == -1) {
            last_share_fd = 0;
            DEBUG(0,("fsSetFullSharePath: open on share %s , path %s failed, %m\n",lp_servicename(snum), handle->conn->connectpath ));
            return -1;
        }
        if (exasmb_get_nfs_fh(last_share_fd, NULL, last_share_nfs_fh) == -1) {
            DEBUG(0,("fsSetFullSharePath: get_fd_nfs_handle failed - %m\n"));
            return -1;
        }
    }
    /* Now we have an updated nfs fh, we can use it to tell the FS the base location */
    // this is probably not true !!!   
	orig_from_name = from_name;
	orig_to_name = to_name;
    if ( (strstr(from_name, "./")) == from_name) 
        from_name += 2;
    if ( (strstr(to_name, "./")) == to_name) 
        to_name += 2;

    DEBUG(4, ("exanet_rename:  from_name %s, to_name %s\n", 
              from_name,to_name));
    DEBUG(4, ("exanet_rename:  from_path %s, to_path %s\n", 
              from_path,to_path));
    
    ret = exasmb_vfs_rename(last_share_nfs_fh, from_name, to_name);
    DEBUG(4, ("exa_rename: result %d\n", ret));
    if (ret == -1) {
        if (errno > 0) {
            DEBUG(0, ("exa_rename failed errno %d %s\n", errno, strerror(errno)));

			/* Protocols supportability: alert if errno==EROFS */
			alert_ro_fs( (handle->conn ? handle->conn->connectpath : NULL), from_name);	
            return -1;
        } else if (errno == VFS_FATAL_ERROR) {
            DEBUG(0, ("exa_rename fatal error, aborting\n"));    
            exit_server("exa_rename fatal error");    
        } else if (errno == VFS_UNEXPECTED_ERROR) {
            DEBUG(0,("exa_rename unexpected error\n"));
            return -1;
        } else if (errno == VFS_NOT_SUPPORTED) {
			DEBUG(3,("rename not supported, calling next\n"));
			return SMB_VFS_NEXT_RENAME(handle, orig_from_name, orig_to_name);
		}
    }
    return ret;
}     


static
void
init_stat_by_readdir_stat(const struct xattr_stat* xstat, SMB_STRUCT_STAT* sbuf){
	sbuf->st_dev = xstat->stat.st_dev;
	sbuf->st_mode = xstat->stat.st_mode;
	sbuf->st_nlink = xstat->stat.st_nlink;
	sbuf->st_uid = xstat->stat.st_uid;
	sbuf->st_gid = xstat->stat.st_gid;
	sbuf->st_rdev = xstat->stat.st_rdev;
	sbuf->st_size = xstat->stat.st_size;
	sbuf->st_blksize = xstat->stat.st_blksize;
	sbuf->st_blocks = xstat->stat.st_blocks;
	sbuf->st_atim = xstat->stat.st_atim;
	sbuf->st_mtim = xstat->stat.st_mtim;
	sbuf->st_ctim = xstat->stat.st_ctim;
	sbuf->st_ex_btime.tv_sec = xstat->createtime_sec;
	sbuf->st_ex_btime.tv_nsec = xstat->createtime_nsec;
}
static
BOOL
use_last_dir_stat(const char *fname, connection_struct *conn, BOOL islstat, SMB_STRUCT_STAT *sbuf)
{
	if (last_dir_read && last_dir_read->conn == conn) {
		exanet_dir *dirp = last_dir_read;
		int snum = SNUM(conn);
		int len;
		last_dir_read = NULL;
		if (lp_hide_special_files(snum) || lp_hideunreadable(snum) ||
				lp_hideunwriteable_files(snum))
			return False;
		len = dirp->path_len;
		DEBUG(3,("comparing fname %s to dirp->path %s\n", fname, dirp->path));
		if (strncmp(dirp->path, fname, len) == 0) {
			const char *s = &fname[len];
			while (*s == '/')
				s++;
			len = dirp->prev_name.len;
			if (strncmp(s, dirp->prev_name.name, len) ||
					s[dirp->prev_name.len])
				return False;;
		}
		if (!islstat && S_ISLNK(dirp->prev_stat->stat.st_mode))
			return False;
		
		init_stat_by_readdir_stat(dirp->prev_stat, sbuf);
		DEBUG(3, ("Got createtime for file %s, seconds %u, nseconds %u\n", fname, dirp->prev_stat->createtime_sec,
				  dirp->prev_stat->createtime_nsec ));
		DEBUG(5, ("use_last_dir_stat: dirp->prev_stat->st_dev=%d dirp->prev_stat->st_ino=%.0f dirp->prev_stat->__st_ino=%.0f\n",
			(unsigned int)dirp->prev_stat->stat.st_dev, (double)dirp->prev_stat->stat.st_ino, (double)dirp->prev_stat->stat.__st_ino));
		// Don't use real dev number from kernel
		// sbuf->st_dev = dirp->st_dev;
		sbuf->st_dev = 1; // Apply DLM hack
		sbuf->st_blksize = dirp->st_blksize;
		sbuf->st_blocks = (sbuf->st_size + (dirp->st_blksize-1)) / dirp->st_blksize;
		DEBUG(5, ("use_last_dir_stat: dev=%x st_ino=%.0f\n",
			(unsigned int)sbuf->st_dev, (double)sbuf->st_ino));
		return True;
	}
    else
        DEBUG(5, ("use_last_dir_stat for %s, last_dir_read=%p, last_dir_read->conn=%p, conn=%p\n", 
                  fname, last_dir_read, last_dir_read ? last_dir_read->conn : 0, conn));
	return False;
}

#define __u32 u_int32_t
#define __u64 u_int64_t

#pragma pack(1)
struct xattr_stat_create_time {
        __u32   st_dev;     /* ID of device containing file */
        __u64   st_ino;     /* inode number */
        __u32   st_mode;    /* protection */
        __u32   st_nlink;   /* number of hard links */
        __u32   st_uid;     /* user ID of owner */
        __u32   st_gid;     /* group ID of owner */
        __u64   st_rdev;    /* device ID (if special file) */
        __u64   st_size;    /* total size, in bytes */
        __u32   st_blksize; /* blocksize for filesystem I/O */
        __u64   st_blocks;  /* number of blocks allocated */
        __u32   st_atime_sec;   /* time of last access */
        __u32   st_atime_nsec;   /* time of last access */
        __u32   st_mtime_sec;   /* time of last modification */
        __u32   st_mtime_nsec;   /* time of last modification */
        __u32   st_ctime_sec;   /* time of last status change */
        __u32   st_ctime_nsec;   /* time of last status change */
        __u32   st_create_time_sec;   /* time of last status change */
        __u32   st_create_time_nsec;   /* time of last status change */
};
#pragma pack()

static
void
init_stat_by_kernel_stat(const struct xattr_stat_create_time* xstat, SMB_STRUCT_STAT* sbuf)
{
	
    sbuf->st_dev = 1;
	sbuf->st_ino = (ino_t) xstat->st_ino;
	sbuf->st_mode = (mode_t) xstat->st_mode;
	sbuf->st_nlink = (nlink_t) xstat->st_nlink;
	sbuf->st_uid = (uid_t)xstat->st_uid;
	sbuf->st_gid = (gid_t)xstat->st_gid;
	sbuf->st_rdev = (dev_t)xstat->st_rdev;
	sbuf->st_size = (off_t)xstat->st_size;
	sbuf->st_blksize = (blksize_t)xstat->st_blksize;
	sbuf->st_blocks = (blkcnt_t) xstat->st_blocks;
	sbuf->st_atim.tv_sec = (time_t) xstat->st_atime_sec;
	sbuf->st_atim.tv_nsec = (long) xstat->st_atime_nsec;
	sbuf->st_mtim.tv_sec = (time_t) xstat->st_mtime_sec;
	sbuf->st_mtim.tv_nsec = (long) xstat->st_mtime_nsec;
	sbuf->st_ctim.tv_sec = (time_t) xstat->st_ctime_sec;
	sbuf->st_ctim.tv_nsec = (long) xstat->st_ctime_nsec;
	sbuf->st_ex_btime.tv_sec = (time_t) xstat->st_create_time_sec;
	sbuf->st_ex_btime.tv_nsec = (long) xstat->st_create_time_nsec;
	sbuf->st_is_snapshot_dir = 0;
}

/*
	returns true
	if path matches (as whole) one of the following patterns

	#/.snapshots/X
	.snapshots/X

	but not

	#/.snapshots/.
	.snapshots/.

	where
		# is arbitrary (0+) number of any characters
		X is arbitrary (1+) number of any character
*/

static
BOOL
is_snapshot_dir_path(const char* path)
{
	const char snapshots[] = ".snapshots/";
	char* p_snapshots;

	p_snapshots = strstr(path, snapshots);

	return
		/* found */
		p_snapshots
		/* at the beginning of the string or has preceeding slash */
		&& (p_snapshots == path || p_snapshots[-1] == '/')
		/* there are chars after the match */
		&& p_snapshots[sizeof(snapshots) - 1]
		/* tail after match is not equal to "." */
		&& !(p_snapshots[sizeof(snapshots) - 1] == '.' && !p_snapshots[sizeof(snapshots)])
		;
}

static
int
fs_stat(struct vfs_handle_struct *handle, const char *fname, SMB_STRUCT_STAT *sbuf)
{
	const char * name = NULL;
	struct xattr_stat_create_time xt_stat;
	int ret;
	if (! fname) {
		return -1;
	}
	// cut off the header './' of fname
	if (strncmp(fname, "./", 2) == 0) {
		name = fname + 2;
	} else{
		name = fname;
	}
	DEBUG(5, ("fs_stat %s\n", name));
	if (use_last_dir_stat(name, handle->conn, False, sbuf))
		return 0;
	// else call to getxattr in kernel which will return both stat and create time.
	DEBUG(5, ("Calling getxattr stat for %s\n", fname));
	ret = getxattr(fname, "exastat", &xt_stat, sizeof(xt_stat));
	if (ret == -1) {
		if (errno == EOPNOTSUPP) {
			DEBUG(5, ("Failed to getxattr on file %s - operation not supported, calling the next VFS_STAT\n", fname));
			return SMB_VFS_NEXT_STAT(handle, fname, sbuf);
		}
		DEBUG(5, ("Failed to getxattr on file %s, error %d(%s)\n", fname, errno, strerror(errno)));
		return ret;
	}
	/* getxattr succeeded */
	DEBUG(5, ("Got reply from getxattr on file %s\n", fname));
	DEBUG(5, ("xstat dev %d ino %lld mode %o nlink %d uid %d gid  %d rdev %lld size %lld blksize %d blocks %lld atime %d %d mtime %d %d ctime %d %d cr time %d %d\n",
			  xt_stat.st_dev, xt_stat.st_ino, xt_stat.st_mode, xt_stat.st_nlink, xt_stat.st_uid, xt_stat.st_gid, xt_stat.st_rdev, xt_stat.st_size, xt_stat.st_blksize, xt_stat.st_blocks,xt_stat.st_atime_sec, xt_stat.st_atime_nsec, xt_stat.st_mtime_sec, xt_stat.st_mtime_nsec, xt_stat.st_ctime_sec, 
			  xt_stat.st_ctime_nsec,xt_stat.st_create_time_sec, xt_stat.st_create_time_nsec));
   
	init_stat_by_kernel_stat(&xt_stat, sbuf); 
	time_t temp_time = convert_timespec_to_time_t(sbuf->st_ex_btime);
	DEBUG(5, ("The result xstat create time on file %s is %s\n", fname, ctime(&temp_time)));
	
    if (S_ISDIR(sbuf->st_mode)) {
		sbuf->st_size = 0;
	}

	sbuf->st_is_snapshot_dir = S_ISDIR(sbuf->st_mode) && is_snapshot_dir_path(fname);

	DEBUG(10, ("fs_stat: %s, st_is_snapshot_dir = %d\n", fname, sbuf->st_is_snapshot_dir));
	 	
    return 0;
}

static
int
fs_lstat(struct vfs_handle_struct *handle, const char *path, SMB_STRUCT_STAT *sbuf)
{
	const char * cutpath = NULL;
	struct xattr_stat_create_time xt_stat;
	int ret;
	if (! path) {
		return -1;
	}
	// cut off the header './' of fname
	if (strncmp(path, "./", 2) == 0) {
		cutpath = path + 2;
	} else{
		cutpath = path;
	}
	if (use_last_dir_stat(cutpath, handle->conn, True, sbuf))
		return 0;

	// else call to getxattr in kernel which will return both stat and create time.
	DEBUG(5, ("Calling lgetxattr stat for %s\n", path));
	ret = lgetxattr(path, "exastat", &xt_stat, sizeof(xt_stat));
	if (ret == -1) {
		if (errno == EOPNOTSUPP) {
			DEBUG(5, ("Failed to lgetxattr on file %s - operation not supported, calling SMB_VFS_NEXT_LSTAT\n", path));
			return SMB_VFS_NEXT_LSTAT(handle, path, sbuf);
		}
		DEBUG(5, ("Failed to lgetxattr on file %s, error %d(%s)\n", path, errno, strerror(errno)));
		return ret;
	}
	/* lgetxattr succeeded */
	
    init_stat_by_kernel_stat(&xt_stat, sbuf); 
	if (S_ISDIR(sbuf->st_mode)) {
		sbuf->st_size = 0;
	}
	sbuf->st_dev = 1;
    return 0;
}

static
int
fs_fstat(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_STRUCT_STAT *sbuf)
{
	DEBUG(5,("Got fs_fstat request for filename %s\n", fsp->fsp_name));
	int ret;
	struct xattr_stat_create_time xt_stat;
	ret = fgetxattr(fd,"exastat", &xt_stat, sizeof(xt_stat));
	if (ret == -1) {
		if (errno == EOPNOTSUPP) {
			DEBUG(5, ("Failed to fgetxattr on file %s - operation not supported, calling SMB_VFS_NEXT_FSTAT\n", fsp->fsp_name));
			return SMB_VFS_NEXT_FSTAT(handle, fsp, fd, sbuf);
		}
		DEBUG(5, ("Failed to fgetxattr on file %s, error %d(%s)\n", fsp->fsp_name, errno, strerror(errno)));
		return ret;
	}
	/* fgetxattr succeeded */
	DEBUG(5, ("Got reply from fgetxattr on file %s\n", fsp->fsp_name));
	DEBUG(5, ("xstat dev %d ino %lld mode %o nlink %d uid %d gid  %d rdev %lld size %lld blksize %d blocks %lld atime %d %d mtime %d %d ctime %d %d cr time %d %d\n",
			  xt_stat.st_dev, xt_stat.st_ino, xt_stat.st_mode, xt_stat.st_nlink, xt_stat.st_uid, xt_stat.st_gid, xt_stat.st_rdev, xt_stat.st_size, xt_stat.st_blksize, xt_stat.st_blocks,xt_stat.st_atime_sec, xt_stat.st_atime_nsec, xt_stat.st_mtime_sec, xt_stat.st_mtime_nsec, xt_stat.st_ctime_sec, 
			  xt_stat.st_ctime_nsec,xt_stat.st_create_time_sec, xt_stat.st_create_time_nsec));
	init_stat_by_kernel_stat(&xt_stat, sbuf); 
	
	if (S_ISDIR(sbuf->st_mode)) {
		sbuf->st_size = 0;
	}
	sbuf->st_dev = 1;
    time_t temp_time = convert_timespec_to_time_t(sbuf->st_ex_btime);
	DEBUG(5, ("The result xstat create time on file %s is %s\n", fsp->fsp_name, ctime(&temp_time)));
	
	return 0;
}

static
int fs_ftruncate(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_OFF_T len)
{
	int result;
	DEBUG(5,("fs_ftruncate: in\n"));
    result = SMB_VFS_NEXT_FTRUNCATE(handle, fsp, fd, len);

    /* resetting the size to force a stat before the next sendfile */
    if (result != -1)
        fsp->size_for_sendfile = 0; 

	DEBUG(5,("fs_ftruncate: result %d\n", result));
    return result;
}

/*get snapshot id
return: 0 for current or id for snapshot*/
static uint32 fs_get_snapshot_id(char nfsfh[32] ){
    
    uint32 i = 0;
    i =  exasmb_vfs_get_snap_id(nfsfh);
    DEBUG(3,("fs_get_snapshot_id =%d\n",i));
    return i;
}
 
/*
 * This function will access the FileSystem and obtain the total size - dsize(in KB) and the free size - dfree(in KB)
 * bsize - block size. e.g. 1024 bytes
 * dsize - total size in bytes, in terms of bsize. e.g. with bsize of 1024 and dsize 512 the overall size is 512*1024 means 512KB
 * dfree - total free size in bytes, in terms of bsize, see example above
 */
SMB_BIG_UINT fs_disk_free(vfs_handle_struct *handle, const char *path, BOOL small_query, SMB_BIG_UINT *bsize, 
			       SMB_BIG_UINT *dfree, SMB_BIG_UINT *dsize)
{
	int result;
    static EXA_NFS_FH last_share_nfs_fh;
    static pstring last_share_path;
    static int last_share_fd = 0;
    pstring current_share_path;
    int snum =  SNUM(handle->conn);
    
    DEBUG(4, ("fs_disk_free: start\n"));

    pstrcpy(current_share_path, handle->conn->connectpath);
    
    if (!last_share_fd || strcmp (current_share_path, last_share_path)) {
        /* First we close any possibly opened fd, to another share */
        if (last_share_fd > 0)
            close(last_share_fd);

        /* Now open an fs to the current share and get the nfs file handle for it */
        pstrcpy(last_share_path, handle->conn->connectpath);
        last_share_fd = open(last_share_path, O_RDONLY);
        if (last_share_fd == -1) {
            last_share_fd = 0;
            DEBUG(0,("fs_disk_free: open on share %s , path %s failed, %m\n", lp_servicename(snum),handle->conn->connectpath ));
            return -1;
        }
        if (exasmb_get_nfs_fh(last_share_fd, NULL, last_share_nfs_fh) == -1) {
            DEBUG(0,("fs_disk_free: get_fd_nfs_handle failed - %m\n"));
            return -1;
        }
    }
    /* Now we have an updated nfs fh, we can use it to tell the FS the base location */
	result = exasmb_vfs_statfs(last_share_nfs_fh, bsize, dfree, dsize);
	return result;
}

static int
fs_fsetxattr(struct vfs_handle_struct *handle, struct files_struct *fsp, int fd, const char *name, const void *value, size_t size, int flags)
{
	int result = 0;
	if (name == NULL) {
		errno = EOPNOTSUPP;
		return  -1;
	}
	if (strcmp(SAMBA_XATTR_CREATE_TIME_EA_NAME, name) != 0) {
		return SMB_VFS_NEXT_FSETXATTR(handle,fsp, fd, name, value, size, flags);
    }
	/* else send create time to the ExaStore file system */
	const struct timespec* p_create_time = (const struct timespec*) value;
	DEBUG(3,("setting createtime: %s on file %s \n", time_to_asc(convert_timespec_to_time_t(*p_create_time)), fsp->fsp_name));
    if (fsp->nfs_fh[0] == 0) {
		DEBUG(5,("Invalid nfs file handle before vfs_set_create_time, attempting to acquire...\n"));
		if (exasmb_get_nfs_fh(fd, NULL, fsp->nfs_fh) == -1) {
			DEBUG(0,("FAILED acquiring nfs file handle before shwrite\n"));
			return -1;
		}
	}
	
	DEBUG(3,("setting createtime: %ld, %ld on file %s \n", p_create_time->tv_sec, p_create_time->tv_nsec, fsp->fsp_name));
    result = exasmb_vfs_set_create_time(fsp->nfs_fh, p_create_time->tv_sec, p_create_time->tv_nsec);
	if (result == 0) {
		// need to refresh the kernel create time cache - with a special refreshing uexastat attribute
		struct xattr_stat_create_time xt_stat;
		DEBUG(5, ("refreshing kernel create time on file %s\n",  fsp->fsp_name));
		int ret = fgetxattr(fd,"uexastat", &xt_stat, sizeof(xt_stat));
		if (ret == -1) {
    		        DEBUG(0, ("Failed to refresh kernel create time on file %s, error %d(%s)\n", fsp->fsp_name, errno, strerror(errno)));
		}
	} else {
		/* Protocols supportability: alert if errno==EROFS */
		alert_ro_fs( (fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name);
	}
	return result;
}



/*
  * ExaNet 2008 / Samba upgrade project / ADS (Alternate Data Streams)
  * Each Samba module a VFS module should have a init_module function if it's a shared module,
  * or smthing like vfs_example_init if it's staticly linked to samba
  */
NTSTATUS init_module(void)
{
    DEBUG(5,("vfs_fs_init:enter\n"));
	NTSTATUS ret = smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "fs", fs_ops);

    DEBUG(5,("vfs_fs_init:enter2\n"));

	if (!NT_STATUS_IS_OK(ret))
		return ret;
	
	vfs_fs_debug_level = debug_add_class("fs");
	if (vfs_fs_debug_level == -1) {
		vfs_fs_debug_level = DBGC_VFS;
		DEBUG(0, ("vfs_fs: Couldn't register custom debugging class!\n"));
	} else {
		DEBUG(10, ("vfs_fs: Debug class number of 'fs': %d\n", vfs_fs_debug_level));
	}

    if (! my_seekdir)
        my_seekdir = wrap_seekdir;
    get_snapshot_id =  fs_get_snapshot_id;

	if ( get_snapshots_timestamps == NULL ) {
		get_snapshots_timestamps = exasmb_get_snapshots_timestamps;
	}
	if ( get_snapshot_name_by_timestamp == NULL ) {
		get_snapshot_name_by_timestamp = exasmb_get_snapshot_name_by_timestamp;
	}

	if ( get_antivirus_info == NULL ) {
		get_antivirus_info = exasmb_get_antivirus_info;
	}
	if ( set_antivirus_info == NULL ) {
		set_antivirus_info = exasmb_set_antivirus_info;
	}
	if ( set_antivirus_info_guarded == NULL ) {
		set_antivirus_info_guarded = exasmb_set_antivirus_info_guarded;
	}

	return ret;
}

