/*
 * Old scanning mechanism for change notify feature.
 * Usefull on clustered systems when the notify tdb cannot be cluster-wide.
 *
 * Copyright (C) 2008, Dina Fine, dina@exanet.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "includes.h"


struct hash_watch_context {
	struct hash_watch_context *prev, *next;
	struct sys_notify_context *sys_ctx;
	void (*callback)(struct sys_notify_context *ctx, 
			 void *private_data,
			 struct notify_event *ev);
	struct notify_list *listel;
	uint32_t filter; /* the windows completion filter */
	struct timed_event *t_event;
	const char *path;
	struct connection_struct *conn;
	SMB_STRUCT_STAT dir_st; 
};
struct notify_list {
	struct notify_list *next, *prev;
	void *private_data;
	void (*callback)(void *, const struct notify_event *);
	void *sys_notify_handle;
	int depth;
};

static struct hash_watch_context *hash_notify_list;
#define NOTIFY_TIMEOUT 10

static BOOL
was_changed(struct hash_watch_context* watch)
{
	SMB_STRUCT_STAT dir_st;
    
	DEBUG(10,("watch %s\n", watch->path));

	if (SMB_VFS_STAT(watch->conn, watch->path, &dir_st) == -1) {
		DEBUG(1,("failed to stat directory %s, errno - %d (%s)\n", watch->path, errno, strerror(errno)));
		return True;
	}

	if ((watch->dir_st.st_mtime !=  dir_st.st_mtime) ||
	    (watch->dir_st.st_ctime != dir_st.st_ctime)) {
		watch->dir_st.st_mtime = dir_st.st_mtime;
		watch->dir_st.st_ctime = dir_st.st_ctime;
		DEBUG(5,("Changed %s\n", watch->path));
		return True;
	}
	DEBUG(5,("No change %s\n", watch->path));
	return False;
}


static void
timed_event_handler(struct event_context *event_ctx,struct timed_event *te,
		    const struct timeval *now,void *private_data)
{

	struct hash_watch_context *watch = (struct hash_watch_context *) private_data;
	TALLOC_FREE(watch->t_event);

	DEBUG(3,("handle watch path %s\n", watch->path));

	if (was_changed(watch)) {
		struct notify_event ne;
		ne.action = NOTIFY_ACTION_MODIFIED;
		ne.path = NULL;	/* will send a general notification */
		watch->callback(watch->sys_ctx, watch->listel, &ne);
	}

	DEBUG(10,("Re-insert the timed event on %s\n", watch->path));
	watch->t_event = event_add_timed(event_ctx, NULL, 
					 timeval_current_ofs(lp_parm_int(SNUM(watch->conn), "hash_notify", "timeout", NOTIFY_TIMEOUT),
							     0),                           
					 "change_notify_check",
					 timed_event_handler, watch);
	if (watch->t_event == NULL) {
		DEBUG(0,("Failed to add change notify timed event for directory %s\n", watch->path));
		return;
	}
}

static void 
dump_watch_list()
{
	struct hash_watch_context *watch_p;

	if (hash_notify_list == NULL) {
		DEBUG(10,("Watch List is empty\n"));
		return;
	}
	watch_p = hash_notify_list;
	while (watch_p != NULL) {
		DEBUG(10,("Watch on path: %s\n", watch_p->path));
		watch_p = watch_p->next;
	}
}

static int 
hash_watch_context_destructor(struct hash_watch_context *watch)
{
	DEBUG(10, ("hash_watch_context_destructor is called on watch path %s\n", watch->path));
	if ( DEBUGLVL( 10 )) {
		dump_watch_list();
	}

	TALLOC_FREE(watch->t_event);
	watch->t_event = NULL; 
	DLIST_REMOVE(hash_notify_list, watch);
	return 0;
}


/*
  add a watch. The watch is removed when the caller calls
  talloc_free() on *handle
*/
static NTSTATUS 
hash_watch(vfs_handle_struct *vfs_handle,
	   struct sys_notify_context *ctx,
	   struct notify_entry *e,
	   void (*callback)(struct sys_notify_context *ctx, 
			    void *private_data,
			    struct notify_event *ev),
	   void *private_data, 
	   void *handle_p)
{

	/* we will support only directory content related notifications */
	const uint32 watch_mask = (FILE_NOTIFY_CHANGE_FILE_NAME| FILE_NOTIFY_CHANGE_DIR_NAME);
	struct hash_watch_context *watch;
	void **handle = (void **)handle_p;

	DEBUG(3,("Add watch on %s\n", e->path));

	if (!(watch = TALLOC_P(ctx, struct hash_watch_context))) {
		DEBUG(0, ("No memory to allocate hash_watch_context\n"));
		return NT_STATUS_NO_MEMORY;
	}

	watch->callback = callback;
	watch->listel = (struct notify_list*) private_data;
	watch->sys_ctx = ctx;
	if (!(watch->path = talloc_strdup(watch, e->path))) {
		DEBUG(0, ("talloc_strdup failed\n"));
		TALLOC_FREE(watch);
		return NT_STATUS_NO_MEMORY;
	}
	watch->filter = watch_mask;
	watch->conn = ctx->conn;
	/* take the directory stat */
	if (SMB_VFS_STAT(ctx->conn, e->path, &(watch->dir_st))== -1) {
		DEBUG(1,("failed to stat directory %s, errno - %d (%s)\n", e->path, errno, strerror(errno)));
		TALLOC_FREE(watch);
		return NT_STATUS_UNSUCCESSFUL;
	}
	DEBUG(3,("Add timed event on %s\n", e->path));
	watch->t_event = event_add_timed(ctx->ev, NULL, 
					 timeval_current_ofs(lp_parm_int(SNUM(watch->conn), "hash_notify", "timeout", NOTIFY_TIMEOUT),
							     0),                                 
					 "change_notify_check",
					 timed_event_handler, watch);
	if (watch->t_event == NULL) {
		DEBUG(0, ("Failed to add change notify timed event for directory %s\n", watch->path));
		TALLOC_FREE(watch);
		return NT_STATUS_UNSUCCESSFUL;
	}

	DLIST_ADD(hash_notify_list, watch);
	talloc_set_destructor(watch, hash_watch_context_destructor);

	*handle = (void *)watch;
	/*  We do nulify all other bits - do not want the local notify.tdb be involved */
	e->filter = 0;
	e->subdir_filter = 0;
	return NT_STATUS_OK;

}



/* VFS operations structure */

static vfs_op_tuple notify_hash_op_tuples[] = {

	{SMB_VFS_OP(hash_watch),    SMB_VFS_OP_NOTIFY_WATCH, SMB_VFS_LAYER_OPAQUE},
	{SMB_VFS_OP(NULL),SMB_VFS_OP_NOOP,SMB_VFS_LAYER_NOOP}

};


NTSTATUS vfs_notify_hash_init(void);
NTSTATUS vfs_notify_hash_init(void)
{
	return  smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "notify_hash",
				 notify_hash_op_tuples);
}
