/*
   Unix SMB/CIFS implementation.
   SMB NT Security Descriptor / Unix permission conversion.
   Copyright (C) Jeremy Allison 1994-2000.
   Copyright (C) Andreas Gruenbacher 2002.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"
#include "smbutil/exasmb_vfsfsclient.h"
#include "smbutil/exasmb_aclclient.h"
#include "exasmb_if.h"
#include "exasmb/exasmb_id_mapping.h"
#include "events/EventsClientAccess_c.h"

#define	MAX_ACE_COUNT	200

extern struct current_user current_user;
extern struct generic_mapping file_generic_mapping;
extern int (*nt_access)(int, uint32, char[32], uint32_t*, char*);
extern int (*nt_path_access)(const char *path, uint32, uint32_t*, char*);
extern int (*allowed_delete)(connection_struct *, const char*);
static int vfs_unixntacl_debug_level = DBGC_VFS;

void my_ads_disp_sd(SEC_DESC *sd);
char *my_get_errno(char *reply, int reply_size ,int my_errno) {

    switch (my_errno) {
    case VFS_UNKNOWN_REQUEST_TYPE:
        snprintf(reply, reply_size, "Unknown Request Type");
        break;
    case VFS_INTERNAL_FS_PROBLEM:
        snprintf(reply, reply_size, "Internal FS Problem");
        break;
    case VFS_PERM_DENIED:
        snprintf(reply, reply_size, "Permission Denied");
        break;
    case VFS_READ_ERROR:
        snprintf(reply, reply_size, "Read Error");
        break;
    case VFS_WRITE_ERROR:
        snprintf(reply, reply_size, "Write Error");
        break;
    case VFS_FATAL_ERROR :
        snprintf(reply, reply_size, "Fatal Error");
        break;
    default:
        snprintf(reply, reply_size, "Don't know");
    }
    return reply;
}

static const char *strerror_extended_by_vfs_messages(int errno_ex)
{
	static char unknown_err_buf[100];
	
	if(errno_ex >= 0)
		return strerror(errno_ex);
	switch(errno_ex)
	{
		case VFS_FATAL_ERROR:
			return "VFS_FATAL_ERROR";
		case VFS_OPERATION_FATAL_ERROR:
			return "VFS_OPERATION_FATAL_ERROR";
		case VFS_CLIENT_DISCONNECTED_ERROR:
			return "VFS_CLIENT_DISCONNECTED_ERROR";
		case VFS_UNEXPECTED_ERROR:
			return "VFS_UNEXPECTED_ERROR";
		case VFS_UNKNOWN_REQUEST_TYPE:
			return "VFS_UNKNOWN_REQUEST_TYPE";
		case VFS_INTERNAL_FS_PROBLEM:
			return "VFS_INTERNAL_FS_PROBLEM";
		case VFS_READ_ERROR:
			return "VFS_READ_ERROR";
		case VFS_WRITE_ERROR:
			return "VFS_WRITE_ERROR";
		case VFS_PERM_DENIED:
			return "VFS_PERM_DENIED";
		case VFS_NOT_SUPPORTED:
			return "VFS_NOT_SUPPORTED";
	}

	/* Error not translated... */
	snprintf(unknown_err_buf, sizeof(unknown_err_buf), "Unknown error #%d", errno_ex);
	return unknown_err_buf;
}

void print_unix_acl(unixntacl_acl *acl)
{
	int i;
	DEBUG(10,("unixntacl_acl: type 0x%x aces: %d\n",
				 acl->type, acl->ace_count));
	for (i = 0; i < acl->ace_count; ++i) {
		unixntacl_acl_ace *ace = &acl->aces[i];
		char type[32];
		const int type_max = sizeof(type) - 1;
		
		switch (ace->type) {
		case ACCESS_ALLOWED_ACE_TYPE:
			safe_strcpy(type, "allow", type_max);
			break;
		case ACCESS_DENIED_ACE_TYPE:
			safe_strcpy(type, "deny", type_max);
			break;
		default:
			snprintf(type, sizeof(type), "unknown: %u", (int) ace->type);
		}
		DEBUGADD(10,("\tace: type %s flags: %08X mask: %08X trustee temp id: %u\n",
					 type, ace->flags, ace->mask, ace->trustee.id));
	}
}

struct knfs_fh {
	unsigned short size;
	unsigned char data[64];
};

static int acl_get_access(int 		h_len, 
						  char 	   *h_data, 
						  uint32_t 	uid, 
						  uint32_t 	gid,
						  int 		ngroups, 
						  uint32_t *groups, 
						  int		checkSpecialPerms,
						  uint32_t *max_access,
						  char		*access_troubleshooting_details)
{
    int ret = 0;
    fstring my_errno_str;
	// Does the user have permissions to change ownership?
	BOOL has_take_ownership_priv = user_has_privileges(current_user.nt_user_token,
												  &se_take_ownership);
	BOOL has_restore_priv = user_has_privileges(current_user.nt_user_token,
										   &se_restore);

   if ( uid == 0 ) {
      *max_access = FILE_GENERIC_ALL;
      DEBUG(6,("root user (uid: 0) is allowed all access (mask: 0x%x)\n", *max_access));
      return 0;
   }

   /* Check that the number of groups is not too big to fit in the VFS ACL message size */
	if (ngroups >= GROUPS_PER_USER) {
		DEBUG(1, ("acl_get_access may return incorrect permissions because number of groups (%d) is bigger than the maximum supported (%d)\n", ngroups, GROUPS_PER_USER));
	}

	// get ACLs from exafs
	ret = exasmb_acl_get_access(h_len, h_data, uid, gid, ngroups, groups,
		max_access, access_troubleshooting_details);
    if (ret != 0) {
        if (errno >= 0) {
            DEBUG(1, ("exasmb_acl_get_access failed, uid: %u gid: %u errno: %d (%s).\n", 
					  uid, gid, errno, strerror(errno) ));
        } else {
            DEBUG(1, ("exasmb_acl_get_access failed, uid: %u gid: %u errno: %d (%s).\n", 
					  uid, gid, errno, my_get_errno(my_errno_str, FSTRING_LEN, errno)));
        }
    } 
	
	// add the information about the user rights about ownership changing
	// root (uid == 0) is also given this permission. The main reason is that
	// later on try_chown will become root and some code paths will call this func again
	if (checkSpecialPerms && (uid == 0 || (has_take_ownership_priv || has_restore_priv))) {
		*max_access = *max_access | STD_RIGHT_WRITE_OWNER_ACCESS;
	}

	return ret;
}
/*
static int get_fd_nfs_handle(int fd, struct knfs_fh *nfs_fh)
{
    if (ioctl(fd, _IOR('u',1, sizeof(*nfs_fh)), nfs_fh) == -1) {
        DEBUG(0,("get_fd_nfs_handle: ioctl failed !  errno %d\n", errno));
        return -1;
    } else if( DEBUGLVL( 3 )) {
        int i;
        pstring nfs_fh_string,temp_str;
        nfs_fh_string[0] = '\0';
        for (i=0; i<32; i++) {
            snprintf (temp_str,PSTRING_LEN,"%s%.2X %s",
                i>0 ? nfs_fh_string : "", (unsigned char) nfs_fh->data[i], (1+i)%8==0 ? "\n" : "");
            strncpy (nfs_fh_string,temp_str, PSTRING_LEN);
        }
        DEBUG(3,("get_fd_nfs_handle: This is the nfs file handle we got:\n%s\n", nfs_fh_string));
    }
	return 0;
}
*/
static
int get_nfs_fh_access(struct knfs_fh *nfs_fh, uint32_t *granted, char *access_troubleshooting_details)
{
	DEBUG(2,("get_nfs_fh_access: \n"));
	if (acl_get_access(nfs_fh->size, (char*)nfs_fh->data, current_user.ut.uid,
					   current_user.ut.gid, current_user.ut.ngroups,
					   current_user.ut.groups, 1, granted, access_troubleshooting_details) == -1) {
		DEBUG(0,("check_nt_access: exasmb_acl_get_access failed\n"));
		return -1;
	}
	return 0;
}

static
int check_desired_access_ext(uint32_t granted, uint32_t access_mask, uint32_t *missing_bits)
{
	uint32_t effective_access;
	// w2k honor FILE_READ_ATTRIBUTES even if only FILE_READ_EA is set
	if (!(granted & FILE_READ_ATTRIBUTES) && (granted & FILE_READ_EA))
		effective_access = access_mask & (granted | FILE_READ_ATTRIBUTES);
	else
		effective_access = access_mask & granted;
	DEBUG(2,("check_desired_access: in desired 0x%08x granted 0x%08x effective 0x%08x\n",
			 access_mask, granted, effective_access));

    /* set missing bit */
    (*missing_bits) = access_mask & (~granted);
	if (effective_access == access_mask) {
		return 0;
	}
	DEBUG(2,("check_desired_access: failed 0x%08x != 0x%08x missing bits 0x%08x\n",
			 access_mask, effective_access, *missing_bits));

	errno = EACCES;
	return -1;
}

static
int check_desired_access(uint32_t granted, uint32_t access_mask)
{
    uint32_t missing_bits;
    return check_desired_access_ext(granted, access_mask, &missing_bits);
}


int unixntacl_access(ino_t inode, uint32 acc_desired, uint32 *acc_granted)
{
	*acc_granted = acc_desired;
	return 0;
}

int unix_to_sec_ace(unixntacl_acl_ace *ace, SEC_ACE *sec_ace)
{
    SEC_ACCESS acc;
	DOM_SID sid;
	int ret = 0;
	
	DEBUG(10,( "Mapping ACE trustee TEMP ID %u (as UID) to SID...\n", ace->trustee.id ));
	uid_to_sid( &sid, ace->trustee.id );
	if ( ret != 0 ) {
		DEBUG(2,( "Failed to map ACE trustee TEMP ID %u to SID. exasmb_map_temp_id_to_sid return value: %d, errno: %d (%s).\n", 
				  ace->trustee.id, ret, errno, strerror_extended_by_vfs_messages(errno) ));
		return ret;
	}
	DEBUG(10,( "Mapped ACE trustee TEMP ID %u to SID %s.\n", 
			   ace->trustee.id, sid_string_static(&sid) ));

	init_sec_access(&acc,ace->mask);
	init_sec_ace(sec_ace, &sid, ace->type, acc, ace->flags);
	
	return 0;
}

int unix_to_sec_acl(SEC_ACL **sec_acl, unixntacl_acl *acl)
{
	int i;
	SEC_ACE *nt_ace_list = (SEC_ACE *) calloc(acl->ace_count, sizeof(SEC_ACE));
	*sec_acl = NULL;
	if (!nt_ace_list)
		goto done;
	for (i = 0; i < acl->ace_count; ++i) {
		if (unix_to_sec_ace(&acl->aces[i], &nt_ace_list[i]) != 0)
			goto done;
	}
	*sec_acl = make_sec_acl(main_loop_talloc_get(), ACL_REVISION,
							acl->ace_count, nt_ace_list);
done:
	SAFE_FREE(nt_ace_list);
	return *sec_acl ? 0 : -1;
}



int sec_ace_to_unix(unixntacl_acl_ace *ace, SEC_ACE *sec_ace)
{
	BOOL map_ret = false;

    if (sec_ace->type != SEC_ACE_TYPE_ACCESS_ALLOWED &&
		sec_ace->type != SEC_ACE_TYPE_ACCESS_DENIED) {
		return 1;
	}

	DEBUG(10,( "Mapping ACE trustee SID %s to TEMP ID...\n", 
			   sid_string_static(&sec_ace->trustee) ));
	map_ret = sid_to_uid(&sec_ace->trustee, &ace->trustee.id);
	if ( ! map_ret ) {
		DEBUG(2,( "Failed to map ACE trustee SID %s to TEMP ID.\n", 
				   sid_string_static(&sec_ace->trustee) ));
		return 1;
	}
	DEBUG(10,( "Mapped ACE trustee SID %s to TEMP ID %u.\n", 
			   sid_string_static(&sec_ace->trustee), ace->trustee.id ));
	
	ace->type = sec_ace->type;
	ace->flags |= sec_ace->flags;
	se_map_generic(&sec_ace->access_mask, &file_generic_mapping);
	ace->mask = sec_ace->access_mask;
	
	return 0;
}

int sec_acl_to_unix(unixntacl_acl **acl, SEC_ACL *sec_acl)
{
	int i, j;
	unixntacl_acl *lacl = exasmb_acl_new_acl(0, sec_acl->num_aces);
	if (!lacl)
		return -1;
	for (i = 0, j = 0; i < sec_acl->num_aces; ++i) {
		switch (sec_ace_to_unix(&lacl->aces[j], &sec_acl->aces[i])) {
		case 0:
			++j;
			break;
		case 1:
			--lacl->ace_count;
			break;
		case -1:
		default:
			exasmb_acl_free_acl(lacl);
			return -1;
		}
	}
	*acl = lacl;
	return 0;
}

	
	
/****************************************************************************
 Function to create owner and group SIDs from a SMB_STRUCT_STAT.
****************************************************************************/

static void create_file_sids(SMB_STRUCT_STAT *psbuf, DOM_SID *powner_sid, DOM_SID *pgroup_sid)
{
	DEBUG(10,( "Mapping file owner UID %u to SID...\n", 
			   psbuf->st_uid ));
	uid_to_sid( powner_sid, psbuf->st_uid );
	DEBUG(10,( "Mapped file owner UID %u to SID %s.\n", 
			   psbuf->st_uid, sid_string_static(powner_sid) ));

	DEBUG(10,( "Mapping file group GID %u to SID...\n", 
			   psbuf->st_gid ));
	gid_to_sid( pgroup_sid, psbuf->st_gid );
	DEBUG(10,( "Mapped file group GID %u to SID %s...\n", 
			   psbuf->st_gid, sid_string_static(pgroup_sid) ));
}
	

/****************************************************************************
 Reply to query a security descriptor from an fsp. If it succeeds it allocates
 the space for the return elements and returns the size needed to return the
 security descriptor. This should be the only external function needed for
 the UNIX style get ACL.
****************************************************************************/

static
size_t my_get_nt_acl(struct vfs_handle_struct *handle, files_struct *fsp,
					 uint32 security_info, SEC_DESC **ppdesc)
{
    SMB_STRUCT_STAT sbuf;
	DOM_SID owner_sid;
	DOM_SID group_sid;
	size_t sd_size = 0;
	SEC_ACL *psa = NULL;
	SEC_DESC *psd = NULL;
	
	unixntacl_acl_ace l_aces[MAX_ACE_COUNT];
	unixntacl_acl l_acl = {0, MAX_ACE_COUNT, l_aces};
	unixntacl_acl *unix_acl = &l_acl;
	struct knfs_fh nfs_fh;
	uint32_t granted_access;
    fstring my_errno_str;

	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
	(void)EMPTY_DETAILS_STR(access_troubleshooting_details);
	
	*ppdesc = NULL;

	DEBUG(10,("get_nt_acl: called for file: %s directory: %s fd: %d\n",
			  fsp->fsp_name, fsp->is_directory ? "yes" : "no", fsp->fh->fd));

	if(fsp->is_directory || fsp->fh->fd == -1) {
		int	fd, ret;

		if ((fd = SMB_VFS_NEXT_OPEN(handle, fsp->fsp_name, fsp, O_RDONLY, 0)) == -1) {
			return 0;
		}
		/* Get the stat struct for the owner info. */
		ret = SMB_VFS_NEXT_FSTAT(handle, fsp, fd, &sbuf);
		if (ret == 0)
			ret = exasmb_get_nfs_fh(fd, &nfs_fh.size, (char*)nfs_fh.data);
		SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
		if (ret != 0)
			return 0;
		if (get_nfs_fh_access(&nfs_fh, &granted_access, access_troubleshooting_details) == -1)
			return 0;

	} else {

		/* Get the stat struct for the owner info. */
		if(SMB_VFS_NEXT_FSTAT(handle, fsp, fsp->fh->fd,&sbuf) != 0) {
			return 0;
		}
		if (exasmb_get_nfs_fh(fsp->fh->fd, &nfs_fh.size, (char*)nfs_fh.data) != 0) {
			return 0;
		}
		granted_access = fsp->access_mask;
	}
	// check access rights
	if (check_desired_access(granted_access, READ_CONTROL_ACCESS) == -1) {
		cifs_access_denied(GRANTED_ACCESS_NOT_ENOUGH,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "get_nt_acl",
			READ_CONTROL_ACCESS, granted_access, access_troubleshooting_details);
		return 0;
	}

	/*
	 * Get the owner, group and world SIDs.
	 */

	if (lp_profile_acls(SNUM(fsp->conn))) {
		/* For WXP SP1 the owner must be administrators. */
		sid_copy(&owner_sid, &global_sid_Builtin_Administrators);
		sid_copy(&group_sid, &global_sid_Builtin_Users);
	} else {
		create_file_sids(&sbuf, &owner_sid, &group_sid);
	}
	DEBUG(10,("my_get_nt_acl: security_info: 0x%x", security_info));
	if ((security_info & DACL_SECURITY_INFORMATION)/* && !(security_info & PROTECTED_DACL_SECURITY_INFORMATION)*/) {
		DEBUG(10,("my_get_nt_acl: name `%s` fd: %d dir: %s\n",
				  fsp->fsp_name, fsp->fh->fd, (fsp->is_directory ? "yes" : "no")));
        if (exasmb_acl_get_acl(nfs_fh.size, (char*) nfs_fh.data, &unix_acl) == -1) {
            if (errno >= 0) {
                DEBUG(1, ("exasmb_acl_get_acl failed errno %d %s\n", errno, strerror(errno)));
            }else{
                DEBUG(1, ("exasmb_acl_get_acl failed errno %d %s\n", errno, my_get_errno(my_errno_str, FSTRING_LEN, errno)));
            }
			return 0;
		}
		
		if (!unix_acl) {
			DEBUG(10,("get_nt_acl: unix_acl is NULL.\n"));
			goto no_acl;
		}
		print_unix_acl(unix_acl);
		if (unix_to_sec_acl(&psa, unix_acl) == -1) {
			DEBUG(10,("get_nt_acl: unix_to_sec_acl failed.\n"));
			goto done;
		}
	} /* security_info & DACL_SECURITY_INFORMATION */
no_acl:

	psd = make_standard_sec_desc( main_loop_talloc_get(),
			(security_info & OWNER_SECURITY_INFORMATION) ? &owner_sid : NULL,
			(security_info & GROUP_SECURITY_INFORMATION) ? &group_sid : NULL,
			psa,
			&sd_size);

	if(!psd) {
		DEBUG(0,("get_nt_acl: Unable to malloc space for security descriptor.\n"));
		sd_size = 0;
	} else {
		/*
		 * Windows 2000: The DACL_PROTECTED flag in the security
		 * descriptor marks the ACL as non-inheriting, i.e., no
		 * ACEs from higher level directories propagate to this
		 * ACL. In the POSIX ACL model permissions are only
		 * inherited at file create time, so ACLs never contain
		 * any ACEs that are inherited dynamically. The DACL_PROTECTED
		 * flag doesn't seem to bother Windows NT.
		 */
		if (unix_acl) {
			if (unix_acl->type & SE_DESC_DACL_PROTECTED)
				psd->type |= SE_DESC_DACL_PROTECTED;
			else if (unix_acl->type & SE_DESC_DACL_AUTO_INHERITED)
				psd->type |= SE_DESC_DACL_AUTO_INHERITED;
		}
		my_ads_disp_sd(psd);
	}

	*ppdesc = psd;

 done:
	 return sd_size;
}



/****************************************************************************
 Unpack a SEC_DESC into a UNIX owner and group.
****************************************************************************/

static BOOL unixntacl_unpack_nt_owners( int snum, SMB_STRUCT_STAT *psbuf, uid_t *puser, gid_t *pgrp, uint32 security_info_sent, SEC_DESC *psd)
{
	DOM_SID owner_sid;
	DOM_SID grp_sid;

	*puser = (uid_t)-1;
	*pgrp = (gid_t)-1;

	if(security_info_sent == 0) {
		DEBUG(0,("unpack_nt_owners: no security info sent !\n"));
		return True;
	}

	/*
	 * Validate the owner and group SID's.
	 */

	memset(&owner_sid, '\0', sizeof(owner_sid));
	memset(&grp_sid, '\0', sizeof(grp_sid));

	DEBUG(5,("unpack_nt_owners: validating owner_sids.\n"));

	/*
	 * Don't immediately fail if the owner sid cannot be validated.
	 * This may be a group chown only set.
	 */

	if (security_info_sent & OWNER_SECURITY_INFORMATION) {
		sid_copy(&owner_sid, psd->owner_sid);
		DEBUG(10,( "Mapping file owner SID %s to TEMP ID...\n", 
				   sid_string_static(&owner_sid) ));
		if (!sid_to_uid(&owner_sid, puser)) {
			DEBUG(2,( "Failed to map file owner SID %s to TEMP ID.\n", 
					   sid_string_static(&owner_sid) ));
            if (lp_force_unknown_acl_user(snum)) {
				/* this allows take ownership to work reasonably */
				*puser = current_user.ut.uid;
			} else {
				DEBUG(3,("unpack_nt_owners: unable to validate owner sid for %s\n",
					 sid_string_static(&owner_sid)));
				return False;
			}
		} else {
			DEBUG(10,( "Mapped file owner SID %s to TEMP ID %u.\n", 
					   sid_string_static(&owner_sid), *puser ));
		
			// Dina: if the sid is domain group sid or local group or well-known group, set both UID and GID of the file
			// for correct quota functionality inside the fsd
			enum lsa_SidType type;
			int win_ret = winbind_lookup_sid(NULL, &owner_sid, NULL, NULL, &type);
			if (win_ret){
				if (type == SID_NAME_DOM_GRP || type == SID_NAME_ALIAS || type == SID_NAME_WKN_GRP) {
                    			*pgrp = *puser;
					DEBUG(2, ("Giving away ownership to group sid %s, setting the group onwerhip as well\n", sid_string_static(&owner_sid)));
				}
			}
		}
 	}

	/*
	 * Don't immediately fail if the group sid cannot be validated.
	 * This may be an owner chown only set.
	 */

	if (security_info_sent & GROUP_SECURITY_INFORMATION) {
		sid_copy(&grp_sid, psd->group_sid);
		DEBUG(10,( "Mapping file group SID %s to TEMP ID...\n", 
				   sid_string_static(&grp_sid) ));
		if (!sid_to_gid( &grp_sid, pgrp)) {
			DEBUG(2,( "Failed to map file group SID %s to TEMP ID.\n", 
					   sid_string_static(&grp_sid) ));
			if (lp_force_unknown_acl_user(snum)) {
				/* this allows take group ownership to work reasonably */
				*pgrp = current_user.ut.gid;;
			} else {
				DEBUG(3,("unpack_nt_owners: unable to validate group sid.\n"));
				return False;
			}
		} else {
			DEBUG(10,( "Mapped file group SID %s to TEMP ID %u.\n", 
					   sid_string_static(&grp_sid), *pgrp ));
		}
	}

	DEBUG(5,("unpack_nt_owners: owner_sids validated.\n"));

	return True;
}

/* Protocols Supportability: dump alerts on access denied with details of failure
	for chown command: here is specific logic of chown flow. */
static void alert_chown_denied(const char *connectpath, const char *fname, uint32_t granted,
	char *access_troubleshooting_details, uid_t uid)
{
	if(uid == current_user.ut.uid) {	/* Case (2) - taking ownership logic of unixntacl_try_chown */
		/* Post supportability alert */
		cifs_access_denied(GRANTED_ACCESS_NOT_ENOUGH, connectpath, fname, "chown", 
			STD_RIGHT_WRITE_OWNER_ACCESS, granted,
			access_troubleshooting_details);
	}
	if(lp_enable_privileges()) {	/* Also alert lack of NTFS privileges */
		const char* nas_volume_name;
		const char* nas_path = nas_path_and_volname(connectpath, fname, &nas_volume_name);

		if(uid == current_user.ut.uid) {
			clientAccess_cifsUserHasNoPrivileges_event(
				"neither se_take_ownership nor se_restore", "chown",
				client_addr(), current_user.ut.uid, nas_path, nas_volume_name);
		} else { /* In case of giving ownership only restore privilege matters */
			clientAccess_cifsUserHasNoPrivileges_event(
				"se_restore", "chown",
				client_addr(), current_user.ut.uid, nas_path, nas_volume_name);
		}
	}
}

/****************************************************************************
 Try to chown a file. We will be able to chown it under the following conditions.

  1) If we have root privileges, then it will just work. - ExaNet fix: skip this case
  2) If we have ACL permissions to take ownership (change to current user only)
  3) If we have SeTakeOwnershipPrivilege we can change the user to the current user.
  4) If we have SeRestorePrivilege we can change the user to any other user.
  5) If we have write permission to the file and dos_filemodes is set
     then allow chown to the currently authenticated user.
****************************************************************************/

static int unixntacl_try_chown(connection_struct *conn, const char *fname, int fd, uid_t uid, gid_t gid)
{
	int ret;
	files_struct *fsp;
	SMB_STRUCT_STAT st;

	// ACL checking vars
	struct knfs_fh nfs_fh;
	uint32_t granted = 0;

	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
	(void)EMPTY_DETAILS_STR(access_troubleshooting_details);

	if(!CAN_WRITE(conn)) {
		return -1;
	}

	/* Case (2): Exanet bug21086 - applicable only for taking ownership, not for giving a file to other user */
	/* Also cannot change only group by such way. */
	if(uid == current_user.ut.uid) {
		/* Check if the user has permissions through ACLs*/
		if (exasmb_get_nfs_fh(fd, &nfs_fh.size, (char*)nfs_fh.data) == -1) {
			DEBUG(0,("try_chown: skipping user ACL test - get_fd_nfs_handle failed - %m\n"));
		} else {
			if (acl_get_access(nfs_fh.size, (char*) nfs_fh.data, current_user.ut.uid,
							   current_user.ut.gid, current_user.ut.ngroups,
							   current_user.ut.groups, 0, &granted,
							   access_troubleshooting_details) == -1) {
				DEBUG(0,("try_chown: skipping user ACL test - exasmb_acl_get_access failed\n"));
			} else {
				DEBUG(2,("try chown - ACL ownership - granted 0x%08x\n", granted));
				if (check_desired_access(granted, STD_RIGHT_WRITE_OWNER_ACCESS) == 0) { 
					become_root();
                    			ret = SMB_VFS_CHOWN(conn, fname, uid, gid);
					unbecome_root();
					return ret;
				} else {
					DEBUG(2,("try chown - ACL ownership - User doesn't have Take Ownership in ACLs\n"));
				}
			}
		}
	}

    /* Case (3) / (4) */
    if (lp_enable_privileges()) {

		BOOL has_take_ownership_priv = user_has_privileges(current_user.nt_user_token,
													  &se_take_ownership);
		BOOL has_restore_priv = user_has_privileges(current_user.nt_user_token,
											   &se_restore);
		DEBUG(6, ("lp_enable_privileges() is ON: has_take_ownership_priv=%u has_restore_priv=%u uid=%u current_user=%u\n",
			has_take_ownership_priv, has_restore_priv, uid, current_user.ut.uid));
			
		/* Case (3) */
		if ( ( has_take_ownership_priv && ( uid == current_user.ut.uid ) ) ||
		/* Case (4) */
			 ( has_restore_priv ) ) {
			 	DEBUG(6, ("making SMB_VFS_CHOWN due (3)/(4)\n"));
		
				become_root();
                		ret = SMB_VFS_CHOWN(conn, fname, uid, gid);
				unbecome_root();
				return ret;
		}
	}

	/* Case (5). */
	if (!lp_dos_filemode(SNUM(conn))) {
		/* Protocols Supportability - Access deined event occured */
		alert_chown_denied(conn->connectpath, fname, granted, access_troubleshooting_details, uid);
		return -1;
	}

	if (SMB_VFS_STAT(conn,fname,&st)) {
		return -1;
	}

	
	if (!NT_STATUS_IS_OK(open_file_fchmod(conn,fname,&st, &fsp))) {
		return -1;
	}

	/* only allow chown to the current user. This is more secure,
	   and also copes with the case where the SID in a take ownership ACL is
	   a local SID on the users workstation 
	*/
	uid = current_user.ut.uid;

	become_root();
	/* Keep the current file gid the same. */
	ret = SMB_VFS_FCHOWN(fsp, fsp->fh->fd, uid, (gid_t)-1);
	unbecome_root();

	close_file_fchmod(fsp);

	return ret;
}

/****************************************************************************
 Reply to set a security descriptor on an fsp. security_info_sent is the
 description of the following NT ACL.
 This should be the only external function needed for the UNIX style set ACL.
****************************************************************************/
static
BOOL my_set_nt_acl(struct vfs_handle_struct *handle, files_struct *fsp,
				   uint32 security_info_sent, SEC_DESC *psd)
{
	/* what user is current? */
	extern struct current_user current_user;
	
	connection_struct *conn = fsp->conn;
	uid_t user = (uid_t)-1;
	gid_t grp = (gid_t)-1;
	SMB_STRUCT_STAT sbuf;  
	mode_t orig_mode = (mode_t)0;
	uid_t orig_uid;
	gid_t orig_gid;
	BOOL need_chown = False;
	unixntacl_acl *unix_acl = NULL;
	BOOL result = True;
	struct knfs_fh nfs_fh;
	uint32_t granted_access;
    fstring my_errno_str;
	int fd = -1;

	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
	(void)EMPTY_DETAILS_STR(access_troubleshooting_details);

	DEBUG(10,("set_nt_acl: called for file %s\n", fsp->fsp_name ));

	if (!CAN_WRITE(conn)) {
		DEBUG(10,("set acl rejected on read-only share\n"));
		return False;
	}

	/*
	 * Get the current state of the file.
	 */

	if(fsp->is_directory || fsp->fh->fd == -1) {
		int ret;

		if ((fd = SMB_VFS_NEXT_OPEN(handle, fsp->fsp_name, fsp, O_RDONLY, 0)) == -1)
			return False;
		ret = SMB_VFS_NEXT_FSTAT(handle, fsp, fd, &sbuf);
		if (ret == 0)
			ret = exasmb_get_nfs_fh(fd, &nfs_fh.size, (char*)nfs_fh.data);
		else {
			SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
			return False;
		}
		if (get_nfs_fh_access(&nfs_fh, &granted_access, access_troubleshooting_details) == -1)
			goto cleanup_and_fail;
	} else {
		if(SMB_VFS_NEXT_FSTAT(handle, fsp, fsp->fh->fd, &sbuf) != 0)
			return False;
		if (exasmb_get_nfs_fh(fsp->fh->fd, &nfs_fh.size, (char*)nfs_fh.data) != 0) {
			return False;
		}
		granted_access = fsp->access_mask;
	}

	/* Save the original elements we check against. */
	orig_mode = sbuf.st_mode;
	orig_uid = sbuf.st_uid;
	orig_gid = sbuf.st_gid;

	/*
	 * Unpack the user/group/world id's.
	 */

	if (!unixntacl_unpack_nt_owners( SNUM(conn), &sbuf, &user, &grp, security_info_sent, psd))
		goto cleanup_and_fail;

	/*
	 * Do we need to chown ?
	 */

	if (((user != (uid_t)-1) && (orig_uid != user)) || (( grp != (gid_t)-1) && (orig_gid != grp)))
		need_chown = True;

	// check all needed access rights before we do anything
	if ((security_info_sent & DACL_SECURITY_INFORMATION) && psd->dacl &&
			(check_desired_access(granted_access, WRITE_DAC_ACCESS) == -1)) {
		DEBUG(2,("my_set_nt_acl: failed - no WRITE_DAC_ACCESS\n"));
		cifs_access_denied(GRANTED_ACCESS_NOT_ENOUGH,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "set_nt_acl", 
			WRITE_DAC_ACCESS, granted_access, access_troubleshooting_details);
		goto cleanup_and_fail;
	}
	if (need_chown && (check_desired_access(granted_access, WRITE_OWNER_ACCESS) == -1)) {
		DEBUG(2,("my_set_nt_acl: failed - no WRITE_OWNER_ACCESS\n"));
		cifs_access_denied(GRANTED_ACCESS_NOT_ENOUGH,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "set_nt_acl",
			WRITE_OWNER_ACCESS, granted_access, access_troubleshooting_details);
		goto cleanup_and_fail;
	}
	/*
	 * Chown before setting ACL only if we don't change the user, or
	 * if we change to the current user, but not if we want to give away
	 * the file.
	 */

	if (need_chown && (user == (uid_t)-1 || user == current_user.ut.uid)) {

		DEBUG(3,("set_nt_acl: chown %s. uid = %u, gid = %u, orig_uid=%u, current_user_uid=%u\n",
				fsp->fsp_name, (unsigned int)user, (unsigned int)grp,
				(unsigned int)orig_uid, (unsigned int)current_user.ut.uid ));

		if(unixntacl_try_chown( fsp->conn, fsp->fsp_name,
			(fsp->fh->fd == -1) ? fd : (fsp->fh->fd), user, grp) == -1) {
			DEBUG(3,("set_nt_acl: chown %s, %u, %u failed. Error = %s.\n",
				fsp->fsp_name, (unsigned int)user, (unsigned int)grp, strerror(errno) ));
			goto cleanup_and_fail;
		}

		/*
		 * Recheck the current state of the file, which may have changed.
		 * (suid/sgid bits, for instance)
		 */

		if(fsp->is_directory || fsp->fh->fd == -1) {
			if(SMB_VFS_NEXT_STAT(handle, fsp->fsp_name, &sbuf) != 0)
				goto cleanup_and_fail;
		} else {
			if(SMB_VFS_NEXT_FSTAT(handle, fsp,fsp->fh->fd,&sbuf) != 0)
				goto cleanup_and_fail;
		}

		/* Save the original elements we check against. */
		orig_mode = sbuf.st_mode;
		orig_uid = sbuf.st_uid;
		orig_gid = sbuf.st_gid;

		/* We did it, don't try again */
		need_chown = False;
	}

	
	/*
	 * If no DACL then this is a chown only security descriptor.
	 */
	if ((security_info_sent & DACL_SECURITY_INFORMATION) && psd->dacl) {
      if (psd->dacl->num_aces > MAX_ACE_COUNT) {
         DEBUG(0, ("set ACL failed: number of ACEs (%u) is bigger than maximum supported (%u).\n",
                   psd->dacl->num_aces, MAX_ACE_COUNT));
         goto cleanup_and_fail;
      }
		if (sec_acl_to_unix(&unix_acl, psd->dacl) == -1)
			goto cleanup_and_fail;
		unix_acl->type = psd->type;
		DEBUG(10,("my_set_nt_acl: sending..."));
		print_unix_acl(unix_acl);
		if (exasmb_acl_set_acl(nfs_fh.size, (char*)nfs_fh.data, unix_acl) == -1){
            if (errno >= 0) {
                DEBUG(1, ("exasmb_acl_set_acl failed errno %d %s\n", errno, strerror(errno)));
				alert_ro_fs(conn->connectpath, fsp->fsp_name);	// Protocols supportability: alert if errno==EROFS
            }else{
                DEBUG(1, ("exasmb_acl_set_acl failed errno %d %s\n", errno, my_get_errno(my_errno_str, FSTRING_LEN, errno)));
            }
            result = False;
        }
		exasmb_acl_free_acl(unix_acl);
	}
	/* Any chown pending? */
	if (result && need_chown) {
		DEBUG(3,("my_set_nt_acl: chown %s. uid = %u, gid = %u, orig_uid=%u, current_user_uid=%u.\n",
			fsp->fsp_name, (unsigned int)user, (unsigned int)grp,
			(unsigned int)orig_uid, (unsigned int)current_user.ut.uid ));

		if(unixntacl_try_chown( fsp->conn, fsp->fsp_name, 
			(fsp->fh->fd == -1) ? fd : (fsp->fh->fd), user, grp) == -1) {
			DEBUG(3,("my_set_nt_acl: chown %s, %u, %u failed. Error = %s.\n",
				fsp->fsp_name, (unsigned int)user, (unsigned int)grp, strerror(errno) ));
			goto cleanup_and_fail;
		}
	}

	if(fd != -1)
		SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
	return result;
	
cleanup_and_fail:
	if(fd != -1)
		SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
	return False;
	
}


static
size_t unixntacl_fget_nt_acl(struct vfs_handle_struct *handle, struct files_struct *fsp, int fd,  uint32 security_info, struct security_descriptor_info **ppdesc)
{
	return my_get_nt_acl(handle, fsp, security_info, ppdesc);
}
static
size_t unixntacl_get_nt_acl(struct vfs_handle_struct *handle, struct files_struct *fsp, const char *name,  uint32 security_info, struct security_descriptor_info **ppdesc)
{
	return my_get_nt_acl(handle, fsp, security_info, ppdesc);
}
static
BOOL unixntacl_fset_nt_acl(struct vfs_handle_struct *handle, struct files_struct *fsp, int fd, uint32 security_info_sent, struct security_descriptor_info *psd)
{
	return my_set_nt_acl(handle, fsp, security_info_sent, psd);
}
static
BOOL unixntacl_set_nt_acl(struct vfs_handle_struct *handle, struct files_struct *fsp, const char *name, uint32 security_info_sent, struct security_descriptor_info *psd)
{
	return my_set_nt_acl(handle, fsp, security_info_sent, psd);
}

static
int check_nt_access(int fd, uint32 access_mask,char nfsfh[32], uint32_t *missing_bits,
	char *access_troubleshooting_details)
{
	struct knfs_fh nfs_fh;
	uint32_t granted;
	
	DEBUG(2,("check_nt_access: fd %d  access_mask 0x%08X\n", fd, access_mask));
	if (exasmb_get_nfs_fh(fd, &nfs_fh.size, (char*)nfs_fh.data) == -1) {
		DEBUG(1,("check_nt_access: get_fd_nfs_handle failed - %m\n"));
		return -1;
	}
    DEBUG(2,("nfs_fh.data %d %d\n", nfs_fh.data[0], nfs_fh.data[1]));
   	if (nfsfh)
		memcpy(nfsfh, nfs_fh.data, 32); // here is the bug !  need to copy after the size
	if (acl_get_access(nfs_fh.size, (char*)nfs_fh.data, current_user.ut.uid,
					   current_user.ut.gid, current_user.ut.ngroups,
					   current_user.ut.groups, 1, &granted, access_troubleshooting_details) == -1) {
		DEBUG(0,("check_nt_access: exasmb_acl_get_access failed\n"));
		return -1;
	}
	
	return check_desired_access_ext(granted, access_mask, missing_bits);
}

static
int check_nt_path_access_wrapped(const char* current_dir, const char *fname, uint32_t access_mask,
	uint32_t *missing_bits, char* access_troubleshooting_details)
{
	uint32_t granted;
	int err = 0;
	struct knfs_fh nfs_fh;
	int fd = -1;
    int ret = 0;

	if (current_dir) {
		DEBUG(2,("check_path_access: %s (%s):\n", fname, current_dir));
	} else {
		DEBUG(2,("check_path_access: %s (error getting pwd):\n", fname));
	}

	fd = open(fname, O_RDONLY);
	if (fd == -1) {
		/* bug19968: ADS streams should NOT claim here (files with ':' actually don't exist) */
		if(errno != ENOENT || strchr(fname, ':')==NULL) {
			DEBUG((errno == ENOENT ? 1 : 0),
				("check_nt_path_access: open failed - %m, fname=%s\n", fname));
			
			/* log_stack_trace();*/ /* bug19968: Don't see any necessity in such backtrace now */
		}
		
		return 0; /* Return 'access granted' if no such file ?! */
	}
	if (exasmb_get_nfs_fh(fd, &nfs_fh.size, (char*)nfs_fh.data) == -1)
		err = errno;
	close(fd);
	if (err) {
		errno = err;
		DEBUG(1,("check_path_access: get_fd_nfs_handle failed - %m\n"));
		return -1;
	}
	if (acl_get_access(nfs_fh.size, (char*)nfs_fh.data, current_user.ut.uid,
					   current_user.ut.gid, current_user.ut.ngroups,
					   current_user.ut.groups, 1, &granted, access_troubleshooting_details) == -1) {
		DEBUG(0,("check_path_access: exasmb_acl_get_access failed\n"));
		return -1;
	}
	ret = check_desired_access_ext(granted, access_mask, missing_bits);

    DEBUG(10,("check_nt_path_access_wrapped returns %d for path [%s] \n", ret, fname));

    return ret;
}

extern int (*win_prev_vers_translate_path_p)(TALLOC_CTX *ctx, const char *path, const char *fname, char **new_path);

static
int check_nt_path_access(const char *fname, uint32_t access_mask,
	uint32_t *missing_bits, char *access_troubleshooting_details)
{
	int ret = 0;
	const char* current_dir = NULL;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* translated_path = fname;

	current_dir = get_current_dir_name();
	if ( current_dir == NULL ) {
		goto path_translated;
	}

	if ( win_prev_vers_translate_path_p == NULL ) {
		goto path_translated;
	}

	local_mem_ctx = talloc_init( __FUNCTION__ );
	if ( local_mem_ctx == NULL ) {
		DEBUG( 0, ( "Failed to allocate local_mem_ctx. Leaving path as it's (\"%s\").\n",
					fname ) );
		goto path_translated;
	}

	ret = win_prev_vers_translate_path_p( local_mem_ctx, current_dir, fname, (char**)( &translated_path ) );
	if ( ret != 0 ) {
		DEBUG(10,( "win_prev_vers_translate_path didn't translate path (\"%s\"+\"%s\").\n", 
				   current_dir, fname ));
		goto path_translated;
	}

path_translated:

	ret = check_nt_path_access_wrapped(current_dir, translated_path, access_mask,
		missing_bits, access_troubleshooting_details);

	free( (char*)(current_dir) );
	TALLOC_FREE( local_mem_ctx );
	return ret;
}

static char *container_path(const char *path)
{
	static char str[1024];
	char *s;
	if (!path)
		return str;
	safe_strcpy(str, path, sizeof(str)-1);
	while ((s = strrchr(str, '/')) && s != str && s[1] == '\0') {
		*s = '\0';
	}
	if (s)
		*s = '\0';
	else
		*(uint16_t *) str = *(uint16_t *) ".";
	return str;
}

static int check_acl_for_delete(connection_struct *conn, const char *path)
{
	char path_container[256];
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];

    DEBUG(2,("check_acl_for_delete: in\n"));
    DEBUG(10,("check_acl_for_delete: path=%s\n",path));

    safe_strcpy(path_container, container_path(path), sizeof(path_container)-1);

    /*to delete file or empty directory the minimum access that need is:
     Parent directory with list
    and one of the two: or parent with delete_sub_folders permission or file with delete permission */

    if (check_nt_path_access(path_container, FILE_LIST_DIRECTORY,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		DEBUG(2,("check_acl_for_delete: check_nt_path_access for FILE_LIST_DIRECTORY is -1. path=%s not a dir\n",path));
		cifs_access_denied(MISSED_BITS,
			(conn ? conn->connectpath : NULL), path_container, "check_acl_for_delete",
			FILE_LIST_DIRECTORY, missing_bits, access_troubleshooting_details);
		return -1;
	}

	if ((check_nt_path_access(path_container, FILE_DELETE_CHILD|FILE_LIST_DIRECTORY,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) ) {
		uint32_t missing_bits2 = 0;
		char access_troubleshooting_details2[MAX_ACCESS_DETAILS_STR_LEN + 1];
        if(check_nt_path_access(path, DELETE_ACCESS,
			&missing_bits2, EMPTY_DETAILS_STR(access_troubleshooting_details2)) == -1) {
			DEBUG(2,("check_acl_for_delete: no acl permissions to delete: '%s'\n", path));

			/* Post two supportability alerts - one for file and one for container */
			cifs_access_denied(MISSED_BITS,
				(conn ? conn->connectpath : NULL), path, "check_acl_for_delete",
				DELETE_ACCESS, missing_bits2, access_troubleshooting_details2);
			cifs_access_denied(MISSED_BITS,
				(conn ? conn->connectpath : NULL), path_container, "check_acl_for_delete",
				FILE_DELETE_CHILD|FILE_LIST_DIRECTORY,
				missing_bits, access_troubleshooting_details);

			return -1;
		}
	}

	DEBUG(10,("check_acl_for_delete: SUCCESS. got acl permissions to delete: '%s'\n",path));

	return 0;
}

const int FCREAT_ACCESS = (FILE_ADD_FILE);
const int READ_DIR_ACCESS =	(FILE_LIST_DIRECTORY);
// currently ingnoring FILE_TRAVERSE
static
DIR *unixntacl_opendir(vfs_handle_struct *handle, const char *fname, const char *mask, uint32 attr)
{
    DIR *result;
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];

	DEBUG(2,("unixntacl_opendir: in (%s)\n", fname));
	if (check_nt_path_access(fname, READ_DIR_ACCESS,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), fname, "opendir",
			READ_DIR_ACCESS, missing_bits, access_troubleshooting_details);
		return NULL;
	}
    result = SMB_VFS_NEXT_OPENDIR(handle, fname, mask, attr);
	DEBUG(2,("unixntacl_opendir: result %p\n", result));
    return result;
}


static
int unixntacl_mkdir(vfs_handle_struct *handle, const char *path, mode_t mode)
{
	const int MKDIR_ACCESS =    FILE_ADD_SUBDIRECTORY/* |
								FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA*/;
	int result;
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
	char *cont_path = container_path(path);

	DEBUG(2,("unixntacl_mkdir: in (%s)\n", path));
	if (check_nt_path_access(cont_path, MKDIR_ACCESS,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), cont_path, "mkdir",
			MKDIR_ACCESS, missing_bits, access_troubleshooting_details);	
		return -1;
	}
    result = SMB_VFS_NEXT_MKDIR(handle, path, mode);
	DEBUG(2,("unixntacl_mkdir: result %d\n", result));
    return result;
}

static
int unixntacl_rmdir(vfs_handle_struct *handle, const char *path)
{
	int result;
	DEBUG(2,("unixntacl_rmdir: in (%s)\n", path));
    if(check_acl_for_delete(handle->conn, path))
		return -1;

    result = SMB_VFS_NEXT_RMDIR(handle, path);
	DEBUG(2,("unixntacl_rmdir: result %d\n", result));
    return result;
}


static
int unixntacl_open(vfs_handle_struct *handle,  const char *fname,
	files_struct *fsp, int flags, mode_t mode)
{
	int result;
	SMB_STRUCT_STAT sbuf;
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];

	DEBUG(2,("unixntacl_open: in (%s)\n", fname));
	if ((flags & O_CREAT) && SMB_VFS_NEXT_STAT(handle, fname, &sbuf) != 0) {
		char *cont_path = container_path(fname);
		DEBUG(2,("unixntacl_open: check_path_access(%s)\n", cont_path));
		if (check_nt_path_access(cont_path, FCREAT_ACCESS,
			&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
			cifs_access_denied(MISSED_BITS,
				(fsp->conn ? fsp->conn->connectpath : NULL), cont_path, "open",
				FCREAT_ACCESS, missing_bits, access_troubleshooting_details);
			return -1;
		}
	}
    result = SMB_VFS_NEXT_OPEN(handle, fname, fsp, flags, mode);
	DEBUG(2,("unixntacl_open: result %d (%d)\n", result, errno));
    return result;
}


const int READ_ACCESS = (FILE_READ_DATA/*|FILE_READ_EA|FILE_READ_ATTRIBUTES*/);

static
ssize_t unixntacl_read(vfs_handle_struct *handle, files_struct *fsp, int fd, void *data, size_t n)
{
	ssize_t result;
	uint32_t missing_bits;
	if (check_desired_access_ext(fsp->access_mask, READ_ACCESS|FILE_EXECUTE, &missing_bits) == -1 &&
		missing_bits == (READ_ACCESS|FILE_EXECUTE)) {
		cifs_access_denied(OPEN_MODE_CONFLICT,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_read",
			READ_ACCESS|FILE_EXECUTE, fsp->access_mask, NULL);
		return -1;
	}
	DEBUG(2,("unixntacl_read: in\n"));
    result = SMB_VFS_NEXT_READ(handle, fsp, fd, data, n);
	DEBUG(2,("unixntacl_read: result %lld\n", (long long)result));
    return result;
}

static
ssize_t unixntacl_pread(vfs_handle_struct *handle, files_struct *fsp, int fd,
						void *data, size_t n, SMB_OFF_T offset)
{
	ssize_t result;
	uint32_t missing_bits;
	if (check_desired_access_ext(fsp->access_mask, READ_ACCESS|FILE_EXECUTE, &missing_bits) == -1 &&
		missing_bits == (READ_ACCESS|FILE_EXECUTE)) {
		cifs_access_denied(OPEN_MODE_CONFLICT,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_pread",
			READ_ACCESS|FILE_EXECUTE, fsp->access_mask, NULL);
		return -1;
	}
	DEBUG(2,("unixntacl_pread: in\n"));
    result = SMB_VFS_NEXT_PREAD(handle, fsp, fd, data, n, offset);
	DEBUG(2,("unixntacl_pread: result %lld\n", (long long)result));
    return result;
}

static
ssize_t unixntacl_write(vfs_handle_struct *handle, files_struct *fsp, int fd,
						const void *data, size_t n)
{
	ssize_t result;
	uint32_t missing_bits;
	DEBUG(2,("unixntacl_write: in\n"));
	if (check_desired_access_ext(fsp->access_mask, FILE_WRITE_DATA|FILE_APPEND_DATA, &missing_bits) == -1 &&
		missing_bits == (FILE_WRITE_DATA|FILE_APPEND_DATA) ) {
		cifs_access_denied(OPEN_MODE_CONFLICT,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_write",
			FILE_WRITE_DATA|FILE_APPEND_DATA, fsp->access_mask, NULL);
		return -1;
	}
    result = SMB_VFS_NEXT_WRITE(handle, fsp, fd, data, n);
	DEBUG(2,("unixntacl_write: result %lld\n", (long long)result));
    return result;
}

static
ssize_t unixntacl_pwrite(vfs_handle_struct *handle, files_struct *fsp, int fd,
						 const void *data, size_t n, SMB_OFF_T offset)
{
	ssize_t result;
	uint32_t missing_bits;
	DEBUG(2,("unixntacl_pwrite: in\n"));
    if (check_desired_access_ext(fsp->access_mask, FILE_WRITE_DATA|FILE_APPEND_DATA, &missing_bits) == -1 &&
		missing_bits == (FILE_WRITE_DATA|FILE_APPEND_DATA) ) {
		cifs_access_denied(OPEN_MODE_CONFLICT,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_pwrite",
			FILE_WRITE_DATA|FILE_APPEND_DATA, fsp->access_mask, NULL);
		return -1;
	}
    result = SMB_VFS_NEXT_PWRITE(handle, fsp, fd, data, n, offset);
	DEBUG(2,("unixntacl_pwrite: result %lld\n", (long long)result));
    return result;
}

/*
static
SMB_OFF_T unixntacl_lseek(vfs_handle_struct *handle, files_struct *fsp, int filedes, SMB_OFF_T offset, int whence)
{
	SMB_OFF_T result;
	DEBUG(2,("unixntacl_lseek: in:\n"));
	if ((check_desired_access(fsp->access_mask, FILE_GENERIC_WRITE) == -1) &&
		(check_desired_access(fsp->access_mask, READ_ACCESS) == -1) &&
		(fsp->desired_access != FILE_EXECUTE)) {
			return -1;
	}
    result = SMB_VFS_NEXT_LSEEK(handle, fsp, filedes, offset, whence);
	DEBUG(2,("unixntacl_lseek: result %lld\n", (long long)result));
    return result;
}
*/
static
ssize_t unixntacl_sendfile(vfs_handle_struct *handle, int tofd, files_struct *fsp, int fromfd, const DATA_BLOB *hdr,
			SMB_OFF_T offset, size_t n)
{
	ssize_t result;
	uint32_t missing_bits;
	DEBUG(2,("unixntacl_sendfile: in\n"));
	if (check_desired_access_ext(fsp->access_mask, READ_ACCESS|FILE_EXECUTE, &missing_bits) == -1 &&
		missing_bits == (READ_ACCESS|FILE_EXECUTE)) {
		cifs_access_denied(OPEN_MODE_CONFLICT,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_sendfile",
			READ_ACCESS|FILE_EXECUTE, fsp->access_mask, NULL);
		return -1;
	}
    result = SMB_VFS_NEXT_SENDFILE(handle, tofd, fsp, fromfd, hdr, offset, n);
	DEBUG(2,("unixntacl_sendfile: result %lld\n", (long long)result));
    return result;
}

static
int unixntacl_rename(vfs_handle_struct *handle, const char *old, const char *new_name)
{
	char new_container[256];
	int result;
    SMB_STRUCT_STAT st;
    int DIR_WRITE_ACCESS;
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];

	DEBUG(2,("unixntacl_rename: in\n")); 
    if(sys_stat(old, &st))
    return -1;  

    /*The needed permissions to rename: resource can be delete and has some access
    To rename a file: container with create files
    To rename directory: container with create subfolders*/
    safe_strcpy(new_container, container_path(new_name), sizeof(new_container)-1);
    if (S_ISDIR(st.st_mode))
        DIR_WRITE_ACCESS = FILE_ADD_SUBDIRECTORY;
    else
        DIR_WRITE_ACCESS = FILE_ADD_FILE;

    if(check_nt_path_access(new_container, DIR_WRITE_ACCESS,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), new_container, "rename",
			DIR_WRITE_ACCESS, missing_bits, access_troubleshooting_details);
		return -1;
	}

	if(check_acl_for_delete(handle->conn, old) == -1) return -1;
	if(check_nt_path_access(old, SYNCHRONIZE_ACCESS,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) != 0) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), old, "rename",
			SYNCHRONIZE_ACCESS, missing_bits, access_troubleshooting_details);
		return -1;
    }
    
    result = SMB_VFS_NEXT_RENAME(handle, old, new_name);
	DEBUG(2,("unixntacl_rename: result %d\n", result));
    return result;
}
/*
int unixntacl_fsync(vfs_handle_struct *handle, files_struct *fsp, int fd)
{
}
*/
const int STAT_ACCESS = (FILE_READ_ATTRIBUTES|FILE_READ_EA);

/*
static
int unixntacl_unlink(vfs_handle_struct *handle, connection_struct *conn, const char *path)
{
	int result;
	DEBUG(2,("unixntacl_unlink: in\n"));
	if (check_path_access(container_path(path), conn, FILE_DELETE_CHILD) == -1) {
		if (check_nt_path_access(path, DELETE_ACCESS) == -1) {
			DEBUG(0,("unixntacl_unlink: no perm to unlink '%s'\n", path));
			return -1;
		}
	}
    result = SMB_VFS_NEXT_UNLINK(handle, conn, path);
	DEBUG(2,("unixntacl_unlink: result %d\n", result));
    return result;
}
*/

static
int unixntacl_chown(vfs_handle_struct *handle, const char *path, uid_t uid, gid_t gid)
{
	int result;
	DEBUG(2,("unixntacl_chown: in\n"));
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];

	if (check_nt_path_access(path, WRITE_OWNER_ACCESS,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), path, "chown",
			WRITE_OWNER_ACCESS, missing_bits, access_troubleshooting_details);
		return -1;
	}
    result = SMB_VFS_NEXT_CHOWN(handle, path, uid, gid);
	DEBUG(2,("unixntacl_chown: result %d\n", result));
    return result;
}

static
int unixntacl_fchown(vfs_handle_struct *handle, files_struct *fsp, int fd, uid_t uid, gid_t gid)
{
	int result;
	DEBUG(2,("unixntacl_fchown: in\n"));
	if (!(fsp->access_mask & WRITE_OWNER_ACCESS)) {
		cifs_access_denied(OPEN_MODE_CONFLICT,
			(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_fchown",
			WRITE_OWNER_ACCESS, fsp->access_mask, NULL);
		DEBUG(0,("failed: desired_access 0x%08x\n", fsp->access_mask));
		return -1;
	}
    result = SMB_VFS_NEXT_FCHOWN(handle, fsp, fd, uid, gid);
	DEBUG(2,("unixntacl_fchown: result %d\n", result));
    return result;
}
/*
int unixntacl_chdir(vfs_handle_struct *handle, connection_struct *conn, const char *path)
{
	int result;
	DEBUG(0,("unixntacl_chdir: in\n"));
    result = SMB_VFS_NEXT_CHDIR(handle, conn, path);
	DEBUG(0,("unixntacl_chdir: result %d\n", result));
    return result;
}

char *unixntacl_getwd(vfs_handle_struct *handle, connection_struct *conn, char *path)
{
}

int unixntacl_utime(vfs_handle_struct *handle, connection_struct *conn, const char *path, struct utimbuf *times)
{
}
*/
static
int unixntacl_ftruncate(vfs_handle_struct *handle, files_struct *fsp, int fd, SMB_OFF_T len)
{
	int result;
	uint32_t missing_bits = 0;
	DEBUG(2,("unixntacl_ftruncate: in\n"));
    DEBUG(2,("unixntacl_ftruncate: access_mask is %08x\n", fsp->access_mask));
	if (check_desired_access_ext(fsp->access_mask, FILE_WRITE_DATA|FILE_WRITE_ATTRIBUTES, &missing_bits) == -1) {
		/* bug1463: allow truncation if either FILE_WRITE_DATA or FILE_WRITE_ATTRIBUTES is set */
		if (missing_bits == (FILE_WRITE_DATA|FILE_WRITE_ATTRIBUTES) &&
			check_desired_access(fsp->access_mask, READ_ACCESS) == -1) {
			/* Post two events to deliver the logic of access check... */
			cifs_access_denied(OPEN_MODE_CONFLICT,
				(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_ftruncate",
				FILE_WRITE_DATA|FILE_WRITE_ATTRIBUTES, fsp->access_mask, NULL);
			cifs_access_denied(OPEN_MODE_CONFLICT,
				(fsp->conn ? fsp->conn->connectpath : NULL), fsp->fsp_name, "unixntacl_ftruncate",
				READ_ACCESS, fsp->access_mask, NULL);
			return -1;
		}
	}
    result = SMB_VFS_NEXT_FTRUNCATE(handle, fsp, fd, len);
	DEBUG(2,("unixntacl_ftruncate: result %d\n", result));
    return result;
}
/*
BOOL unixntacl_lock(vfs_handle_struct *handle, files_struct *fsp, int fd, int op, SMB_OFF_T offset, SMB_OFF_T count, int type)
{
}
*/
static
int unixntacl_symlink(vfs_handle_struct *handle, const char *oldpath, const char *newpath)
{
	int result;
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
	char *cont_path = container_path(newpath);

	DEBUG(2,("unixntacl_symlink: in\n"));
	if (check_nt_path_access(cont_path, FILE_ADD_FILE,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), cont_path, "symlink",
			FILE_ADD_FILE, missing_bits, access_troubleshooting_details);
		return -1;
	}
    result = SMB_VFS_NEXT_SYMLINK(handle, oldpath, newpath);
	DEBUG(2,("unixntacl_symlink: result %d\n", result));
    return result;
}

static
int unixntacl_readlink(vfs_handle_struct *handle, const char *path, char *buf, size_t bufsiz)
{
	int result;
	uint32_t missing_bits = 0;
	char access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
	char *cont_path = container_path(path);

	DEBUG(2,("unixntacl_readlink: in\n"));
	if (check_nt_path_access(cont_path, READ_ACCESS,
		&missing_bits, EMPTY_DETAILS_STR(access_troubleshooting_details)) == -1) {
		cifs_access_denied(MISSED_BITS,
			(handle->conn ? handle->conn->connectpath : NULL), cont_path, "readlink",
			READ_ACCESS, missing_bits,
			access_troubleshooting_details);
		return -1;
	}
    result = SMB_VFS_NEXT_READLINK(handle, path, buf, bufsiz);
	DEBUG(2,("unixntacl_readlink: result %d\n", result));
    return result;
}

static vfs_op_tuple unixntacl_ops[] = {
		/* Directory operations */
	
	{SMB_VFS_OP(unixntacl_opendir),		SMB_VFS_OP_OPENDIR,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(unixntacl_mkdir),		SMB_VFS_OP_MKDIR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_rmdir),		SMB_VFS_OP_RMDIR,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_closedir),	SMB_VFS_OP_CLOSEDIR,	SMB_VFS_LAYER_TRANSPARENT},
	
		/* File operations */
	
	{SMB_VFS_OP(unixntacl_open),		SMB_VFS_OP_OPEN,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_close),		SMB_VFS_OP_CLOSE,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_read),		SMB_VFS_OP_READ,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_pread),		SMB_VFS_OP_PREAD,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_write),		SMB_VFS_OP_WRITE,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_pwrite),		SMB_VFS_OP_PWRITE,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_lseek),		SMB_VFS_OP_LSEEK,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_sendfile),	SMB_VFS_OP_SENDFILE,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_rename),		SMB_VFS_OP_RENAME,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_fsync),		SMB_VFS_OP_FSYNC,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_stat),		SMB_VFS_OP_STAT,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_fstat),		SMB_VFS_OP_FSTAT,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_lstat),		SMB_VFS_OP_LSTAT,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_unlink),		SMB_VFS_OP_UNLINK,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_chmod),		SMB_VFS_OP_CHMOD,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_fchmod),		SMB_VFS_OP_FCHMOD,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_chown),		SMB_VFS_OP_CHOWN,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_fchown),		SMB_VFS_OP_FCHOWN,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_chdir),		SMB_VFS_OP_CHDIR,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_getwd),		SMB_VFS_OP_GETWD,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_utime),		SMB_VFS_OP_UTIME,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_ftruncate),	SMB_VFS_OP_FTRUNCATE,	SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_lock),		SMB_VFS_OP_LOCK,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_symlink),		SMB_VFS_OP_SYMLINK,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_readlink),	SMB_VFS_OP_READLINK,	SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_link),		SMB_VFS_OP_LINK,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_mknod),		SMB_VFS_OP_MKNOD,		SMB_VFS_LAYER_TRANSPARENT},
//	{SMB_VFS_OP(unixntacl_realpath),	SMB_VFS_OP_REALPATH,	SMB_VFS_LAYER_TRANSPARENT},
	
		/* Windows ACL operations. */
	{SMB_VFS_OP(unixntacl_fget_nt_acl),	SMB_VFS_OP_FGET_NT_ACL,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_get_nt_acl),	SMB_VFS_OP_GET_NT_ACL,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_fset_nt_acl),	SMB_VFS_OP_FSET_NT_ACL,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_set_nt_acl),	SMB_VFS_OP_SET_NT_ACL,	SMB_VFS_LAYER_TRANSPARENT},
#if 0
	{SMB_VFS_OP(unixntacl_opendir),	SMB_VFS_OP_OPENDIR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(unixntacl_open),	SMB_VFS_OP_OPEN,		SMB_VFS_LAYER_TRANSPARENT},
	/* NT ACL operations. */
	{SMB_VFS_OP(fget_unixntacl_acl),SMB_VFS_OP_FGET_NT_ACL,	SMB_VFS_LAYER_OPAQUE},
	{SMB_VFS_OP(get_unixntacl_acl),	SMB_VFS_OP_GET_NT_ACL,	SMB_VFS_LAYER_OPAQUE},
	{SMB_VFS_OP(fset_unixntacl_acl),SMB_VFS_OP_FSET_NT_ACL,	SMB_VFS_LAYER_OPAQUE},
	{SMB_VFS_OP(set_unixntacl_acl),	SMB_VFS_OP_SET_NT_ACL,	SMB_VFS_LAYER_OPAQUE},
#endif
	{SMB_VFS_OP(NULL),		SMB_VFS_OP_NOOP,	SMB_VFS_LAYER_NOOP}
};


/*
* ExaNet 2008 / Samba upgrade project / acl support
* Each Samba module a VFS module should have a init_module function if it's a shared module,
* or smthing like vfs_example_init if it's staticly linked to samba
*/
NTSTATUS init_module(void)
{
	NTSTATUS ret = smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "unixntacl", unixntacl_ops);

	if (!NT_STATUS_IS_OK(ret))
		return ret;
	
	vfs_unixntacl_debug_level = debug_add_class("unixntacl");
	if (vfs_unixntacl_debug_level == -1) {
		vfs_unixntacl_debug_level = DBGC_VFS;
		DEBUG(0, ("vfs_unixntacl: Couldn't register custom debugging class!\n"));
	} else {
		DEBUG(10, ("vfs_unixntacl: Debug class number of 'unixntacl': %d\n", vfs_unixntacl_debug_level));
	}

	nt_access = check_nt_access;
	nt_path_access = check_nt_path_access;
    allowed_delete = check_acl_for_delete;
    
	return ret;
}





/* 
   Unix SMB/CIFS implementation.
   Samba utility functions. ADS stuff
   Copyright (C) Alexey Kotovich 2002
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"

static struct perm_mask_str {
	uint32  mask;
	const char   *str;
} perms[] = {
	{SEC_RIGHTS_FULL_CTRL,		"[Full Control]"},

	{SEC_RIGHTS_LIST_CONTENTS,	"[List Contents]"},
	{SEC_RIGHTS_LIST_OBJECT,	"[List Object]"},

	{SEC_RIGHTS_READ_ALL_PROP,	"[Read All Properties]"},	
	{SEC_RIGHTS_READ_PERMS,		"[Read Permissions]"},	

	{SEC_RIGHTS_WRITE_ALL_VALID,	"[All validate writes]"},
	{SEC_RIGHTS_WRITE_ALL_PROP,  	"[Write All Properties]"},

	{SEC_RIGHTS_MODIFY_PERMS,	"[Modify Permissions]"},
	{SEC_RIGHTS_MODIFY_OWNER,	"[Modify Owner]"},

	{SEC_RIGHTS_CREATE_CHILD,	"[Create All Child Objects]"},

	{SEC_RIGHTS_DELETE,		"[Delete]"},
	{SEC_RIGHTS_DELETE_SUBTREE,	"[Delete Subtree]"},
	{SEC_RIGHTS_DELETE_CHILD,	"[Delete All Child Objects]"},

	{SEC_RIGHTS_CHANGE_PASSWD,	"[Change Password]"},	
	{SEC_RIGHTS_RESET_PASSWD,	"[Reset Password]"},
	{0,				0}
};

/* convert a security permissions into a string */
static void ads_disp_perms(uint32 type)
{
	int i = 0;
	int j = 0;

	DEBUGADD(10,("Permissions: "));
	
	if (type == SEC_RIGHTS_FULL_CTRL) {
		DEBUGADD(10,("%s\n", perms[j].str));
		return;
	}

	for (i = 0; i < 32; i++) {
		if (type & (1 << i)) {
			for (j = 1; perms[j].str; j ++) {
				if (perms[j].mask == (((unsigned) 1) << i)) {
					DEBUGADD(10,("\n\t%s", perms[j].str));
				}	
			}
			type &= ~(1 << i);
		}
	}

	/* remaining bits get added on as-is */
	if (type != 0) {
		DEBUGADD(10,("[%08x]", type));
	}
	DEBUGADD(10,("\n"));
}

/* display ACE */
static void ads_disp_ace(SEC_ACE *sec_ace)
{
	const char *access_type = "UNKNOWN";

	if (!sec_ace_object(sec_ace->type)) {
		DEBUGADD(10,("------- ACE (type: 0x%02x, flags: 0x%02x, size: 0x%02x, mask: 0x%x)\n",
					 sec_ace->type,
					 sec_ace->flags,
					 sec_ace->size,
					 sec_ace->access_mask));
	} else {
		DEBUGADD(10,("------- ACE (type: 0x%02x, flags: 0x%02x, size: 0x%02x, mask: 0x%x, object flags: 0x%x)\n",
					 sec_ace->type,
					 sec_ace->flags,
					 sec_ace->size,
					 sec_ace->access_mask,
					 sec_ace->obj_flags));
	}
	
	if (sec_ace->type == SEC_ACE_TYPE_ACCESS_ALLOWED) {
		access_type = "ALLOWED";
	} else if (sec_ace->type == SEC_ACE_TYPE_ACCESS_DENIED) {
		access_type = "DENIED";
	} else if (sec_ace->type == SEC_ACE_TYPE_SYSTEM_AUDIT) {
		access_type = "SYSTEM AUDIT";
	} else if (sec_ace->type == SEC_ACE_TYPE_ACCESS_ALLOWED_OBJECT) {
		access_type = "ALLOWED OBJECT";
	} else if (sec_ace->type == SEC_ACE_TYPE_ACCESS_DENIED_OBJECT) {
		access_type = "DENIED OBJECT";
	} else if (sec_ace->type == SEC_ACE_TYPE_SYSTEM_AUDIT_OBJECT) {
		access_type = "AUDIT OBJECT";
	}

	DEBUGADD(10,("access SID:  %s\naccess type: %s\n",
				 sid_string_static(&sec_ace->trustee), access_type));

	ads_disp_perms(sec_ace->access_mask);
}

/* display ACL */
static void ads_disp_acl(SEC_ACL *sec_acl, const char *type)
{
	if (!sec_acl)
		DEBUGADD(10,("------- (%s) ACL not present\n", type));
	else {
		DEBUGADD(10,("------- (%s) ACL (revision: %d, size: %d, number of ACEs: %d)\n",
					 type,
					 sec_acl->revision,
					 sec_acl->size,
					 sec_acl->num_aces));			
	}
}

/* display SD */
void my_ads_disp_sd(SEC_DESC *sd)
{
	int i;
	
	DEBUG(10,("-------------- Security Descriptor (revision: %d, type: 0x%02x)\n", 
               sd->revision,
               sd->type));
	DEBUGADD(10,("owner SID: %s\n", sid_string_static(sd->owner_sid)));
	DEBUGADD(10,("group SID: %s\n", sid_string_static(sd->group_sid)));

	if (sd->sacl) {
		ads_disp_acl(sd->sacl, "system");
		for (i = 0; i < sd->sacl->num_aces; i ++)
			ads_disp_ace(&sd->sacl->aces[i]);
	}

    if (sd->dacl) {
        ads_disp_acl(sd->dacl, "user");
        for (i = 0; i < sd->dacl->num_aces; i ++)
            ads_disp_ace(&sd->dacl->aces[i]);
    }
    
    DEBUGADD(10,("-------------- End Of Security Descriptor\n"));
}
