/* 
 * Implementation of an Windows Previous Versions (AKA Shadow Copy) module
 *
 * Copyright (C) Sergey Kleyman, sergey.kleyman@exanet.com 	2009
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "includes.h"
#include "smbutil/win_prev_vers.h"

/*
 * log level legend:
 *		0 - error we always want to know about, preventing the module from functioning correctly (out of memory, internal error, etc.)
 *		1 - error we would like to know about, but the module can still function correctly (cannot register custom log class, etc.)
 *		2 - error that is very rare but possible as part of normal flow
 *		3 - event that is rare and interesting as execution flow indicator
 *      4 - error that is rare but maybe caused by invalid user input 
 *      5 - event that is interesting as execution flow indicator
 *     10 - trace that helps to understand execution flow
 */

static int this_VFS_module_debug_class = DBGC_VFS;
static const char* this_VFS_module_name = "win_prev_vers";

#undef DBGC_CLASS
#define DBGC_CLASS this_VFS_module_debug_class

static const uint32 max_number_of_snapshots_to_list_default = 64;
static uint32 max_number_of_snapshots_to_list = 0;

extern int (*get_snapshots_timestamps)( const char* volume_name, 
										uint32 max_number_of_snapshots,
										uint32 snapshots_timestamps[],
										uint32* number_of_snapshots );
extern int (*get_snapshot_name_by_timestamp)( const char* volume_name,
											  uint32 snapshot_timestamp,
											  int* snapshot_found,
											  char snapshot_name[] );

static const char previous_version_timestamp_prefix[] = "@GMT-";
static const size_t previous_version_timestamp_prefix_length = STATIC_ARRAY_LENGTH( previous_version_timestamp_prefix ) - 1;
static const char previous_version_timestamp_sample[] = "@GMT-2004.02.18-15.44.00";
static const size_t previous_version_timestamp_length = STATIC_ARRAY_LENGTH( previous_version_timestamp_sample ) - 1;
static const char timestamp_to_label_format[] = "@GMT-%04u.%02u.%02u-%02u.%02u.%02u";

static const char absolute_path_prefix[] = "/mnt/cifs/";
static const size_t absolute_path_prefix_length = STATIC_ARRAY_LENGTH( absolute_path_prefix ) - 1;
static const char snapshots_sub_dir_literal[] = ".snapshots";
static const char path_delimiter = '/';

typedef uint32 timestamp_t;

typedef struct path_split_result {
	const char* abs_prefix_begin;
	const char* abs_prefix_end;
	const char* volume_name_begin;
	const char* volume_name_end;
	const char* path_to_share_begin;
	const char* path_to_share_end;
	const char* path_to_file_before_timestamp_begin;
	const char* path_to_file_before_timestamp_end;
	timestamp_t timestamp;
	const char* path_to_file_after_timestamp_begin;
	const char* path_to_file_after_timestamp_end;
} path_split_result;

static BOOL convert_timestamp_to_label( timestamp_t timestamp, SHADOW_COPY_LABEL label ) 
{
	int ret = 0;
	struct tm broken_down_timestamp;
	struct tm* ret_tm = NULL;
	time_t timestamp_time_t = timestamp;

    ret_tm = gmtime_r( &timestamp_time_t, &broken_down_timestamp );
	if ( ret_tm == NULL ) {
		DEBUG(0,( "Failed to break down timestamp (%u), gmtime_r returned NULL. errno: %d (%s).\n", 
				  timestamp, errno, strerror(errno) ));
		return False;
	}

	if ( 1900 + broken_down_timestamp.tm_year > 9999 ) {
		DEBUG(0,( "Cannot convert timestamp (%u) to label because year (%u) doesn't fit in 4 decimal digits.\n",
				  timestamp, 1900 + broken_down_timestamp.tm_year ));
		return False;
	}

	ret = snprintf( label, sizeof( SHADOW_COPY_LABEL ), timestamp_to_label_format, 
					1900 + broken_down_timestamp.tm_year,
					broken_down_timestamp.tm_mon + 1,
					broken_down_timestamp.tm_mday,
					broken_down_timestamp.tm_hour,
					broken_down_timestamp.tm_min,
					broken_down_timestamp.tm_sec );
	if ( ret != ( sizeof( SHADOW_COPY_LABEL ) - 1 ) ) {
		DEBUG(0,( "Failed to print broken down timestamp (%u) as \"@GMT-%04u.%02u.%02u-%02u.%02u.%02u\". snprintf returned: %d, errno: %d (%s).\n", 
					timestamp,
					1900 + broken_down_timestamp.tm_year,
					broken_down_timestamp.tm_mon + 1,
					broken_down_timestamp.tm_mday,
					broken_down_timestamp.tm_hour,
					broken_down_timestamp.tm_min,
					broken_down_timestamp.tm_sec,
					ret, errno, strerror(errno) ));
		return False;
	}

	return True;
}

static const char* dbg_timestamp_to_string( timestamp_t timestamp ) 
{
	static SHADOW_COPY_LABEL label_buffer;
	BOOL ret_bool = False;

	ret_bool = convert_timestamp_to_label( timestamp, label_buffer );
	if ( ret_bool ) {
		return label_buffer;
	}

    return "<N/A>";
}

static BOOL convert_label_to_timestamp( const char* timestamp_begin, 
										timestamp_t* snap_timestamp ) 
{
	unsigned int parsed_year = 0;
	unsigned int parsed_month = 0;
	unsigned int parsed_day = 0;
	unsigned int parsed_hour = 0;
	unsigned int parsed_min = 0;
	unsigned int parsed_sec = 0;
	static const unsigned int sscanf_success_return_value = 6;
	struct tm broken_down_timestamp;
	int ret = 0;
	time_t ret_time_t = (time_t)(0);


	DEBUG(10,( "BEGIN: timestamp_begin: \"%.*s\".\n", 
			   previous_version_timestamp_length, timestamp_begin ));

	ret = sscanf( timestamp_begin, timestamp_to_label_format, 
				  &parsed_year,
				  &parsed_month,
				  &parsed_day,
				  &parsed_hour,
				  &parsed_min,
				  &parsed_sec );
	if ( ret != sscanf_success_return_value ) {
		DEBUG(4,( "Failed to parse label \"%.*s\". Successfully scanned only %u items out of %u. errno = %d (%s).\n", 
				  previous_version_timestamp_length, timestamp_begin, 
				  ret, sscanf_success_return_value, errno, strerror(errno) ));
		return False;
	}

	if ( parsed_year < 1900 ) {
		DEBUG(4,( "Cannot convert label (\"%.*s\") to timestamp because year (%u) is less then 1900.\n",
				  previous_version_timestamp_length, timestamp_begin, parsed_year ));
		return False;
	}
	broken_down_timestamp.tm_year = parsed_year - 1900;

	if ( ( 1 > parsed_month ) || ( parsed_month > 12 ) ) {
		DEBUG(4,( "Cannot convert label (\"%.*s\") to timestamp because month (%u) is not in the range [1-12].\n",
				  previous_version_timestamp_length, timestamp_begin, parsed_month ));
		return False;
	}
	broken_down_timestamp.tm_mon = parsed_month - 1;

	if ( ( 1 > parsed_day ) || ( parsed_day > 31 ) ) {
		DEBUG(4,( "Cannot convert label (\"%.*s\") to timestamp because day (%u) is not in the range [1-31].\n",
				  previous_version_timestamp_length, timestamp_begin, parsed_day ));
		return False;
	}
	broken_down_timestamp.tm_mday = parsed_day;

	if ( parsed_hour > 23 ) {
		DEBUG(4,( "Cannot convert label (\"%.*s\") to timestamp because hour (%u) is not in the range [0-23].\n",
				  previous_version_timestamp_length, timestamp_begin, parsed_hour ));
		return False;
	}
	broken_down_timestamp.tm_hour = parsed_hour;

	if ( parsed_min > 59 ) {
		DEBUG(4,( "Cannot convert label (\"%.*s\") to timestamp because minute (%u) is not in the range [0-59].\n",
					previous_version_timestamp_length, timestamp_begin, parsed_min ));
		return False;
	}
	broken_down_timestamp.tm_min = parsed_min;

	if ( parsed_sec > 61 ) {
		DEBUG(4,( "Cannot convert label (\"%.*s\") to timestamp because second (%u) is not in the range [0-61].\n",
				  previous_version_timestamp_length, timestamp_begin, parsed_sec ));
		return False;
	}
	broken_down_timestamp.tm_sec = parsed_sec;

    broken_down_timestamp.tm_isdst = 0;

	ret_time_t = timegm( &broken_down_timestamp );
	if ( ret_time_t == (time_t)(-1) ) {
		DEBUG(4,( "Failed to convert broken down time for label (\"%.*s\") to time_t. errno = %d (%s).\n", 
				  previous_version_timestamp_length, timestamp_begin, errno, strerror(errno) ));
		return False;
	}

	{
		SHADOW_COPY_LABEL label;
		size_t label_len = 0;
		BOOL ret_bool = false;

		ret_bool = convert_timestamp_to_label( (timestamp_t)(ret_time_t), label );
		if ( ! ret_bool ) {
			DEBUG(0,( "Failed to convert timestamp (%u) built from label (\"%.*s\") back to label for validity check.\n", 
					  (timestamp_t)(ret_time_t), 
					  previous_version_timestamp_length, timestamp_begin ));
			return False;
		}

		label_len = strnlen( label, sizeof(SHADOW_COPY_LABEL) ); 
		if ( label_len != previous_version_timestamp_length ) {
			DEBUG(0,( "Label (\"%s\") converted from built timestamp (%u) has length %u while should be %u.\n", 
					  label, (timestamp_t)(ret_time_t), label_len, previous_version_timestamp_length ));
			return False;
		}

		if ( strncmp( timestamp_begin, label, previous_version_timestamp_length ) != 0 ) {
			DEBUG(4,( "Label (\"%.*s\") extracted from path differs from label (\"%s\") converted from built timestamp (%u). Probably malformed date/time.\n", 
					  previous_version_timestamp_length, timestamp_begin,
					  label,
					  (timestamp_t)(ret_time_t) ));
			return False;
		}
	}

	*snap_timestamp = (timestamp_t)(ret_time_t);
	DEBUG(10,( "END: timestamp: %u.\n", *snap_timestamp ));

	return True;
}

static BOOL concat_paths( TALLOC_CTX* mem_ctx, 
						  const char* path_part_1,
						  const char* path_part_2_begin,
						  const char* path_part_2_end,
						  const char** new_path )
{
	const char* result = NULL;

	if ( path_part_2_begin == path_part_2_end ) {
		*new_path = path_part_1;
		return True;
	}

	result = talloc_asprintf( mem_ctx, "%s%c%.*s", 
							  path_part_1, 
							  path_delimiter, 
							  path_part_2_end - path_part_2_begin, path_part_2_begin );
	if ( result == NULL ) {
		DEBUG(0,( "Failed to concatenate \"%s\" + '%c' + \"%.*s\". talloc_asprintf return NULL. errno = %d (%s).\n",
				  path_part_1, path_delimiter, path_part_2_end - path_part_2_begin, path_part_2_begin,
				  errno, strerror(errno) ));
		return False;
	}

	*new_path = result;
	return True;
}

static const char* skip_path_delimiters( const char* path )
{
	const char* current = path;

	for ( ; *current != '\0' ; ++current ) {
		if ( *current != path_delimiter ) {
			return current;
		}
	}

	return current;
}

static const char* skip_path_delimiters_backwards( const char* path, const char* begin )
{
	const char* current = path;

	for ( ; current != (begin - 1) ; --current ) {
		if ( *current != path_delimiter ) {
			return current;
		}
	}

	return current;
}

static const char* find_timestamp( const char* path, timestamp_t* timestamp )
{
	const char* path_tail = path;
	const char* timestamp_begin = NULL;
	const char* path_end = NULL;
	BOOL ret_bool = False;

	DEBUG(10,( "BEGIN: path: \"%s\".\n", path ));

	path_end = path + strlen( path );

	for (;;) {
		timestamp_begin = strstr( path_tail, previous_version_timestamp_prefix );
		if ( timestamp_begin == NULL ) {
			DEBUG(10,( "Path tail (\"%s\") doesn't contain timestamp prefix (\"%s\").\n",
					  path_tail, previous_version_timestamp_prefix ));
			return NULL;
		}

		if ( timestamp_begin + previous_version_timestamp_length > path_end ) {
			DEBUG(4,( "Subpath (\"%s\") after previous version timestamp prefix (\"%s\") contains only partial timestamp.\n",
					   timestamp_begin, previous_version_timestamp_prefix ));
			return NULL;
		}

		if ( (timestamp_begin > path) &&
			 (*(timestamp_begin-1) != path_delimiter) ) {
			DEBUG(4,( "Char ('%c') before timestamp (\"%s\") is not path_delimiter ('%c'). Path: \"%s\".\n",
					  *(timestamp_begin-1), timestamp_begin, path_delimiter, path ));
			goto next_step;
		}

		if ( (timestamp_begin + previous_version_timestamp_length < path_end) &&
			 (*(timestamp_begin+previous_version_timestamp_length) != path_delimiter) ) {
			DEBUG(4,( "Char ('%c') after timestamp (\"%.*s\") is not path_delimiter ('%c'). Path: \"%s\".\n",
					   *(timestamp_begin+previous_version_timestamp_length), 
					   previous_version_timestamp_length, timestamp_begin,
					   path_delimiter, path ));
			goto next_step;
		}

		ret_bool = convert_label_to_timestamp( timestamp_begin, timestamp );
		if ( ret_bool ) {
			DEBUG(10,( "BEGIN: timestamp_begin: \"%s\", timestamp: %u.\n", 
					   timestamp_begin, *timestamp ));
			return timestamp_begin;
		}
		DEBUG(10,( "Path tail (\"%s\") doesn't contain timestamp prefix (\"%s\").\n",
				  path_tail, previous_version_timestamp_prefix ));

	next_step:
		path_tail += previous_version_timestamp_prefix_length;
		if ( path_tail > path_end ) {
			DEBUG(0,( "Parsing error: path_tail (%u) > path_end (%u). Path: \"%s\".\n",
					  path_tail - path, path_end - path, path ));
			return NULL;
		}
	}

	return NULL;
}

static BOOL split_path_to_parts( const char* path_to_share, 
								 const char* path_to_file,
								 path_split_result* split_result )
{
#define DEBUG_SUBSTR( substr, str ) DEBUG(10,( "%s: (%d) \"%.*s\" | \"%s\".\n", #substr, substr - str, substr - str, str, substr ))
	/* /mnt/cifs/<IP>/vol_name(/<path_share>)(/<path_file>)(@GMT-2009.03.04-20.40.15)(/<path_file>)	*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver                          									*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/                         									*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/sub_dir/prev_ver_share_1 									*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/@GMT-2009.03.04-20.40.15   								*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/@GMT-2009.03.04-20.40.15/   								*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/@GMT-2009.03.04-20.40.15/dir/f   							*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/sub_dir/prev_ver_share_1/@GMT-2009.03.04-20.40.15   		*/
	/* /mnt/cifs/10.64.151.30/vol_prev_ver/sub_dir/prev_ver_share_1/@GMT-2009.03.04-20.40.15/dir/f   */

	const char* path_to_share_end = NULL;
	const char* path_to_file_end = NULL;
    const char* timestamp_begin = NULL;
	path_split_result res;

	DEBUG(10,( "BEGIN: path_to_share: \"%s\", path_to_file: \"%s\".\n", 
			   path_to_share, path_to_file ));

	timestamp_begin = strstr( path_to_file, previous_version_timestamp_prefix );
	if ( timestamp_begin == NULL ) {
		DEBUG(10,( "Path to file (\"%s\") doesn't contain timestamp prefix (\"%s\").\n",
				  path_to_file, previous_version_timestamp_prefix ));
		return False;
	}

	if ( strncmp( path_to_share, absolute_path_prefix, absolute_path_prefix_length ) != 0 ) {
		DEBUG(10,( "Path to share \"%s\" doesn't begin with absolute path prefix (\"%s\"). Returning it as it is.\n", 
				  path_to_share, absolute_path_prefix ));
		return False;
	}

	res.abs_prefix_begin = path_to_share;
	DEBUG_SUBSTR( res.abs_prefix_begin, path_to_share );

	path_to_share_end = path_to_share + strlen( path_to_share );
	res.path_to_share_end = path_to_share_end;
	DEBUG_SUBSTR( res.path_to_share_end, path_to_share );

	path_to_file_end = path_to_file + strlen( path_to_file );
	res.path_to_file_after_timestamp_end = path_to_file_end;
	DEBUG_SUBSTR( res.path_to_file_after_timestamp_end, path_to_file );

	if ( path_to_file[0] != path_delimiter ) {
		res.path_to_file_before_timestamp_begin = path_to_file;
	} else {
		if ( ((path_to_file_end - path_to_file) < (path_to_share_end - path_to_share)) ||
			 (strncmp( path_to_file, path_to_share, path_to_share_end - path_to_share ) != 0) ) {
			DEBUG(10,( "Path to file (\"%s\") is absolute but doesn't begin with path to share (\"%s\").\n", 
					  path_to_file, path_to_share ));
			return False;
		}
		res.path_to_file_before_timestamp_begin = 
			skip_path_delimiters( path_to_file + ( path_to_share_end - path_to_share ) );
	}
	DEBUG_SUBSTR( res.path_to_file_before_timestamp_begin, path_to_file );

	timestamp_begin = find_timestamp( res.path_to_file_before_timestamp_begin, 
									  &(res.timestamp) );
	if ( timestamp_begin == NULL ) {
		DEBUG(10,( "Relative path to file (\"%s\") doesn't contain valid timestamp.\n",
				   res.path_to_file_before_timestamp_begin ));
		return False;
	}
	DEBUG_SUBSTR( timestamp_begin, path_to_file );

	res.abs_prefix_end = strchr( path_to_share + absolute_path_prefix_length, path_delimiter );
	if ( res.abs_prefix_end == NULL ) {
		DEBUG(0,( "Invalid input: subpath (\"%s\", \"%s\"[%d-]) after absolute path prefix (\"%s\") doesn't contain path delimiter ('%c').\n",
				  path_to_share + absolute_path_prefix_length, 
				  path_to_share, strlen( absolute_path_prefix ), 
				  absolute_path_prefix,
				  path_delimiter ));
		return False;
	}
	DEBUG_SUBSTR( res.abs_prefix_end, path_to_share );

	res.volume_name_begin = skip_path_delimiters( res.abs_prefix_end );
	if ( *(res.volume_name_begin) == '\0' ) {
		DEBUG(0,( "Invalid input: subpath \"%s\" (\"%s\"[%d-]) doesn't contain volume name.\n",
				  res.abs_prefix_end, path_to_share, res.abs_prefix_end - path_to_share ));
		return False;
	}
	DEBUG_SUBSTR( res.volume_name_begin, path_to_share );

	res.volume_name_end = strchr( res.volume_name_begin, path_delimiter );
	if ( res.volume_name_end == NULL ) {
		res.volume_name_end = path_to_share_end;
	}
	DEBUG_SUBSTR( res.volume_name_end, path_to_share );

	res.path_to_share_begin = skip_path_delimiters( res.volume_name_end );
	DEBUG_SUBSTR( res.path_to_share_begin, path_to_share );

	if ( res.path_to_file_before_timestamp_begin < timestamp_begin ) {
		res.path_to_file_before_timestamp_end = 
			skip_path_delimiters_backwards( timestamp_begin - 1, res.path_to_file_before_timestamp_begin ) + 1;
		if ( res.path_to_file_before_timestamp_end <= res.path_to_file_before_timestamp_begin ) {
			DEBUG(0,( "Parsing error: res.path_to_file_before_timestamp_end <= res.path_to_file_before_timestamp_begin. res.path_to_file_before_timestamp_end: \"%.*s\"|\"%s\", res.path_to_file_before_timestamp_begin: \"%.*s\"|\"%s\".\n",
					  res.path_to_file_before_timestamp_end - path_to_file, path_to_file, res.path_to_file_before_timestamp_end,
					  res.path_to_file_before_timestamp_begin - path_to_file, path_to_file, res.path_to_file_before_timestamp_begin ));
			return False;
		}
	} else {
		res.path_to_file_before_timestamp_end = timestamp_begin;
	}
	DEBUG_SUBSTR( res.path_to_file_before_timestamp_end, path_to_file );

	res.path_to_file_after_timestamp_begin = 
		skip_path_delimiters( timestamp_begin + previous_version_timestamp_length );
	DEBUG_SUBSTR( res.path_to_file_after_timestamp_begin, path_to_file );

	memcpy( split_result, &res, sizeof( res ) );

	return True;
}

static BOOL build_path( TALLOC_CTX* mem_ctx, 
						const path_split_result* split_result,
						const char* snapshot_name,
						const char** new_path )
{
	BOOL ret_bool = False;
	const char* result = NULL;

	result = talloc_asprintf( mem_ctx, "%.*s%c%.*s%c%s%c%s", 
							  split_result->abs_prefix_end - split_result->abs_prefix_begin, 
								split_result->abs_prefix_begin,
							  path_delimiter,
							  split_result->volume_name_end - split_result->volume_name_begin, 
								split_result->volume_name_begin,
							  path_delimiter,
							  snapshots_sub_dir_literal,
							  path_delimiter,
							  snapshot_name );
	if ( result == NULL ) {
		DEBUG( 0, ( "Failed to build path to snapshot (\"%.*s%c%.*s%c%s%c%s\").\n",
					split_result->abs_prefix_end - split_result->abs_prefix_begin, 
						split_result->abs_prefix_begin,
					path_delimiter,
					split_result->volume_name_end - split_result->volume_name_begin, 
						split_result->volume_name_begin,
					path_delimiter,
					snapshots_sub_dir_literal,
					path_delimiter,
					snapshot_name ) );
		return False;
	}

	ret_bool = concat_paths( mem_ctx, 
							 result, 
							 split_result->path_to_share_begin, 
							 split_result->path_to_share_end, 
							 &result );
	if ( ! ret_bool ) {
		DEBUG(0,( "Failed to concatenate path to snapshot (\"%s\") with path to share (\"%.*s\").\n",
				  result, 
				  split_result->path_to_share_end - split_result->path_to_share_begin, 
					split_result->path_to_share_begin ));
		return False;
	}

	ret_bool = concat_paths( mem_ctx, 
							 result, 
							 split_result->path_to_file_before_timestamp_begin, 
							 split_result->path_to_file_before_timestamp_end, 
							 &result );
	if ( ! ret_bool ) {
		DEBUG(0,( "Failed to concatenate path to snapshot (\"%s\") with path to file before timestamp (\"%.*s\").\n",
				  result, 
				  split_result->path_to_file_before_timestamp_end - split_result->path_to_file_before_timestamp_begin, 
					split_result->path_to_file_before_timestamp_begin ));
		return False;
	}

	ret_bool = concat_paths( mem_ctx, 
							 result, 
							 split_result->path_to_file_after_timestamp_begin, 
							 split_result->path_to_file_after_timestamp_end, 
							 &result );
	if ( ! ret_bool ) {
		DEBUG(0,( "Failed to concatenate path to snapshot (\"%s\") with path to file after timestamp (\"%.*s\").\n",
				  result, 
				  split_result->path_to_file_after_timestamp_end - split_result->path_to_file_after_timestamp_begin, 
					split_result->path_to_file_after_timestamp_begin ));
		return False;
	}

	*new_path = result;
	return True;
}

static BOOL extract_volume_name( TALLOC_CTX* mem_ctx,
								 const char* path_to_share, 
								 char** volume_name )
{
	BOOL ret_bool = False;
	path_split_result split_result;
	char* volume_name_res = NULL;

	ret_bool = split_path_to_parts( path_to_share, 
									previous_version_timestamp_sample, 
									&split_result );
	if ( ! ret_bool ) {
		DEBUG(0,( "Failed to split dummy path: \"%s\" + \"%s\".\n", 
				  path_to_share, previous_version_timestamp_sample ));
		return False;
	}

	volume_name_res = talloc_strndup( mem_ctx, 
									  split_result.volume_name_begin, 
									  split_result.volume_name_end - split_result.volume_name_begin );
	if ( volume_name_res == NULL ) {
		DEBUG(0,( "Failed to duplicate volume name (\"%.*s\").\n", 
				  split_result.volume_name_end - split_result.volume_name_begin, 
					split_result.volume_name_begin ));
		return False;
	}

	*volume_name = volume_name_res;
	return True;
}

#define CHECK_IF_EXTERN_FUNCTION_INITED( f ) (check_if_extern_func_inited( (f), (#f) ))

static int check_if_extern_func_inited( void* func_ptr, const char* func_name )
{
	static BOOL message_printed = False;

	if ( func_ptr != NULL ) {
		return 0;
	}

	if ( ! message_printed ) {
		DEBUG(0,( "ERROR: extern pointer '%s' is not initialized, probably because \"fs\" VFS module is not loaded. This VFS module will fail back to disabled mode.\n",
				  func_name ));
		message_printed = True;
	}
	
	errno = ENOPROTOOPT;
	return -1;
}

static int fetch_snapshots_timestamps( const char* volume_name, 
									   uint32 max_number_of_snapshots,
									   timestamp_t snapshots_timestamps[],
									   uint32* number_of_snapshots )
{
	int ret = 0;
	int dbg_level = 0;

	DEBUG(10,( "Calling extern get_snapshots_timestamps (%p) with:", 
			  get_snapshots_timestamps));
	DEBUGADD(10,( " volume_name = \"%s\"", volume_name ));
	DEBUGADD(10,( ", max_number_of_snapshots = %u...\n", max_number_of_snapshots ));

	ret = CHECK_IF_EXTERN_FUNCTION_INITED( get_snapshots_timestamps );
	if ( ret != 0 ) {
		return ret;
	}

	ret = get_snapshots_timestamps( volume_name,
									max_number_of_snapshots,
									snapshots_timestamps,
									number_of_snapshots );
	dbg_level = ((ret == 0) ? 10 : 2);
	DEBUG(dbg_level,( "Call to extern get_snapshots_timestamps completed. ret: %d", ret ));
	if ( ret == 0 ) {
		DEBUGADD(10,( " (SUCCESS)" ));
		DEBUGADD(10,( ", number_of_snapshots = %u", *number_of_snapshots ));
		if ( number_of_snapshots > 0 ) {
			uint32 i;

			DEBUGADD(10,( ", snapshots_timestamps:\n" ));
			for ( i = 0 ; i < (*number_of_snapshots) ; ++i ) {
				DEBUGADD(10,( "snapshots_timestamps[%u] = %u", i, snapshots_timestamps[i] ));
				if ( i != ((*number_of_snapshots)-1) ) {
					DEBUGADD(10,( ",\n" ));
				}
			}
		}
	} else {
		DEBUGADD(2,( " (FAILURE), errno = %d (%s)", errno, strerror(errno) ));
	}
	DEBUGADD(dbg_level,( ".\n" ));
	return ret;
}

static BOOL fetch_snapshot_name_by_timestamp( const char* volume_name,
											  timestamp_t snapshot_timestamp,
											  int* snapshot_found,
											  char snapshot_name[] )
{
	int ret = 0;
	int dbg_level = 0;

	DEBUG(10,( "Calling extern get_snapshot_name_by_timestamp (%p) with:", 
			  get_snapshots_timestamps));
	DEBUGADD(10,( "volume_name = \"%s\"", volume_name ));
	DEBUGADD(10,( ", snapshot_timestamp = %u...\n", snapshot_timestamp ));

	ret = CHECK_IF_EXTERN_FUNCTION_INITED( get_snapshot_name_by_timestamp );
	if ( ret != 0 ) {
		return ret;
	}

	ret = get_snapshot_name_by_timestamp( volume_name,
										  snapshot_timestamp,
										  snapshot_found,
										  snapshot_name );
	dbg_level = ((ret == 0) ? 10 : 2);
	DEBUG(dbg_level,( "Call to extern get_snapshot_name_by_timestamp completed. ret: %d", ret ));
	if ( ret == 0 ) {
		DEBUGADD(10,( " (SUCCESS)\n"));
		DEBUGADD(10,( ", snapshot_found = " ));
		if ( (*snapshot_found) != 0 ) {
			DEBUGADD(10,( "Yes\n" ));
			DEBUGADD(10,( ", snapshot_name = \"%s\"", snapshot_name ));
		} else {
			DEBUGADD(10,( "No" ));
		}
	} else {
		DEBUGADD(2,( " (FAILURE), errno = %d (%s)", errno, strerror(errno) ));
	}
	DEBUGADD(dbg_level,( ".\n" ));
	return ret;

}

static int translate_path( TALLOC_CTX* mem_ctx, 
						   const char* path_to_share, 
						   const char* path_to_file, 
						   char** new_path )
{
	BOOL ret_bool = False;
	int ret_int = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	path_split_result split_result;
	const char* volume_name = NULL;
	char snapshot_name[ VFSFS_MAX_SNAPSHOT_NAME_LENGTH ];
	int snapshot_found = 0;
	const char* temp_new_path = NULL;
	char* ret_new_path = NULL;

	DEBUG(10,( "BEGIN: path_to_share: \"%s\", path_to_file: \"%s\".\n", 
			   path_to_share, path_to_file ));

	local_mem_ctx = talloc_init( __FUNCTION__ );
	if ( local_mem_ctx == NULL ) {
		DEBUG(0,( "Failed to allocate local_mem_ctx.\n" ));
		return -1;
	}

	ret_bool = split_path_to_parts( path_to_share, path_to_file, &split_result );
	if ( ! ret_bool ) {
		DEBUG(10,( "Path (\"%s\" + \"%s\") cannot be splitted.\n", 
				   path_to_share, path_to_file ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	if ( split_result.volume_name_end - split_result.volume_name_begin > VFSFS_MAX_VOLUME_NAME_LENGTH ) {
		DEBUG(0,( "Volume name is too long. name =\"%.*s\", length = %u, max length = %u.\n", 
				  split_result.volume_name_end - split_result.volume_name_begin, 
					split_result.volume_name_begin,
				  split_result.volume_name_end - split_result.volume_name_begin, 
				  VFSFS_MAX_VOLUME_NAME_LENGTH ));
		errno = ENOBUFS;
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}
	volume_name = talloc_strndup( local_mem_ctx, 
								  split_result.volume_name_begin, 
								  split_result.volume_name_end - split_result.volume_name_begin );
	if ( volume_name == NULL ) {
		DEBUG(0,( "Failed to duplicate volume name (\"%.*s\").\n", 
				  split_result.volume_name_end - split_result.volume_name_begin, 
					split_result.volume_name_begin ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	ret_int = fetch_snapshot_name_by_timestamp( volume_name, 
												split_result.timestamp, 
												&snapshot_found, 
												snapshot_name );
	if ( ret_int != 0 ) {
		DEBUG(2,( "Failed to carry out search for snapshot name by timestamp (%u, \"%s\"). volume: \"%s\".\n",
				  split_result.timestamp, dbg_timestamp_to_string( split_result.timestamp ), volume_name ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}
	if ( snapshot_found == 0 ) {
		DEBUG(4,( "Snapshot with timestamp (%u, \"%s\") was not found.\n",
				  split_result.timestamp, dbg_timestamp_to_string( split_result.timestamp ) ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	ret_bool = build_path( local_mem_ctx, &split_result, snapshot_name, &temp_new_path );
	if ( ! ret_bool ) {
		DEBUG(0,( "Failed to build new path. Path to share = \"%s\", to file = \"%s\".\n",
				  path_to_share, path_to_file ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	ret_new_path = talloc_strdup( mem_ctx, temp_new_path );
	if ( ret_new_path == NULL ) {
		DEBUG(0,( "Failed to duplicate built path (\"%s\").\n", temp_new_path ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	DEBUG(6,( "Successfully translated: \"%s\" + \"%s\" -> \"%s\".\n", 
			  path_to_share, path_to_file, ret_new_path ));

	*new_path = ret_new_path;
	TALLOC_FREE( local_mem_ctx );
	return 0;
}

//static int check_access_for_get_shadow_copy_data( vfs_handle_struct* handle, files_struct* fsp )
//{
//    static const uint32 access_mask = FILE_READ_DATA;
//    BOOL access_allowed = False;
//
//    if ( fsp->fh->fd != -1 ) {
//        if ( nt_access != NULL ) {
//            access_allowed = ( nt_access( fsp->fh->fd, access_mask, fsp->nfs_fh ) == 0 );
//        } else {
//            access_allowed = True;
//        }
//    } else {
//        if ( nt_path_access != NULL ) {
//            TALLOC_CTX* local_mem_ctx = NULL;
//            BOOL ret_bool = False;
//            const char* abs_path_to_file = NULL;
//
//            local_mem_ctx = talloc_init( __FUNCTION__ );
//            if ( local_mem_ctx == NULL ) {
//                DEBUG( 0, ( "Failed to allocate local_mem_ctx.\n" ) );
//                return -1;
//            }
//
//            ret_bool = build_abs_path_to_file( local_mem_ctx,
//                                               handle->conn->connectpath,
//                                               fsp->fsp_name,
//                                               &abs_path_to_file );
//            if ( ! ret_bool ) {
//                DEBUG( 0, ( "Failed to build absolute path to file. (handle->conn->connectpath = \"%s\", fsp->fsp_name = \"%s\").\n",
//                             handle->conn->connectpath, fsp->fsp_name ) );
//                TALLOC_FREE( local_mem_ctx );
//                return -1;
//            }
//
//            access_allowed = ( nt_path_access( abs_path_to_file, access_mask ) == 0 );
//            TALLOC_FREE( local_mem_ctx );
//        } else {
//            access_allowed = True;
//        }
//    }
//
//    if ( ! access_allowed ) {
//        DEBUG(2,( "Cannot get list of snapshots for file \"%s\" + \"%s\": access denied (necessary access_mask = 0x%x).\n",
//                  handle->conn->connectpath, fsp->fsp_name, access_mask ));
//        errno = EACCES;
//        return -1;
//    }
//
//    return 0;
//}

static int get_labels_for_volume( vfs_handle_struct* handle, 
								  files_struct* fsp, 
								  SHADOW_COPY_DATA* shadow_copy_data )
{
	TALLOC_CTX* local_mem_ctx = NULL;
	BOOL ret_bool = False;
	const char* volume_name = NULL;
	timestamp_t timestamps[ VFSFS_MAX_LIMIT_NUMBER_OF_SNAPSHOTS_TO_LIST ];
	uint32 number_of_snapshots = 0;
	SHADOW_COPY_LABEL* labels = NULL;
	uint32 i = 0;
	int ret_int = 0;

//  DEBUG(0,( "SergeyKl: file = \"%s\" + \"%s\", access granted = %s, fsp->access_mask = 0x%x, FILE_READ_DATA = 0x%x.\n",
//            handle->conn->connectpath, fsp->fsp_name, BOOLSTR(CONTAINS_MASK( fsp->access_mask, FILE_READ_DATA )), fsp->access_mask, FILE_READ_DATA ) );

//  ret_int = SMB_VFS_NEXT_GET_SHADOW_COPY_DATA( handle, fsp, shadow_copy_data, should_fill_in_labels );
//  if ( ( ret_int == -1 ) && ( errno == EACCES ) ) {
//      return -1;
//  }

	local_mem_ctx = talloc_init( __FUNCTION__ );
	if ( local_mem_ctx == NULL ) {
		DEBUG(0,( "Failed to allocate local_mem_ctx.\n" ));
		errno = ENOMEM;
		return -1;
	}

	ret_bool = extract_volume_name( local_mem_ctx, 
									handle->conn->connectpath, 
									(char**)&volume_name );
	if ( ! ret_bool ) {
		DEBUG(0,( "Failed to extract volume name from path to share \"%s\".\n",
					 handle->conn->connectpath ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	ret_int = fetch_snapshots_timestamps( volume_name,
										  max_number_of_snapshots_to_list,
										  timestamps,
										  &number_of_snapshots );
	if ( ret_int != 0 ) {
		DEBUG(2,( "Failed to get snapshots' timestamps for volume \"%s\". (path to share = \"%s\").\n",
				  volume_name, handle->conn->connectpath ));
		TALLOC_FREE( local_mem_ctx );
		return -1;
	}

	if ( number_of_snapshots != 0 ) {
		labels = TALLOC_ARRAY( shadow_copy_data->mem_ctx, SHADOW_COPY_LABEL, number_of_snapshots );
		if ( labels == NULL ) {
			DEBUG(0,( "Failed to allocated array for snapshot timestamps (total size = %u)\n",
					  number_of_snapshots * sizeof(SHADOW_COPY_LABEL) ));
			TALLOC_FREE( local_mem_ctx );
			errno = ENOMEM;
			return -1;
		}
	}

	for( i = 0 ; i < number_of_snapshots ; ++i ) {
		ret_bool = convert_timestamp_to_label( timestamps[i], labels[i] );
		if ( ! ret_bool ) {
			DEBUG(0,( "Failed to convert timestamp %u to label. (i = %u, number_of_snapshots = %u)\n", 
					  timestamps[i], i, number_of_snapshots ));
			TALLOC_FREE( local_mem_ctx );
			return -1;
		}
	}

	shadow_copy_data->num_volumes = number_of_snapshots;
	shadow_copy_data->labels = labels;

	TALLOC_FREE( local_mem_ctx );
	return 0;
}

static int vfs_get_shadow_copy_data( vfs_handle_struct* handle, 
									 files_struct* fsp, 
									 SHADOW_COPY_DATA* shadow_copy_data, 
									 BOOL should_fill_in_labels )
{
	int ret_int = 0;

	DEBUG(10,( "Starting to process GET_SHADOW_COPY_DATA request for volume with share \"%s\". (should_fill_in_labels = %s).\n", 
			   handle->conn->connectpath, BOOLSTR( should_fill_in_labels ) ));

	if ( should_fill_in_labels ) {
		 ret_int = get_labels_for_volume( handle, fsp, shadow_copy_data );
	} else {
		shadow_copy_data->num_volumes = max_number_of_snapshots_to_list;
		shadow_copy_data->labels = NULL;
	}

	if ( ret_int == 0 ) {
		DEBUG(10,( "Successfully finished processing GET_SHADOW_COPY_DATA request. number_of_snapshots = %u. (path share = \"%s\", should_fill_in_labels = %s).\n", 
				   shadow_copy_data->num_volumes, handle->conn->connectpath, BOOLSTR( should_fill_in_labels ) ));
	}

	return 0;
}

static void vfs_base_impl_pre( const char* calling_function_name,
							   vfs_handle_struct* handle, 
							   const char* fname, 
							   TALLOC_CTX** p_local_mem_ctx,
							   const char** p_path_after_translation )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;
	const char* path_to_share = NULL; 

	*p_path_after_translation = fname;
	*p_local_mem_ctx = NULL;

	if ( handle->conn == NULL ) {
		DEBUG(2,( "%s: No translation will be attempted for path \"%s\" because handle->conn == NULL.\n", 
				  calling_function_name, fname ));
		return;
	}
	path_to_share = handle->conn->connectpath; 

	DEBUG(10,( "%s: Translating path \"%s\" + \"%s\"...\n", 
			   calling_function_name, path_to_share, fname ));

	local_mem_ctx = talloc_init( __FUNCTION__ );
	if ( local_mem_ctx == NULL ) {
		DEBUG(0,( "%s: Failed to allocate local_mem_ctx. Leaving path as it's (\"%s\" + \"%s\").\n",
				  calling_function_name, path_to_share, fname ));
		return;
	}
	*p_local_mem_ctx = local_mem_ctx;

	ret = translate_path( local_mem_ctx, path_to_share, fname, (char**)( &path_after_translation ) );
	if ( ret == -1 ) {
		DEBUG(10,( "%s: Path (\"%s\" + \"%s\") was not translated. Leaving it as it is.\n",
				  calling_function_name, path_to_share, fname ));
		return;
	}
	*p_path_after_translation = path_after_translation;
}

static void vfs_base_impl_post( const char* calling_function_name,
								vfs_handle_struct* handle, 
								const char* fname, 
								const char* path_after_translation,
								BOOL is_vfs_next_successful,
								TALLOC_CTX** p_local_mem_ctx )
{
	const char* path_to_share = NULL; 

	if ( (handle->conn == NULL) && (path_after_translation == fname) ) {
		DEBUG(10,( "%s: No translation was be attempted for path \"%s\" because handle->conn == NULL.\n", 
					 calling_function_name, fname ));
		TALLOC_FREE( *p_local_mem_ctx );
		return;
	}
	path_to_share = handle->conn->connectpath; 

	if ( fname != path_after_translation ) {
		if ( is_vfs_next_successful ) {
			DEBUG(10,( "%s: VFS_NEXT returned successfully after translation \"%s\" + \"%s\" -> \"%s\".\n",
					   calling_function_name, 
					   path_to_share, fname, path_after_translation ));
		} else {
			DEBUG(5,( "%s: VFS_NEXT returned with failure (errno = %d (%s)) after translation \"%s\" + \"%s\" -> \"%s\".\n",
					  calling_function_name, 
					  errno, strerror( errno ),
					  path_to_share, fname, path_after_translation ));
		}
	} else {
		if ( is_vfs_next_successful ) {
			DEBUG(10,( "%s: VFS_NEXT returned successfully with not translated path \"%s\" + \"%s\".\n",
					   calling_function_name, 
					   path_to_share, fname ));
		} else {
			DEBUG(10,( "%s: VFS_NEXT returned with failure (errno = %d (%s)) with not translated path \"%s\" + \"%s\".\n",
					   calling_function_name, 
					   errno, strerror( errno ),
					   path_to_share, fname ));
		}
	}
	TALLOC_FREE( *p_local_mem_ctx );
}

static SMB_BIG_UINT vfs_disk_free( vfs_handle_struct* handle,  
								   const char* path, 
								   BOOL small_query, 
								   SMB_BIG_UINT* bsize,
								   SMB_BIG_UINT* dfree, 
								   SMB_BIG_UINT* dsize)
{
	SMB_BIG_UINT ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle, 
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_DISK_FREE( handle, path_after_translation, small_query, bsize, dfree, dsize );

	vfs_base_impl_post( __FUNCTION__, 
						handle, 
						path, 
						path_after_translation, 
						( ret != (SMB_BIG_UINT)(-1) ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_statvfs( vfs_handle_struct* handle, const char* path, vfs_statvfs_struct* statbuf)
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_STATVFS( handle, path_after_translation, statbuf );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static DIR* vfs_opendir( vfs_handle_struct* handle, const char* fname, const char* mask, uint32 attr )
{
	DIR* ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   fname, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_OPENDIR( handle, path_after_translation, mask, attr );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						fname, 
						path_after_translation, 
						( ret != NULL ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_mkdir( vfs_handle_struct* handle, const char* path, mode_t mode )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_MKDIR( handle, path_after_translation, mode );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_rmdir( vfs_handle_struct* handle, const char* path )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_RMDIR( handle, path_after_translation );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_open( vfs_handle_struct* handle, const char* fname, files_struct* fsp, int flags, mode_t mode )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   fname, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_OPEN( handle, path_after_translation, fsp, flags, mode );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						fname, 
						path_after_translation, 
						( ret != -1 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_rename( vfs_handle_struct* handle, const char* oldname, const char* newname )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   oldname, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_RENAME( handle, path_after_translation, newname );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						oldname, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_stat( vfs_handle_struct* handle, const char* fname, SMB_STRUCT_STAT* sbuf )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   fname, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_STAT( handle, path_after_translation, sbuf );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						fname, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

int vfs_lstat( vfs_handle_struct* handle, const char* path, SMB_STRUCT_STAT* sbuf )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LSTAT( handle, path_after_translation, sbuf );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_unlink( vfs_handle_struct* handle, const char* path )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_UNLINK( handle, path_after_translation );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_chmod( vfs_handle_struct* handle, const char* path, mode_t mode )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_CHMOD( handle, path_after_translation, mode );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_chown( vfs_handle_struct* handle, const char* path, uid_t uid, gid_t gid )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_CHOWN( handle, path_after_translation, uid, gid );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_chdir( vfs_handle_struct* handle, const char* path )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_CHDIR( handle, path_after_translation );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_ntimes( vfs_handle_struct* handle, const char* path, const struct timespec ts[2] )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_NTIMES( handle, path_after_translation, ts );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret != -1 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_symlink( vfs_handle_struct* handle, const char* oldpath, const char* newpath )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   oldpath, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_SYMLINK( handle, path_after_translation, newpath );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						oldpath, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_readlink( vfs_handle_struct *handle, const char* path, char* buf, size_t bufsiz )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_READLINK( handle, path_after_translation, buf, bufsiz );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret != -1 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_link( vfs_handle_struct* handle, const char* oldpath, const char* newpath )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   oldpath, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LINK( handle, path_after_translation, newpath );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						oldpath, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_mknod( vfs_handle_struct* handle, const char* pathname, mode_t mode, SMB_DEV_T dev )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   pathname, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_MKNOD( handle, path_after_translation, mode, dev );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						pathname, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static char* vfs_realpath( vfs_handle_struct* handle, const char* path, char* resolved_path )
{
	char* ret = NULL;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle, 
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_REALPATH( handle, path_after_translation, resolved_path );

	vfs_base_impl_post( __FUNCTION__, 
						handle, 
						path, 
						path_after_translation, 
						( ret != NULL ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_chflags( vfs_handle_struct* handle, const char* path, int flags )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_CHFLAGS( handle, path_after_translation, flags );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static ssize_t vfs_getxattr( vfs_handle_struct *handle, const char* path, const char* name, void* value, size_t size )
{
	ssize_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_GETXATTR( handle, path_after_translation, name, value, size );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret != (ssize_t)(-1) ), 
						&local_mem_ctx );

	return ret;
}

static ssize_t vfs_lgetxattr( vfs_handle_struct* handle, const char* path, const char* name, void* value, size_t size )
{
	ssize_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LGETXATTR( handle, path_after_translation, name, value, size );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret != (ssize_t)(-1) ), 
						&local_mem_ctx );

	return ret;
}

static ssize_t vfs_listxattr( vfs_handle_struct* handle, const char* path, char* list, size_t size )
{
	ssize_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LISTXATTR( handle, path_after_translation, list, size );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret != (ssize_t)(-1) ), 
						&local_mem_ctx );

	return ret;
}

ssize_t vfs_llistxattr( vfs_handle_struct* handle, const char* path, char* list, size_t size )
{
	ssize_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LLISTXATTR( handle, path_after_translation, list, size );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret != (ssize_t)(-1) ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_removexattr( vfs_handle_struct* handle, const char* path, const char* name )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_REMOVEXATTR( handle, path_after_translation, name );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_lremovexattr( vfs_handle_struct* handle, const char* path, const char* name )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LREMOVEXATTR( handle, path_after_translation, name );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_setxattr( vfs_handle_struct* handle, const char* path, const char* name, const void* value, size_t size, int flags )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_SETXATTR( handle, path_after_translation, name, value, size, flags );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static int vfs_lsetxattr( vfs_handle_struct* handle, const char* path, const char* name, const void* value, size_t size, int flags )
{
	int ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__, 
					   handle,
					   path, 
					   &local_mem_ctx, 
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LSETXATTR( handle, path_after_translation, name, value, size, flags );

	vfs_base_impl_post( __FUNCTION__, 
						handle,
						path, 
						path_after_translation, 
						( ret == 0 ), 
						&local_mem_ctx );

	return ret;
}

static ssize_t vfs_listads( vfs_handle_struct* handle, const char* path, char* list, size_t size )
{
	ssize_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;

	vfs_base_impl_pre( __FUNCTION__,
					   handle,
					   path,
					   &local_mem_ctx,
					   &path_after_translation );

	ret = SMB_VFS_NEXT_LISTADS( handle, path_after_translation, list, size );

	vfs_base_impl_post( __FUNCTION__,
						handle,
						path,
						path_after_translation,
						(ret != -1),
						&local_mem_ctx );

	return ret;
}

static
size_t vfs_get_nt_acl( vfs_handle_struct* handle, files_struct* fsp, const char* name, uint32 security_info, struct security_descriptor_info** ppdesc )
{
	size_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;
	char* saved_fsp_name = NULL;
	
	vfs_base_impl_pre( __FUNCTION__,
					   handle,
					   fsp->fsp_name,
					   &local_mem_ctx,
					   &path_after_translation );

	saved_fsp_name = fsp->fsp_name;
	fsp->fsp_name = (char*)path_after_translation;
	ret = SMB_VFS_NEXT_GET_NT_ACL( handle, fsp, path_after_translation, security_info, ppdesc );
	fsp->fsp_name = saved_fsp_name;

	vfs_base_impl_post( __FUNCTION__,
						handle,
						fsp->fsp_name,
						path_after_translation,
						(ret != 0),
						&local_mem_ctx );

	return ret;
}

static
size_t vfs_fget_nt_acl( vfs_handle_struct* handle, files_struct* fsp, int fd, uint32 security_info, struct security_descriptor_info** ppdesc )
{
	size_t ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;
	char* saved_fsp_name = NULL;

	vfs_base_impl_pre( __FUNCTION__,
					   handle,
					   fsp->fsp_name,
					   &local_mem_ctx,
					   &path_after_translation );

	saved_fsp_name = fsp->fsp_name;
	fsp->fsp_name = (char*)path_after_translation;
	ret = SMB_VFS_NEXT_FGET_NT_ACL( handle, fsp, fd, security_info, ppdesc );
	fsp->fsp_name = saved_fsp_name;

	vfs_base_impl_post( __FUNCTION__,
						handle,
						fsp->fsp_name,
						path_after_translation,
						(ret != 0),
						&local_mem_ctx );

	return ret;
}

static
BOOL vfs_set_nt_acl( vfs_handle_struct* handle, files_struct* fsp, const char* name, uint32 security_info_sent, struct security_descriptor_info* psd )
{
	BOOL ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;
	char* saved_fsp_name = NULL;

	vfs_base_impl_pre( __FUNCTION__,
					   handle,
					   fsp->fsp_name,
					   &local_mem_ctx,
					   &path_after_translation );

	saved_fsp_name = fsp->fsp_name;
	fsp->fsp_name = (char*)path_after_translation;
	ret = SMB_VFS_NEXT_SET_NT_ACL( handle, fsp, path_after_translation, security_info_sent, psd );
	fsp->fsp_name = saved_fsp_name;

	vfs_base_impl_post( __FUNCTION__,
						handle,
						fsp->fsp_name,
						path_after_translation,
						ret,
						&local_mem_ctx );

	return ret;
}

static
BOOL vfs_fset_nt_acl(struct vfs_handle_struct *handle, struct files_struct *fsp, int fd, uint32 security_info_sent, struct security_descriptor_info *psd)
{
	BOOL ret = 0;
	TALLOC_CTX* local_mem_ctx = NULL;
	const char* path_after_translation = NULL;
	char* saved_fsp_name = NULL;

	vfs_base_impl_pre( __FUNCTION__,
					   handle,
					   fsp->fsp_name,
					   &local_mem_ctx,
					   &path_after_translation );

	saved_fsp_name = fsp->fsp_name;
	fsp->fsp_name = (char*)path_after_translation;
	ret = SMB_VFS_NEXT_FSET_NT_ACL( handle, fsp, fd, security_info_sent, psd );
	fsp->fsp_name = saved_fsp_name;

	vfs_base_impl_post( __FUNCTION__,
						handle,
						fsp->fsp_name,
						path_after_translation,
						ret,
						&local_mem_ctx );

	return ret;
}

/* VFS operations structure */
static vfs_op_tuple implemented_ops[] = {
	/* 	
	* void* op; 								vfs_op_type type;					vfs_op_layer layer;
	*/
	{ SMB_VFS_OP( vfs_get_shadow_copy_data ),	SMB_VFS_OP_GET_SHADOW_COPY_DATA,	SMB_VFS_LAYER_TRANSPARENT },

	{ SMB_VFS_OP( vfs_disk_free ),				SMB_VFS_OP_DISK_FREE, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_statvfs ),				SMB_VFS_OP_STATVFS,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_opendir ),				SMB_VFS_OP_OPENDIR,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_mkdir ),					SMB_VFS_OP_MKDIR,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_rmdir ),					SMB_VFS_OP_RMDIR,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_open ),					SMB_VFS_OP_OPEN,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_rename ),					SMB_VFS_OP_RENAME, 					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_lstat ),					SMB_VFS_OP_LSTAT, 					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_stat ),					SMB_VFS_OP_STAT,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_unlink ),					SMB_VFS_OP_UNLINK,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_chmod ),					SMB_VFS_OP_CHMOD,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_chown ),					SMB_VFS_OP_CHOWN,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_chdir ),					SMB_VFS_OP_CHDIR, 					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_ntimes ),					SMB_VFS_OP_NTIMES,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_symlink ),				SMB_VFS_OP_SYMLINK,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_readlink ),				SMB_VFS_OP_READLINK,				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_link ),					SMB_VFS_OP_LINK, 					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_mknod ),					SMB_VFS_OP_MKNOD, 					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_realpath ),				SMB_VFS_OP_REALPATH,				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_chflags ),				SMB_VFS_OP_CHFLAGS,					SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_getxattr ),				SMB_VFS_OP_GETXATTR, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_lgetxattr ),				SMB_VFS_OP_LGETXATTR, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_listxattr ),				SMB_VFS_OP_LISTXATTR, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_llistxattr ),				SMB_VFS_OP_LLISTXATTR, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_removexattr ),			SMB_VFS_OP_REMOVEXATTR, 			SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_lremovexattr ),			SMB_VFS_OP_LREMOVEXATTR, 			SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_setxattr ),				SMB_VFS_OP_SETXATTR, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_lsetxattr ),				SMB_VFS_OP_LSETXATTR, 				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_listads ),               	SMB_VFS_OP_LISTADS,                 SMB_VFS_LAYER_TRANSPARENT },

	{ SMB_VFS_OP( vfs_get_nt_acl ),				SMB_VFS_OP_GET_NT_ACL,				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_fget_nt_acl ),			SMB_VFS_OP_FGET_NT_ACL,				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_fset_nt_acl ),			SMB_VFS_OP_FSET_NT_ACL,				SMB_VFS_LAYER_TRANSPARENT },
	{ SMB_VFS_OP( vfs_set_nt_acl ),				SMB_VFS_OP_SET_NT_ACL,				SMB_VFS_LAYER_TRANSPARENT },

	{ SMB_VFS_OP( NULL ),						SMB_VFS_OP_NOOP,					SMB_VFS_LAYER_NOOP 		  }
};

static void init_log_level( void )
{
	int registered_class = -1;

	registered_class = debug_add_class( this_VFS_module_name );
	if ( registered_class != -1 ) {
		DEBUG(3,( "%s: Successfully registered custom debug class: %d.\n", 
				  __FUNCTION__, registered_class ));
		this_VFS_module_debug_class = registered_class;
	} else {
		DEBUG(1,( "%s: Failed to register custom debug class \"%s\". debug_add_class returned: %d.\n", 
				  __FUNCTION__, this_VFS_module_name, this_VFS_module_debug_class ));
		this_VFS_module_debug_class = DBGC_VFS;
	}

	DEBUG(3,( "%s: Debug class for this module (\"%s\"): %d. Debug level for this class: %d.\n", 
			   __FUNCTION__, this_VFS_module_name, this_VFS_module_debug_class, 
			   DEBUGLEVEL_CLASS[ this_VFS_module_debug_class ] ));
}

static void load_config( void )
{
	const char* param_name = NULL;
	uint32 max_number_of_snapshots_to_list_conf = 0;

	DEBUG(3,( "Configuration:" ));

	param_name = "max number of snapshots to list";
	max_number_of_snapshots_to_list_conf = lp_parm_int( GLOBAL_SECTION_SNUM, 
														this_VFS_module_name, 
														param_name, 
														max_number_of_snapshots_to_list_default );
	max_number_of_snapshots_to_list = MIN( max_number_of_snapshots_to_list_conf, 
										   VFSFS_MAX_LIMIT_NUMBER_OF_SNAPSHOTS_TO_LIST );
	DEBUGADD(3,( " %s = %u", param_name, max_number_of_snapshots_to_list ));
	if ( max_number_of_snapshots_to_list != max_number_of_snapshots_to_list_conf ) {
		DEBUGADD(3,( " (reduced from %u specified by user)", max_number_of_snapshots_to_list_conf ));
	}

	DEBUGADD(3,( ".\n" ));
}

extern int (*win_prev_vers_translate_path_p)( TALLOC_CTX* ctx,
											  const char* path, 
											  const char* fname, 
											  char** new_path );

NTSTATUS vfs_win_prev_vers_init( void );
NTSTATUS vfs_win_prev_vers_init( void )
{
	NTSTATUS ret = NT_STATUS_OK;

	if ( previous_version_timestamp_length != (sizeof(SHADOW_COPY_LABEL)-1) ) {
		DEBUG(0,( "%s: Previous version timestamp (for instance \"%s\") length (%u) should be equal to sizeof(SHADOW_COPY_LABEL)-1 (%u).\n", 
				  __FUNCTION__, 
				  previous_version_timestamp_sample, 
				  previous_version_timestamp_length, 
				  sizeof(SHADOW_COPY_LABEL)-1 ));
		return NT_STATUS_INTERNAL_ERROR;
	}

	if ( win_prev_vers_translate_path_p == NULL ){
		win_prev_vers_translate_path_p = translate_path;
	}

	ret = smb_register_vfs( SMB_VFS_INTERFACE_VERSION, this_VFS_module_name, implemented_ops );
	if ( ! NT_STATUS_IS_OK( ret ) ) {
		DEBUG(0,( "%s: Failed to register VFS module \"%s\". smb_register_vfs returned 0x%x.\n", 
				  __FUNCTION__, this_VFS_module_name, NT_STATUS_V(ret) ));
		return ret;
	}
	DEBUG(3,( "%s: Successfully registered VFS module \"%s\".\n", __FUNCTION__, this_VFS_module_name ));

	init_log_level();

	load_config();

	return ret;
}

