/* 
   Unix SMB/CIFS implementation.
   session handling for utmp and PAM

   Copyright (C) tridge@samba.org       2001
   Copyright (C) abartlet@samba.org     2001
   Copyright (C) Gerald (Jerry) Carter  2006   
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* a "session" is claimed when we do a SessionSetupX operation
   and is yielded when the corresponding vuid is destroyed.

   sessions are used to populate utmp and PAM session structures
*/

#include "includes.h"

static TDB_CONTEXT *tdb;

/********************************************************************
********************************************************************/

BOOL session_init(void)
{
	if (tdb)
		return True;

    /**
     * ExaNet 2008 / Samba upgrade project / MMC support
     *
     * call lock_shared_path() instead of lock_path() for cluster awarness (if enabled)
     **/
    tdb = tdb_open_log(lock_shared_path("sessionid.tdb"), 0, TDB_CLEAR_IF_FIRST|TDB_DEFAULT, 
               O_RDWR | O_CREAT, 0644);
	if (!tdb) {
		DEBUG(1,("session_init: failed to open sessionid tdb\n"));
		return False;
	}

	return True;
}

/********************************************************************
 called when a session is created
********************************************************************/

BOOL session_claim(user_struct *vuser)
{
	int i = 0;
	TDB_DATA data;
	struct sockaddr sa;
	struct in_addr *client_ip;
	struct sessionid sessionid;
	uint32 pid = (uint32)sys_getpid();
	TDB_DATA key;		
	fstring keystr;
	char * hostname;
	int tdb_store_flag;  /* If using utmp, we do an inital 'lock hold' store,
				but we don't need this if we are just using the 
				(unique) pid/vuid combination */

    /**
     * ExaNet 2008 / Samba upgrade project / MMC support
     *
     * nodeid - 
     **/
    fstring nodeid; 

	vuser->session_keystr = NULL;

	/* don't register sessions for the guest user - its just too
	   expensive to go through pam session code for browsing etc */
	if (vuser->guest) {
		return True;
	}

	if (!session_init())
		return False;

	ZERO_STRUCT(sessionid);

	data.dptr = NULL;
	data.dsize = 0;

	if (lp_utmp()) {
		for (i=1;i<MAX_SESSION_ID;i++) {
			slprintf(keystr, sizeof(keystr)-1, "ID/%d", i);
			key.dptr = keystr;
			key.dsize = strlen(keystr)+1;
			
			if (tdb_store(tdb, key, data, TDB_INSERT) == 0) break;
		}
		
		if (i == MAX_SESSION_ID) {
			DEBUG(1,("session_claim: out of session IDs (max is %d)\n", 
				 MAX_SESSION_ID));
			return False;
		}
		slprintf(sessionid.id_str, sizeof(sessionid.id_str)-1, SESSION_UTMP_TEMPLATE, i);
		tdb_store_flag = TDB_MODIFY;
	} else
	{
		slprintf(keystr, sizeof(keystr)-1, "ID/%lu/%u", 
			 (long unsigned int)sys_getpid(), 
			 vuser->vuid);
		slprintf(sessionid.id_str, sizeof(sessionid.id_str)-1, 
			 SESSION_TEMPLATE, (long unsigned int)sys_getpid(), 
			 vuser->vuid);

		key.dptr = keystr;
		key.dsize = strlen(keystr)+1;
			
		tdb_store_flag = TDB_REPLACE;
	}

	/* If 'hostname lookup' == yes, then do the DNS lookup.  This is
           needed because utmp and PAM both expect DNS names 
	   
	   client_name() handles this case internally.
	*/

	hostname = client_name();
	if (strcmp(hostname, "UNKNOWN") == 0) {
		hostname = client_addr();
	}

    /**
     * ExaNet 2008 / Samba upgrade project / MMC support
     *
     * set nodeid field for cluster wide awareness
     **/
    if (gethostname(nodeid, sizeof(nodeid) -1) != 0) {
        DEBUG(0, ("session_claim: error getting local node name (nodeid)\n"));
    }
    fstrcpy(sessionid.nodeid, nodeid);

	fstrcpy(sessionid.username, vuser->user.unix_name);
	fstrcpy(sessionid.hostname, hostname);
	sessionid.id_num = i;  /* Only valid for utmp sessions */
	sessionid.pid = pid;
	sessionid.uid = vuser->uid;
	sessionid.gid = vuser->gid;
	fstrcpy(sessionid.remote_machine, get_remote_machine_name());
	fstrcpy(sessionid.ip_addr, client_addr());
	sessionid.connect_start = time(NULL);

	client_ip = client_inaddr(&sa);

	if (!smb_pam_claim_session(sessionid.username, sessionid.id_str, sessionid.hostname)) {
		DEBUG(1,("pam_session rejected the session for %s [%s]\n",
				sessionid.username, sessionid.id_str));
		if (tdb_store_flag == TDB_MODIFY) {
			tdb_delete(tdb, key);
		}
		return False;
	}

	data.dptr = (char *)&sessionid;
	data.dsize = sizeof(sessionid);
	if (tdb_store(tdb, key, data, tdb_store_flag) != 0) {
		DEBUG(1,("session_claim: unable to create session id record\n"));
		return False;
	}

	if (lp_utmp()) {
		sys_utmp_claim(sessionid.username, sessionid.hostname, 
			       client_ip,
			       sessionid.id_str, sessionid.id_num);
	}

	vuser->session_keystr = SMB_STRDUP(keystr);
	if (!vuser->session_keystr) {
		DEBUG(0, ("session_claim:  strdup() failed for session_keystr\n"));
		return False;
	}
	return True;
}

/********************************************************************
 called when a session is destroyed
********************************************************************/

void session_yield(user_struct *vuser)
{
	TDB_DATA dbuf;
	struct sessionid sessionid;
	struct in_addr *client_ip;
	TDB_DATA key;

	if (!tdb) return;

	if (!vuser->session_keystr) {
		return;
	}

	key.dptr = vuser->session_keystr;
	key.dsize = strlen(vuser->session_keystr)+1;

	dbuf = tdb_fetch(tdb, key);

	if (dbuf.dsize != sizeof(sessionid))
		return;

	memcpy(&sessionid, dbuf.dptr, sizeof(sessionid));

	client_ip = interpret_addr2(sessionid.ip_addr);

	SAFE_FREE(dbuf.dptr);

	if (lp_utmp()) {
		sys_utmp_yield(sessionid.username, sessionid.hostname, 
			       client_ip,
			       sessionid.id_str, sessionid.id_num);
	}

	smb_pam_close_session(sessionid.username, sessionid.id_str, sessionid.hostname);

	tdb_delete(tdb, key);
}

/********************************************************************
********************************************************************/

BOOL session_traverse(int (*fn)(TDB_CONTEXT *, TDB_DATA, TDB_DATA, void *),
		      void *state)
{
	if (!session_init()) {
		DEBUG(3, ("No tdb opened\n"));
		return False;
	}

	tdb_traverse(tdb, fn, state);
	return True;
}

/********************************************************************
********************************************************************/

struct session_list {
	int count;
	struct sessionid *sessions;
};

static int gather_sessioninfo(TDB_CONTEXT *stdb, TDB_DATA kbuf, TDB_DATA dbuf, void *state)
{
	uint32 i;	
	struct session_list *sesslist = (struct session_list *) state;
	const struct sessionid *current = (const struct sessionid *) dbuf.dptr;

	i = sesslist->count;
	
	sesslist->sessions = SMB_REALLOC_ARRAY(sesslist->sessions, struct sessionid, i+1);
	if (!sesslist->sessions) {
		sesslist->count = 0;
		return -1;
	}

	memcpy(&sesslist->sessions[i], current, sizeof(struct sessionid));
	sesslist->count++;

	DEBUG(7,("gather_sessioninfo session from %s@%s\n", 
		 current->username, current->remote_machine));

	return 0;
}

/********************************************************************
********************************************************************/

/**
 * ExaNet 2008 / Samba upgrade project / MMC support
 *
 * This function has been changed in order to support multi nodes.										
 * The solution is based on the file system being distributed, 
 * and the sessionid.tdb file being accessible from all nodes.																	
 * Once all files are accesible, the new location is changed from lp_lockdir() to lp_lockshareddir()
 * which includes the ndoe name (%h --> host). The normal operation will be kept (session_traverse)		
 * but there will be a scan (readdir) of the parent dir for other nodes and all sessionid.tdb files		
 * from other hosts (local will be filtered) will be added.												
 **/
int list_sessions(struct sessionid **session_list)
{
	struct session_list sesslist;

	sesslist.count = 0;
	sesslist.sessions = NULL;
	
    /**
     * ExaNet 2008 / Samba upgrade project / MMC support
     *
     * get session information from current node first
     **/
	if (!session_traverse(gather_sessioninfo, (void *) &sesslist)) {
		DEBUG(3, ("Session traverse failed for current node\n"));
		SAFE_FREE(sesslist.sessions);
		*session_list = NULL;
		return 0;
	}

    /**
     * ExaNet 2008 / Samba upgrade project / MMC support
     *
     * get information from the other nodes if enabled
     **/
    if (lp_cluster_wide_awareness()) {
        DIR *baseDir;
        struct dirent *dirEnt;
        pstring baseSharedPath;
        pstring thisNode;
        pstring currentPath;
        pstring currentPathFresh;
        struct stat statbuf;
        char *lastSlash;
        TDB_CONTEXT *tdb;

        /* Get the shared dir path, including the current node path */
        pstrcpy(baseSharedPath,lp_lockshareddir());
        /* get rid of the trailing slash */
        trim_char(baseSharedPath,'\0','/');
        /* find last slash to find base of shared path (no node) */
        lastSlash = strrchr_m(baseSharedPath, '/');
        /* if not found, exit */
        if (lastSlash == NULL) {
            DEBUG(0, ("list_sessions - error while parsing base shared path\n"));
            SAFE_FREE(sesslist.sessions);
            *session_list = NULL;
            return 0;
        }
        /* should contain the current node name */
        pstrcpy(thisNode, lastSlash + 1); 
        *lastSlash = '\0';
        baseDir = opendir(baseSharedPath);
        if (baseDir == NULL) {
            DEBUG(0, ("list_sessions - error opening baseDir\n"));
            SAFE_FREE(sesslist.sessions);
            *session_list = NULL;
            return 0;
        }
        /* go over all directories that start with node, except for the current one */
        while ((dirEnt = readdir(baseDir)) != NULL) {
            if ((strncmp("node", dirEnt->d_name, strlen("node")) == 0) &&
                (strncmp(thisNode, dirEnt->d_name, strlen(thisNode)) != 0)) 
            {
                snprintf(currentPath, PSTRING_LEN, "%s/%s/sessionid.tdb", baseSharedPath, dirEnt->d_name);
                snprintf(currentPathFresh, PSTRING_LEN, "%s/%s/fresh", baseSharedPath, dirEnt->d_name);
                /* check if fresh to know if samba is up */
                if (stat(currentPathFresh, &statbuf) != 0) {
                    DEBUG(1, ("Failure on stat of %s. Skipping this node. error = %m\n", currentPathFresh));
                    continue;
                } else {
                    time_t now = time(NULL);
                    if (now - statbuf.st_mtime > 300) { /* test interval is 120, so we take 120 times 2 and then some... */
                        DEBUG(1, ("Stat of %s shown that node is not fresh (%ld seconds). Node might be down, skipping\n", currentPathFresh, (now - statbuf.st_mtime)));
                        continue;
                    }
                }

                /* this is the session_traverse code per node */
                tdb = tdb_open_log(currentPath, 0, TDB_DEFAULT, O_RDONLY, 0644);
                if (!tdb) {
                    DEBUG(1,("list_sessions: error opening tdbfile %s\n", currentPath));
                    continue;
                }

                tdb_traverse(tdb, gather_sessioninfo, (void *) &sesslist);
                tdb_close(tdb);
            }
        }
        closedir(baseDir);
    }

	*session_list = sesslist.sessions;
	return sesslist.count;
}
