#include <netinet/in.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/**/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
/**/

// declare dlm_client non-const
#define	EXASMB_DLMCLIENT_SOURCE

#define XNT_USING_STD_CERR	// dumping errors to std::cerr
#include "xnt_error_logging.h"

#include "exasmb_acl_client.h"
#include "exasmb_aclclient.h"
#include "exasmb_vfsfsclient.h"
#include "smb_acl_message.h"
#include "unixntacl.h"
#include "unixntacl_xdr.h"

const int MAX_BUFF	= 8*1024;

// Remap 'BACKOFF_ABORTED' event to some existent SMBACL error
#define BACKOFF_ABORTED_ERROR SMBACL_MESSAGE_STALE

exasmb_acl_client::exasmb_acl_client(int a_domain) throw() :
	conn_type(CONN_TYPE_DIRECT),
	sockfd(-1),
	ctrl_buf(NULL),
	data_size(0)
,	domain(a_domain)
{
    ctrl_buf = new uint32_t[4096/sizeof(uint32_t)];
	XNT_ERROR("pid " << getpid());
	return;
}

exasmb_acl_client::~exasmb_acl_client()
{
	delete[] ctrl_buf;
	ctrl_buf = NULL;
}

smb_acl_error
exasmb_acl_client::call(
	const smb_acl_request *request, xdr_encode encode,
	smb_acl_message *reply, xdr_decode decode)
{
	smb_acl_error r;
	XNT_WARN("encoding...");
	const uint32_t *p = encode(request, ctrl_buf);
	XNT_WARN("sending: " << request->code << " p: " << p << " ctrl_buf: " <<
			 reinterpret_cast<const void*>(ctrl_buf));
	if ((r = send(reinterpret_cast<const char *>(p) -
				  reinterpret_cast<char *>(ctrl_buf))) != SMBACL_MESSAGE_OK)
		return r;
	if ((r = recv()) != SMBACL_MESSAGE_OK)
		return r;
	p = decode(reply, ctrl_buf);
	XNT_WARN("decoded: type: " << reply->type << " error: " << reply->error <<
			 " len: " << (char*) p - (char*) ctrl_buf);
	switch (conn_type) {
	case CONN_TYPE_DIRECT:
		break;
	case CONN_TYPE_REMOTE:
		break;
	}
	XNT_DEBUG("client: " << this << " done request: " << request <<
			  " result: " << reply->error);
    if ((smb_acl_error)reply->error == SMBACL_MESSAGE_COMM_ERROR ) {
		return backoff_result(SMBACL_MESSAGE_COMM_ERROR, BACKOFF_ABORTED_ERROR);
    }
	reset_backoff_timeout();
    return (smb_acl_error)reply->error;
}

smb_acl_error
exasmb_acl_client::send(int len)
{
	uint32_t msg_len = htonl(len);
	struct iovec iov[2] = {
		{&msg_len, sizeof(msg_len)},
		{ctrl_buf, len}
	};
	ssize_t sent = writev(iov, 2);
	XNT_DEBUG("msg len: " << len << " len: " << len+sizeof(msg_len) << " sent: " << sent);
	if (sent < len + int(sizeof(msg_len))) {
		XNT_DEBUG("writev failed - sent: " << sent << " - "<< strerror(errno));
		return backoff_result(SMBACL_MESSAGE_COMM_ERROR, BACKOFF_ABORTED_ERROR);
	}
	XNT_WARN("sent: " << sent << " len: " << len+sizeof(msg_len));
    	return SMBACL_MESSAGE_OK;
}

smb_acl_error
exasmb_acl_client::recv()
{
	uint32_t left;
	ssize_t n = read(&left, sizeof(left));
	if (n < static_cast<ssize_t>(sizeof(left))) {
		backoff();	// Note - backoff shouldn't abort here to preserve correct error code (not COMM_ERROR)
		return SMBACL_MESSAGE_GENERAL_ERROR;
	}
	left = ntohl(left);
	XNT_WARN("left: " << left);
	for (rcved = 0; left > 0; left -= n, rcved += n) {
		n = read(&ctrl_buf[rcved], left);
		if (n < 0) {
			XNT_WARN("read failed: n: " << n << " - " << strerror(errno));
			return backoff_result(SMBACL_MESSAGE_COMM_ERROR, BACKOFF_ABORTED_ERROR);
		}
	}
	XNT_WARN("rcved: " << rcved << " left: " << left);
    return SMBACL_MESSAGE_OK;
}

static inline int
exasmb_sun_len(const struct sockaddr_un& addr)
{
	int len = SUN_LEN(&addr);
	if (!addr.sun_path[0]) {
		len += strlen(&addr.sun_path[1]) + 2;
	}
	return len;
}

const char *exasmb_acl_client::socket_name() const
{
	static char name[64] = "";
	
	if (!name[0])
		sprintf(name, "smbacl.%d", domain);
	return name;
}

void exasmb_acl_client::connect()
{
	char buf[128];
	struct sockaddr_un addr;
	
	while (true) {
		if ((sockfd = ::socket(PF_UNIX, SOCK_STREAM, 0)) > 0) {
			::memset(&addr, 0, sizeof(addr));
			addr.sun_family = AF_LOCAL;
			::strcpy(&addr.sun_path[1], socket_name());
			if (::connect(sockfd, (struct sockaddr *) &addr,
						  exasmb_sun_len(addr)) == 0) {
				XNT_DEBUG("connected");
				break;
			} else  {
				sprintf(buf, "::connect() failed: %m - (this=%p,uid=%u,gid=%u)",
						this, geteuid(), getegid());
			}
		} else {
			sprintf(buf, "::socket(PF_UNIX, SOCK_STREAM, 0) failed");
			goto error_;
		}
error_:
		fprintf(stderr, "%s\n", buf);
		disconnect();
		sleep(1);
	}
}

void exasmb_acl_client::disconnect()
{
	XNT_HERE;
	if (sockfd != -1) {
		close(sockfd);
		sockfd = -1;
		XNT_DEBUG("disconnected");
	}
}

int exasmb_acl_client::read(void *buf, size_t count)
{
	if (sockfd == -1)
		return -1;
	int n = ::read(sockfd, buf, count);
	if (n < 0) {
		disconnect();
	}
	return n;
}

int exasmb_acl_client::writev(const struct iovec *vector, int count)
{
	if (sockfd == -1)
		connect();
	int n = ::writev(sockfd, vector, count);
	if (n < 0) {
		XNT_DEBUG("writev failed: n:" << n << " - " << strerror(errno));
		disconnect();
	}
	return n;
}


/******************************************************************************/

static exasmb_acl_client *cli = NULL;


extern "C" int32_t exasmb_getsmbid(void);
static void init_acl_client()
{
	if (!cli)
		cli = new exasmb_acl_client(exasmb_getsmbid());
}

int exasmb_acl_get_acl(int h_len, char *h_data, unixntacl_acl **acl)
{
	init_acl_client();
	smb_acl_get_acl_request req;
	smb_acl_get_acl_reply rep;
	smb_acl_error err;
	do {
		req.code = MESSAGE_CODE_GET_ACL;
		req.fh_len = h_len;
		req.fh_data = h_data;
		rep.acl = *acl;
		err = cli->call(&req, encode_get_acl_request, &rep, decode_get_acl_reply);
   } while (err == SMBACL_MESSAGE_COMM_ERROR);
	
    switch (err) {
    case SMBACL_MESSAGE_OK:
        return 0;
    case SMBACL_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
    case SMBACL_MESSAGE_STALE:
        errno = ESTALE;
        break;
    case SMBACL_MESSAGE_GENERAL_ERROR:
        errno = VFS_UNKNOWN_REQUEST_TYPE;
        break;
    case SMBACL_MESSAGE_NO_SEC_STYLE:
        errno = VFS_INTERNAL_FS_PROBLEM;
        break;
    case SMBACL_MESSAGE_READ_ERROR:
        errno = VFS_READ_ERROR;
        break;
	case SMBACL_MESSAGE_PERM:
        errno = VFS_PERM_DENIED;
        break;	
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }

    return -1;
}

int exasmb_acl_set_acl(int h_len, char *h_data, const unixntacl_acl *acl)
{
	init_acl_client();
	smb_acl_set_acl_request req;
	smb_acl_set_acl_reply rep;
	smb_acl_error err;
    do {
        req.code = MESSAGE_CODE_SET_ACL;
        req.fh_len = h_len;
        req.fh_data = h_data;
        req.acl = const_cast<unixntacl_acl *>(acl);
        err = cli->call(&req, encode_set_acl_request, &rep, decode_set_acl_reply);
    } while (err == SMBACL_MESSAGE_COMM_ERROR);

    switch (err) {
    case SMBACL_MESSAGE_OK:
        return 0;
    case SMBACL_MESSAGE_VOL_NO_SPACE_ERROR:
        errno = ENOSPC; 
        break;
    case SMBACL_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
    case SMBACL_MESSAGE_STALE:
        errno = ESTALE;
        break;
    case SMBACL_MESSAGE_GENERAL_ERROR:
        errno = VFS_UNKNOWN_REQUEST_TYPE;
        break;
    case SMBACL_MESSAGE_NO_SEC_STYLE:
        errno = VFS_INTERNAL_FS_PROBLEM;
        break;
    case SMBACL_MESSAGE_WRITE_ERROR:
        errno = VFS_READ_ERROR;
        break;
	case SMBACL_MESSAGE_RO_FS:
		errno = EROFS;
		break;
    case SMBACL_MESSAGE_PERM:
        errno = VFS_PERM_DENIED;
        break;
    
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }

    return -1;
}

int exasmb_acl_get_access(int h_len, char *h_data, uint32_t uid, uint32_t gid,
						  int ngroups, uint32_t *groups, uint32_t *max_access, char *access_troubleshooting_details)
{
	init_acl_client();
	smb_acl_get_access_request req;
	smb_acl_get_access_reply rep;
	smb_acl_error err;
	XNT_WARN("h_len: " << h_len << " h_data: " << (void*) h_data <<
			 " uid: " << uid << " gid: " << gid << " ngroups: " << ngroups);
	do {
		req.code = MESSAGE_CODE_GET_ACCESS;
		req.fh_len = h_len;
		req.fh_data = h_data;
		req.uid = uid;
		req.ngroups = 1;
		req.groups[0] = gid; /* the gid is the first group, which is why the next loop is limited odly */
		for (int i = 0; i < ngroups && i < (GROUPS_PER_USER - 1); ++i) {			
			if (groups[i] != gid) {
				req.groups[req.ngroups++] = groups[i];
			}
		}
		err = cli->call(&req, encode_get_access_request, &rep, decode_get_access_reply);
    } while (err == SMBACL_MESSAGE_COMM_ERROR);
	XNT_DEBUG("err: " << err << " rep.access: " << std::hex << rep.access);
	
    switch (err) {
    case SMBACL_MESSAGE_OK:
        *max_access = rep.access;
		strcpy(access_troubleshooting_details, rep.access_troubleshooting_details);
        return 0;
    case SMBACL_MESSAGE_VOL_NO_SPACE_ERROR:
        errno = ENOSPC; 
        break;
    case SMBACL_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
    case SMBACL_MESSAGE_STALE:
        errno = ESTALE;
        break;
    case SMBACL_MESSAGE_GENERAL_ERROR:
        errno = VFS_UNKNOWN_REQUEST_TYPE;
        break;
    case SMBACL_MESSAGE_NO_SEC_STYLE:
        errno = VFS_INTERNAL_FS_PROBLEM;
        break;
    case SMBACL_MESSAGE_READ_ERROR:
        errno = VFS_READ_ERROR;
        break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }

    return -1;
}

unixntacl_acl *exasmb_acl_new_acl(uint32_t type, int max_aces)
{
	unixntacl_acl	*acl = (unixntacl_acl *) calloc(1, sizeof(unixntacl_acl));
	if (!acl)
		return NULL;
	acl->aces = (unixntacl_acl_ace *)
		calloc(max_aces, sizeof(unixntacl_acl_ace));
	if (!acl->aces) {
		free(acl);
		return NULL;
	}
	acl->type = type;
	acl->ace_count = max_aces;
	return acl;
}

void exasmb_acl_free_acl(unixntacl_acl *acl)
{
	if (!acl)
		return;
	if (acl->aces) {
		free(acl->aces);
		acl->aces = NULL;
	}
	free(acl);
}

