#ifndef EXASMB_CLIENT_BACKOFF_H
#define EXASMB_CLIENT_BACKOFF_H

#include <time.h>
#include "xnt_error_logging.h"

/* These values are in nanoseconds, must be less than 1 sec */
#define MAXBACKOFF 999999999
#define STARTBACKOFF 1000000

/* Timeout to abort backoff algorithm, in milliseconds (to avoid infinite loops) */
#define ABORTBACKOFF 90000

struct exasmb_client_backoff {
	exasmb_client_backoff(	long int startbackoff = STARTBACKOFF,
							long int maxbackoff = MAXBACKOFF,
							long int abortbackoff = ABORTBACKOFF)
		: _startbackoff(startbackoff)
		, _maxbackoff(maxbackoff)
		, _abortbackoff(abortbackoff)
		, _totalbackoff(0)
	{
		_backoff_timer.tv_nsec = startbackoff;
		_backoff_timer.tv_sec = 0;
	}

    bool backoff()
	{
		_backoff_timer.tv_nsec += rand() % (_startbackoff/10);
		_backoff_timer.tv_nsec *= 2;
		if (_backoff_timer.tv_nsec > _maxbackoff) { /* Never go above max limit */
			_backoff_timer.tv_nsec = _maxbackoff;
		}
		_totalbackoff += _backoff_timer.tv_nsec / 1000000;	// Add current delay to total counter

		XNT_WARN("backing off - " << _backoff_timer.tv_nsec << " nanoseconds.");
		if(nanosleep(&_backoff_timer, NULL) != 0) {
			// nanosleep can be interrupted by a signal - that's not so critical for overall performance, ignore it
			XNT_WARN("backing off has failed: tried to sleep " << _backoff_timer.tv_nsec << " nanoseconds, errno=" << errno);
		}
		if(_totalbackoff >= _abortbackoff) {
			XNT_ERROR("backoff reached critical total timeout, aborting backoff algorithm to avoid infinite loops");
			return false;	// Return false if backoff should be aborted
		}
		return true;	// backoff ok
	}

	// Small convenience code to avoid condition operator duplication
	template<class ErrorType>
	inline ErrorType backoff_result(ErrorType backoff_ok_err, ErrorType abort_backoff_err)
	{
		return (backoff() == true) ? backoff_ok_err : abort_backoff_err;
	}

	inline void reset_backoff_timeout()
	{
        _backoff_timer.tv_nsec = _startbackoff;
		_totalbackoff = 0;
        XNT_TRACE("reset backoff counter to " << _startbackoff);
    }
private:
    const long int _startbackoff;
    const long int _maxbackoff;
	const long int _abortbackoff;

    struct timespec _backoff_timer;
	long int _totalbackoff;
};


#endif // EXASMB_CLIENT_BACKOFF_H
