#ifndef SMB_VFSFS_MESSAGE_H
#define SMB_VFSFS_MESSAGE_H

#include <stdint.h>
#include <sys/types.h>
#include "smb_acl_message.h"
#include "win_prev_vers.h"
#include "sid_struct.h"

struct smb_vfsfs_message {
	uint32_t	type;
    union {
		uint32_t		code;
		smb_acl_error	error;
    };
    
	smb_vfsfs_message() {}
	smb_vfsfs_message(const smb_vfsfs_message& msg)
	:	type(msg.type)
	{ 
        switch (type) {
		case MESSAGE_TYPE_REQUEST:
			code = msg.code;
			break;
		case MESSAGE_TYPE_REPLY:
			error = msg.error;
			break;
        }
	}
};

struct smb_vfsfs_request : smb_vfsfs_message {
	uint32_t	fh_len;
	char		*fh_data;
	
	smb_vfsfs_request() {}
	smb_vfsfs_request(const smb_vfsfs_request& req)
	:	smb_vfsfs_message(req)
	,	fh_len(req.fh_len)
	,	fh_data(req.fh_data)
	{
	}
};

struct smb_vfs_sendfile_request : smb_acl_request {
	uint64_t	offset;
	uint64_t	count;
	int		id;
	
	smb_vfs_sendfile_request() {}
	smb_vfs_sendfile_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_sendfile_reply : smb_acl_message {
	uint64_t	offset;
	uint64_t	count;
	
	smb_vfs_sendfile_reply() {}
	smb_vfs_sendfile_reply(const smb_acl_request& req)
	:	smb_acl_message(req)
	{
	}
};


/************************************/
/* Read File						*/
/************************************/

struct smb_vfs_readfile_request : smb_acl_request {
	uint64_t	offset;
	uint64_t	count;
	int			shmid;
	int		id;
	
	smb_vfs_readfile_request() {}
	smb_vfs_readfile_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_readfile_reply : smb_acl_message {
	uint64_t	offset;
	uint64_t	count;
	
	smb_vfs_readfile_reply(){}
    
    smb_vfs_readfile_reply(const smb_acl_request& req)
	:smb_acl_message(req)
	{
	}
};

/*********************************/
/*   write shared memory request */
/*********************************/
struct smb_vfs_shwrite_request : smb_acl_request {
	uint64_t	offset; // file offset
    uint64_t	buff_offset;// buffer offset
	size_t	count;
    int     shmid;
	int		id;
    int     unstable_flag;
	int		id_must_exist;
	int		update_mtime;
	
	smb_vfs_shwrite_request() {}
	smb_vfs_shwrite_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_shwrite_reply : smb_acl_message {
	size_t	count;
	int		unstable_pending;
	
	smb_vfs_shwrite_reply() {}
	smb_vfs_shwrite_reply(const smb_acl_request& req)
	:	smb_acl_message(req)
	{
	}
};

/*********************************/
/*   rename request */
/*********************************/
struct smb_vfs_rename_request : smb_acl_request {
	char from_name[1024];
    uint32_t from_name_len;
    char to_name[1024];
    uint32_t to_name_len;

	
	smb_vfs_rename_request() {}
    smb_vfs_rename_request(const smb_acl_request& req)
   :	smb_acl_request(req)
   {
   }
};

struct smb_vfs_rename_reply:smb_acl_message {
    smb_vfs_rename_reply() {}
	smb_vfs_rename_reply(const smb_acl_request& req)
	:	smb_acl_message(req)
	{
	}
};

/*********************************/
/*   fsync request */
/*********************************/
struct smb_vfs_fsync_request : smb_acl_request {
	int	id;
	
	smb_vfs_fsync_request() {}
	smb_vfs_fsync_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_fsync_reply : smb_acl_message {
	smb_vfs_fsync_reply() {}
	smb_vfs_fsync_reply(const smb_vfs_fsync_reply& rep)
	:	smb_acl_message(rep)
	{
	}
};

/*********************************/
/*   read dir plus request */
/*********************************/
struct smb_vfs_dirplus_request : smb_acl_request {
	uint64_t	cookie; // file offset
	size_t	size;
	int     dir_shmid;
	
	smb_vfs_dirplus_request() {}
	smb_vfs_dirplus_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_dirplus_reply : smb_acl_message {
	uint64_t	cookie;
	
	smb_vfs_dirplus_reply() {}
	smb_vfs_dirplus_reply(const smb_acl_request& req)
	:	smb_acl_message(req)
	{
	}
};

/**************************************/
/*   get_snapshots_timestamps request */
/**************************************/
struct smb_vfs_get_snapshots_timestamps_request : smb_acl_request {
	char volume_name[VFSFS_MAX_VOLUME_NAME_LENGTH + 1];
	uint32_t max_number_of_snapshots;

	smb_vfs_get_snapshots_timestamps_request()
	{}
	smb_vfs_get_snapshots_timestamps_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{}
};

struct smb_vfs_get_snapshots_timestamps_reply : smb_acl_message {
	uint32_t number_of_snapshots;
	uint32_t snapshots_timestamps[ VFSFS_MAX_LIMIT_NUMBER_OF_SNAPSHOTS_TO_LIST ];

	smb_vfs_get_snapshots_timestamps_reply() 
	: 	number_of_snapshots(0)
	{}
	smb_vfs_get_snapshots_timestamps_reply(const smb_vfs_get_snapshots_timestamps_reply& rep)
	:	smb_acl_message(rep)
	{}
	void add_timestamp( uint32_t timestamp_to_add ){
		snapshots_timestamps[number_of_snapshots++] = timestamp_to_add;
	}
};

struct smb_vfs_get_snapshot_name_by_timestamp_request : smb_acl_request {
	char volume_name[VFSFS_MAX_VOLUME_NAME_LENGTH + 1];
	uint32_t snapshot_timestamp;

	smb_vfs_get_snapshot_name_by_timestamp_request() 
	{}
	smb_vfs_get_snapshot_name_by_timestamp_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{}
};

struct smb_vfs_get_snapshot_name_by_timestamp_reply : smb_acl_message {
	bool snapshot_found;
	char snapshot_name[VFSFS_MAX_SNAPSHOT_NAME_LENGTH + 1];

	smb_vfs_get_snapshot_name_by_timestamp_reply() 
	{}
	smb_vfs_get_snapshot_name_by_timestamp_reply(const smb_vfs_get_snapshot_name_by_timestamp_reply& rep)
	:	smb_acl_message(rep)
	{}
};

/**************************************/
/*     antivirus requests/replies     */
/**************************************/

struct smb_vfs_get_antivirus_info_request : smb_acl_request {
	smb_vfs_get_antivirus_info_request() {}
	smb_vfs_get_antivirus_info_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_get_antivirus_info_reply : smb_acl_message {
	uint32_t val;

	smb_vfs_get_antivirus_info_reply() {}
	smb_vfs_get_antivirus_info_reply(const smb_vfs_get_antivirus_info_reply& rep)
	:	smb_acl_message(rep)
	{
	}
};

struct smb_vfs_set_antivirus_info_request : smb_acl_request {
	uint32_t val;

	smb_vfs_set_antivirus_info_request() {}
	smb_vfs_set_antivirus_info_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_set_antivirus_info_reply : smb_acl_message {
	smb_vfs_set_antivirus_info_reply() {}
	smb_vfs_set_antivirus_info_reply(const smb_vfs_set_antivirus_info_reply& rep)
	:	smb_acl_message(rep)
	{
	}
};

struct smb_vfs_set_antivirus_info_guarded_request : smb_acl_request {
	uint32_t val;
	uint32_t guard;

	smb_vfs_set_antivirus_info_guarded_request() {}
	smb_vfs_set_antivirus_info_guarded_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_vfs_set_antivirus_info_guarded_reply : smb_acl_message {
	uint32_t val;

	smb_vfs_set_antivirus_info_guarded_reply() {}
	smb_vfs_set_antivirus_info_guarded_reply(const smb_vfs_set_antivirus_info_guarded_reply& rep)
	:	smb_acl_message(rep)
	{
	}
};

/**************************************/
/*   id mapping requests / replies    */
/**************************************/
struct smb_vfs_map_temp_id_to_sid_request : smb_acl_request {
	uint32_t temp_id;

	smb_vfs_map_temp_id_to_sid_request()
	{}
	smb_vfs_map_temp_id_to_sid_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{}
};

struct smb_vfs_map_temp_id_to_sid_reply : smb_acl_message {
	sid_struct sid;

	smb_vfs_map_temp_id_to_sid_reply() 
	{}
	smb_vfs_map_temp_id_to_sid_reply(const smb_vfs_map_temp_id_to_sid_reply& rep)
	:	smb_acl_message(rep)
	{}
};

struct smb_vfs_map_sid_to_temp_id_request : smb_acl_request {
	sid_struct sid;

	smb_vfs_map_sid_to_temp_id_request()
	{}
	smb_vfs_map_sid_to_temp_id_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{}
};

struct smb_vfs_map_sid_to_temp_id_reply : smb_acl_message {
	uint32_t temp_id;

	smb_vfs_map_sid_to_temp_id_reply() 
	{}
	smb_vfs_map_sid_to_temp_id_reply(const smb_vfs_map_sid_to_temp_id_reply& rep)
	:	smb_acl_message(rep)
	{}
};

/******************************************/
/*   generic request (passing only NFS FH */
/******************************************/
// do we really need this ?  why not use smb_acl_request ?
/*
struct smb_vfs_request : smb_acl_request {
	smb_vfs_request() {}
	smb_vfs_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};
*/

/*********************************/
/*   statfs reply                */
/*********************************/
struct smb_vfs_statfs_reply : smb_acl_message {
	unsigned long long bsize;
	unsigned long long dfree;
    unsigned long long dsize;
	smb_vfs_statfs_reply() {}
	smb_vfs_statfs_reply(const smb_vfs_statfs_reply& rep)
	:	smb_acl_message(rep)
	{
	}
};


/**************************************/
/*   set create time request    */
/**************************************/
struct smb_vfs_set_create_time_request : smb_acl_request {
	uint32_t create_time_sec;
	uint32_t create_time_nsec;

	smb_vfs_set_create_time_request()
	{}
	smb_vfs_set_create_time_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{}
};

struct smb_vfs_set_create_time_reply:smb_acl_message {
    smb_vfs_set_create_time_reply() {}
	smb_vfs_set_create_time_reply(const smb_acl_request& req)
	:	smb_acl_message(req)
	{
	}
};


#ifdef	__cplusplus
extern "C" {
#endif


uint32_t *encode_read_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_read_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_read_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_read_reply(smb_acl_message *rep, const uint32_t *p);
	
uint32_t *encode_sendfile_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_sendfile_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_sendfile_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_sendfile_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_shwrite_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_shwrite_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_shwrite_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_shwrite_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_rename_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_rename_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_rename_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_rename_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_fsync_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_fsync_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_fsync_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_fsync_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_dirplus_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_dirplus_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_dirplus_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_dirplus_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_statfs_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_statfs_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_statfs_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_statfs_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_get_snapshots_timestamps_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_get_snapshots_timestamps_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_get_snapshots_timestamps_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_get_snapshots_timestamps_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_get_snapshot_name_by_timestamp_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_get_snapshot_name_by_timestamp_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_get_snapshot_name_by_timestamp_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_get_snapshot_name_by_timestamp_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_get_antivirus_info_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_get_antivirus_info_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_get_antivirus_info_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_get_antivirus_info_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_set_antivirus_info_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_set_antivirus_info_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_set_antivirus_info_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_set_antivirus_info_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_set_antivirus_info_guarded_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_set_antivirus_info_guarded_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_set_antivirus_info_guarded_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_set_antivirus_info_guarded_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_map_sid_to_temp_id_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_map_sid_to_temp_id_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_map_sid_to_temp_id_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_map_sid_to_temp_id_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_map_temp_id_to_sid_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_map_temp_id_to_sid_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_map_temp_id_to_sid_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_map_temp_id_to_sid_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_set_create_time_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_set_create_time_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_set_create_time_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_set_create_time_reply(smb_acl_message *rep, const uint32_t *p);

#ifdef	__cplusplus
}
#endif

#endif // #ifndef SMB_VFSFS_MESSAGE_H


