#include <netinet/in.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/**/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
/**/

// declare dlm_client non-const
#define	EXASMB_DLMCLIENT_SOURCE

#define XNT_USING_STD_CERR	// dumping errors to std::cerr
#include "xnt_error_logging.h"

#include "exasmb_vfsfs_client.h"
#include "exasmb_vfsfsclient.h"
#include "exasmb_vfsfs_message.h"

#ifdef XNT_CRITICAL
#undef XNT_CRITICAL
#endif

#define XNT_CRITICAL(x) PRINT_WARN(x)

const int MAX_BUFF	= 8*1024;

// Remap 'BACKOFF_ABORTED' event to some existent SMBVFSFS error
#define BACKOFF_ABORTED_ERROR SMBVFSFS_MESSAGE_STALE_ERROR

exasmb_vfsfs_client::exasmb_vfsfs_client(int a_domain) throw() :
	conn_type(CONN_TYPE_DIRECT),
	sockfd(-1),
	ctrl_buf(NULL),
	data_size(0)
,	domain(a_domain)
,	ancillary_fd(-1)
,	ancillary_fd_sent(false)

{
    ctrl_buf = new uint32_t[1024/sizeof(uint32_t)];
	XNT_ERROR("pid " << getpid());
    XNT_ERROR("domain " << domain);
	return;
}

exasmb_vfsfs_client::~exasmb_vfsfs_client()
{
	delete[] ctrl_buf;
	ctrl_buf = NULL;
}

smb_acl_error
exasmb_vfsfs_client::call(
	const smb_acl_request *request, xdr_encode encode,
	smb_acl_message *reply, xdr_decode decode, int fd, bool overide_fd_sent)
{
    /*The ancillary has to be update only if the negotiate partner do not have it
     Or if for some reason the ancillary fd has changed 
    */
    if (overide_fd_sent)
        ancillary_fd_sent = false;

    if ((!ancillary_fd_sent) || ((ancillary_fd != fd) && (fd != -1))) {
        ancillary_fd = fd;
        ancillary_fd_sent = false;
    }

	smb_acl_error r;
	XNT_WARN("encoding...");
	const uint32_t *p = encode(request, ctrl_buf);
	XNT_WARN("sending: " << request->code << " p: " << p << " ctrl_buf: " <<
			 reinterpret_cast<const void*>(ctrl_buf));
	if ((r = send(reinterpret_cast<const char *>(p) -
				  reinterpret_cast<char *>(ctrl_buf))) != SMBVFSFS_MESSAGE_OK)
		return r;
	if ((r = recv()) != SMBVFSFS_MESSAGE_OK)
		return r;
	p = decode(reply, ctrl_buf);
	XNT_WARN("decoded: type: " << reply->type << " error: " << reply->error <<
			 " len: " << (char*) p - (char*) ctrl_buf);
	switch (conn_type) {
	case CONN_TYPE_DIRECT:
		break;
	case CONN_TYPE_REMOTE:
		break;
	}
	XNT_DEBUG("client: " << this << " done request: " << request <<
			  " result: " << reply->error);
	if (reply->error == SMBVFSFS_MESSAGE_COMM_ERROR) {
        if(backoff() == false) return BACKOFF_ABORTED_ERROR;
    } else
        reset_backoff_timeout();
    XNT_DEBUG("reply->error is " << reply->error);
    return reply->error;
}

smb_acl_error
exasmb_vfsfs_client::send(int len)
{
	uint32_t msg_len = htonl(len);
	struct iovec iov[2] = {
		{&msg_len, sizeof(msg_len)},
		{ctrl_buf, len}
	};
	ssize_t sent = writev(iov, 2);
	XNT_DEBUG("msg len: " << len << " len: " << len+sizeof(msg_len) << " sent: " << sent);
	if (sent < len + int(sizeof(msg_len))) {
		XNT_DEBUG("writev failed - sent: " << sent << " - "<< strerror(errno));
		return backoff_result(SMBVFSFS_MESSAGE_COMM_ERROR, BACKOFF_ABORTED_ERROR);
	}
    XNT_WARN("sent: " << sent << " len: " << len+sizeof(msg_len));
	return SMBVFSFS_MESSAGE_OK;
}

smb_acl_error
exasmb_vfsfs_client::recv()
{
	uint32_t left;
	ssize_t n = read(&left, sizeof(left));
	if ((uint32_t)n < sizeof(left)) {
		backoff();  // Note - backoff shouldn't abort here to preserve correct error code (not COMM_ERROR)
		return SMBVFSFS_MESSAGE_SOCKET_CLOSED_ERROR;
	}
	left = ntohl(left);
	XNT_WARN("left: " << left);
	for (rcved = 0; left > 0; left -= n, rcved += n) {
		n = read(&ctrl_buf[rcved], left);
		if (n < 0) {
			XNT_WARN("read failed: n: " << n << " - " << strerror(errno));
			return backoff_result(SMBVFSFS_MESSAGE_COMM_ERROR, BACKOFF_ABORTED_ERROR);
		}
	}
	XNT_WARN("rcved: " << rcved << " left: " << left);
	return SMBVFSFS_MESSAGE_OK;
}

static inline int
exasmb_sun_len(const struct sockaddr_un& addr)
{
	int len = SUN_LEN(&addr);
	if (!addr.sun_path[0]) {
		len += strlen(&addr.sun_path[1]) + 2;
	}
	return len;
}

const char *exasmb_vfsfs_client::socket_name() const
{
	static char name[64] = "";
	
	if (!name[0])
		sprintf(name, "smbacl.%d", domain);
	return name;
}

void exasmb_vfsfs_client::connect()
{
	char buf[128];
	struct sockaddr_un addr;
	XNT_DEBUG("enter");
	while (true) {
		if ((sockfd = ::socket(PF_UNIX, SOCK_STREAM, 0)) > 0) {
			::memset(&addr, 0, sizeof(addr));
			addr.sun_family = AF_LOCAL;
			::strcpy(&addr.sun_path[1], socket_name());
			if (::connect(sockfd, (struct sockaddr *) &addr,
						  exasmb_sun_len(addr)) == 0) {
				XNT_DEBUG("connected");
				break;
			} else  {
				sprintf(buf, "::connect() failed: %m - (this=%p,uid=%u,gid=%u)",
						this, geteuid(), getegid());
			}
		} else {
			sprintf(buf, "::socket(PF_UNIX, SOCK_STREAM, 0) failed");
			goto error_;
		}
error_:
		fprintf(stderr, "%s\n", buf);
		disconnect();
		sleep(1);
	}
}

void exasmb_vfsfs_client::disconnect()
{
	XNT_HERE;
	if (sockfd != -1) {
		close(sockfd);
		sockfd = -1;
		XNT_DEBUG("disconnected");
	}
}

int exasmb_vfsfs_client::read(void *buf, size_t count)
{
    int n;
	if (sockfd == -1)
		return -1;
    while (1)
    {
        n = ::read(sockfd, buf, count);
        if (n < 0 && errno == EINTR)
            continue;
        else
            break;
    }
	
	if (n < 0) {
		disconnect();
	}
	return n;
}

int exasmb_vfsfs_client::writev(const struct iovec *vector, int count)
{
	char cmsg_buf[CMSG_SPACE(sizeof(int))]; // in case we need it
	struct msghdr msg;
	
	if (sockfd == -1)
		connect();
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_iov = const_cast<struct iovec*>(vector);
	msg.msg_iovlen = count;
	if (!ancillary_fd_sent && (ancillary_fd != -1)) {
		msg.msg_control = cmsg_buf;
		msg.msg_controllen = sizeof cmsg_buf;
		
		cmsghdr *cmsg = CMSG_FIRSTHDR(&msg);
		
		cmsg->cmsg_len = CMSG_LEN(sizeof(int));
		cmsg->cmsg_level = SOL_SOCKET;
		cmsg->cmsg_type = SCM_RIGHTS;
		memcpy(CMSG_DATA(cmsg), &ancillary_fd, sizeof(int));
	} else {
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
	}
	msg.msg_flags = 0;
    XNT_DEBUG("writev before sendmsg");
	int n = sendmsg(sockfd, &msg, 0);
    if (n < 0) {
		XNT_DEBUG("writev failed: n:" << n << " - " << strerror(errno));
		disconnect();
        ancillary_fd_sent = false;
	}
    else {
        /*If Send had succeeded, there is no need to send it again later*/
        ancillary_fd_sent = true;
    }
	return n;
}



/******************************************************************************/

static exasmb_vfsfs_client *cli = NULL;

extern "C" int32_t exasmb_getsmbid(void);
extern "C" void* exasmb_vfsfs_client_ctx;
static void init_vfsfs_client()
{
	if ( exasmb_vfsfs_client_ctx != NULL ) {
		cli = reinterpret_cast<exasmb_vfsfs_client*>( exasmb_vfsfs_client_ctx );
	} else {
		cli = new exasmb_vfsfs_client(exasmb_getsmbid());
		exasmb_vfsfs_client_ctx = reinterpret_cast<void*>( cli );
	}
}

ssize_t exasmb_vfs_read(const EXA_NFS_FH fh, int shmid, int64_t offset, size_t count, int id)
{
	XNT_DEBUG("exasmb_vfs_read:enter\n");
	init_vfsfs_client();
	smb_vfs_readfile_request req;
	smb_vfs_readfile_reply rep;
	smb_acl_error err;

	XNT_WARN("EXA_NFS_FH: " << (void*) fh << " offset: " << offset << " count: " << count);
	do {
		req.code = MESSAGE_CODE_READ;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
		req.offset = offset;
		req.count = count;
		req.shmid = shmid;
		req.id = id; 
		err = cli->call(&req, encode_read_request,
						&rep, decode_read_reply);
		XNT_WARN("call err:" << err);
	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	/* read file returns for now only the below errors */
	switch (err) {
	case SMBVFSFS_MESSAGE_OK:
		return rep.count;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_STALE_ERROR:
		errno = ESTALE;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1; 
}


ssize_t exasmb_vfs_sendfile(int tofd, const EXA_NFS_FH fh, int64_t offset, size_t count, int id)
{
    XNT_DEBUG("exasmb_vfs_sendfile:enter\n");
	init_vfsfs_client();
    XNT_DEBUG("exasmb_vfs_sendfile:after init_vfsfs_client\n");
	smb_vfs_sendfile_request req;
	smb_vfs_sendfile_reply rep;
	smb_acl_error err = SMBVFSFS_MESSAGE_OK;
    bool first_try = true;
    
    /*The ancillary fd cannot be -1 in sendfile*/
    if (tofd == -1) {
        XNT_WARN("exasmb_vfs_sendfile: tofd is -1");
        errno = VFS_UNEXPECTED_ERROR;
        return -1;
    }
	XNT_WARN("tofd: " << tofd << " EXA_NFS_FH: " << (void*) fh <<
			 " offset: " << offset << " count: " << count);
    do {
        req.code = MESSAGE_CODE_SENDFILE;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
		req.offset = offset;
		req.count = count;
		req.id = id;
        XNT_DEBUG("exasmb_vfs_sendfile: before call\n");
		err = cli->call(&req, encode_sendfile_request,
						&rep, decode_sendfile_reply, tofd, 
						(err == SMBVFSFS_MESSAGE_ANCILLARY_COMM_ERROR));
        XNT_WARN("call err:" << err);
        first_try = false;
    } while ((err == SMBVFSFS_MESSAGE_COMM_ERROR) ||
             /*If an ancillary comm_error returned, than probably a new fd could not be opened*/
             (err == SMBVFSFS_MESSAGE_ANCILLARY_COMM_ERROR));
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}
    
    switch (err) {
    case SMBVFSFS_MESSAGE_OK:
        return req.count;
    case SMBVFSFS_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	case BACKOFF_ABORTED_ERROR:
		/* On backoff abort returning some distinctive error code to detect it in upper layers */
		errno = VFS_OPERATION_FATAL_ERROR;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
    return -1;
}

/*
 * write request from fs vfs layer
*/
ssize_t exasmb_vfs_shwrite(int shmid, const EXA_NFS_FH fh, off_t offset, size_t buff_offset, size_t count, int id,
						   int *id_must_exist, int unstable_flag, bool update_mtime)
{
    XNT_DEBUG("enter\n");
	init_vfsfs_client();
    smb_vfs_shwrite_request req;
	smb_vfs_shwrite_reply rep;
	smb_acl_error err;
    
	XNT_WARN("shmid " << shmid << " EXA_NFS_FH: " << (void*) fh <<
			 " offset: " << offset << " count: " << count << " buff_offset: " << buff_offset);
	do {
		req.code = MESSAGE_CODE_SHWRITE;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
		req.offset = (uint64_t) offset; //file offset
        req.buff_offset = (uint64_t) buff_offset; // buffer offset
		req.count = count;
        req.shmid = shmid;
		req.id = id;
        req.unstable_flag = unstable_flag;
		req.id_must_exist = *id_must_exist;
		req.update_mtime = update_mtime;
        XNT_DEBUG("before call");
		err = cli->call(&req, encode_shwrite_request,
						&rep, decode_shwrite_reply);
        XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

    switch (err) {
    case  SMBVFSFS_MESSAGE_OK:
        *id_must_exist = rep.unstable_pending;
		return ssize_t(rep.count);
    case SMBVFSFS_MESSAGE_STALE_ERROR:
        errno = ESTALE;
        break;
    case SMBVFSFS_MESSAGE_FILE_MAXIMUM_SIZE_EXCEED_ERROR:
        errno = EFBIG;
        break;
    case SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR:
        errno = ENOSPC; 
        break;
    case SMBVFSFS_MESSAGE_DQUOTA_ERROR:
        errno = EDQUOT;
        break;
    case SMBVFSFS_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
    case SMBVFSFS_MESSAGE_OPERATION_FATAL_ERROR:
        errno = VFS_OPERATION_FATAL_ERROR;
        break;
    case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
        errno = VFS_CLIENT_DISCONNECTED_ERROR;
        break;
	case SMBVFSFS_MESSAGE_RO_FS:
		errno = EROFS;
		break;
	case SMBVFSFS_MESSAGE_PERM_ERROR:
		errno = VFS_PERM_DENIED;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	case SMBVFSFS_MESSAGE_SOCKET_CLOSED_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
	return -1;
}


int exasmb_vfs_rename(const EXA_NFS_FH nfs_fh, const char *from_name, const char *to_name){
    XNT_DEBUG("enter\n");
    init_vfsfs_client();
    smb_vfs_rename_request req;
    smb_vfs_rename_reply rep;
    smb_acl_error err;

    XNT_WARN("from name "<< from_name << " to_name " << to_name);
    do {
		req.code = MESSAGE_CODE_RENAME;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(nfs_fh);
        req.from_name_len = strlen(from_name);
        strncpy(req.from_name, from_name, 1024);
        req.to_name_len = strlen(to_name);
        strncpy(req.to_name, to_name, 1024);
        
        XNT_DEBUG("before call");
		err = cli->call(&req, encode_rename_request, &rep, decode_rename_reply);
        XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

    switch (err) {
    case  SMBVFSFS_MESSAGE_OK:
        return 0;
    case SMBVFSFS_MESSAGE_FILE_NOT_FOUND_ERROR:
        errno = ENOENT;
        break;
    case SMBVFSFS_MESSAGE_STALE_ERROR:
        errno = ESTALE;
        break;
    case SMBVFSFS_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
	case SMBVFSFS_MESSAGE_RO_FS:
		errno = EROFS;
		break;
	case SMBVFSFS_MESSAGE_PERM_ERROR:
		errno = VFS_PERM_DENIED;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
   
    return -1;       
}

int exasmb_vfs_fsync(const EXA_NFS_FH fh, int id)
{
    XNT_DEBUG("enter\n");
	init_vfsfs_client();
    smb_vfs_fsync_request req;
	smb_vfs_fsync_reply rep;
	smb_acl_error err;
    
	XNT_WARN("EXA_NFS_FH: " << (void*) fh << " id: " << (void*)id);
	do {
		req.code = MESSAGE_CODE_FSYNC;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
		req.id = id;
        XNT_DEBUG("before call");
		err = cli->call(&req, encode_fsync_request,
						&rep, decode_fsync_reply);
        XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

    switch (err) { 
    case SMBVFSFS_MESSAGE_OK:
        return 0;
    case SMBVFSFS_MESSAGE_STALE_ERROR:
        errno = ESTALE;
        break;
    case SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR:
        errno = ENOSPC; 
        break;
    case SMBVFSFS_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
    case SMBVFSFS_MESSAGE_OPERATION_FATAL_ERROR:
        errno = VFS_OPERATION_FATAL_ERROR;
        break;
    case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
        errno = VFS_CLIENT_DISCONNECTED_ERROR;
        break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	case SMBVFSFS_MESSAGE_RO_FS:
		errno = EROFS;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
	
	return -1;
}

int exasmb_vfs_dirplus(int shmid, size_t shmsize, const EXA_NFS_FH fh, uint64_t *cookie)
{
    XNT_DEBUG("enter\n");
	init_vfsfs_client();
    smb_vfs_dirplus_request req;
	smb_vfs_dirplus_reply rep;
	smb_acl_error err;
    
    do {
		req.code = MESSAGE_CODE_DIRPLUS;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
		req.cookie = *cookie;
		req.size = shmsize;
		req.dir_shmid = shmid;
        XNT_DEBUG("before call");
		err = cli->call(&req, encode_dirplus_request,
						&rep, decode_dirplus_reply);
        XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}
	    
    switch (err) {
    case SMBVFSFS_MESSAGE_OK:
        *cookie = rep.cookie;
		if (*cookie != ~0ull)
			(*cookie)++;
		return 0;
    case SMBVFSFS_MESSAGE_STALE_ERROR:
        errno = ESTALE;
        break;
    case SMBVFSFS_MESSAGE_FATAL_ERROR:
        errno = VFS_FATAL_ERROR;
        break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
    return -1;
	
}

uint32_t
exasmb_vfs_get_snap_id(const EXA_NFS_FH fh)
{
        return ntohl(*reinterpret_cast<const uint32_t*>(fh+8));
}


int exasmb_get_snapshots_timestamps( const char* volume_name, 
									 uint32_t max_number_of_snapshots,
									 uint32_t snapshots_timestamps[],
									 uint32_t* number_of_snapshots )
{
	XNT_DEBUG("enter\n");
	init_vfsfs_client();
	smb_vfs_get_snapshots_timestamps_request req;
	smb_vfs_get_snapshots_timestamps_reply rep;
	smb_acl_error err;

	do {
		req.code = MESSAGE_CODE_GET_SNAPSHOTS_TIMESTAMPS;
		req.fh_len = 0;
		strcpy( req.volume_name, volume_name );
		req.max_number_of_snapshots = max_number_of_snapshots;
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_get_snapshots_timestamps_request,
						&rep, decode_get_snapshots_timestamps_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}
	
	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		*number_of_snapshots = rep.number_of_snapshots;
		for ( uint32_t i = 0; i < (*number_of_snapshots) ; ++i ) {
			snapshots_timestamps[i] = rep.snapshots_timestamps[i];
		}
		return 0;
	case SMBVFSFS_MESSAGE_VOLUME_NOT_FOUND:
		errno = ENOENT;
		break;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

int exasmb_get_snapshot_name_by_timestamp( const char* volume_name,
										   uint32_t snapshot_timestamp,
										   int* snapshot_found,
										   char snapshot_name[] )
{
	XNT_DEBUG("enter\n");
	init_vfsfs_client();
	smb_vfs_get_snapshot_name_by_timestamp_request req;
	smb_vfs_get_snapshot_name_by_timestamp_reply rep;
	smb_acl_error err;

	do {
		req.code = MESSAGE_CODE_GET_SNAPSHOT_NAME_BY_TIMESTAMP;
		req.fh_len = 0;
		strcpy( req.volume_name, volume_name );
		req.snapshot_timestamp = snapshot_timestamp;
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_get_snapshot_name_by_timestamp_request,
						&rep, decode_get_snapshot_name_by_timestamp_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		if ( rep.snapshot_found ) {
			*snapshot_found = 1;
			strcpy( snapshot_name, rep.snapshot_name );
		} else {
			*snapshot_found = 0;
		}
		return 0;
	case SMBVFSFS_MESSAGE_VOLUME_NOT_FOUND:
		errno = ENOENT;
		break;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

int exasmb_get_antivirus_info( const char* nfs_fh_data, 
							   uint32_t nfs_fh_len, 
							   uint32_t* p_info_out )
{
	XNT_DEBUG("enter\n");

	if ( p_info_out == NULL ) {
		errno = EINVAL;
		return -1;
	}

	init_vfsfs_client();
	smb_vfs_get_antivirus_info_request req;
	smb_vfs_get_antivirus_info_reply rep;
	smb_acl_error err;

	XNT_WARN("EXA_NFS_FH: " << (void*) fh);
	do {
		req.code = MESSAGE_CODE_GET_ANTIVIRUS_INFO;
		req.fh_len = nfs_fh_len;
		req.fh_data = const_cast< char* >( nfs_fh_data );
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_get_antivirus_info_request,
						&rep, decode_get_antivirus_info_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		*p_info_out = rep.val;
		return 0;
	case SMBVFSFS_MESSAGE_STALE_ERROR:
		errno = ESTALE;
		break;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

int exasmb_set_antivirus_info( const char* nfs_fh_data, 
							   uint32_t nfs_fh_len, 
							   uint32_t val )
{
	XNT_DEBUG("enter\n");

	if ( exasmb_vfs_get_snap_id( nfs_fh_data ) != 0 ) {
		errno = EROFS;
		return -1;
	}

	init_vfsfs_client();
	smb_vfs_set_antivirus_info_request req;
	smb_vfs_set_antivirus_info_reply rep;
	smb_acl_error err;

	XNT_WARN("EXA_NFS_FH: " << (void*) fh << " id: " << (void*)id);
	do {
		req.code = MESSAGE_CODE_SET_ANTIVIRUS_INFO;
		req.fh_len = nfs_fh_len;
		req.fh_data = const_cast< char* >( nfs_fh_data );
		req.val = val;
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_set_antivirus_info_request,
						&rep, decode_set_antivirus_info_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		return 0;
	case SMBVFSFS_MESSAGE_STALE_ERROR:
		errno = ESTALE;
		break;
	case SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR:
		errno = ENOSPC; 
		break;
	case SMBVFSFS_MESSAGE_RO_FS:
		errno = EROFS;
		break;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

int exasmb_set_antivirus_info_guarded( const char* nfs_fh_data, 
									   uint32_t nfs_fh_len, 
									   uint32_t val, 
									   uint32_t guard, 
									   uint32_t* p_info_out )
{
	XNT_DEBUG("enter\n");

	if ( exasmb_vfs_get_snap_id( nfs_fh_data ) != 0 ) {
		errno = EROFS;
		return -1;
	}

	init_vfsfs_client();
	smb_vfs_set_antivirus_info_guarded_request req;
	smb_vfs_set_antivirus_info_guarded_reply rep;
	smb_acl_error err;

	XNT_WARN("EXA_NFS_FH: " << (void*) fh << " id: " << (void*)id);
	do {
		req.code = MESSAGE_CODE_SET_ANTIVIRUS_INFO_GUARDED;
		req.fh_len = nfs_fh_len;
		req.fh_data = const_cast< char* >( nfs_fh_data );
		req.val = val;
		req.guard = guard;
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_set_antivirus_info_guarded_request,
						&rep, decode_set_antivirus_info_guarded_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		*p_info_out = rep.val;
		return 0;
	case SMBVFSFS_MESSAGE_STALE_ERROR:
		errno = ESTALE;
		break;
	case SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR:
		errno = ENOSPC; 
		break;
	case SMBVFSFS_MESSAGE_RO_FS:
		errno = EROFS;
		break;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

char* exasmb_vfs_sid_to_string( const sid_struct* p_sid_struct, char* buffer, uint32_t buffer_size )
{
	std::string sid_str = p_sid_struct->to_string();
	if ( sid_str.length() + 1 > buffer_size ) {
		return "<BUFFER IS TO SMALL FOR SID>";
	}

	::strcpy( buffer, sid_str.c_str() );
	return buffer;
}

int exasmb_vfs_map_sid_to_temp_id( const sid_struct* p_sid, uint32_t* p_temp_id_out )
{
	XNT_DEBUG("enter\n");

	if ( p_sid == NULL ) {
		errno = EINVAL;
		return -1;
	}
	if ( p_temp_id_out == NULL ) {
		errno = EINVAL;
		return -1;
	}

	init_vfsfs_client();
	smb_vfs_map_sid_to_temp_id_request req;
	smb_vfs_map_sid_to_temp_id_reply rep;
	smb_acl_error err;

	do {
		req.code = MESSAGE_CODE_MAP_SID_TO_TEMP_ID;
		req.fh_len = 0;
		req.sid = *p_sid;
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_map_sid_to_temp_id_request,
						&rep, decode_map_sid_to_temp_id_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		*p_temp_id_out = rep.temp_id;
		return 0;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

int exasmb_vfs_map_temp_id_to_sid( uint32_t temp_id, sid_struct* p_sid_out )
{
	XNT_DEBUG("enter\n");

	if ( p_sid_out == NULL ) {
		errno = EINVAL;
		return -1;
	}

	init_vfsfs_client();
	smb_vfs_map_temp_id_to_sid_request req;
	smb_vfs_map_temp_id_to_sid_reply rep;
	smb_acl_error err;

	do {
		req.code = MESSAGE_CODE_MAP_TEMP_ID_TO_SID;
		req.fh_len = 0;
		req.temp_id = temp_id;
		XNT_DEBUG("before call");
		err = cli->call(&req, encode_map_temp_id_to_sid_request,
						&rep, decode_map_temp_id_to_sid_reply);
		XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

	switch (err) { 
	case SMBVFSFS_MESSAGE_OK:
		*p_sid_out = rep.sid;
		return 0;
	case SMBVFSFS_MESSAGE_FATAL_ERROR:
		errno = VFS_FATAL_ERROR;
		break;
	case SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR:
		errno = VFS_CLIENT_DISCONNECTED_ERROR;
		break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
	default:
		errno = VFS_UNEXPECTED_ERROR;
	}
	return -1;
}

int exasmb_vfs_statfs(const EXA_NFS_FH fh, uint64_t *bsize, uint64_t *dfree, uint64_t*dsize)
{
    XNT_DEBUG("enter\n");
	init_vfsfs_client();
    smb_acl_request req;
	smb_vfs_statfs_reply rep;
	smb_acl_error err;
    
	do {
		req.code = MESSAGE_CODE_STATFS;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
        XNT_DEBUG("before call");
		err = cli->call(&req, encode_statfs_request,
						&rep, decode_statfs_reply);
        XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}

    switch (err) { 
    case SMBVFSFS_MESSAGE_OK:
		*bsize = rep.bsize;
		*dfree = rep.dfree;
		*dsize = rep.dsize;
        return 0;
    case SMBVFSFS_MESSAGE_STALE_ERROR:
        errno = ESTALE;
        break;
    case SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR:
        errno = ENOSPC; 
        break;
    case SMBVFSFS_MESSAGE_OPERATION_FATAL_ERROR:
        errno = VFS_OPERATION_FATAL_ERROR;
        break;
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
	
	return -1;
}

int exasmb_vfs_set_create_time(const EXA_NFS_FH fh, uint32_t create_time_sec, uint32_t create_time_nsec)
{
	XNT_DEBUG("enter\n");
	init_vfsfs_client();
	smb_vfs_set_create_time_request req;
	smb_vfs_set_create_time_reply rep;
	smb_acl_error err;
	
	do {
		req.code = MESSAGE_CODE_SET_CREATE_TIME;
		req.fh_len = 32;
		req.fh_data = const_cast<char *>(fh);
		req.create_time_sec = create_time_sec;
		req.create_time_nsec = create_time_nsec;
        XNT_DEBUG("before call");
		err = cli->call(&req, encode_set_create_time_request,
						&rep, decode_set_create_time_reply);
        XNT_WARN("call err:" << err);

	} while (err == SMBVFSFS_MESSAGE_COMM_ERROR);
	
	if ( err != SMBVFSFS_MESSAGE_OK ) {
		XNT_CRITICAL("while exited with err = " << err);
	}
	switch (err) { 
    case SMBVFSFS_MESSAGE_OK:
        return 0;
    case SMBVFSFS_MESSAGE_STALE_ERROR:
        errno = ESTALE;
        break;
    case SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR:
        errno = ENOSPC; 
        break;
    case SMBVFSFS_MESSAGE_OPERATION_FATAL_ERROR:
        errno = VFS_OPERATION_FATAL_ERROR;
        break;
	case SMBVFSFS_MESSAGE_RO_FS:
		errno = EROFS;
		break;
	case SMBVFSFS_MESSAGE_PERM_ERROR:
		errno = VFS_PERM_DENIED;
		break;	
	case SMBVFSFS_MESSAGE_NOT_SUPPORTED:
		errno = VFS_NOT_SUPPORTED;
		break;
    default:
        errno = VFS_UNEXPECTED_ERROR;
    }
	
	return -1;

}

