/**
 * @(#) sid_struct.cpp
 */

#include "sid_struct.h"
#ifdef __cplusplus
#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <cassert>
#include <errno.h>

sid_struct& 
sid_struct::operator=(const sid_struct& sid)
{
	if (&sid != this) {
	   _revision = sid._revision;
	   _sub_authority_count = sid._sub_authority_count;
	   assert(_sub_authority_count <= SID_SIZE);
	   _identifier_authority = sid._identifier_authority;
	   for (int i = 0; i < _sub_authority_count; i++) { 
		   _sub_authority[i] = sid._sub_authority[i];
	   }
	}
	return *this;
}

sid_struct::sid_struct(const sid_struct& other)
{
	_revision = other._revision;
	_sub_authority_count = other._sub_authority_count;
	assert(_sub_authority_count <= SID_SIZE);
	_identifier_authority = other._identifier_authority;
	for (int i = 0; i < _sub_authority_count; i++) {
		_sub_authority[i] = other._sub_authority[i];
	}
}

bool 
sid_struct::operator==(const sid_struct& other) const { 

	if ( _sub_authority_count != other._sub_authority_count ) {
		return false;
	}
	for (int i=_sub_authority_count-1;i>=0 ;i--) {
		if (_sub_authority[i] != other._sub_authority[i]) {
			return false;
		}
	}
	return(_identifier_authority == other._identifier_authority) && (_revision == other._revision);
}



bool
sid_struct::sid_from_string(const std::string& sidstr) {

	// first setting the fields to 0 for the case of returning due to error. 
	_revision = 0;
	_sub_authority_count = 0;

	assert(sidstr != "");
	if ((sidstr == "") || ((sidstr.find("S-") == std::string::npos))) { //string must be of form S-Revision-IdentAuth-<subauth>
		// SID string doesn't start with S- 		
		return false;
	}

	std::string::size_type tmppos1,tmppos2;
	tmppos2 = 0;
	std::vector<std::string> tokens;
	char *str_end;
	while (  ( tmppos1 = sidstr.find_first_of("-",tmppos2+1) ) != std::string::npos ) {

		//tmppos1 now points to the end of the string, tmppos2
		// points to the beginning (including the "-").
		tokens.push_back(sidstr.substr(tmppos2+1, tmppos1-(tmppos2+1)));
		tmppos2 = tmppos1;
    }
	tokens.push_back(sidstr.substr(tmppos2+1));

	// S-Revision-IdentAuth-<subauth>
    // 0-  1      - 2       - 3 -> 18
	if (tokens.size() < 4) { // we need at least 4 fields (including the S) for a valid SID 
		// SID string doesn't contain enough fields		
		return false;
	}
	errno = 0;

	_revision = strtol(tokens[1].c_str(), &str_end, 10);
	// check for successful conversion  
	if ((errno) || (str_end != (tokens[1].c_str() + tokens[1].size()))) {
		// SID string is not a legal SID string. Couldn't get Revision field.		
		return false;
	}

	unsigned long long tmpAuthID;
	tmpAuthID = strtoll(tokens[2].c_str(), &str_end, 10);
	// check for successful conversion  
	if ((errno) || (str_end != (tokens[2].c_str() + tokens[2].size()))) {
		// SID string is not a legal SID string. Couldn't get IdentAuth field.		
		return false;
	}

 	_sub_authority_count = tokens.size()-3;
    assert(_sub_authority_count <= SID_SIZE);
	if (_sub_authority_count > SID_SIZE) {
		// Too many fields in SID string.		
		return false;
	}
    
	for (int i=0;i<_sub_authority_count;i++) {
		_sub_authority[i]=strtoll(tokens[i+3].c_str(), &str_end, 10);
		// check for successful conversion  
		if ((errno) || (str_end != (tokens[i+3].c_str() + tokens[i+3].size()))) {
			// SID SubtAuth field isn't a number.		
			return false;
		}
	}

    /* Store the identifier authority in big endian format */
    _identifier_authority[5] = (0x00000000000000FFll & tmpAuthID);   
    _identifier_authority[4] = (0x000000000000FF00ll & tmpAuthID) >> 8;
    _identifier_authority[3] = (0x0000000000FF0000ll & tmpAuthID) >> 16;
    _identifier_authority[2] = (0x00000000FF000000ll & tmpAuthID) >> 24;
    _identifier_authority[1] = (0x000000FF00000000ll & tmpAuthID) >> 32;
    _identifier_authority[0] = (0x0000FF0000000000ll & tmpAuthID) >> 40;

   	return true;
}

std::string
sid_struct::to_string() const
{
	// Adapted from MS KB 198907
	//
	std::stringstream sid_stream;
	sid_stream << "S-" << (unsigned int)_revision << "-";
	unsigned long long val=0;
	for (int i=0; i<6; i++) {
		val = val << 8;
        val |= _identifier_authority[i]; 
	}
	sid_stream << val << "-";
	for (int i=0; i < _sub_authority_count - 1; i++) {
		sid_stream << _sub_authority[i] << "-";
	}
	sid_stream << _sub_authority[_sub_authority_count - 1];
	return sid_stream.str();
}  

uint32_t
sid_struct::hash()const
{
	uint32 hval = _revision;
	for (int i = 0; i < SID_AUTHORITY_SIZE; i++) {
		hval += _identifier_authority[i];
	}
	for (int i=0; i < _sub_authority_count ; i++) {
		hval += _sub_authority[i];
	}
	return hval;

}

std::ostream& operator<< (std::ostream& os, const sid_struct& sidst) {

	os << sidst.to_string();
	return os;
}


#endif

