#ifndef SMB_ACL_MESSAGE_H
#define SMB_ACL_MESSAGE_H

#include <stdint.h>
#include <iostream>
#include "unixntacl.h"

enum smb_acl_message_type {
	MESSAGE_TYPE_REQUEST = 1,
	MESSAGE_TYPE_REPLY = 2,
};

enum smb_acl_message_code {
	MESSAGE_CODE_NULL = 0,
	MESSAGE_CODE_SET_ACL = 1,
	MESSAGE_CODE_GET_ACL = 2,
	MESSAGE_CODE_GET_ACCESS = 3,
    MESSAGE_CODE_SENDFILE = 4,
    MESSAGE_CODE_SHWRITE = 5,
    MESSAGE_CODE_RENAME = 6,
    MESSAGE_CODE_FSYNC = 7,    
    MESSAGE_CODE_DIRPLUS = 9,
	MESSAGE_CODE_READ = 10,
	MESSAGE_CODE_GET_SNAPSHOTS_TIMESTAMPS = 11,
	MESSAGE_CODE_GET_SNAPSHOT_NAME_BY_TIMESTAMP = 12,
	MESSAGE_CODE_GET_ANTIVIRUS_INFO = 13,
	MESSAGE_CODE_SET_ANTIVIRUS_INFO = 14,
	MESSAGE_CODE_SET_ANTIVIRUS_INFO_GUARDED = 15,
	MESSAGE_CODE_MAP_SID_TO_TEMP_ID = 16,
	MESSAGE_CODE_MAP_TEMP_ID_TO_SID = 17,
	MESSAGE_CODE_SET_CREATE_TIME = 18,
    MESSAGE_CODE_STATFS = 90,
};

enum smb_acl_error {
	SMBACL_MESSAGE_OK = 0,
	SMBACL_MESSAGE_GENERAL_ERROR = -1,
	SMBACL_MESSAGE_NO_SEC_STYLE = -2,		// Cannot get security style
	SMBACL_MESSAGE_STALE = -3,
	SMBACL_MESSAGE_PERM = -4,				// Permissions error
	SMBACL_MESSAGE_COMM_ERROR = -5,			// Client should retry
	SMBACL_MESSAGE_READ_ERROR = -6,			// ACL read error
	SMBACL_MESSAGE_WRITE_ERROR = -7,		// ACL write error
	SMBACL_MESSAGE_VOL_NO_SPACE_ERROR = -8,	// No space left on volume
	SMBACL_MESSAGE_FATAL_ERROR	 = -12,  	// fatal error -> see comment below
	SMBACL_MESSAGE_NOT_SUPPORTED = -15,		// Not supported
	SMBACL_MESSAGE_RO_FS = -16,				// Read only file system


	SMBVFSFS_MESSAGE_OK = SMBACL_MESSAGE_OK, // 0
	/* the file is stale */
	SMBVFSFS_MESSAGE_STALE_ERROR	= SMBACL_MESSAGE_STALE,	// -3

	SMBVFSFS_MESSAGE_PERM_ERROR = SMBACL_MESSAGE_PERM, // -4

	/* soft error, samba should re try */
	SMBVFSFS_MESSAGE_COMM_ERROR = SMBACL_MESSAGE_COMM_ERROR, // -5
	
	SMBVFSFS_MESSAGE_OUT_OF_VOLUME_SPACE_ERROR	= SMBACL_MESSAGE_VOL_NO_SPACE_ERROR, // -8

	/* user is out of quota */
	SMBVFSFS_MESSAGE_DQUOTA_ERROR = -9,
		   					
	SMBVFSFS_MESSAGE_FILE_MAXIMUM_SIZE_EXCEED_ERROR = -10,
	SMBVFSFS_MESSAGE_FILE_NOT_FOUND_ERROR = -11,
    
	
	/* 
		fatal error in the protocol		
		1. working with an volume with the pre rollback version 
		2. bad shared memory segments passed on the line
		3. using wrong fh (not one created by the fs)						 
		4. any error on sendfile, when no retries are allowed  
	*/
	SMBVFSFS_MESSAGE_FATAL_ERROR = SMBACL_MESSAGE_FATAL_ERROR,
	/* this operation state is not consistent there is no point in retry */
	SMBVFSFS_MESSAGE_OPERATION_FATAL_ERROR = -13,

	//sent to the void
	SMBVFSFS_MESSAGE_CLIENT_DISCONNECTED_ERROR = -14,

	// Reply to samba that the server didn't implement --> use next vfs layer
	SMBVFSFS_MESSAGE_NOT_SUPPORTED = SMBACL_MESSAGE_NOT_SUPPORTED, // -15
	
	// Read only file system
	SMBVFSFS_MESSAGE_RO_FS = SMBACL_MESSAGE_RO_FS, // -16 

	/*  soft error, samba should re try sending the ancillary_fd  
		this can happen as the fsd file limit was breached by previous operations
		and the fsd must inc his fd table, and cause samba to re send it 
	 */
	SMBVFSFS_MESSAGE_ANCILLARY_COMM_ERROR = -20,

	/* socket closed */
	SMBVFSFS_MESSAGE_SOCKET_CLOSED_ERROR = -21,

	/* volume with given name not found */
	SMBVFSFS_MESSAGE_VOLUME_NOT_FOUND = -22
};

std::ostream& operator<<(std::ostream& out, smb_acl_message_code code);

struct smb_acl_message {
	uint32_t	type;
	union {
		smb_acl_message_code	code;
		/* can be acl error or vfs error  */
		smb_acl_error	error;
	};
	
	smb_acl_message() {}
	smb_acl_message(const smb_acl_message& msg)
	:	type(msg.type)
	{
		switch (type) {
		case MESSAGE_TYPE_REQUEST:
			code = msg.code;
			break;
		case MESSAGE_TYPE_REPLY:
			error = msg.error;
			break;
		}
	}
};

struct smb_acl_request : smb_acl_message {
	uint32_t	fh_len;
	char		*fh_data;
	
	smb_acl_request() {}
	smb_acl_request(const smb_acl_request& req)
	:	smb_acl_message(req)
	,	fh_len(req.fh_len)
	,	fh_data(req.fh_data)
	{
	}
};

struct smb_acl_get_acl_request : smb_acl_request {
	smb_acl_get_acl_request() {}
	smb_acl_get_acl_request(const smb_acl_request& req)
	: smb_acl_request(req)
	{
	}
};

inline const uint32_t *decode_get_acl_request(smb_acl_request *, const uint32_t *p)
	{
		return p;
	}

struct smb_acl_get_acl_reply : smb_acl_message {
	struct unixntacl_acl	*acl;
};


struct smb_acl_set_acl_request : smb_acl_request {
	struct unixntacl_acl	*acl;
	
	smb_acl_set_acl_request() {}
	smb_acl_set_acl_request(const smb_acl_request& a_req)
	:	smb_acl_request(a_req)
	{
	}
};

struct smb_acl_set_acl_reply : smb_acl_message {
};

struct smb_acl_get_access_request : smb_acl_request {
	uint32_t	uid;
	uint32_t	ngroups;
	uint32_t	groups[GROUPS_PER_USER];
	
	smb_acl_get_access_request() {}
	smb_acl_get_access_request(const smb_acl_request& req)
	:	smb_acl_request(req)
	{
	}
};

struct smb_acl_get_access_reply : smb_acl_message {
	uint32_t	access;
	char 		access_troubleshooting_details[MAX_ACCESS_DETAILS_STR_LEN + 1];
};

#ifdef	__cplusplus
extern "C" {
#endif

uint32_t *encode_get_access_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_get_access_reply(smb_acl_message *rep, const uint32_t *p);

const uint32_t *decode_message(smb_acl_message *msg, const uint32_t *p);
const uint32_t *decode_request(smb_acl_request *req, const uint32_t *p);
uint32_t *encode_reply(smb_acl_error error, uint32_t *p);
uint32_t *encode_request(smb_acl_message_code code, uint32_t *p, const smb_acl_request& req);
const uint32_t *decode_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_get_acl_request(const smb_acl_request *req, uint32_t *p);

uint32_t *encode_get_acl_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_get_acl_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_set_acl_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_set_acl_request(smb_acl_request *req, const uint32_t *p);

uint32_t *encode_set_acl_reply(const smb_acl_message *rep, uint32_t *p);
const uint32_t *decode_set_acl_reply(smb_acl_message *rep, const uint32_t *p);

uint32_t *encode_get_access_request(const smb_acl_request *req, uint32_t *p);
const uint32_t *decode_get_access_request(smb_acl_request *req, const uint32_t *p);

#ifdef	__cplusplus
}
#endif

#endif // #ifndef SMB_ACL_MESSAGE_H

