#ifndef	XNT_XDR_MISC_H
#define	XNT_XDR_MISC_H

#include <endian.h>
#include <stdint.h>
#include <netinet/in.h>

#include "sid_struct.h"

/******************************************************************************/
/*                          Basic types decode/encode functions			 	  */
/******************************************************************************/

static inline uint32_t* 
xnt_enc32(const uint32_t *val, uint32_t *p)
{
    *p = htonl(*val);
   
    return p + 1;
}

static inline uint32_t* 
xnt_enc32(const uint32_t val, uint32_t *p)
{
    *p = htonl(val);
   
    return p + 1;
}

static inline const uint32_t * 
xnt_dec32(uint32_t *val, const uint32_t *p)
{
    *val = ntohl(*p);
    return p + 1;
}

static inline const uint32_t *
xnt_dec64(uint64_t *val, const uint32_t *p)
{
# if __BYTE_ORDER == __BIG_ENDIAN
	*val = *(uint64_t *) p;
# elif __BYTE_ORDER == __LITTLE_ENDIAN
	uint32_t *val32 = (uint32_t *)val;
	*val32++ = ntohl(p[1]);
	*val32 = ntohl(p[0]);
# else
# error "only little/big endian byte order supported"
# endif
	return p + 2;
}

static inline uint32_t *
xnt_enc64(const uint64_t *val, uint32_t *p)
{
# if __BYTE_ORDER == __BIG_ENDIAN
	*(uint64_t *) p = *val;
# elif __BYTE_ORDER == __LITTLE_ENDIAN
	const uint32_t *val32 = (const uint32_t *)val;
	p[1] = htonl(*val32); val32++;
	p[0] = htonl(*val32);
# else
# error "only little/big endian byte order supported"
# endif
	return p + 2;
}

static inline uint32_t* 
xnt_encode_bool(bool val, uint32_t *p)
{
	*p = htonl( val ? 1 : 0 );
	return p + 1;
}

static inline const uint32_t* 
xnt_decode_bool(bool* val, const uint32_t *p)
{
	*val = ( ntohl( *p ) != 0 );
	return p + 1;
}

template< typename T >
static inline T* ptr_plus_n_bytes( T* p, uint32_t n )
{
	return p + ( n / sizeof(T) ) + ( (n % sizeof(T)) == 0 ? 0 : 1 );
}

static inline uint32_t* 
xnt_encode_string( const char* str_in, uint32_t* p_buffer_out )
{
	uint32_t number_of_bytes_to_copy = strlen( str_in ) + 1;

	memcpy( p_buffer_out, str_in, number_of_bytes_to_copy );
	return ptr_plus_n_bytes( p_buffer_out, number_of_bytes_to_copy );
}

static inline const uint32_t* 
xnt_decode_string( const uint32_t* p_buffer_in, char* str_out )
{
	uint32_t number_of_bytes_to_copy = strlen( reinterpret_cast<const char*>(p_buffer_in) ) + 1;

	memcpy( str_out, p_buffer_in, number_of_bytes_to_copy );
	return ptr_plus_n_bytes( p_buffer_in, number_of_bytes_to_copy );
}

static inline uint32_t* 
xnt_encode_known_len_byte_seq( const uint8_t byte_seq_in[], uint32_t len_in, uint32_t* p_buffer_out )
{
	memcpy( p_buffer_out, byte_seq_in, len_in );
	return ptr_plus_n_bytes( p_buffer_out, len_in );
}

static inline const uint32_t* 
xnt_decode_known_len_byte_seq( const uint32_t* p_buffer_in, uint32_t len_in, uint8_t byte_seq_out[] )
{
	memcpy( byte_seq_out, p_buffer_in, len_in );
	return ptr_plus_n_bytes( p_buffer_in, len_in );
}

typedef uint8_t revision_and_sacount_and_sidauth_byte_array_t[8];

static inline uint32_t* 
xnt_encode_sid( const sid_struct& sid_in, uint32_t* p_buffer_out )
{
	revision_and_sacount_and_sidauth_byte_array_t revision_and_sacount_and_sidauth;
	// revision_and_sacount_and_sidauth  	
	// 		byte 0		- revision
	// 		byte 1 		- sub_authority_count
	//		bytes 2-7 	- identifier_authority
	revision_and_sacount_and_sidauth[0] = sid_in._revision;

	uint32_t sub_authority_count = sid_in._sub_authority_count;
	if ( sub_authority_count > SID_SIZE ) { 
		sub_authority_count = 0;
	}
	revision_and_sacount_and_sidauth[1] = sub_authority_count;

	for ( uint32_t i = 0 ; i < SID_AUTHORITY_SIZE ; ++i ) {
		revision_and_sacount_and_sidauth[i+2] = sid_in._identifier_authority[i];
	}

	p_buffer_out = xnt_encode_known_len_byte_seq( 
		revision_and_sacount_and_sidauth,
		sizeof( revision_and_sacount_and_sidauth_byte_array_t ),
		p_buffer_out );
	for ( uint32_t i = 0 ; i < sub_authority_count ; ++i ) {
		p_buffer_out = xnt_enc32( sid_in._sub_authority[i], p_buffer_out );
	}

	return p_buffer_out;
}

static inline const uint32_t* 
xnt_decode_sid( const uint32_t* p_buffer_in, sid_struct& sid_out )
{
	revision_and_sacount_and_sidauth_byte_array_t revision_and_sacount_and_sidauth;
	const uint32_t* p_buffer_in_current = p_buffer_in;

	p_buffer_in_current = xnt_decode_known_len_byte_seq( 
		p_buffer_in_current,
		sizeof( revision_and_sacount_and_sidauth_byte_array_t ),
		revision_and_sacount_and_sidauth );

	sid_out._revision = revision_and_sacount_and_sidauth[0];

	sid_out._sub_authority_count = revision_and_sacount_and_sidauth[1];
	if ( sid_out._sub_authority_count > SID_SIZE ) { 
		sid_out._sub_authority_count = 0;
	}

	for ( uint32_t i = 0 ; i < SID_AUTHORITY_SIZE ; ++i ) { 
		sid_out._identifier_authority[i] = revision_and_sacount_and_sidauth[i+2];
	}

	for ( uint32_t i = 0 ; i < sid_out._sub_authority_count ; ++i ) {
		p_buffer_in_current = xnt_dec32( &(sid_out._sub_authority[i]), p_buffer_in_current );
	}

	return p_buffer_in_current;
}

/******************************************************************************/
#endif // #ifndef	XNT_XDR_MISC_H

