/*
 *  Unix SMB/Netbios implementation.
 *  Utility for managing share permissions
 *
 *  Copyright (C) Tim Potter                    2000
 *  Copyright (C) Jeremy Allison                2000
 *  Copyright (C) Jelmer Vernooij               2003
 *  Copyright (C) Gerald (Jerry) Carter         2005.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "includes.h"

static TALLOC_CTX *ctx;

typedef struct set_of_strings_struct {
	char** array_of_strings;
	size_t first_free_index;
	size_t size;
} set_of_strings_struct;
typedef set_of_strings_struct* set_of_strings; 

void set_of_strings_delete( set_of_strings* pset_to_delete ) {
	if ( (*pset_to_delete) != NULL ) {
		set_of_strings set_to_delete = *pset_to_delete;

		if ( set_to_delete->array_of_strings != NULL ) {
			int i;
			for ( i = 0 ; i < set_to_delete->first_free_index ; ++i ) {
				free( set_to_delete->array_of_strings[i] );
			}
			memset( set_to_delete->array_of_strings, 0, set_to_delete->size );
			free( set_to_delete->array_of_strings );
			set_to_delete->array_of_strings = NULL;
		}
		ZERO_STRUCT(*set_to_delete);
		free( set_to_delete );
		set_to_delete = NULL;
		*pset_to_delete = NULL;
	}
}

BOOL set_of_strings_new( set_of_strings* pset_of_strings, size_t initial_size ) {
	BOOL ret = False;
	set_of_strings new_set = NULL;

	new_set = (set_of_strings)calloc( 1, sizeof( set_of_strings_struct ) );
	if ( new_set != NULL ) {
		DEBUG(10,("Successfully allocated struct for set_of_strings_struct.\n"));
	} else {
		DEBUG(0,("Failed to allocate struct for set_of_strings_struct.\n"));
		goto out;
	}

	new_set->first_free_index = 0;
	new_set->array_of_strings = (char**)calloc( initial_size, sizeof( char* ) );
	if ( new_set->array_of_strings != NULL ) {
		DEBUG(10,("Successfully allocated array of char* of size %d.\n", initial_size));
	} else {
		DEBUG(0,("Failed to allocate array of char* of size %d.\n", initial_size));
		goto out;
	}
	new_set->size = initial_size;

	*pset_of_strings = new_set;
	new_set = NULL;
	ret = True;

out:

	if ( new_set != NULL ) {
		set_of_strings_delete( &new_set );
	}
	return ret;
}

BOOL set_of_strings_add( set_of_strings this_obj, const char* str_to_add ) {
	BOOL ret = False;
	char* str_to_add_copy = NULL;

	str_to_add_copy = strdup( str_to_add );
	if ( str_to_add_copy != NULL ) {
		DEBUG(10,("Successfully duplicated string \"%s\".\n", str_to_add));
	} else {
		DEBUG(0,("Failed to duplicate string \"%s\".\n", str_to_add));
		goto out;
	}

	if ( this_obj->first_free_index == this_obj->size ) {
		char** new_array = NULL;
		size_t new_size = 2 * this_obj->size;

		new_array = (char**)realloc( this_obj->array_of_strings, new_size );
		if ( new_array != NULL ) {
			DEBUG(10,("Successfully realloc-ed internal array from size %d to %d.\n", this_obj->size, new_size));
		} else {
			DEBUG(0,("Failed to realloc internal array from size %d to %d.\n", this_obj->size, new_size));
			goto out;
		}
		this_obj->array_of_strings = new_array;
	}

	this_obj->array_of_strings[ this_obj->first_free_index ] = str_to_add_copy;
	++this_obj->first_free_index;
	str_to_add_copy = NULL;
	ret = True;

out:

	if ( str_to_add_copy != NULL ) {
		free( str_to_add_copy );
		str_to_add_copy = NULL;
	}

	return ret;
}

BOOL set_of_strings_contains( set_of_strings this_obj, const char* str_to_check ) {
	int i;
	for ( i = 0 ; i < this_obj->first_free_index ; ++i ) {
		if ( strcmp( this_obj->array_of_strings[i], str_to_check ) == 0 ) {
			return True;
		}
	}
	return False;
}

enum acl_mode {
	SMB_ACL_INVALID = 0, 
	SMB_ACL_REMOVE_ACE, 
	SMB_ACL_MODIFY, 
	SMB_ACL_ADD, 
	SMB_ACL_SET, 
	SMB_ACL_VIEW, 
	SMB_ACL_BACKUP, 
	SMB_ACL_RESTORE,
	SMB_ACL_CREATE,
	SMB_ACL_DELETE };

struct perm_value {
	const char *perm;
	uint32 mask;
};

/* These values discovered by inspection */

static const struct perm_value special_values[] = {
	{ "R", SEC_RIGHTS_FILE_READ },
	{ "W", SEC_RIGHTS_FILE_WRITE },
	{ "X", SEC_RIGHTS_FILE_EXECUTE },
	{ "D", SEC_STD_DELETE },
	{ "P", SEC_STD_WRITE_DAC },
	{ "O", SEC_STD_WRITE_OWNER },
	{ NULL, 0 },
};

#define SEC_RIGHTS_DIR_CHANGE ( SEC_RIGHTS_DIR_READ|SEC_STD_DELETE|SEC_RIGHTS_DIR_WRITE|SEC_DIR_TRAVERSE )

static const struct perm_value standard_values[] = {
	{ "READ",   SEC_RIGHTS_DIR_READ|SEC_DIR_TRAVERSE },
	{ "CHANGE", SEC_RIGHTS_DIR_CHANGE },
	{ "FULL",   SEC_RIGHTS_DIR_ALL },
	{ NULL, 0 },
};

/********************************************************************
 print an ACE on a FILE
********************************************************************/

static void print_ace(FILE *f, SEC_ACE *ace)
{
	const struct perm_value *v;
	int do_print = 0;
	uint32 got_mask;

	fprintf(f, "%s:", sid_string_static(&ace->trustee));

	/* Ace type */

	if (ace->type == SEC_ACE_TYPE_ACCESS_ALLOWED) {
		fprintf(f, "ALLOWED");
	} else if (ace->type == SEC_ACE_TYPE_ACCESS_DENIED) {
		fprintf(f, "DENIED");
	} else {
		fprintf(f, "%d", ace->type);
	}

	/* Not sure what flags can be set in a file ACL */

	fprintf(f, "/%d/", ace->flags);

	/* Standard permissions */

	for (v = standard_values; v->perm; v++) {
		if (ace->access_mask == v->mask) {
			fprintf(f, "%s", v->perm);
			return;
		}
	}

	/* Special permissions.  Print out a hex value if we have
	   leftover bits in the mask. */

	got_mask = ace->access_mask;

 again:
	for (v = special_values; v->perm; v++) {
		if ((ace->access_mask & v->mask) == v->mask) {
			if (do_print) {
				fprintf(f, "%s", v->perm);
			}
			got_mask &= ~v->mask;
		}
	}

	if (!do_print) {
		if (got_mask != 0) {
			fprintf(f, "0x%08x", ace->access_mask);
		} else {
			do_print = 1;
			goto again;
		}
	}
}

/********************************************************************
 print a ascii version of a security descriptor on a FILE handle
********************************************************************/

static void sec_desc_print(FILE *f, SEC_DESC *sd, BOOL only_ACL)
{
	uint32 i;

	if (!only_ACL) {
		fprintf(f, "REVISION:%d\n", sd->revision);
		/* Print owner and group sid */

		fprintf(f, "OWNER:%s\n", sid_string_static(sd->owner_sid));

		fprintf(f, "GROUP:%s\n", sid_string_static(sd->group_sid));
	}


	/* Print aces */
	for (i = 0; sd->dacl && i < sd->dacl->num_aces; i++) {
		SEC_ACE *ace = &sd->dacl->aces[i];
		if (!only_ACL) {
			fprintf(f, "ACL:");
		} else {
			if ( i != 0 ) {
				fprintf(f, ",");
			}
		}
		print_ace(f, ace);
		if (!only_ACL) {
			fprintf(f, "\n");
		}
	}

}

/********************************************************************
    parse an ACE in the same format as print_ace()
********************************************************************/

static BOOL parse_ace(SEC_ACE *ace, const char *orig_str)
{
	char *p;
	const char *cp;
	fstring tok;
	unsigned int atype = 0;
	unsigned int aflags = 0;
	unsigned int amask = 0;
	DOM_SID sid;
	SEC_ACCESS mask;
	const struct perm_value *v;
	char *str = SMB_STRDUP(orig_str);

	if (!str) {
		return False;
	}

	ZERO_STRUCTP(ace);
	p = strchr_m(str,':');
	if (!p) {
		printf("ACE '%s': missing ':'.\n", orig_str);
		SAFE_FREE(str);
		return False;
	}
	*p = '\0';
	p++;
	/* Try to parse numeric form */

	if (sscanf(p, "%i/%i/%i", &atype, &aflags, &amask) == 3 &&
	    string_to_sid(&sid, str)) {
		goto done;
	}

	/* Try to parse text form */

	if (!string_to_sid(&sid, str)) {
		printf("ACE '%s': failed to convert '%s' to SID\n",
			orig_str, str);
		SAFE_FREE(str);
		return False;
	}

	cp = p;
	if (!next_token(&cp, tok, "/", sizeof(fstring))) {
		printf("ACE '%s': failed to find '/' character.\n",
			orig_str);
		SAFE_FREE(str);
		return False;
	}

	if (strncmp(tok, "ALLOWED", strlen("ALLOWED")) == 0) {
		atype = SEC_ACE_TYPE_ACCESS_ALLOWED;
	} else if (strncmp(tok, "DENIED", strlen("DENIED")) == 0) {
		atype = SEC_ACE_TYPE_ACCESS_DENIED;
	} else {
		printf("ACE '%s': missing 'ALLOWED' or 'DENIED' entry at '%s'\n",
			orig_str, tok);
		SAFE_FREE(str);
		return False;
	}

	/* Only numeric form accepted for flags at present */
	/* no flags on share permissions */

	if (!(next_token(&cp, tok, "/", sizeof(fstring)) &&
	      sscanf(tok, "%i", &aflags) && aflags == 0)) {
		printf("ACE '%s': bad integer flags entry at '%s'\n",
			orig_str, tok);
		SAFE_FREE(str);
		return False;
	}

	if (!next_token(&cp, tok, "/", sizeof(fstring))) {
		printf("ACE '%s': missing / at '%s'\n",
			orig_str, tok);
		SAFE_FREE(str);
		return False;
	}

	if (strncmp(tok, "0x", 2) == 0) {
		if (sscanf(tok, "%i", &amask) != 1) {
			printf("ACE '%s': bad hex number at '%s'\n",
				orig_str, tok);
			SAFE_FREE(str);
			return False;
		}
		goto done;
	}

	for (v = standard_values; v->perm; v++) {
		if (strcmp(tok, v->perm) == 0) {
			amask = v->mask;
			goto done;
		}
	}

	p = tok;

	while(*p) {
		BOOL found = False;

		for (v = special_values; v->perm; v++) {
			if (v->perm[0] == *p) {
				amask |= v->mask;
				found = True;
			}
		}

		if (!found) {
			printf("ACE '%s': bad permission value at '%s'\n",
				orig_str, p);
			SAFE_FREE(str);
		 	return False;
		}
		p++;
	}

	if (*p) {
		SAFE_FREE(str);
		return False;
	}

 done:
	mask = amask;
	init_sec_ace(ace, &sid, atype, mask, aflags);
	SAFE_FREE(str);
	return True;
}


/********************************************************************
********************************************************************/

static SEC_DESC* parse_acl_string(TALLOC_CTX *mem_ctx, const char *szACL, size_t *sd_size )
{
	SEC_DESC *sd = NULL;
	SEC_ACE *ace;
	SEC_ACL *acl;
	int num_ace;
	const char *pacl;
	int i;
	
	if ( !szACL )
		return NULL;

	pacl = szACL;
	num_ace = count_chars( pacl, ',' ) + 1;
	
	if ( !(ace = TALLOC_ZERO_ARRAY( mem_ctx, SEC_ACE, num_ace )) ) 
		return NULL;
	
	for ( i=0; i<num_ace; i++ ) {
		char *end_acl = strchr_m( pacl, ',' );
		fstring acl_string;

		strncpy( acl_string, pacl, MIN( PTR_DIFF( end_acl, pacl ), sizeof(fstring)-1) );
		acl_string[MIN( PTR_DIFF( end_acl, pacl ), sizeof(fstring)-1)] = '\0';
		
		if ( !parse_ace( &ace[i], acl_string ) )
			return NULL;

		pacl = end_acl;
		pacl++;
	}
	
	if ( !(acl = make_sec_acl( mem_ctx, NT4_ACL_REVISION, num_ace, ace )) )
		return NULL;
		
	sd = make_sec_desc( mem_ctx, SEC_DESC_REVISION, SEC_DESC_SELF_RELATIVE, 
		NULL, NULL, NULL, acl, sd_size);

	return sd;
}

/* add an ACE to a list of ACEs in a SEC_ACL */
static BOOL add_ace(TALLOC_CTX *mem_ctx, SEC_ACL **the_acl, SEC_ACE *ace)
{
	SEC_ACL *new_ace;
	SEC_ACE *aces;
	if (! *the_acl) {
		return (((*the_acl) = make_sec_acl(mem_ctx, 3, 1, ace)) != NULL);
	}

	if (!(aces = SMB_CALLOC_ARRAY(SEC_ACE, 1+(*the_acl)->num_aces))) {
		return False;
	}
	memcpy(aces, (*the_acl)->aces, (*the_acl)->num_aces * sizeof(SEC_ACE));
	memcpy(aces+(*the_acl)->num_aces, ace, sizeof(SEC_ACE));
	new_ace = make_sec_acl(mem_ctx,(*the_acl)->revision,1+(*the_acl)->num_aces, aces);
	SAFE_FREE(aces);
	(*the_acl) = new_ace;
	return True;
}

/* The MSDN is contradictory over the ordering of ACE entries in an ACL.
   However NT4 gives a "The information may have been modified by a
   computer running Windows NT 5.0" if denied ACEs do not appear before
   allowed ACEs. */

static int ace_compare(SEC_ACE *ace1, SEC_ACE *ace2)
{
	if (sec_ace_equal(ace1, ace2)) 
		return 0;

	if (ace1->type != ace2->type) 
		return ace2->type - ace1->type;

	if (sid_compare(&ace1->trustee, &ace2->trustee)) 
		return sid_compare(&ace1->trustee, &ace2->trustee);

	if (ace1->flags != ace2->flags) 
		return ace1->flags - ace2->flags;

	if (ace1->access_mask != ace2->access_mask) 
		return ace1->access_mask - ace2->access_mask;

	if (ace1->size != ace2->size) 
		return ace1->size - ace2->size;

	return memcmp(ace1, ace2, sizeof(SEC_ACE));
}

static void sort_acl(SEC_ACL *the_acl)
{
	uint32 i;
	if (!the_acl) return;

	qsort(the_acl->aces, the_acl->num_aces, sizeof(the_acl->aces[0]), QSORT_CAST ace_compare);

	for (i=1;i<the_acl->num_aces;) {
		if (sec_ace_equal(&the_acl->aces[i-1], &the_acl->aces[i])) {
			int j;
			for (j=i; j<the_acl->num_aces-1; j++) {
				the_acl->aces[j] = the_acl->aces[j+1];
			}
			the_acl->num_aces--;
		} else {
			i++;
		}
	}
}


static int change_share_sec(TALLOC_CTX *mem_ctx, 
							const char *tdb_file_path,	
							const char *sharename, 
							const char *the_acl, 
							enum acl_mode mode,
							FILE* output_file, 
							BOOL only_ACLs)
{
	SEC_DESC *sd = NULL;
	SEC_DESC *old = NULL;
	size_t sd_size = 0;
	uint32 i, j;
	
	if (mode != SMB_ACL_SET) {
	    if (!(old = get_share_security_for_tdb( mem_ctx, tdb_file_path, sharename, &sd_size )) ) {
		fprintf(stderr, "Unable to retrieve permissions for share [%s]\n", sharename);
		return -1;
	    }
	}

	if ( (mode != SMB_ACL_VIEW) && !(sd = parse_acl_string(mem_ctx, the_acl, &sd_size )) ) {
		fprintf( stderr, "Failed to parse acl\n");
		return -1;
	}
	
	switch (mode) {
	case SMB_ACL_VIEW:
		sec_desc_print(output_file, old, only_ACLs);
		return 0;
	case SMB_ACL_REMOVE_ACE:
	    for (i=0;sd->dacl && i<sd->dacl->num_aces;i++) {
		BOOL found = False;

		for (j=0;old->dacl && j<old->dacl->num_aces;j++) {
		    if (sec_ace_equal(&sd->dacl->aces[i], &old->dacl->aces[j])) {
			uint32 k;
			for (k=j; k<old->dacl->num_aces-1;k++) {
			    old->dacl->aces[k] = old->dacl->aces[k+1];
			}
			old->dacl->num_aces--;
			found = True;
			break;
		    }
		}

		if (!found) {
		printf("ACL for ACE:");
		print_ace(stdout, &sd->dacl->aces[i]);
		printf(" not found\n");
		}
	    }

	    break;
	case SMB_ACL_MODIFY:
	    for (i=0;sd->dacl && i<sd->dacl->num_aces;i++) {
		BOOL found = False;

		for (j=0;old->dacl && j<old->dacl->num_aces;j++) {
		    if (sid_equal(&sd->dacl->aces[i].trustee,
			&old->dacl->aces[j].trustee)) {
			old->dacl->aces[j] = sd->dacl->aces[i];
			found = True;
		    }
		}

		if (!found) {
		    printf("ACL for SID %s not found\n", sid_string_static(&sd->dacl->aces[i].trustee));
		}
	    }

	    if (sd->owner_sid) {
		old->owner_sid = sd->owner_sid;
	    }

	    if (sd->group_sid) {
		old->group_sid = sd->group_sid;
	    }
	    break;
	case SMB_ACL_ADD:
	    for (i=0;sd->dacl && i<sd->dacl->num_aces;i++) {
		add_ace(mem_ctx, &old->dacl, &sd->dacl->aces[i]);
	    }
	    break;
	case SMB_ACL_SET:
	    old = sd;
	    break;
	default:
		fprintf( stderr, "Unsupported mode = %d.\n", mode );
		return -1;
	}

	/* Denied ACE entries must come before allowed ones */
	sort_acl(old->dacl);

	if ( !set_share_security_for_tdb( tdb_file_path, sharename, old ) ) {
	    fprintf( stderr, "Failed to store acl for share [%s]\n", sharename );
	    return 2;
	}
	return 0;
}


static const char* BACKUP_VERSION_HEADER = "share permissions backup version 1";
static const char* BACKUP_RESULT_END_OF_FILE_MARKER = "share permissions backup end of file";

static BOOL read_line( FILE* file_to_read_from, char* buffer, size_t buffer_size )
{
	char* rc_str = NULL;
	size_t length_stored = 0;

	rc_str = fgets( buffer, (int)buffer_size, file_to_read_from );
	if ( rc_str == NULL ) {
		return False;
	}
	
	length_stored = strlen( buffer );
	if ( buffer[ length_stored - 1 ] == '\n' ) {
		buffer[ length_stored - 1 ] = '\0';
	}

	return True;
}

static BOOL skip_line( FILE* file_to_read_from, 
					   const char* file_name, 
					   const char* file_desc, 
					   const char* line_to_comp, 
					   const char* line_desc )
{
	BOOL rc_BOOL = False;
	pstring line_read;

	pstrcpy( line_read, "" );

	rc_BOOL = read_line( file_to_read_from, line_read, PSTRING_LEN );
	if ( rc_BOOL ) {
		DEBUG(10,("Successfully read %s from file %s \"%s\": \"%s\".\n", 
				  line_desc, file_desc, file_name, line_read));
	} else {
		DEBUG(0,("Failed read %s from file %s \"%s\", errno = %d(%s).\n",
				 line_desc, file_desc, file_name, errno, strerror(errno)));
		return False;
	}

	if ( strcmp( line_read, line_to_comp ) == 0 ) {
		DEBUG(10,("Successfully matched read %s \"%s\" against expected \"%s\".\n",
				 line_desc, line_read, line_to_comp));
	} else {
		DEBUG(0,("Failure: read %s \"%s\" does not match expected \"%s\".\n",
				 line_desc, line_read, line_to_comp));
		return False;
	}

	return True;
}

/********************************************************************
  back
********************************************************************/

static int backup_shares_sec( TALLOC_CTX *mem_ctx, 
							  const char* share_names_file_name, 
							  const char* backup_result_file_name )
{
	int ret = 0;
	int rc_int = 0;
	FILE* share_names_file = NULL;
	FILE* backup_result_file = NULL;

	/* open file containing list of share names */
    share_names_file = fopen( share_names_file_name, "r" );
	if ( share_names_file != NULL ) {
		DEBUG(10,("Successfully opened file containing list of share names \"%s\".\n",
				 share_names_file_name));
	} else {
		DEBUG(0,("Failed to open file containing list of share names \"%s\", errno = %d(%s).\n",
				 share_names_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}
	/* open file for backup result */
	backup_result_file = fopen( backup_result_file_name, "w" );
	if ( backup_result_file != NULL ) {
		DEBUG(10,("Successfully opened for write file for backup result \"%s\".\n",
				 backup_result_file_name));
	} else {
		DEBUG(0,("Failed to open for write file for backup result \"%s\", errno = %d(%s).\n",
				 backup_result_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

	/* write version header */
	rc_int = fprintf( backup_result_file, "%s\n\n", BACKUP_VERSION_HEADER );
	if ( rc_int > 0 ) {
		DEBUG(10,("Successfully writen version header \"%s\" to backup result file \"%s\".\n",
				  BACKUP_VERSION_HEADER, backup_result_file_name));
	} else {
		DEBUG(0,("Failed to write version header \"%s\" to backup result file \"%s\", errno = %d(%s).\n",
				 BACKUP_VERSION_HEADER, backup_result_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

	for ( ; ; ) {
		pstring share_name;	
		BOOL rc_BOOL = False;

		pstrcpy( share_name, "" );

		rc_BOOL = read_line( share_names_file, share_name, PSTRING_LEN );
		if ( rc_BOOL ) {
			DEBUG(10,("current share = \"%s\".\n", share_name));
		} else {
			DEBUG(10,("Reached end of file with list of share names \"%s\".\n", 
					  share_names_file_name));
			break;
		}

		rc_int = fprintf( backup_result_file, "[%s]\n", share_name );
		if ( rc_int > 0 ) {
			DEBUG(10,("Successfully writen name sub-header for share \"%s\" to backup result file \"%s\".\n",
					  share_name, backup_result_file_name));
		} else {
			DEBUG(0,("Failed to write name sub-header for share \"%s\" to backup result file \"%s\", errno = %d(%s).\n",
					 share_name, backup_result_file_name, errno, strerror(errno)));
			ret = 4;
			goto out;
		}

		rc_int = change_share_sec( 
					mem_ctx, 
					lock_path("share_info.tdb"), 
					share_name, 
					NULL, 
					SMB_ACL_VIEW,
					backup_result_file,
					True );
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully writen ACLs for share \"%s\" to backup result file \"%s\".\n",
					  share_name, backup_result_file_name));
		} else {
			DEBUG(0,("Failed to write ACLs for share \"%s\" to backup result file \"%s\", errno = %d(%s).\n",
					 share_name, backup_result_file_name, errno, strerror(errno)));
			ret = 4;
			goto out;
		}

		rc_int = fprintf( backup_result_file, "\n\n" );
		if ( rc_int > 0 ) {
			DEBUG(10,("Successfully writen EOL after section for share \"%s\" to backup result file \"%s\".\n",
					  share_name, backup_result_file_name));
		} else {
			DEBUG(0,("Failed to write EOL after section for share \"%s\" to backup result file \"%s\", errno = %d(%s).\n",
					 share_name, backup_result_file_name, errno, strerror(errno)));
			ret = 4;
			goto out;
		}
	}

	/* write end of file marker */
	rc_int = fprintf( backup_result_file, "%s\n", BACKUP_RESULT_END_OF_FILE_MARKER );
	if ( rc_int > 0 ) {
		DEBUG(10,("Successfully writen end of file marker \"%s\" to backup result file \"%s\".\n",
				  BACKUP_RESULT_END_OF_FILE_MARKER, backup_result_file_name));
	} else {
		DEBUG(0,("Failed to write end of file marker \"%s\" to backup result file \"%s\", errno = %d(%s).\n",
				 BACKUP_RESULT_END_OF_FILE_MARKER, backup_result_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

out:

	if ( share_names_file != NULL ) {
        int rc_int = 0;
		rc_int = fclose( share_names_file ); 
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully closed file containing list of share names \"%s\".\n",
					 share_names_file_name));
		} else {
			DEBUG(1,("Failed to close file containing list of share names \"%s\", errno = %d(%s).\n",
					 share_names_file_name, errno, strerror(errno)));
		}
	}
	if ( backup_result_file != NULL ) {
		int rc_int = 0;
		rc_int = fclose( backup_result_file ); 
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully closed file for backup result \"%s\".\n",
					 backup_result_file_name));
		} else {
			DEBUG(1,("Failed to close file for backup result \"%s\", errno = %d(%s).\n",
					 backup_result_file_name, errno, strerror(errno)));
		}
	}

	return ret;
}


/********************************************************************
  restore
********************************************************************/

static int restore_shares_sec( TALLOC_CTX* mem_ctx, 
							   const char* backup_result_file_name, 
							   const char* share_names_file_name, 
							   const char* share_info_to_write_to_file_name )
{
	int ret = 0;
	int rc_int = 0;
	FILE* share_names_file = NULL;
	FILE* backup_result_file = NULL;
	BOOL rc_BOOL = False;
	pstring current_line;
	char* buffer_for_ACL = NULL;
	const size_t MAX_ACE_SIZE_ASCII = strlen( "S-d-d-dd-ddddddddd-dddddddddd-dddddddddd-dddd:ALLOWED/ddd/0xdddddddd" );
	const size_t MAX_DELIMITER_SIZE_ASCII = strlen( "," );
	const size_t MAX_NUMBER_OF_ACES_IN_ACL = 2000;
	const size_t BUFFER_FOR_ACL_SIZE = 1 + MAX_NUMBER_OF_ACES_IN_ACL * ( MAX_ACE_SIZE_ASCII + MAX_DELIMITER_SIZE_ASCII );
	set_of_strings shares_restored = NULL;

	pstrcpy( current_line, "" );

	/* open file for backup result */
	backup_result_file = fopen( backup_result_file_name, "r" );
	if ( backup_result_file != NULL ) {
		DEBUG(10,("Successfully opened for read file with backup result \"%s\".\n",
				 backup_result_file_name));
	} else {
		DEBUG(0,("Failed to open for read file with backup result \"%s\", errno = %d(%s).\n",
				 backup_result_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

	rc_BOOL = set_of_strings_new( &shares_restored, 100 /* initial_size */ );
	if ( ! rc_BOOL ) {
		ret = 4;
		goto out;
	}

	/* parse version header */
	rc_BOOL = skip_line( backup_result_file, 
					     backup_result_file_name, 
				         "with backup result", 
				         BACKUP_VERSION_HEADER, 
				         "version header line" );
	if ( ! rc_BOOL ) {
		ret = 4;
		goto out;
	}
	rc_BOOL = skip_line( backup_result_file, 
						 backup_result_file_name, 
						 "with backup result", 
						 "", 
						 "empty line" );
	if ( ! rc_BOOL ) {
		ret = 4;
		goto out;
	}

	buffer_for_ACL = (char*)malloc( BUFFER_FOR_ACL_SIZE );
	if ( buffer_for_ACL != NULL ) {
		DEBUG(10,("Successfully allocated buffer of size %d for ACL in ASCII.\n", BUFFER_FOR_ACL_SIZE));
	} else {
		DEBUG(0,("Failed to allocated buffer of size %d for ACL in ASCII.\n", BUFFER_FOR_ACL_SIZE));
		ret = 4;
		goto out;
	}

	for ( ; ; ) {
		pstring share_name;	
		BOOL rc_BOOL = False;
		size_t current_line_len = 0;

		pstrcpy( share_name, "" );

		rc_BOOL = read_line( backup_result_file, current_line, PSTRING_LEN );
		if ( rc_BOOL ) {
			DEBUG(10,("Successfully read line from file with backup result \"%s\": \"%s\".\n", 
					  backup_result_file_name, current_line));
		} else {
			DEBUG(0,("Failed to read line from file with backup result \"%s\", errno = %d(%s).\n",
					 backup_result_file_name, errno, strerror(errno)));
			ret = 4;
			goto out;
		}
		if ( strcmp( current_line, BACKUP_RESULT_END_OF_FILE_MARKER ) == 0 ) {
			DEBUG(10,("Successfully matched end of file marker from file with backup result \"%s\": \"%s\".\n", 
					  backup_result_file_name, current_line));
			break;
		} else {
			DEBUG(10,("Parsing current line \"%s\" as share name sub-header...\n", current_line));
		}

		current_line_len = strlen( current_line );
		if ( ( current_line[0] == '[' ) && ( current_line[current_line_len-1] == ']' ) && ( current_line_len >= 3 ) ) {
			strncpy( share_name, current_line + 1, current_line_len - 2 );
			share_name[ current_line_len - 2 ] = '\0';
			DEBUG(10,("Current share name \"%s\".\n", share_name));
		} else {
			DEBUG(0,("Current line \"%s\" does not have required share name sub-header format: [<SHARE_NAME>].\n",
					 current_line));
			ret = 4;
			goto out;
		}

		rc_BOOL = read_line( backup_result_file, buffer_for_ACL, BUFFER_FOR_ACL_SIZE );
		if ( rc_BOOL ) {
			DEBUG(10,("Successfully read line with ACL for share \"%s\" from file with backup result \"%s\": \"%s\".\n", 
					  share_name, backup_result_file_name, buffer_for_ACL));
		} else {
			DEBUG(0,("Failed to read line with ACL for share \"%s\" from file with backup result \"%s\", errno = %d(%s).\n",
					 share_name, backup_result_file_name, errno, strerror(errno)));
			ret = 4;
			goto out;
		}

		rc_int = change_share_sec( 
					mem_ctx, 
					share_info_to_write_to_file_name, 
					share_name, 
					buffer_for_ACL, 
					SMB_ACL_SET,
					stdout,
					False );
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully stored ACL \"%s\" for share \"%s\" to share info file \"%s\".\n",
					  buffer_for_ACL, share_name, share_info_to_write_to_file_name));
		} else {
			DEBUG(0,("Failed to store ACL \"%s\" for share \"%s\" to backup result file \"%s\", rc_int = %d.\n",
					 buffer_for_ACL, share_name, share_info_to_write_to_file_name, rc_int));
			ret = 4;
			goto out;
		}

		rc_BOOL = set_of_strings_add( shares_restored, share_name );
		if ( ! rc_BOOL ) {
			ret = 4;
			goto out;
		}

		rc_BOOL = skip_line( backup_result_file, 
							 backup_result_file_name, 
							 "with backup result", 
							 "", 
							 "empty line" );
		if ( ! rc_BOOL ) {
			ret = 4;
			goto out;
		}
	}

	/* 
	 * go over current list of shares and remove permissions for unrestored shares
	 */

	/* open file containing list of share names */
	share_names_file = fopen( share_names_file_name, "r" );
	if ( share_names_file != NULL ) {
		DEBUG(10,("Successfully opened file containing list of share names \"%s\".\n",
				 share_names_file_name));
	} else {
		DEBUG(0,("Failed to open file containing list of share names \"%s\", errno = %d(%s).\n",
				 share_names_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

	/* go over share list */
	for ( ; ; ) {
		pstring share_name;	
		BOOL rc_BOOL = False;

		pstrcpy( share_name, "" );

		rc_BOOL = read_line( share_names_file, share_name, PSTRING_LEN );
		if ( rc_BOOL ) {
			DEBUG(10,("current share = \"%s\".\n", share_name));
		} else {
			DEBUG(10,("Reached end of file with list of share names \"%s\".\n", 
					  share_names_file_name));
			break;
		}

		/* if share was restored go to the next in the list */
		if ( set_of_strings_contains( shares_restored, share_name ) ) {
			continue;
		}

		/* if share was not restored then delete its permissions */
		rc_BOOL = delete_share_security_for_tdb( 
					share_info_to_write_to_file_name, 
					share_name );
		if ( rc_BOOL ) {
			DEBUG(10,("Successfully deleted ACL for share \"%s\" in share info file \"%s\".\n",
					  share_name, share_info_to_write_to_file_name));
		} else {
			DEBUG(0,("Failed to delete ACL for share \"%s\" in share info file \"%s\".\n",
					 share_name, share_info_to_write_to_file_name));
			return 4;
		}
	}

out:

	if ( share_names_file != NULL ) {
        int rc_int = 0;
		rc_int = fclose( share_names_file ); 
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully closed file containing list of share names \"%s\".\n",
					 share_names_file_name));
		} else {
			DEBUG(1,("Failed to close file containing list of share names \"%s\", errno = %d(%s).\n",
					 share_names_file_name, errno, strerror(errno)));
		}
	}

	if ( buffer_for_ACL != NULL ) {
		free( buffer_for_ACL ); 
		buffer_for_ACL = NULL;  
		DEBUG(10,("Successfully freed buffer of size %d for ACL in ASCII.\n", BUFFER_FOR_ACL_SIZE));
	}

	set_of_strings_delete( &shares_restored );

	if ( backup_result_file != NULL ) {
		int rc_int = 0;
		rc_int = fclose( backup_result_file ); 
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully closed file with backup result \"%s\".\n",
					 backup_result_file_name));
		} else {
			DEBUG(1,("Failed to close file with backup result \"%s\", errno = %d(%s).\n",
					 backup_result_file_name, errno, strerror(errno)));
		}
	}

	return ret;
}

static int create_shares_sec( TALLOC_CTX* mem_ctx, 
							  const char* share_names_file_name, 
							  const char* share_info_to_write_to_file_name )
{
	int ret = 0;
	int rc_int = 0;
	FILE* share_names_file = NULL;
	pstring current_line;
	static const char* default_share_ACL = "S-1-1-0:ALLOWED/0/FULL";

	pstrcpy( current_line, "" );

	/* open file containing list of share names */
	share_names_file = fopen( share_names_file_name, "r" );
	if ( share_names_file != NULL ) {
		DEBUG(10,("Successfully opened file containing list of share names \"%s\".\n",
				 share_names_file_name));
	} else {
		DEBUG(0,("Failed to open file containing list of share names \"%s\", errno = %d(%s).\n",
				 share_names_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

	/* go over share list */
	for ( ; ; ) {
		pstring share_name;	
		BOOL rc_BOOL = False;

		pstrcpy( share_name, "" );

		rc_BOOL = read_line( share_names_file, share_name, PSTRING_LEN );
		if ( rc_BOOL ) {
			DEBUG(10,("current share = \"%s\".\n", share_name));
		} else {
			DEBUG(10,("Reached end of file with list of share names \"%s\".\n", 
					  share_names_file_name));
			break;
		}

		rc_int = change_share_sec( 
					mem_ctx, 
					share_info_to_write_to_file_name, 
					share_name, 
					default_share_ACL, 
					SMB_ACL_SET,
					stdout,
					False );
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully stored default ACL \"%s\" for share \"%s\" to share info file \"%s\".\n",
					  default_share_ACL, share_name, share_info_to_write_to_file_name));
		} else {
			DEBUG(0,("Failed to store default ACL \"%s\" for share \"%s\" to backup result file \"%s\", rc_int = %d.\n",
					 default_share_ACL, share_name, share_info_to_write_to_file_name, rc_int));
			ret = 4;
			goto out;
		}
	}

out:

	if ( share_names_file != NULL ) {
        int rc_int = 0;
		rc_int = fclose( share_names_file ); 
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully closed file containing list of share names \"%s\".\n",
					 share_names_file_name));
		} else {
			DEBUG(1,("Failed to close file containing list of share names \"%s\", errno = %d(%s).\n",
					 share_names_file_name, errno, strerror(errno)));
		}
	}

	return ret;
}

static int delete_shares_sec( TALLOC_CTX* mem_ctx, 
							  const char* share_names_file_name, 
							  const char* share_info_to_write_to_file_name )
{
	int ret = 0;
	FILE* share_names_file = NULL;
	pstring current_line;

	pstrcpy( current_line, "" );

	/* open file containing list of share names */
	share_names_file = fopen( share_names_file_name, "r" );
	if ( share_names_file != NULL ) {
		DEBUG(10,("Successfully opened file containing list of share names \"%s\".\n",
				 share_names_file_name));
	} else {
		DEBUG(0,("Failed to open file containing list of share names \"%s\", errno = %d(%s).\n",
				 share_names_file_name, errno, strerror(errno)));
		ret = 4;
		goto out;
	}

	/* go over share list */
	for ( ; ; ) {
		pstring share_name;	
		BOOL rc_BOOL = False;

		pstrcpy( share_name, "" );

		rc_BOOL = read_line( share_names_file, share_name, PSTRING_LEN );
		if ( rc_BOOL ) {
			DEBUG(10,("current share = \"%s\".\n", share_name));
		} else {
			DEBUG(10,("Reached end of file with list of share names \"%s\".\n", 
					  share_names_file_name));
			break;
		}

		rc_BOOL = delete_share_security_for_tdb( 
					share_info_to_write_to_file_name, 
					share_name );
		if ( rc_BOOL ) {
			DEBUG(10,("Successfully deleted ACL for share \"%s\" in share info file \"%s\".\n",
					  share_name, share_info_to_write_to_file_name));
		} else {
			DEBUG(0,("Failed to delete ACL for share \"%s\" in share info file \"%s\".\n",
					 share_name, share_info_to_write_to_file_name));
         ret = 4;
         goto out;
		}
	}

out:

	if ( share_names_file != NULL ) {
        int rc_int = 0;
		rc_int = fclose( share_names_file ); 
		if ( rc_int == 0 ) {
			DEBUG(10,("Successfully closed file containing list of share names \"%s\".\n",
					 share_names_file_name));
		} else {
			DEBUG(1,("Failed to close file containing list of share names \"%s\", errno = %d(%s).\n",
					 share_names_file_name, errno, strerror(errno)));
		}
	}

	return ret;
}


/********************************************************************
  main program
********************************************************************/

int main(int argc, const char *argv[])
{
	int opt;
	int retval = 0;
	enum acl_mode mode = SMB_ACL_INVALID;
	const char* used_option = NULL;
	static char *the_acl = NULL;
	fstring sharename;
	BOOL force_acl = False;
	int snum;
	poptContext pc;
	BOOL initialize_sid = False;
	struct poptOption long_options[] = {
		POPT_AUTOHELP
		{ "remove", 'r', POPT_ARG_STRING, &the_acl, 'r', "Delete an ACE", "ACL" },
		{ "modify", 'm', POPT_ARG_STRING, &the_acl, 'm', "Modify an acl", "ACL" },
		{ "add", 'a', POPT_ARG_STRING, &the_acl, 'a', "Add an ACE", "ACL" },
		{ "replace", 'R', POPT_ARG_STRING, &the_acl, 'R', "Set share permission ACL", "ACLS" },
		{ "view", 'v', POPT_ARG_NONE, NULL, 'v', "View current share permissions" },
		{ "machine-sid", 'M', POPT_ARG_NONE, NULL, 'M', "Initialize the machine SID" },
		{ "force", 'F', POPT_ARG_NONE, NULL, 'F', "Force storing the ACL", "ACLS" },
		{ "backup", 'b', POPT_ARG_NONE, NULL, 'b', "Backup current share permissions for list of shares" },
		{ "restore", 't', POPT_ARG_NONE, NULL, 't', "Restore share permissions" },
		{ "create", 'c', POPT_ARG_NONE, NULL, 'c', "Set default share permissions for list of shares" },
		{ "delete", 'e', POPT_ARG_NONE, NULL, 'e', "Delete share permissions for list of shares" },
		POPT_COMMON_SAMBA
		{ NULL }
	};

	if ( !(ctx = talloc_init("main")) ) {
		fprintf( stderr, "Failed to initialize talloc context!\n");
		return -1;
	}

	/* set default debug level to 1 regardless of what smb.conf sets */
	setup_logging( "sharesec", True );
	DEBUGLEVEL_CLASS[DBGC_ALL] = 1;
	dbf = x_stderr;
	x_setbuf( x_stderr, NULL );

	pc = poptGetContext("sharesec", argc, argv, long_options, 0);
	
	poptSetOtherOptionHelp(pc, "sharename\n");

	while ((opt = poptGetNextOpt(pc)) != -1) {
		switch (opt) {
		case 'r':
			the_acl = smb_xstrdup(poptGetOptArg(pc));
			mode = SMB_ACL_REMOVE_ACE;
			break;

		case 'm':
			the_acl = smb_xstrdup(poptGetOptArg(pc));
			mode = SMB_ACL_MODIFY;
			break;

		case 'a':
			the_acl = smb_xstrdup(poptGetOptArg(pc));
			mode = SMB_ACL_ADD;
			break;
		case 'R':
			the_acl = smb_xstrdup(poptGetOptArg(pc));
			mode = SMB_ACL_SET;
			break;

		case 'v':
			mode = SMB_ACL_VIEW;
			break;

		case 'F':
			force_acl = True;
			break;
			
		case 'M':
			initialize_sid = True;
			break;

		case 'b':
			used_option = "-b|--backup";
			if ( mode != SMB_ACL_INVALID ) {
				DEBUG(0,("Option %s cannot be combined with any other option.\n", used_option));
				return -1;
			}
			mode = SMB_ACL_BACKUP;
			break;

		case 't':
			used_option = "-t|--restore";
			if ( mode != SMB_ACL_INVALID ) {
				DEBUG(0,("Option %s cannot be combined with any other option.\n", used_option));
				return -1;
			}
			mode = SMB_ACL_RESTORE;
			break;

		case 'c':
			used_option = "-c|--create";
			if ( mode != SMB_ACL_INVALID ) {
				DEBUG(0,("Option %s cannot be combined with any other option.\n", used_option));
				return -1;
			}
			mode = SMB_ACL_CREATE;
			break;

		case 'e':
			used_option = "-e|--delete";
			if ( mode != SMB_ACL_INVALID ) {
				DEBUG(0,("Option %s cannot be combined with any other option.\n", used_option));
				return -1;
			}
			mode = SMB_ACL_DELETE;
			break;
		}
	}
	
	setlinebuf(stdout);

	if ( mode == SMB_ACL_BACKUP || mode == SMB_ACL_RESTORE || 
		 mode == SMB_ACL_CREATE || mode == SMB_ACL_DELETE ) {
		if ( force_acl ) {
			DEBUG(0,("Invalid combination of -F and %s.\n", used_option));
			return -1;
		}
		if ( initialize_sid ) {
			DEBUG(0,("Invalid combination of -M and %s.\n", used_option));
			return -1;
		}
	}

	load_case_tables();

	lp_load( dyn_CONFIGFILE, False, False, False, True );

	/* check for initializing secrets.tdb first */
	
	if ( initialize_sid ) {
		DOM_SID *sid = get_global_sam_sid();
		
		if ( !sid ) {
			fprintf( stderr, "Failed to retrieve Machine SID!\n");
			return 3;
		}
		
		printf ("%s\n", sid_string_static( sid ) );
		return 0;
	}

	if ( mode == SMB_ACL_VIEW && force_acl ) {
		fprintf( stderr, "Invalid combination of -F and -v\n");
		return -1;
	}

	if ( mode == SMB_ACL_BACKUP ) {
		pstring share_names_file_name;
		pstring backup_result_file_name;
		const char* errorMessage = "parameter to option -b is missing.\nFORMAT: -b|--backup <share_names_file> <backup_result_file>";

		pstrcpy(share_names_file_name, "");
		pstrcpy(backup_result_file_name, "");

		if(!poptPeekArg(pc)) { 
			DEBUG(0,("1st %s\n", errorMessage));
			return -1;
		}
		pstrcpy(share_names_file_name, poptGetArg(pc));

		if(!poptPeekArg(pc)) { 
			DEBUG(0,("2nd %s\n", errorMessage));
			return -1;
		}
		pstrcpy(backup_result_file_name, poptGetArg(pc));

		retval = backup_shares_sec(ctx, share_names_file_name, backup_result_file_name);

		goto out;
	}

	if ( mode == SMB_ACL_RESTORE ) {
		pstring backup_result_file_name;
		pstring share_names_file_name;
		pstring merge_result_share_info_file_name;
		const char* errorMessage = "parameter to option -t is missing.\nFORMAT: -t|--restore <backup_result_file> <share_names_file> <merge_result_share_info.tdb>";

		pstrcpy(backup_result_file_name, "");
		pstrcpy(merge_result_share_info_file_name, "");

		if(!poptPeekArg(pc)) {
			DEBUG(0,("1st %s\n", errorMessage));
			return -1;
		}
		pstrcpy(backup_result_file_name, poptGetArg(pc));

		if(!poptPeekArg(pc)) { 
			DEBUG(0, ("2nd %s\n", errorMessage));
			return -1;
		}
		pstrcpy(share_names_file_name, poptGetArg(pc));

		if(!poptPeekArg(pc)) { 
			DEBUG(0, ("3rd %s\n", errorMessage));
			return -1;
		}
		pstrcpy(merge_result_share_info_file_name, poptGetArg(pc));

		retval = restore_shares_sec(ctx, 
									backup_result_file_name, 
									share_names_file_name, 
									merge_result_share_info_file_name);

		goto out;
	}

	if ( mode == SMB_ACL_CREATE || mode == SMB_ACL_DELETE ) {
		pstring share_names_file_name;
		pstring merge_result_share_info_file_name;

		pstrcpy(share_names_file_name, "");
		pstrcpy(merge_result_share_info_file_name, "");

		if(!poptPeekArg(pc)) { 
			DEBUG(0,("1st parameter to option %s is missing.\nFORMAT: %s <share_names_file> <merge_result_share_info.tdb>\n", used_option, used_option));
			return -1;
		}
		pstrcpy(share_names_file_name, poptGetArg(pc));

		if(!poptPeekArg(pc)) { 
			DEBUG(0,("2nd parameter to option %s is missing.\nFORMAT: %s <share_names_file> <merge_result_share_info.tdb>\n", used_option, used_option));
			return -1;
		}
		pstrcpy(merge_result_share_info_file_name, poptGetArg(pc));

		if ( mode == SMB_ACL_CREATE ) {
			retval = create_shares_sec(ctx, share_names_file_name, merge_result_share_info_file_name);
		} else {
			retval = delete_shares_sec(ctx, share_names_file_name, merge_result_share_info_file_name);
		}

		goto out;
	}

	/* get the sharename */

	if(!poptPeekArg(pc)) { 
		poptPrintUsage(pc, stderr, 0);	
		return -1;
	}
	
	fstrcpy(sharename, poptGetArg(pc));
	
	snum = lp_servicenumber( sharename );
	
	if ( snum == -1 && !force_acl ) {
		fprintf( stderr, "Invalid sharename: %s\n", sharename);
		return -1;
	}
		
	retval = change_share_sec(ctx, lock_path("share_info.tdb"), sharename, the_acl, mode, stdout, False);

out:

	talloc_destroy(ctx);

	return retval;
}
